theory CoSP imports
  Complex_Main
  "$AFP/CryptHOL/GPV_Applicative"
  "$AFP/CryptHOL/Computational_Model"
  "$AFP/CryptHOL/Negligible"
  Markov_Search
  "~~/src/HOL/Library/Sublist"
  "~~/src/HOL/Library/BNF_Corec"
begin

lemma rel_pmf_bindI_strong: (* Move to Probability_More *)
  assumes "rel_pmf R p q"
  and "\<And>x y. \<lbrakk> R x y; x \<in> set_pmf p; y \<in> set_pmf q \<rbrakk> \<Longrightarrow> rel_pmf P (f x) (g y)"
  shows "rel_pmf P (bind_pmf p f) (bind_pmf q g)"
apply(rule rel_pmf_bindI)
 apply(rule pmf.rel_mono_strong)
  apply(rule assms)
 apply(erule conjI|rule TrueI)+
apply(simp add: assms)
done


lemma spmf_of_pmf_map_the: "lossless_spmf p \<Longrightarrow> spmf_of_pmf (map_pmf the p) = p"
apply(simp add: spmf_of_pmf_def pmf.map_comp)
apply(rewrite in "_ = \<hole>" pmf.map_id[symmetric])
apply(rule pmf.map_cong[OF refl])
apply(auto intro!: option.collapse iff del: not_None_eq simp add: lossless_iff_set_pmf_None)
done

lemma list_case_distrib_lambda: "(\<lambda>x. case_list (nil x) (cons x) xs) = case_list nil (\<lambda>x y z. cons z x y) xs"
by(cases xs) simp_all

lemma list_case_distrib_apply: "case_list nil cons xs y = case_list (nil y) (\<lambda>x xs. cons x xs y) xs"
by(cases xs) simp_all

lemma append_conv_snoc: "xs @ x # ys = (xs @ [x]) @ ys"
by simp

lemma results_map_gpv_id [simp]:
  "results_gpv \<I> (map_gpv f id gpv) = f ` results_gpv \<I> gpv" (is "?lhs = ?rhs")
proof(rule set_eqI iffI)+
  show "x \<in> ?rhs" if "x \<in> ?lhs" for x using that
    by(induction gpv'\<equiv>"map_gpv f id gpv" arbitrary: gpv)(fastforce intro: results_gpv.Pure results_gpv.IO simp add: gpv.map_sel)+
next
  fix x
  assume "x \<in> ?rhs"
  then obtain y where "x = f y" and "y \<in> results_gpv \<I> gpv" by auto
  from this(2) show "x \<in> ?lhs" unfolding \<open>x = f y\<close>
    by(induction)(auto 4 4 intro: results_gpv.Pure results_gpv.IO rev_image_eqI simp add: gpv.map_sel)
qed

lemma Some_eq_map_option: "Some x = map_option f y \<longleftrightarrow> (\<exists>z. y = Some z \<and> x = f z)"
  by(cases y) auto





definition select :: "'a \<Rightarrow> 'a set \<Rightarrow> 'a"
where "select x A = (if x \<in> A then x else SOME y. y \<in> A)"

lemma select_in [simp, intro]: "A \<noteq> {} \<Longrightarrow> select x A \<in> A"
by(auto simp add: select_def intro: someI)

lemma select_simps [simp]: "x \<in> A \<Longrightarrow> select x A = x"
by(simp add: select_def)

lemma selectI:
  "\<lbrakk> A \<noteq> {}; x \<in> A \<Longrightarrow> P x; \<And>x'. \<lbrakk> x \<notin> A; x' \<in> A \<rbrakk> \<Longrightarrow> P x' \<rbrakk>
  \<Longrightarrow> P (select x A)"
by(auto simp add: select_def intro: someI2)

lemma select_empty [simp]: fixes x :: 'a shows "select x {} = (SOME x :: 'a. False)"
by(simp add: select_def)

lemma ennreal_inject: "ennreal x = ennreal y \<longleftrightarrow> (x \<le> 0 \<longleftrightarrow> y \<le> 0) \<and> (x \<ge> 0 \<longrightarrow> y \<ge> 0 \<longrightarrow> x = y)"
including ennreal.lifting by transfer(auto simp add: ennreal_def max_def)

function enat_ceiling :: "ennreal \<Rightarrow> enat"
where
  "enat_ceiling (ennreal x) = enat (nat \<lceil>x\<rceil>)"
| "enat_ceiling \<top> = \<infinity>"
subgoal for P x by(cases x) auto
by(auto simp add: ennreal_inject)
termination by lexicographic_order

lemma enat_ceiling_0 [simp]: "enat_ceiling 0 = 0"
by(simp only: ennreal_0[symmetric] enat_ceiling.simps) simp

lemma enat_ceiling_1 [simp]: "enat_ceiling 1 = 1"
by(simp only: ennreal_1[symmetric] enat_ceiling.simps)(simp add: one_enat_def)

lemma enat_ceiling_numeral [simp]: "enat_ceiling (numeral n) = numeral n"
by(simp only: ennreal_numeral[symmetric] enat_ceiling.simps)(simp add: enat_numeral)

definition size_map :: "('b \<Rightarrow> nat) \<Rightarrow> ('a \<rightharpoonup> 'b) \<Rightarrow> enat"
where "size_map f m = enat_ceiling (\<integral>\<^sup>+ x. f (the (m x)) \<partial>count_space (dom m))"

lemma size_map_empty [simp]: "size_map f Map.empty = 0"
by(simp add: size_map_def)

lemma enat_iless_iff: "n < enat m \<longleftrightarrow> (\<exists>n'. n = enat n' \<and> n' < m)"
by(cases n) simp_all


context fixes f :: "'a \<Rightarrow> 'a set" begin

qualified inductive iteratep :: "'a \<Rightarrow> 'a list \<Rightarrow> bool"
where
  Nil: "iteratep s []"
| Cons: "\<lbrakk> s' \<in> f s; iteratep s' xs \<rbrakk> \<Longrightarrow> iteratep s (s' # xs)"

definition iterate :: "'a \<Rightarrow> 'a list set"
where "iterate s = {xs. iteratep s xs}"

lemma iteratep_iterate_eq [pred_set_conv]: "iteratep s = (\<lambda>xs. xs \<in> iterate s)"
by(simp add: iterate_def)

context begin
local_setup {* Local_Theory.map_background_naming (Name_Space.mandatory_path "iterate") *}
-- \<open>These theorems will shadow existing names only after all contexts have been left.\<close>
qualified lemmas intros = iteratep.intros[to_set]
  and LNil  = Nil[to_set]
  and LCons = Cons[to_set]
  and induct [consumes 1, case_names Nil Cons, induct set: iterate] = iteratep.induct[to_set]
  and cases [consumes 1, case_names Nil Cons, cases set: iterate] = iteratep.cases[to_set]
  and simps = iteratep.simps[to_set]
end

inductive_simps iterate_simps [to_set, iff]:
  "iteratep s []"
  "iteratep s (Cons x xs)"

declare iterate.intros[intro!]

lemma iterate_unfold: "iterate s = insert [] (\<Union>s' \<in> f s. op # s' ` iterate s')"
unfolding set_eq_iff by(subst CoSP.iterate.simps) auto

lemma iterate_all_nthI: "(\<And>n. n < length xs \<Longrightarrow> xs ! n \<in> f ((s # xs) ! n)) \<Longrightarrow> xs \<in> iterate s"
proof(induction xs arbitrary: s)
  case Nil show ?case by simp
next
  case (Cons x xs)
  from Cons.prems[of 0] have "x \<in> f s" by simp
  moreover have "xs \<in> iterate x"
  proof(rule Cons.IH)
    fix n
    assume "n < length xs"
    hence "n + 1 < length (x # xs)" by simp
    from Cons.prems[OF this] show "xs ! n \<in> f ((x # xs) ! n)" by simp
  qed
  ultimately show ?case ..
qed

lemma iterate_nthD:
  assumes "xs \<in> iterate s" "n < length xs"
  shows "xs ! n \<in> f ((s # xs) ! n)"
using assms
proof(induction arbitrary: n)
  case (Cons s' s xs) then show ?case by(cases n) simp_all
qed simp



qualified coinductive literatep :: "'a \<Rightarrow> 'a llist \<Rightarrow> bool"
where
  LNil: "literatep s LNil"
| LCons: "\<lbrakk> s' \<in> f s; literatep s' xs \<rbrakk> \<Longrightarrow> literatep s (LCons s' xs)"

definition literate :: "'a \<Rightarrow> 'a llist set"
where "literate s = {xs. literatep s xs}"

lemma literatep_literate_eq [pred_set_conv]: "literatep s = (\<lambda>xs. xs \<in> literate s)"
by(simp add: literate_def)

context begin
local_setup {* Local_Theory.map_background_naming (Name_Space.mandatory_path "literate") *}
lemmas intros[intro!] = literatep.intros[to_set]
  and LNil  = LNil[to_set]
  and LCons = LCons[to_set]
  and coinduct [consumes 1, case_names literate, case_conclusion literate LNil LCons, coinduct set: literate] = literatep.coinduct[to_set]
  and cases [consumes 1, case_names literate, cases set: literate] = literatep.cases[to_set]
  and simps = literatep.simps[to_set]
end

inductive_simps literate_simps [to_set, iff]:
  "literatep s LNil"
  "literatep s (LCons x xs)"

lemma literate_unfold: "literate s = insert LNil (\<Union>s' \<in> f s. LCons s' ` literate s')"
unfolding set_eq_iff by(subst CoSP.literate.simps) auto

end

context partial_function_definitions begin

lemma fixp_strong_induct_uc':
  fixes F :: "'c \<Rightarrow> 'c"
    and U :: "'c \<Rightarrow> 'b \<Rightarrow> 'a"
    and C :: "('b \<Rightarrow> 'a) \<Rightarrow> 'c"
    and P :: "('b \<Rightarrow> 'a) \<Rightarrow> bool"
  assumes mono: "\<And>x. mono_body (\<lambda>f. U (F (C f)) x)"
    and eq: "f \<equiv> C (fixp_fun (\<lambda>f. U (F (C f))))"
    and inverse: "\<And>f. U (C f) = f"
    and adm: "ccpo.admissible lub_fun le_fun P"
    and bot: "P (\<lambda>_. lub {})"
    and step: "\<And>f'. \<lbrakk> P (U f'); \<And>x. leq (U f' x) (U f x) \<rbrakk> \<Longrightarrow> P (U (F f'))"
  shows "P (U f)"
unfolding eq inverse
apply (rule ccpo.fixp_strong_induct[OF ccpo adm])
apply (insert mono, auto simp: monotone_def fun_ord_def bot fun_lub_def)[2]
apply (rule_tac f'5="C x" in step)
apply (simp_all add: inverse eq fun_ord_def)
done

end

declaration \<open>Partial_Function.init "spmf_strong" @{term spmf.fixp_fun}
  @{term spmf.mono_body} @{thm spmf.fixp_rule_uc} @{thm spmf.fixp_induct_strong2_uc}
  NONE\<close>

context fixes K :: "'s \<Rightarrow> 's option spmf" notes [[function_internals]] begin

partial_function (spmf_strong) iterate_spmf :: "'s \<Rightarrow> 's list spmf"
where
  "iterate_spmf s = 
   bind_spmf (K s) 
     (\<lambda>so. case so of None \<Rightarrow> return_spmf []
           | Some s' \<Rightarrow> map_spmf (op # s') (iterate_spmf s'))"

text \<open>We assume a call-by-value setting here: @{term iterate_spmf} returns the lift of visited states 
  as a whole instead of lazily value by value. Thus, if the iteration does not stop, it collapses into
  the unassigned probability mass, rather than producing an infinite lazy list of states.\<close>

lemma iterate_spmf_fixp_induct [case_names adm bottom step]:
  assumes "ccpo.admissible (fun_lub lub_spmf) (fun_ord (ord_spmf op =)) P"
  and "P (\<lambda>_. return_pmf None)"
  and "\<And>iterate_spmf'.
    \<lbrakk> \<And>s. ord_spmf op = (iterate_spmf' s) (iterate_spmf s); 
      \<And>s. ord_spmf op = (iterate_spmf' s) (bind_spmf (K s) (\<lambda>so. case so of None \<Rightarrow> return_spmf [] | Some s' \<Rightarrow> map_spmf (op # s') (iterate_spmf' s')));
      P iterate_spmf' \<rbrakk>
    \<Longrightarrow> P (\<lambda>s. bind_spmf (K s) (\<lambda>so. case so of None \<Rightarrow> return_spmf [] | Some s' \<Rightarrow> map_spmf (op # s') (iterate_spmf' s')))"
  shows "P iterate_spmf"
using assms
by(rule iterate_spmf.fixp_induct)(unfold fun_ord_def, blast+)

lemma iterate_spmf_subset_iterate: "set_spmf (iterate_spmf s) \<subseteq> iterate (\<lambda>s. \<Union>x\<in>set_spmf (K s). set_option x) s"
proof(induction arbitrary: s rule: iterate_spmf_fixp_induct)
  case adm show ?case by(rule cont_intro ccpo_class.admissible_leI)+
  case bottom show ?case by simp
  case (step iterate_spmf')
  then show ?case
    by(subst iterate_unfold)(fastforce split: option.split_asm simp add: in_set_spmf)
qed

lemma iterate_spmf_induct_set [consumes 1, case_names stop step]:
  assumes *: "xs \<in> set_spmf (iterate_spmf s)"
  and stop: "\<And>s. None \<in> set_spmf (K s) \<Longrightarrow> P [] s"
  assumes step: "\<And>xs s s'. \<lbrakk> Some s' \<in> set_spmf (K s); xs \<in> set_spmf (iterate_spmf s'); P xs s' \<rbrakk> \<Longrightarrow> P (s' # xs) s"
  shows "P xs s"
proof -
  have "set_spmf (iterate_spmf s) \<subseteq> {xs. P xs s}"
  proof(induction arbitrary: s rule: iterate_spmf_fixp_induct)
    case adm show ?case by(rule cont_intro ccpo_class.admissible_leI)+
    case bottom show ?case by simp
    case step': (step iterate_spmf')
    from step'.hyps(1) have "set_spmf (iterate_spmf' s) \<subseteq> set_spmf (iterate_spmf s)" for s
      by(rule monotone_set_spmf[THEN monotoneD])
    with step'.IH show ?case
      by(fastforce simp add: set_spmf_bind_pmf in_set_spmf[symmetric] lossless_iff_set_pmf_None image_iff split: option.split_asm intro!: step stop)
  qed
  with * show ?thesis by blast
qed

end

lemma iterate_spmf_parametric [transfer_rule]:
  includes lifting_syntax shows
  "((S ===> rel_spmf (rel_option S)) ===> S ===> rel_spmf (list_all2 S))
   iterate_spmf iterate_spmf"
unfolding iterate_spmf_def[abs_def]
apply(rule rel_funI)
apply(rule fixp_spmf_parametric[OF iterate_spmf.mono iterate_spmf.mono])
subgoal premises [transfer_rule] by transfer_prover
done

(* context fixes K :: "'s \<Rightarrow> 's spmf" notes [[function_internals]] begin

declaration \<open>Partial_Function.init "spmf_strong" @{term spmf.fixp_fun}
  @{term spmf.mono_body} @{thm spmf.fixp_rule_uc} @{thm spmf.fixp_induct_strong2_uc}
  NONE\<close>

partial_function (spmf_strong) iterate_spmf :: "'s \<Rightarrow> 's list spmf"
where
  "iterate_spmf s = 
   bind_pmf (K s) 
     (\<lambda>so. case so of None \<Rightarrow> return_spmf []
           | Some s' \<Rightarrow> map_spmf (op # s') (iterate_spmf s'))"

lemma iterate_spmf_fixp_induct [case_names adm bottom step]:
  assumes "ccpo.admissible (fun_lub lub_spmf) (fun_ord (ord_spmf op =)) P"
  and "P (\<lambda>_. return_pmf None)"
  and "\<And>iterate_spmf'.
    \<lbrakk> \<And>s. ord_spmf op = (iterate_spmf' s) (iterate_spmf s); 
      \<And>s. ord_spmf op = (iterate_spmf' s) (bind_pmf (K s) (\<lambda>so. case so of None \<Rightarrow> return_spmf [] | Some s' \<Rightarrow> map_spmf (op # s') (iterate_spmf' s')));
      P iterate_spmf' \<rbrakk>
    \<Longrightarrow> P (\<lambda>s. bind_pmf (K s) (\<lambda>so. case so of None \<Rightarrow> return_spmf [] | Some s' \<Rightarrow> map_spmf (op # s') (iterate_spmf' s')))"
  shows "P iterate_spmf"
using assms
by(rule iterate_spmf.fixp_induct)(unfold fun_ord_def, blast+)

lemma iterate_spmf_subset_iterate: "set_spmf (iterate_spmf s) \<subseteq> iterate (set_spmf \<circ> K) s"
proof(induction arbitrary: s rule: iterate_spmf_fixp_induct)
  case adm show ?case by(rule cont_intro ccpo_class.admissible_leI)+
  case bottom show ?case by simp
  case (step iterate_spmf')
  then show ?case
    by(subst iterate_unfold)(fastforce split: option.split_asm simp add: in_set_spmf)
qed

lemma iterate_spmf_induct_set [consumes 1, case_names stop step]:
  assumes *: "xs \<in> set_spmf (iterate_spmf s)"
  and stop: "\<And>s. \<not> lossless_spmf (K s) \<Longrightarrow> P [] s"
  assumes step: "\<And>xs s s'. \<lbrakk> s' \<in> set_spmf (K s); xs \<in> set_spmf (iterate_spmf s'); P xs s' \<rbrakk> \<Longrightarrow> P (s' # xs) s"
  shows "P xs s"
proof -
  have "set_spmf (iterate_spmf s) \<subseteq> {xs. P xs s}"
  proof(induction arbitrary: s rule: iterate_spmf_fixp_induct)
    case adm show ?case by(rule cont_intro ccpo_class.admissible_leI)+
    case bottom show ?case by simp
    case step': (step iterate_spmf')
    from step'.hyps(1) have "set_spmf (iterate_spmf' s) \<subseteq> set_spmf (iterate_spmf s)" for s
      by(rule monotone_set_spmf[THEN monotoneD])
    with step'.IH show ?case
      by(fastforce simp add: set_spmf_bind_pmf in_set_spmf[symmetric] lossless_iff_set_pmf_None image_iff split: option.split_asm intro!: step stop)
  qed
  with * show ?thesis by blast
qed

end

lemma iterate_spmf_parametric [transfer_rule]:
  includes lifting_syntax shows
  "((S ===> rel_spmf S) ===> S ===> rel_spmf (list_all2 S))
   iterate_spmf iterate_spmf"
unfolding iterate_spmf_def[abs_def]
apply(rule rel_funI)
apply(rule fixp_spmf_parametric[OF iterate_spmf.mono iterate_spmf.mono])
subgoal premises [transfer_rule] by transfer_prover
done
 *)


lemma exec_gpv_map:
  fixes callee :: "('call, 'ret, 's) callee"
  shows "exec_gpv callee (map_gpv f g gpv) s = map_spmf (apfst f) (exec_gpv (\<lambda>s c. callee s (g c)) gpv s)"
using exec_gpv_parametric[of "op = :: 's \<Rightarrow> _" "BNF_Def.Grp UNIV g" "BNF_Def.Grp UNIV f", where ?'ret='ret]
unfolding relator_eq
apply(subst (asm) (2 4) eq_alt)
apply(unfold prod.rel_Grp option.rel_Grp pmf.rel_Grp apfst_def gpv.rel_Grp)
apply(simp add: rel_fun_def Grp_def)
done

lemma ord_spmf_mapI:
  includes lifting_syntax shows
  "\<lbrakk> ord_spmf R p q; (R ===> R') f g \<rbrakk> \<Longrightarrow> ord_spmf R' (map_spmf f p) (map_spmf g q)"
apply(simp add: pmf.rel_map ord_option_map2 ord_option_map1[abs_def])
apply(erule rel_pmf_mono)
apply(erule ord_option_mono)
subgoal premises [transfer_rule] by transfer_prover
done

text \<open>
  We need a more refined version of @{const exec_gpv} which distinguishes between failures
  in the gpv and those in the callee. If the GPV fails after some interactions with the callee,
  @{const exec_gpv} does not return the new state of the callee.
\<close>

context fixes callee :: "('call, 'ret, 's) callee" notes [[function_internals]] begin

partial_function (spmf) exec_gpv' :: "('a, 'call, 'ret) gpv \<Rightarrow> 's \<Rightarrow> ('a option \<times> 's) spmf"
where
  "exec_gpv' c s = bind_pmf (the_gpv c) (\<lambda>x.
    case x of None \<Rightarrow> return_spmf (None, s)
    | Some (Pure y) \<Rightarrow> return_spmf (Some y, s)
    | Some (IO out rpv) \<Rightarrow> bind_spmf (callee s out) (\<lambda>(input, s'). exec_gpv' (rpv input) s'))"

lemma exec_gpv'_fixp_induct [case_names adm bottom step]:
  assumes "ccpo.admissible (fun_lub lub_spmf) (fun_ord (ord_spmf op =)) (\<lambda>f. P (\<lambda>c s. f (c, s)))"
  and "P (\<lambda>_ _. return_pmf None)"
  and "\<And>exec_gpv'. P exec_gpv' \<Longrightarrow> 
     P (\<lambda>c s. the_gpv c \<bind> (\<lambda>x. case x of None \<Rightarrow> return_spmf (None, s) | Some (Pure y) \<Rightarrow> return_spmf (Some y, s) | Some (IO out c) \<Rightarrow> callee s out \<bind> (\<lambda>(x, y). exec_gpv' (c x) y)))"
  shows "P exec_gpv'"
using assms(1)
by(rule exec_gpv'.fixp_induct[unfolded curry_conv[abs_def]])(simp_all add: assms(2-))

end

lemma exec_gpv'_parametric [transfer_rule]:
  includes lifting_syntax shows
  "((S ===> C ===> rel_spmf (rel_prod (op = :: 'ret \<Rightarrow> _) S))
   ===> rel_gpv A C ===> S ===> rel_spmf (rel_prod (rel_option A) S))
   exec_gpv' exec_gpv'"
unfolding exec_gpv'_def[abs_def]
apply(rule rel_funI)
apply(rule curry_transfer[THEN rel_funD])
apply(rule fixp_spmf_parametric[OF exec_gpv'.mono exec_gpv'.mono])
subgoal premises [transfer_rule] by transfer_prover
done

lemma exec_gpv'_map:
  fixes callee :: "('call, 'ret, 's) callee"
  shows "exec_gpv' callee (map_gpv f g gpv) s = map_spmf (apfst (map_option f)) (exec_gpv' (\<lambda>s c. callee s (g c)) gpv s)"
using exec_gpv'_parametric[of "op = :: 's \<Rightarrow> _" "BNF_Def.Grp UNIV g" "BNF_Def.Grp UNIV f", where ?'ret='ret]
unfolding relator_eq
apply(subst (asm) (2 4) eq_alt)
apply(unfold prod.rel_Grp option.rel_Grp pmf.rel_Grp apfst_def gpv.rel_Grp)
apply(simp add: rel_fun_def Grp_def)
done

lemma exec_gpv_conv_exec_gpv':
  "exec_gpv callee gpv s = 
   bind_spmf (exec_gpv' callee gpv s)
     (\<lambda>(xo, s'). case xo of None \<Rightarrow> return_pmf None | Some x \<Rightarrow> return_spmf (x, s'))"
  (is "?lhs = ?rhs")
proof(rule spmf.leq_antisym)
  show "ord_spmf op = ?lhs ?rhs"
  proof(induction arbitrary: gpv s rule: exec_gpv_fixp_induct)
    case adm show ?case by simp
    case bottom show ?case by simp
    case (step exec_gpv')
    show ?case
      apply(rewrite exec_gpv'.simps)
      apply(rewrite bind_spmf_def)
      apply(auto simp add: bind_spmf_pmf_assoc intro!: rel_pmf_bind_reflI ord_spmf_bind_reflI step.IH split: option.split generat.split)
      done
  qed
  show "ord_spmf op = ?rhs ?lhs"
  proof(induction arbitrary: gpv s rule: exec_gpv'_fixp_induct)
    case adm show ?case by simp
    case bottom show ?case by simp
    case (step exec_gpv'')
    show ?case
      apply(rewrite exec_gpv.simps)
      apply(rewrite in "ord_spmf _ _ \<hole>" bind_spmf_def)
      apply(auto simp add: bind_spmf_pmf_assoc intro!: rel_pmf_bind_reflI ord_spmf_bind_reflI step.IH split: option.split generat.split)
      done
  qed
qed

primrec list_of_option :: "'a option \<Rightarrow> 'a list"
where 
  "list_of_option None = []"
| "list_of_option (Some x) = [x]"
  
lemma case_list_of_option [simp]:
  "case_list nil cons (list_of_option x) = case_option nil (\<lambda>x. cons x []) x"
  by(cases x) simp_all

definition cr_list_option :: "'a list \<Rightarrow> 'a option \<Rightarrow> bool"
where "cr_list_option xs y \<longleftrightarrow> xs = list_of_option y"

lemma cr_list_optionE [cases pred]:
  assumes "cr_list_option xs y"
  obtains (None) "xs = []" "y = None"
  | (Some) x where "xs = [x]" "y = Some x"
using assms by(cases y)(auto simp add: cr_list_option_def)

lemma cr_list_optionI [simp, intro!]:
  "cr_list_option [] None"
  "cr_list_option [x] (Some x)"
by(simp_all add: cr_list_option_def)

lemma cr_list_option_transfer:
  includes lifting_syntax shows
  "cr_list_option [] None"
  "(op = ===> cr_list_option) (\<lambda>x. [x]) Some"
  "(op = ===> (op = ===> op =) ===> cr_list_option ===> op =) (\<lambda>nil cons. case_list nil (\<lambda>x _. cons x)) case_option"
by(auto simp add: rel_fun_def elim: cr_list_optionE)

lemma exec_gpv_try:
  defines "exec_gpv1 \<equiv> exec_gpv" and "exec_gpv2 \<equiv> exec_gpv"
  shows "exec_gpv1 callee (try_gpv gpv gpv') s = 
   bind_spmf (exec_gpv' callee gpv s) (\<lambda>(xo, s'). case xo of None \<Rightarrow> exec_gpv2 callee gpv' s' | Some x \<Rightarrow> return_spmf (x, s'))"
  (is "?lhs = ?rhs")
proof(rule spmf.leq_antisym)
  show "ord_spmf op = ?lhs ?rhs" unfolding exec_gpv1_def
  proof(induction arbitrary: gpv s rule: exec_gpv_fixp_induct_strong)
    case adm show ?case by simp
    case bottom show ?case by simp
    case (step exec_gpv'')
    show ?case
      apply(rewrite exec_gpv'.simps)
      apply(clarsimp simp add: exec_gpv2_def bind_spmf_pmf_assoc bind_map_spmf try_spmf_def bind_map_pmf o_def intro!: rel_pmf_bind_reflI split: option.split generat.split)
      apply(rule conjI; clarsimp)
      subgoal by(subst exec_gpv.simps)(clarsimp intro!: ord_spmf_bind_reflI step.hyps split: generat.split)
      subgoal by(clarsimp intro!: ord_spmf_bind_reflI step.IH)
      done
  qed
  show "ord_spmf op = ?rhs ?lhs" unfolding exec_gpv2_def exec_gpv1_def
  proof(induction arbitrary: gpv s rule: exec_gpv'_fixp_induct)
    case adm show ?case by simp
    case bottom show ?case by simp
    case (step exec_gpv'')
    show ?case
      apply(rewrite in "ord_spmf _ _ \<hole>" exec_gpv.simps)
      apply(clarsimp simp add: bind_map_spmf bind_spmf_pmf_assoc try_spmf_def bind_map_pmf o_def intro!: rel_pmf_bind_reflI split: option.split)
      apply(rule conjI; clarsimp)
      subgoal by(subst exec_gpv.simps)(clarsimp intro!: ord_spmf_bind_reflI split: generat.split)
      subgoal by(clarsimp split: generat.split intro!: ord_spmf_bind_reflI step.IH)
      done
  qed
qed

lemma mono2mono_map_spmf [THEN spmf.mono2mono, cont_intro, simp]:
  shows monotone_map_spmf: "monotone (ord_spmf op =) (ord_spmf op =) (map_spmf f)"
by(subst map_spmf_conv_bind_spmf[abs_def])(rule monotone_bind_spmf1)

lemma rel_gpv_tryI:
  "\<lbrakk> rel_gpv A C gpv1 gpv1'; rel_gpv A C gpv2 gpv2' \<rbrakk>
  \<Longrightarrow> rel_gpv A C (TRY gpv1 ELSE gpv2) (TRY gpv1' ELSE gpv2')"
subgoal premises [transfer_rule] by transfer_prover
done








text \<open>We need a search function that looks for the first interaction to ensure guardedness of the corecursion.\<close>

context fixes K :: "'s \<Rightarrow> ('s option, 'call, 'ret) gpv" notes [[function_internals]] begin
partial_function (spmf) iterate_search :: "'s \<Rightarrow> ('s list + 's list \<times> 'call \<times> ('s option, 'call, 'ret) rpv) spmf" 
where
  "iterate_search s =
   bind_spmf (the_gpv (K s)) (\<lambda>generat.
     case generat of Pure None \<Rightarrow> return_spmf (Inl [])
     | Pure (Some s') \<Rightarrow> map_spmf (map_sum (op # s') (apfst (op # s'))) (iterate_search s')
     | IO out c \<Rightarrow> return_spmf (Inr ([], out, c)))"

lemmas iterate_search_fixp_induct[case_names adm bottom step] = iterate_search.fixp_induct
end

lemma iterate_search_parametric [transfer_rule]:
  includes lifting_syntax shows
  "((A ===> rel_gpv (rel_option A) C) ===> A ===> rel_spmf (rel_sum (list_all2 A) (rel_prod (list_all2 A) (rel_prod C (rel_rpv (rel_option A) C)))))
   iterate_search iterate_search"
unfolding iterate_search_def[abs_def]
apply(rule rel_funI)
apply(rule fixp_spmf_parametric[OF iterate_search.mono iterate_search.mono])
subgoal premises [transfer_rule] by transfer_prover
done

(* context fixes K :: "'s \<Rightarrow> ('s, 'call, 'ret) gpv" notes [[function_internals]] begin
partial_function (spmf) iterate_search :: "'s \<Rightarrow> ('s list + 's list \<times> 'call \<times> ('s, 'call, 'ret) rpv) spmf" 
where
  "iterate_search s =
   bind_pmf (the_gpv (K s)) (\<lambda>x.
     case x of None \<Rightarrow> return_spmf (Inl [])
     | Some (Pure s') \<Rightarrow> map_spmf (map_sum (op # s') (apfst (op # s'))) (iterate_search s')
     | Some (IO out c) \<Rightarrow> return_spmf (Inr ([], out, c)))"

lemmas iterate_search_fixp_induct[case_names adm bottom step] = iterate_search.fixp_induct
end

lemma iterate_search_parametric [transfer_rule]:
  includes lifting_syntax shows
  "((A ===> rel_gpv A C) ===> A ===> rel_spmf (rel_sum (list_all2 A) (rel_prod (list_all2 A) (rel_prod C (rel_rpv A C)))))
   iterate_search iterate_search"
unfolding iterate_search_def[abs_def]
apply(rule rel_funI)
apply(rule fixp_spmf_parametric[OF iterate_search.mono iterate_search.mono])
subgoal premises [transfer_rule] by transfer_prover
done
 *)

context notes [[bnf_internals]] begin

friend_of_corec map_gpv :: "('a \<Rightarrow> 'a) \<Rightarrow> ('b \<Rightarrow> 'b) \<Rightarrow> ('a, 'b, 'c) gpv \<Rightarrow> ('a, 'b, 'c) gpv"
where "map_gpv f g gpv = GPV (map_spmf (map_generat f g (op \<circ> (map_gpv f g))) (the_gpv gpv))"
subgoal by(rule gpv.expand; simp add: gpv.map_sel)
subgoal by transfer_prover
done

context notes [friend_of_corec_simps] = map_spmf_bind_spmf bind_map_spmf begin
friend_of_corec bind_gpv :: "('a, 'out, 'in) gpv \<Rightarrow> ('a \<Rightarrow> ('a, 'out, 'in) gpv) \<Rightarrow> ('a, 'out, 'in) gpv" where 
  "bind_gpv r f = GPV (map_spmf (map_generat id id (op \<circ> (case_sum id (\<lambda>r. bind_gpv r f))))
     (the_gpv r \<bind> (case_generat
        (\<lambda>x. map_spmf (map_generat id id (op \<circ> Inl)) (the_gpv (f x)))
        (\<lambda>out c. return_spmf (IO out (\<lambda>input. Inr (c input)))))))"
subgoal by(rule bind_gpv.ctr)
subgoal by transfer_prover
subgoal by transfer_prover
subgoal by transfer_prover
done
end

corec iterate_gpv :: "('s \<Rightarrow> ('s option, 'call, 'ret) gpv) \<Rightarrow> 's \<Rightarrow> ('s list, 'call, 'ret) gpv"
where
  "iterate_gpv K s = GPV (map_spmf (\<lambda>x. case x of Inl ss \<Rightarrow> Pure ss 
    | Inr (ss, out, rpv) \<Rightarrow> IO out (\<lambda>input. bind_gpv (map_gpv list_of_option id (rpv input))
        (\<lambda>ss'. case ss' of [] \<Rightarrow> Done ss
               | s' # _ \<Rightarrow> map_gpv (op @ (ss @ [s'])) id (iterate_gpv K s'))))
    (iterate_search K s))"
  
(* 
corec iterate_gpv :: "('s \<Rightarrow> ('s option, 'call, 'ret) gpv) \<Rightarrow> 's \<Rightarrow> ('s list, 'call, 'ret) gpv"
where
  "iterate_gpv K s = GPV (map_spmf
    (case_generat Pure (\<lambda>(out, ss) f. IO out (\<lambda>input. bind_gpv (f input)
        (\<lambda>ss'. case ss' of [] \<Rightarrow> Done ss
               | s' # _ \<Rightarrow> map_gpv (op @ (ss @ [s'])) id (iterate_gpv K s')))))
    (bind_spmf (iterate_search K s) (\<lambda>result.
      case result of Inl ss \<Rightarrow> return_spmf (Pure ss)
      | Inr (ss, out, rpv) \<Rightarrow> return_spmf (IO (out, ss) (\<lambda>input. try_gpv (map_gpv (\<lambda>x. [x]) id (rpv input)) (Done []))))))"
 *)
end

(* term rel_sig_gpv_v0

term rel_sig_gpv_v2
term rel_pre_gpv
term gpv.v2.corecUU
term gpv.v1.Sig

term rel_pre_gpv

find_consts "_ \<Rightarrow> ('c, 'd, 'e, 'a) ssig_gpv_v1"

context includes lifting_syntax begin

inductive rel_sig_gpv_v1 :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> ('c \<Rightarrow> 'd \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'f \<Rightarrow> bool)
  \<Rightarrow> ('a, 'c, 'g, 'e) sig_gpv_v1 \<Rightarrow> ('b, 'd, 'g, 'f) sig_gpv_v1 \<Rightarrow> bool"
for R S T
where "rel_sum (rel_sig_gpv_v0 R S T) (rel_prod (R ===> R) (rel_prod (S ===> S) T)) x y
  \<Longrightarrow> rel_sig_gpv_v1 R S T (sig_gpv_v1.gpv.v1.Sig x) (sig_gpv_v1.gpv.v1.Sig y)"

lemma gpv_v1_Sig_parametric [transfer_rule]:
  "(rel_sum (rel_sig_gpv_v0 R S T) (rel_prod (R ===> R) (rel_prod (S ===> S) T)) ===> rel_sig_gpv_v1 R S T)
  sig_gpv_v1.gpv.v1.Sig sig_gpv_v1.gpv.v1.Sig"
by(simp add: rel_fun_def rel_sig_gpv_v1.simps)

lemma rel_sig_gpv_v1_eq [relator_eq]: "rel_sig_gpv_v1 op = op = op = = op =" sorry

lemma rel_sig_gpv_v1_map:
  "rel_sig_gpv_v1 R S (\<lambda>x y. T (f x) (g y)) x y
  \<Longrightarrow> rel_sig_gpv_v1 R S T (map_sig_gpv_v1 f x) (map_sig_gpv_v1 g y)"
apply(erule rel_sig_gpv_v1.cases)
apply(hypsubst)
apply(unfold sig_gpv_v1.map)
apply(rule rel_sig_gpv_v1.intros)
apply(unfold sum.rel_map)
apply(erule sum.rel_mono_strong)
 apply(unfold sig_gpv_v0.rel_map id_apply)
 apply assumption
apply(unfold prod.rel_map)
apply assumption
done

lemma map_sig_gpv_v1_parametric [transfer_rule]:
  "((A ===> B) ===> rel_sig_gpv_v1 C D A ===> rel_sig_gpv_v1 C D B) map_sig_gpv_v1 map_sig_gpv_v1" sorry

term map_sig_gpv_v1

lemma rel_sig_gpv_v1_mono [mono]:
  "(\<And>x y. T x y \<longrightarrow> T' x y) \<Longrightarrow> rel_sig_gpv_v1 R S T x y \<longrightarrow> rel_sig_gpv_v1 R S T' x y"
by(auto simp add: rel_sig_gpv_v1.simps elim!: sum.rel_mono_strong sig_gpv_v0.rel_mono_strong)

inductive rel_ssig_gpv_v1 :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> ('c \<Rightarrow> 'd \<Rightarrow> bool) \<Rightarrow> ('e \<Rightarrow> 'f \<Rightarrow> bool)
  \<Rightarrow> ('a, 'c, 'g, 'e) ssig_gpv_v1 \<Rightarrow> ('b, 'd, 'g, 'f) ssig_gpv_v1 \<Rightarrow> bool"
for R S T
where Oper: "rel_sig_gpv_v1 R S (rel_ssig_gpv_v1 R S T) x y \<Longrightarrow> rel_ssig_gpv_v1 R S T (gpv.v1.Oper x) (gpv.v1.Oper y)"
| CLeaf: "rel_gpv R S x y \<Longrightarrow> rel_ssig_gpv_v1 R S T (gpv.v1.CLeaf x) (gpv.v1.CLeaf y)"
| VLeaf: "T x y \<Longrightarrow> rel_ssig_gpv_v1 R S T (gpv.v1.VLeaf x) (gpv.v1.VLeaf y)"

lemma [transfer_rule]:
  shows gpv_v1_Oper_parametric: "(rel_sig_gpv_v1 R S (rel_ssig_gpv_v1 R S T) ===> rel_ssig_gpv_v1 R S T) gpv.v1.Oper gpv.v1.Oper"
  and gpv_v1_CLeaf_parametric: "(rel_gpv R S ===> rel_ssig_gpv_v1 R S T) gpv.v1.CLeaf gpv.v1.CLeaf"
  and gpv_v1_VLeaf_parametric: "(T ===> rel_ssig_gpv_v1 R S T) gpv.v1.VLeaf gpv.v1.VLeaf"
by(rule rel_funI; erule rel_ssig_gpv_v1.intros)+

thm gpv.v1.Lam_def

lemma rel_ssig_gpv_v1_mono [mono]:
  "(\<And>x y. T x y \<longrightarrow> T' x y) \<Longrightarrow> rel_ssig_gpv_v1 R S T x y \<longrightarrow> rel_ssig_gpv_v1 R S T' x y"
apply(rule impI)
apply(erule rel_ssig_gpv_v1.induct)
apply(rule rel_ssig_gpv_v1.intros; blast intro: rel_sig_gpv_v1_mono[rule_format, rotated])+
done

lemma rel_ssig_gpv_v1_eq [relator_eq]: "rel_ssig_gpv_v1 op = op = op = = op =" sorry

declare convol_transfer[transfer_rule]

lemma map_ssig_gpv_v1_transfer [transfer_rule]:
  "((A ===> B) ===> rel_ssig_gpv_v1 R S A ===> rel_ssig_gpv_v1 R S B)
  map_ssig_gpv_v1 map_ssig_gpv_v1"
apply(rule rel_funI)
subgoal premises prems for f g
  apply(rule rel_funI)
  apply(erule rel_ssig_gpv_v1.induct)
    apply(unfold ssig_gpv_v1.map)
    apply(rule rel_ssig_gpv_v1.intros)
    apply(rule rel_sig_gpv_v1_mono[rule_format, rotated])
     apply(rule rel_sig_gpv_v1_map)
     apply(erule rel_sig_gpv_v1_mono[rule_format, rotated])
     apply(erule conjunct2)
    apply assumption
   apply(erule rel_ssig_gpv_v1.intros)
  apply(rule rel_ssig_gpv_v1.intros)
  apply(erule prems[THEN rel_funD])
  done
done

lemma rec_ssig_gpv_v1_parametric [transfer_rule]:
  "((rel_sig_gpv_v1 A C (rel_prod (rel_ssig_gpv_v1 A C S) E) ===> E) ===> (S ===> E) ===> (rel_gpv A C ===> E) ===> rel_ssig_gpv_v1 A C S ===> E)
  rec_ssig_gpv_v1 rec_ssig_gpv_v1"
apply(rule rel_funI)+
apply(erule rel_ssig_gpv_v1.induct)
  apply(unfold ssig_gpv_v1.rec)
  apply(erule rel_funD)
  apply(rule rel_sig_gpv_v1_map)
  apply(erule rel_sig_gpv_v1_mono[rule_format, rotated])
  apply clarsimp
 apply(erule (1) rel_funD)
apply(erule (1) rel_funD)
done


term gpv.v1.flat
thm gpv.v1.flat_def

term gpv.v1.eval_core
thm gpv.v1.eval_core_def
thm gpv.v1.unsig_def
thm map_gpv.rho_def
find_theorems map_sig_gpv_v0 rel_fun

term case_sig_gpv_v1

lemma case_sig_gpv_v1_parametric [transfer_rule]:
  "((rel_sum (rel_sig_gpv_v0 R S T) (rel_prod (R ===> R) (rel_prod (S ===> S) T)) ===> X) ===> rel_sig_gpv_v1 R S T ===> X)
  case_sig_gpv_v1 case_sig_gpv_v1"
sorry

lemmas [transfer_rule] = pre_gpv.map_transfer gpv.v0.dtor_transfer sig_gpv_v0.map_transfer gpv.dtor_unfold_transfer

lemma gpv_v1_cutSsig_parametric [transfer_rule]:
  "((S ===> rel_pre_gpv A C (rel_ssig_gpv_v1 A C S)) ===> rel_ssig_gpv_v1 A C S ===> rel_pre_gpv A C (rel_ssig_gpv_v1 A C S))
  gpv.v1.cutSsig gpv.v1.cutSsig"
unfolding gpv.v1.cutSsig_def gpv.v1.flat_def gpv.v1.eval_core_def gpv.v1.Lam_def gpv.v1.unsig_def map_gpv.rho_def gpv.v0.Lam_def gpv.v1.embL_def
apply(rule rel_funI)
subgoal premises [transfer_rule] 
apply transfer_prover_start
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
prefer 59
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_prover_end
done
done

lemma gpv_v1_corecU_parametric [transfer_rule]:
  "((S ===> rel_pre_gpv A C (rel_ssig_gpv_v1 A C S)) ===> S ===> rel_gpv A C) gpv.v1.corecU gpv.v1.corecU"
unfolding gpv.v1.corecU_def
apply transfer_prover_start
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_prover_end
done

lemma gpv_v1_corecUU_parametric [transfer_rule]:
  "((S ===> rel_ssig_gpv_v1 A C (rel_pre_gpv A C (rel_ssig_gpv_v1 A C S))) ===> S ===> rel_gpv A C) gpv.v1.corecUU gpv.v1.corecUU"
unfolding gpv.v1.corecUU_def gpv.v1.eval_core_def gpv.v1.sctr_def gpv.v1.proto_sctr_def gpv.v1.Lam_def gpv.v1.unsig_def gpv.v0.Lam_def gpv.v1.embL_def map_gpv.rho_def gpv.v1.flat_def
apply transfer_prover_start
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_prover_end
done

term gpv.v1.corecUU

find_theorems gpv.v1.corecU
thm gpv.v1.corecU_def
find_theorems dtor_unfold_gpv


(* lemma gpv_corecUU2_parametric [transfer_rule]:
  includes lifting_syntax shows
  "((A ===> rel_ssig_gpv_v2 (rel_pre_gpv
 *)

lemma iterate_gpv_parametric [transfer_rule]:
  includes lifting_syntax shows
  "((S ===> rel_gpv S C) ===> S ===> rel_gpv (list_all2 S) C) iterate_gpv iterate_gpv"
unfolding iterate_gpv_def
apply transfer_prover_start
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
apply transfer_step
sorry
 *)

lemma iterate_gpv_unfold:
  "iterate_gpv K s = GPV (bind_spmf (iterate_search K s) (\<lambda>result.
    case result of Inl ss \<Rightarrow> return_spmf (Pure ss)
    | Inr (ss, out, rpv) \<Rightarrow> return_spmf (IO out (\<lambda>input.
      bind_gpv (rpv input)
        (\<lambda>ss'. case ss' of None \<Rightarrow> Done ss
               | Some s' \<Rightarrow> map_gpv (op @ (ss @ [s'])) id (iterate_gpv K s'))))))"
  apply(subst iterate_gpv.code)
  apply(clarsimp simp: map_spmf_bind_spmf split_def map_spmf_conv_bind_spmf fun_eq_iff bind_map_gpv split: sum.splits intro!: bind_spmf_cong[OF refl] bind_gpv_cong[OF refl])
  done

(* lemma iterate_gpv_sel_TRY:
  "the_gpv (iterate_gpv K s) = bind_spmf (iterate_search K s) (\<lambda>result.
    case result of Inl ss \<Rightarrow> return_spmf (Pure ss)
    | Inr (ss, out, rpv) \<Rightarrow> return_spmf (IO out (\<lambda>input.
      bind_gpv (TRY map_gpv (\<lambda>x. [x]) id (rpv input) ELSE Done [])
        (\<lambda>ss'. case ss' of [] \<Rightarrow> Done ss
               | s' # _ \<Rightarrow> map_gpv (op @ (ss @ [s'])) id (iterate_gpv K s')))))"
by(subst iterate_gpv.code)(clarsimp simp: map_spmf_bind_spmf bind_gpv'_def split: sum.splits intro!: bind_spmf_cong[OF refl])
 *)

lemma iterate_gpv_sel:
  "the_gpv (iterate_gpv K s) = bind_spmf (iterate_search K s) (\<lambda>result.
    case result of Inl ss \<Rightarrow> return_spmf (Pure ss)
    | Inr (ss, out, rpv) \<Rightarrow> return_spmf (IO out (\<lambda>input.
      bind_gpv (rpv input)
        (\<lambda>ss'. case ss' of None \<Rightarrow> Done ss
               | Some s' \<Rightarrow> map_gpv (op @ (ss @ [s'])) id (iterate_gpv K s')))))"
by(subst iterate_gpv_unfold) simp

text \<open>This is the specification we have in mind, but it does not have a unique solution.
  For @{term "K = Done \<circ> Some"}, both sides reduce to the vacuous @{term "iterate_gpv K s = iterate_gpv K s"}.
  In fact, we pick the "least".\<close>
lemma iterate_gpv_unroll:
  "iterate_gpv K s = 
   bind_gpv (K s) (\<lambda>x. case x of None \<Rightarrow> Done [] | Some s' \<Rightarrow> map_gpv (op # s') id (iterate_gpv K s'))"
  apply(rule gpv.expand)
  apply(rewrite in "\<hole> = _" iterate_gpv_sel)
  apply(rewrite in "\<hole> = _" iterate_search.simps)
  apply(clarsimp simp add: map_spmf_bind_spmf bind_gpv_sel list.case_distrib[where h="the_gpv"] gpv.map_sel simp del: bind_gpv_sel' cong del: generat.case_cong_weak option.case_cong_weak intro!: bind_spmf_cong[OF refl] split: generat.split option.split cong: gpv.map_cong)
  apply(rewrite in "_ = \<hole>" iterate_gpv_sel)
  apply(clarsimp simp add: map_spmf_conv_bind_spmf map_gpv_bind_gpv id_def[symmetric] gpv.map_id gpv.map_comp o_def intro!: bind_spmf_cong[OF refl] ext bind_gpv_cong[OF refl] split: sum.split option.split)
  done

lemma iterate_search_Done [simp]: "iterate_search (Done \<circ> Some) s = return_pmf None"
by(induction arbitrary: s rule: iterate_search_fixp_induct)(simp_all add: bind_return_pmf)

lemma iterate_gpv_Done [simp]: "iterate_gpv (Done \<circ> Some) s = Fail"
by(rule gpv.expand)(simp add: iterate_gpv_sel)

lemma (in ccpo)
  assumes mono: "monotone op \<le> op \<le> f"
  and *: "\<And>x. \<lbrakk> x \<in> Complete_Partial_Order.iterates f; \<And>y. \<lbrakk> y \<le> x; y \<in> Complete_Partial_Order.iterates f \<rbrakk> \<Longrightarrow> P y \<rbrakk> \<Longrightarrow> P (f x)"
  shows "P (ccpo_class.fixp f)"
proof -
  def y \<equiv> "ccpo_class.fixp f"
  have "y \<le> ccpo_class.fixp f" "y \<in> Complete_Partial_Order.iterates f" unfolding y_def
    by(simp_all add: iterates_fixp[OF mono])
  with iterates_fixp[OF mono] show "P y"
  proof(induction arbitrary: y)
    case (step x)
    show ?case
 (*       apply(rule * )
      subgoal using step.prems by simp
      subgoal apply(rule step.IH)
 *)    

thm ccpo.iterates_fixp
  oops
  

lemma exec_gpv_iterate:
  fixes exec_gpv1 exec_gpv2 and K :: "'s \<Rightarrow> ('s option, 'out, 'ret) gpv" and \<sigma> :: "'\<sigma>"
  defines "exec_gpv1 \<equiv> exec_gpv" and "exec_gpv2 \<equiv> exec_gpv"
  shows "exec_gpv1 callee (iterate_gpv K s) \<sigma> =
    bind_spmf (iterate_spmf (\<lambda>(so, \<sigma>). case so of None \<Rightarrow> return_spmf None | Some s \<Rightarrow> map_spmf Some (exec_gpv2 callee (K s) \<sigma>)) (Some s, \<sigma>))
      (\<lambda>ss. bind_spmf (assert_spmf (ss \<noteq> [] \<and> Option.is_none (fst (last ss)))) (\<lambda>_. return_spmf (map (the \<circ> fst) (butlast ss), last (map snd ss))))"
  (is "?lhs = ?rhs" is "_ = bind_spmf (iterate_spmf ?iter _) ?f")
proof(rule spmf.leq_antisym)
  note [[show_variants]]
  note [simp del] = last.simps
    and [simp] = bind_return_pmf
  
  have assert: "ss \<noteq> [] \<and> Option.is_none (fst (last ss))" if "ss \<in> set_spmf (iterate_spmf ?iter (Some s, \<sigma>))" for ss s \<sigma>
  proof -
    have "set_spmf (iterate_spmf ?iter (Some s, \<sigma>)) \<subseteq> {ss. ss \<noteq> [] \<and> Option.is_none (fst (last ss))}"
      and "ord_spmf op = (iterate_spmf ?iter (None, \<sigma>)) (return_spmf [])"
    proof(induction arbitrary: s \<sigma> rule: iterate_spmf_fixp_induct)
      case adm show ?case by simp
      case bottom case 1 show ?case by simp
      case bottom case 2 show ?case by simp
      case (step iterate_spmf') case 1
      show ?case
        apply clarsimp
        subgoal for so \<sigma>' ss'
          apply(cases so)
          subgoal using step.IH(2)[of \<sigma>', THEN monotoneD[OF monotone_set_spmf]] by(auto simp add: last.simps)
          subgoal for s' using step.IH(1)[of s' \<sigma>'] by(auto simp add: last.simps)
          done
        done
      case step case 2 show ?case by simp
    qed
    then show ?thesis using that by auto
  qed

  def exec_gpv3 \<equiv> "exec_gpv :: ('\<sigma> \<Rightarrow> 'out \<Rightarrow> ('ret \<times> '\<sigma>) spmf) \<Rightarrow> ('s list, 'out, 'ret) gpv \<Rightarrow> _"
  have "ord_spmf op = ?lhs ?rhs"
    and "exec_gpv1 callee (map_gpv g id gpv) \<sigma> = map_spmf (apfst g) (exec_gpv1 callee gpv \<sigma>)"
    -- \<open>Stupid! We again have to do yet another induction to show that @{const exec_gpv}
        commutes with bind and map! And we must use @{typ "'s list \<Rightarrow> 's list"} instead of
        @{typ "'s option => 's list"} here, because otherwise we'd have different types for @{const exec_gpv}. :-(\<close>
    and "ord_spmf op = (exec_gpv1 callee (bind_gpv gpv f) \<sigma>)
         (bind_spmf (exec_gpv3 callee gpv \<sigma>) (\<lambda>(x, s') \<Rightarrow> exec_gpv callee (f x) s'))"
    for gpv :: "('s list, 'out, 'ret) gpv" and f and g :: "'s list \<Rightarrow> 's list"
   unfolding exec_gpv1_def
  proof(induction arbitrary: s \<sigma> gpv f g rule: exec_gpv_fixp_induct_strong2)
    case adm show ?case
      apply(rule cont_intro)+
      subgoal for a b c d by(cases d; clarsimp)
      done
    case bottom case 1 show ?case by simp
    case bottom case 2 show ?case by simp
    case bottom case 3 show ?case by simp
  next
    case exec_gpv: (step exec_gpv')
    case 1
    have search: "ord_spmf op =
      (bind_spmf (iterate_search K s) (\<lambda>x. case x of Inl ss \<Rightarrow> return_spmf (ss, \<sigma>)
        | Inr (ss, out, rpv) \<Rightarrow> bind_spmf (callee \<sigma> out) (\<lambda>(input, \<sigma>').
            exec_gpv' (bind_gpv (rpv input) (\<lambda>so. case so of None \<Rightarrow> Done ss | Some s' \<Rightarrow> map_gpv (\<lambda>ss'. ss @ [s'] @ ss') id (iterate_gpv K s'))) \<sigma>')))
      (bind_spmf (iterate_spmf ?iter (Some s, \<sigma>)) ?f)"
      (is "ord_spmf _ (?left s) (?right s)")
    proof(induction arbitrary: s rule: iterate_search_fixp_induct)
      case adm show ?case by simp
      case bottom show ?case by simp
    next
      case search: (step iterate_search')
      show ?case unfolding exec_gpv2_def
        apply(simp add: bind_spmf_pmf_assoc)
        apply(subst iterate_spmf.simps)
        apply simp
        apply(subst exec_gpv.simps)
        apply(simp add: bind_map_spmf)
        apply(rule ord_spmf_bind_reflI)
        apply(clarsimp split: generat.split)
        apply(intro conjI strip)
         apply(clarsimp split: option.split)
         apply(intro conjI strip)
          apply(subst iterate_spmf.simps)
          apply(simp add: last.simps)
         apply(clarsimp simp add: bind_map_spmf o_def case_sum_map_sum)
        subgoal for s' using search.IH[of s', THEN monotoneD[OF monotone_map_spmf], where ?f1="apfst (op # s')"]
          apply(simp add: map_spmf_bind_spmf o_def sum.case_distrib[where h="map_spmf _"] split_def cong del: sum.case_cong)
          apply(subst (asm) exec_gpv.IH(2)[symmetric])
          apply(simp add: map_gpv_id_bind_gpv o_def option.case_distrib[where h="map_gpv _ _"] gpv.map_comp id_def[symmetric] exec_gpv2_def cong del: option.case_cong)
          apply(erule spmf.leq_trans)
          apply(rule ord_spmf_bind_reflI)
          apply(drule assert[simplified exec_gpv2_def split_def])
          apply(auto simp add: last.simps)
          done
        apply(rule ord_spmf_bind_reflI)
        apply clarsimp
        subgoal for out rpv input \<sigma>'
          using exec_gpv.IH(3)[of "map_gpv list_of_option id (rpv input)" "\<lambda>xs. case xs of [] \<Rightarrow> Done [] | s' # _ \<Rightarrow> map_gpv (op # s') id (iterate_gpv K s')" \<sigma>']
          apply(simp add: bind_map_gpv o_def)
          apply(erule spmf.leq_trans)
          apply(simp add: exec_gpv3_def exec_gpv_map_gpv_id bind_map_spmf)
          apply(rule ord_spmf_bind_reflI)
          apply(clarsimp split!: option.split)
           using exec_gpv.hyps(2)
           apply(rule spmf.leq_trans)
           apply(subst iterate_spmf.simps)
           apply(clarsimp simp add: o_def bind_eq_return_spmf last.simps)
          apply(subst exec_gpv.IH(2))
          apply(simp add: map_spmf_conv_bind_spmf) 
          apply(rule spmf.leq_trans)
           apply(rule ord_spmf_bindI)
            apply(rule exec_gpv.IH(1))
           apply simp
          apply(simp add: exec_gpv2_def o_def cong del: option.case_cong)
          apply(clarsimp intro!: ord_spmf_bind_reflI simp add: last.simps)
          done
        done
    qed
    show ?case unfolding exec_gpv2_def
      apply(subst iterate_gpv_sel)
      apply(subst bind_spmf_assoc)
      apply(subst prod.case_distrib[where h=return_spmf, symmetric, abs_def])+
      apply(subst sum.case_distrib[where h=return_spmf, symmetric])
      apply simp
      apply(subst sum.case_distrib[where h="case_generat _ _"])
      apply(simp cong del: sum.case_cong)
      apply(subst prod.case_distrib[where h="case_generat _ _"])+
      apply(simp cong del: sum.case_cong)
      apply(subst append_assoc[abs_def])
      using search
      apply(simp cong del: list.case_cong option.case_cong if_weak_cong sum.case_cong add: exec_gpv2_def)
      done
  next
    case step case 3 show ?case unfolding exec_gpv3_def
      apply(rewrite exec_gpv.simps)
      apply(simp del: bind_gpv_sel' add: bind_gpv_sel)
      apply(rule ord_spmf_bind_reflI)
      apply(clarsimp split: generat.split)
      apply(rule ord_spmf_bind_reflI)
      apply clarsimp
      apply(rule spmf.leq_trans)
       apply(rule step.IH)
      apply(simp add: exec_gpv3_def)
      apply(rule ord_spmf_bind_reflI)
      apply clarsimp
      apply(rule step.hyps)
      done
  next
    case step case 2 show ?case
      apply(simp add: gpv.map_sel map_spmf_bind_spmf bind_map_spmf o_def del: id_apply)
      apply(rule bind_spmf_cong[OF refl])
      apply(clarsimp split: generat.split simp add: map_spmf_bind_spmf o_def)
      apply(rule bind_spmf_cong[OF refl])
      apply clarsimp
      apply(rule step.IH)
      done
  qed
  then show "ord_spmf op = ?lhs ?rhs" by -

  show "ord_spmf op = ?rhs ?lhs" unfolding exec_gpv1_def exec_gpv2_def
  proof(induction arbitrary: s \<sigma> rule: iterate_spmf_fixp_induct)
    case adm show ?case by simp
    case bottom show ?case by simp
    case iterate_spmf: (step iterate_spmf')

    show ?case
      apply(simp add: bind_map_spmf)
      apply(subst iterate_gpv_unroll)
      apply(simp add: exec_gpv_bind)
      apply(rule ord_spmf_bind_reflI)
      apply(clarsimp simp add: o_def)
      subgoal for so \<sigma>'
        apply(cases so)
        subgoal
          apply(rule spmf.leq_trans)
           apply(use iterate_spmf.hyps(2) in \<open>rule ord_spmf_bindI\<close>)
           apply(hypsubst)
           apply(rule spmf.leq_refl)
          apply(simp add: last.simps)
          done
        subgoal for s'
          apply(simp add: exec_gpv_map)
          using iterate_spmf.IH(1)[of s' \<sigma>', THEN monotoneD[OF monotone_map_spmf], where ?f1="apfst (op # s')"]
          apply(rule spmf.leq_trans[rotated])
          apply(simp add: map_spmf_bind_spmf o_def)
          apply(rule ord_spmf_bind_reflI)
          apply(drule iterate_spmf.hyps(1)[THEN monotoneD[OF monotone_set_spmf], of "(Some s', \<sigma>')", THEN subsetD])
          apply(drule assert[simplified exec_gpv2_def])
          apply(clarsimp simp add: last.simps)
          done
        done
      done
  qed
qed

(* lemma exec_gpv_iterate:
  fixes exec_gpv1 and K :: "'s \<Rightarrow> ('s, 'out, 'ret) gpv" and \<sigma> :: "'\<sigma>"
  defines "exec_gpv1 \<equiv> exec_gpv"
  shows "exec_gpv1 callee (iterate_gpv K s) \<sigma> =
   bind_spmf (iterate_spmf (\<lambda>(so, \<sigma>). case so of None \<Rightarrow> return_pmf None | Some s \<Rightarrow> exec_gpv' callee (K s) \<sigma>) (Some s, \<sigma>))
     (\<lambda>ss. if ss \<noteq> [] \<and> Option.is_none (fst (last ss)) then return_spmf (map (the \<circ> fst) (butlast ss), last (map snd ss)) else return_pmf None)"
  (is "?lhs = ?rhs" is "_ = bind_spmf (iterate_spmf ?iter _) ?f")
proof(rule spmf.leq_antisym)
  note [[show_variants]]
  note [simp del] = last.simps
    and [simp] = bind_return_pmf

  def exec_gpv3 \<equiv> "exec_gpv :: ('\<sigma> \<Rightarrow> 'out \<Rightarrow> ('ret \<times> '\<sigma>) spmf) \<Rightarrow> ('s list, 'out, 'ret) gpv \<Rightarrow> _"
  have "ord_spmf op = ?lhs ?rhs"
    -- \<open>Stupid! We again have to do yet another induction to show that @{const exec_gpv}
        commutes with bind and map! And we cannot use @{const bind_gpv'} here, because then we'd
        have different types for @{const exec_gpv}. :-(\<close>
    and "ord_spmf op = (exec_gpv1 callee (bind_gpv gpv f) \<sigma>)
         (bind_spmf (exec_gpv3 callee gpv \<sigma>) (\<lambda>(x, s') \<Rightarrow> exec_gpv callee (f x) s'))"
    and "exec_gpv callee (map_gpv g id gpv) \<sigma> = map_spmf (apfst g) (exec_gpv callee gpv \<sigma>)"
    for gpv :: "('s list, 'out, 'ret) gpv" and f and g :: "'s list \<Rightarrow> 's list"
    unfolding exec_gpv1_def
  proof(induction arbitrary: s \<sigma> gpv f g rule: exec_gpv_fixp_induct_strong2)
    case adm show ?case
      apply(rule cont_intro)+
      subgoal for a b c d by(cases d; clarsimp)
      apply simp
      done
  next
    case bottom 
    { case 1 show ?case by simp }
    { case 2 show ?case by simp }
    { case 3 show ?case by simp }
  next
    case exec_gpv: (step exec_gpv'')
    case 1
    have search: "ord_spmf op =
      (bind_spmf (iterate_search K s) (\<lambda>x. case x of Inl ss \<Rightarrow> return_spmf (ss, \<sigma>)
         | Inr (ss, out, rpv) \<Rightarrow> bind_spmf (callee \<sigma> out)
         (\<lambda>(input, \<sigma>'). exec_gpv''
            (bind_gpv (TRY map_gpv (\<lambda>x. [x]) id (rpv input) ELSE Done [])
              (\<lambda>ss'. case ss' of [] \<Rightarrow> Done ss 
                     | s' # _ \<Rightarrow> map_gpv (\<lambda>ss'. ss @ [s'] @ ss') id (iterate_gpv K s')))
            \<sigma>')))
      (bind_spmf (iterate_spmf ?iter (Some s, \<sigma>)) ?f)"
      (is "ord_spmf _ (?left s) (?right s)")
      for s
    proof(induction arbitrary: s rule: iterate_search_fixp_induct)
      case adm show ?case by simp
    next
      case bottom show ?case by simp
    next
      case search: (step iterate_search')

      have "try": "rel_spmf (rel_prod cr_list_option op =)
        (exec_gpv callee (TRY map_gpv (\<lambda>x. [x]) id gpv ELSE Done []) \<sigma>)
        (exec_gpv' callee gpv \<sigma>)" (is "rel_spmf ?R ?lhs ?rhs")
        for gpv \<sigma>
      proof -
        have "rel_spmf ?R ?lhs (bind_spmf (exec_gpv' callee gpv \<sigma>) return_spmf)"
          by(auto simp add: exec_gpv_try exec_gpv'_map bind_map_spmf simp del: bind_return_spmf intro!: rel_spmf_bind_reflI split: option.split)
        thus ?thesis by simp
      qed

      show ?case
        apply(simp add: bind_spmf_pmf_assoc)
        apply(subst iterate_spmf.simps)
        apply simp
        apply(subst exec_gpv'.simps)
        apply(simp add: bind_spmf_pmf_assoc bind_assoc_pmf)
        apply(rule rel_pmf_bind_reflI)
        apply(case_tac x)
         apply(simp)
         apply(subst iterate_spmf.simps)
         apply(simp add: last.simps)
        apply(clarsimp simp add: in_set_spmf[symmetric])
        subgoal for generat
          apply(cases generat)
          subgoal for s'
            apply clarsimp
            apply(simp add: bind_map_spmf spmf.map_comp o_def case_sum_map_sum cong del: sum.case_cong option.case_cong)
            apply(rule spmf.leq_trans)
             using search.IH[of s', THEN monotoneD[OF monotone_map_spmf], where ?f1="apfst (op # s')"]
             apply(simp cong del: sum.case_cong list.case_cong add: id_def o_def map_spmf_bind_spmf)
             apply(subst (asm) sum.case_distrib[where h="map_spmf _"])
             apply(simp cong del: sum.case_cong)
             apply(subst (asm) prod.case_distrib[where h="map_spmf _"])+
             apply(simp add: map_spmf_bind_spmf o_def cong del: sum.case_cong)
             apply(subst (asm) prod.case_distrib[where h="map_spmf _"])
             apply(subst (asm) exec_gpv.IH(3)[symmetric])
             apply(simp add: map_gpv_bind_gpv gpv.map_id cong del: sum.case_cong)
             apply(subst (asm) list.case_distrib[where h="map_gpv _ _"])
             apply(simp add: gpv.map_comp o_def cong del: sum.case_cong list.case_cong)
            apply(simp add: split_def last.simps)
            apply(rule ord_spmf_bind_reflI)
            apply(simp add: last.simps)
            done
          apply(simp add:)
          apply(subst bind_spmf_def)
          apply(subst bind_spmf_def)
          apply(subst bind_assoc_pmf)
          apply(rule rel_pmf_bind_reflI)
          apply(case_tac x)
           apply clarsimp
          apply(clarsimp simp add: in_set_spmf[symmetric])
          apply(rule spmf.leq_trans)
           apply(rule exec_gpv.IH)
          apply(subst bind_spmf_def)
          subgoal for out rpv input \<sigma>'
          apply(rule rel_pmf_bindI[where R="rel_option (rel_prod cr_list_option op =)"])
           subgoal using "try" by(simp add: exec_gpv3_def)

          apply(case_tac x)
           apply simp
          apply clarsimp
          apply(case_tac y)
           apply simp
          apply clarsimp
          apply(erule cr_list_optionE)
           apply clarsimp
           apply(rule spmf.leq_trans)
            apply(rule exec_gpv.hyps)
           apply(subst iterate_spmf.simps)
           apply(simp add: last.simps)
          apply clarsimp
          apply(subst exec_gpv.IH)
          apply(simp add: bind_map_spmf)
          apply(simp add: spmf.map_comp o_def)
          subgoal for \<sigma>'' s''
            using exec_gpv.IH(1)[of s'' \<sigma>'', THEN monotoneD[OF monotone_map_spmf], where ?f1="apfst (op # s'')"]
            apply(rule spmf.leq_trans)
            apply(simp add: map_spmf_bind_spmf)
            apply(rule ord_spmf_bind_reflI)
            apply(simp add: last.simps)
            done
          done
        done
      done
    qed

    show ?case
      apply(subst iterate_gpv_sel_TRY)
      apply(subst bind_spmf_assoc)
      apply(subst prod.case_distrib[where h=return_spmf, symmetric, abs_def])+
      apply(subst sum.case_distrib[where h=return_spmf, symmetric])
      apply simp
      apply(subst sum.case_distrib[where h="case_generat _ _"])
      apply(simp cong del: sum.case_cong)
      apply(subst prod.case_distrib[where h="case_generat _ _"])+
      apply(simp cong del: sum.case_cong)
      apply(subst append_assoc[abs_def])
      using search
      apply(simp cong del: list.case_cong if_weak_cong sum.case_cong)
      done

    case 2 show ?case unfolding exec_gpv3_def
      apply(rewrite exec_gpv.simps)
      apply(simp del: bind_gpv_sel' add: bind_gpv_sel)
      apply(rule ord_spmf_bind_reflI)
      apply(clarsimp split: generat.split)
      apply(rule ord_spmf_bind_reflI)
      apply clarsimp
      apply(rule spmf.leq_trans)
       apply(rule exec_gpv.IH)
      apply(simp add: exec_gpv3_def)
      apply(rule ord_spmf_bind_reflI)
      apply clarsimp
      apply(rule exec_gpv.hyps)
      done

    case 3 show ?case
      apply(simp add: gpv.map_sel map_spmf_bind_spmf bind_map_spmf o_def del: id_apply)
      apply(rule bind_spmf_cong[OF refl])
      apply(clarsimp split: generat.split simp add: map_spmf_bind_spmf o_def)
      apply(rule bind_spmf_cong[OF refl])
      apply clarsimp
      apply(rule exec_gpv.IH)
      done
  qed
  then show "ord_spmf op = ?lhs ?rhs" by -

  show "ord_spmf op = ?rhs ?lhs" unfolding exec_gpv1_def
  proof(induction arbitrary: s \<sigma> rule: iterate_spmf_fixp_induct)
    case adm show ?case by simp
    case bottom show ?case by simp
    case iterate_spmf: (step iterate_spmf')

    show ?case
      apply(simp add: map_bind_pmf bind_spmf_pmf_assoc)
      apply(subst iterate_gpv_unroll)
      apply(subst bind_gpv'_def)
      apply(simp add: exec_gpv_bind)
      apply(subst exec_gpv_try)
      apply(subst exec_gpv'_map)
      apply(simp add: bind_map_spmf)
      apply(subst option.case_distrib[where h="\<lambda>x. bind_spmf x _"])
      apply(rewrite in "ord_spmf op = _ \<hole>" bind_spmf_def)
      apply(rule rel_pmf_bind_reflI)
      apply(split option.split)
      apply(rule conjI; clarsimp)
      apply(simp add: bind_map_spmf)
      apply(split option.split)
      apply(rule conjI; clarsimp)
      subgoal
        apply(rule spmf.leq_trans)
         apply(rule monotone_bind_spmf1[THEN monotoneD])
         apply(rule iterate_spmf.hyps)
        apply(subst iterate_spmf.simps)
        apply simp
        apply(simp add: last.simps)
        done
      subgoal for \<sigma>' s'
        apply(subst exec_gpv_map)
        apply(unfold id_apply)
        using iterate_spmf.IH[of s' \<sigma>', THEN monotoneD[OF monotone_map_spmf], where ?f1="apfst (op # s')"]
        apply(rule spmf.leq_trans[rotated])
        apply(simp add: map_spmf_bind_spmf o_def)
        apply(rule ord_spmf_bind_reflI)
        apply(simp add: last.simps)
        done
      done
  qed
qed
 *)
lemma ord_spmf_reflI': "x = y \<Longrightarrow> ord_spmf op = x y" (* Move to SPMF *)
by simp

(*
lemma iterate_spmf_exec_gpv:
  defines "iterate_spmf1 \<equiv> iterate_spmf" and "iterate_spmf2 \<equiv> iterate_spmf"
  shows
  "iterate_spmf1 (\<lambda>(s, \<sigma>). exec_gpv callee (K s) \<sigma>) (s, \<sigma>) =
   map_spmf (\<lambda>ss. map (apfst the) (if ss \<noteq> [] \<and> Option.is_none (fst (last ss)) then butlast ss else ss))
     (iterate_spmf2 (\<lambda>(so, \<sigma>). case so of None \<Rightarrow> return_pmf None | Some s \<Rightarrow> exec_gpv' callee (K s) \<sigma>) (Some s, \<sigma>))"
  (is "?lhs = ?rhs" is "iterate_spmf1 ?p _ = map_spmf ?f (iterate_spmf2 ?q _)")
proof(rule spmf.leq_antisym)
  note [simp] = bind_return_pmf and [[show_variants]]

  show "ord_spmf op = ?lhs ?rhs" unfolding iterate_spmf1_def
  proof(induction arbitrary: s \<sigma> rule: iterate_spmf_fixp_induct)
    case adm show ?case by simp
    case bottom show ?case by simp
  next
    case (step iterate_spmf')
    show ?case unfolding iterate_spmf2_def
      apply(rewrite in "ord_spmf _ _ \<hole>" iterate_spmf.simps)
      apply(simp add: map_bind_pmf exec_gpv_conv_exec_gpv' bind_spmf_def bind_assoc_pmf)
      apply(rule rel_pmf_bind_reflI)
      apply(clarsimp split: option.split; rule conjI; clarsimp)
      subgoal for \<sigma>' by(rewrite iterate_spmf.simps) simp
      subgoal premises for \<sigma>' s'
        apply(rule spmf.leq_trans)
         apply(rule monotone_map_spmf[THEN monotoneD])
         apply(rule step.IH[unfolded iterate_spmf2_def])
        apply(auto simp add: pmf.map_comp o_def option.map_comp intro!: ord_spmf_reflI' map_spmf_cong)
        done
      done
  qed

  show "ord_spmf op = ?rhs ?lhs" unfolding iterate_spmf2_def
  proof(induction arbitrary: s \<sigma> rule: iterate_spmf_fixp_induct)
    case adm show ?case by simp
    case bottom show ?case by simp
  next
    case (step iterate_spmf')
    show ?case unfolding iterate_spmf1_def
      apply(rewrite in "ord_spmf _ _ \<hole>" iterate_spmf.simps)
      apply(simp add: map_bind_pmf exec_gpv_conv_exec_gpv' bind_spmf_def bind_assoc_pmf)
      apply(rule rel_pmf_bind_reflI)
      apply(clarsimp split: option.split; rule conjI; clarsimp simp add: spmf.map_comp)
      subgoal for \<sigma>'
        apply(rule spmf.leq_trans)
         apply(rule monotone_map_spmf[THEN monotoneD])
         apply(rule step.hyps)
        apply(subst iterate_spmf.simps)
        apply simp
        done
      subgoal premises for \<sigma>' s'
        apply(rule spmf.leq_trans[rotated])
         apply(rule monotone_map_spmf[THEN monotoneD])
         apply(rule step.IH[unfolded iterate_spmf1_def])
        apply(auto simp add: pmf.map_comp o_def option.map_comp intro!: ord_spmf_reflI' map_spmf_cong)
        done
      done
  qed
qed
*)

















      






typedef 'a neset = "{A :: 'a set. A \<noteq> {}}" 
  morphisms neset NE_set by blast
setup_lifting type_definition_neset

lemma neset_neq_empty [simp]: "neset A \<noteq> {}"
using neset[of A] by simp

lemma select_nesetI [intro!, simp]: "select x (neset A) \<in> neset A"
by simp

experiment begin -- \<open>unused\<close>
lift_definition member_ne :: "'a \<Rightarrow> 'a neset \<Rightarrow> bool"  (infix "\<in>ne" 50) is "op \<in>" .
lift_definition UNION_ne1 :: "'a neset \<Rightarrow> ('a \<Rightarrow> 'b set) \<Rightarrow> 'b set" is "UNION" .
(*
syntax
  "_UNION_ne"      :: "pttrn => 'a neset => 'b set => 'b set"  ("(3\<Union>_\<in>ne_./ _)" [0, 0, 10] 10)
translations
  "\<Union>x\<in>ne A. B"   == "CONST UNION_ne1 A (%x. B)"
*)
lemmas UN_ne1_I[intro] = UN_I[Transfer.transferred]
  and UN_ne1_E[elim!] = UN_E[Transfer.transferred]
end














section {* Formalisiation of CoSP by Backes et al. *}

subsection {* Symbolic model *}

text {* We encode the set of terms with the unspecified type @{typ "'t"} *}

locale symbolic_model =
  fixes \<C> :: "'c \<Rightarrow> 't list \<Rightarrow> 't"
  and \<N> :: "'n \<Rightarrow> 't"
  and \<D> :: "'d \<Rightarrow> 't list \<Rightarrow> 't option"
  and \<T> :: "'t \<Rightarrow> bool"
  and deduce :: "'t set \<Rightarrow> 't \<Rightarrow> bool"
begin
notation deduce ("_/ \<turnstile> (2_)" [120, 120] 65)
end

datatype ('c, 'n, 'd) computation =
  Constructor 'c
| Nonce 'n
| Destructor 'd

type_synonym bitstring = "bool list"

text \<open>
  Node identifiers serve two purposes. First, they are used to reference the arguments of
  computations. Second, they serve as the observable events in the traces for the trace
  properties. Uniqueness of identifiers is only needed for the referencing. To avoid this
  invariant, we split the two concepts. The arguments of computations are referenced
  by giving the number of the node on the trace from the root to the node (only nodes that can be referenced are counted).
  This then corresponds to the position of the event in the hybrid trace.
\<close>
type_synonym backlink = nat

codatatype ('id, 'c, 'n, 'd) protocol =
  Computation
    ("identifier": 'id) 
    (computation: "('c, 'n, 'd) computation")
    (c_data: "backlink list")
    (yes_successor: "('id, 'c, 'n, 'd) protocol")
    (no_successor: "('id, 'c, 'n, 'd) protocol")
| Output ("identifier": 'id) (o_data: backlink) (successor: "('id, 'c, 'n, 'd) protocol")
| Input ("identifier": 'id) (successor: "('id, 'c, 'n, 'd) protocol")
| Control ("identifier": 'id) (out_metatdata: bitstring) (successor_labels: "bitstring neset") (successors: "bitstring \<Rightarrow> ('id, 'c, 'n, 'd) protocol")
| Probabilistic ("identifier": 'id) (successor_probs: "bitstring pmf") (successors: "bitstring \<Rightarrow> ('id, 'c, 'n, 'd) protocol")

context begin
qualified inductive annotationsp :: "backlink \<Rightarrow> ('id, 'c, 'n, 'd) protocol \<Rightarrow> bool"
  for x
where
  Computation: "x \<in> set data \<Longrightarrow> annotationsp x (Computation ident lab data yes no)"
| Computation_yes: "annotationsp x yes \<Longrightarrow> annotationsp x (Computation ident lab data yes no)"
| Computation_no: "annotationsp x no \<Longrightarrow> annotationsp x (Computation ident lab data yes no)"
| Output: "annotationsp x (Output ident x succ)"
| Output_succ: "annotationsp x succ \<Longrightarrow> annotationsp x (Output ident data succ)"
| Input_succ: "annotationsp x succ \<Longrightarrow> annotationsp x (Input ident succ)"
(*| Control: "annotationsp x (Control ident x A succs)"*)
-- \<open>We do not consider the out-metadata to be part of the annotations, because this is a literal string being
  sent out to the adversary rather than a reference to a node in the tree.\<close>
| Control_succ: "\<lbrakk> annotationsp x (succs y); y \<in> neset A \<rbrakk> \<Longrightarrow> annotationsp x (Control ident out A succs)"
| Probabilistic_succ: "\<lbrakk> annotationsp x (succs y); y \<in> set_pmf p \<rbrakk> \<Longrightarrow> annotationsp x (Probabilistic ident p succs)"

definition annotations :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> backlink set"
where "annotations prot \<equiv> {x. annotationsp x prot}"

lemma annotationsp_annotations_eq [pred_set_conv]: "annotationsp x prot = (x \<in> annotations prot)"
by(simp add: annotations_def)

context begin
local_setup {* Local_Theory.map_background_naming (Name_Space.mandatory_path "annotations") *}

lemmas intros [intro?] = annotationsp.intros[to_set]
  and Computation        = Computation[to_set]
  and Computation_yes    = Computation_yes[to_set]
  and Computation_no     = Computation_no[to_set]
  and Output             = Output[to_set]
  and Output_succ        = Output_succ[to_set]
  and Input_succ         = Input_succ[to_set]
(*  and Control            = Control[to_set] *)
  and Control_succ       = Control_succ[to_set]
  and Probabilistic_succ = Probabilistic_succ[to_set]
  and induct [consumes 1, case_names 
    Computation Computation_yes Computation_no
    Output Output_succ
    Input
    Control Control_succ
    Probabilistic_succ
    , induct set: annotations] = annotationsp.induct[to_set]
  and cases [consumes 1, case_names
    Computation Computation_yes Computation_no
    Output Output_succ
    Input
    Control Control_succ
    Probabilistic_succ
    , cases set: annotations] = annotationsp.cases[to_set]
  and simps = annotationsp.simps[to_set]
end

end

lemma annotations_simps [simp]:
  "\<And>data. annotations (Computation ident lab data yes no) = set data \<union> annotations yes \<union> annotations no"
  "\<And>data. annotations (Output ident data succ) = insert data (annotations succ)"
  "annotations (Input ident succ) = annotations succ"
  "annotations (Control ident out A succs) = (*insert out*) (\<Union>x\<in>neset A. annotations (succs x))"
  "annotations (Probabilistic ident p succs) = (\<Union>x\<in>set_pmf p. annotations (succs x))"
unfolding set_eq_iff by(subst annotations.simps; auto)+

context begin
qualified inductive identifiersp :: "'id \<Rightarrow> ('id, 'c, 'n, 'd) protocol \<Rightarrow> bool"
  for x
where
  Computation: "identifiersp x (Computation x lab data yes no)"
| Computation_yes: "identifiersp x yes \<Longrightarrow> identifiersp x (Computation ident lab data yes no)"
| Computation_no: "identifiersp x no \<Longrightarrow> identifiersp x (Computation ident lab data yes no)"
| Output: "identifiersp x (Output x data succ)"
| Output_succ: "identifiersp x succ \<Longrightarrow> identifiersp x (Output ident data succ)"
| Input: "identifiersp x (Input x succ)"
| Input_succ: "identifiersp x succ \<Longrightarrow> identifiersp x (Input ident succ)"
| Control: "identifiersp x (Control x out A succs)"
| Control_succ: "\<lbrakk> identifiersp x (succs y); y \<in> neset A \<rbrakk> \<Longrightarrow> identifiersp x (Control ident out A succs)"
| Probabilistic: "identifiersp x (Probabilistic x p succs)"
| Probabilistic_succ: "\<lbrakk> identifiersp x (succs y); y \<in> set_pmf p \<rbrakk> \<Longrightarrow> identifiersp x (Probabilistic ident p succs)"

definition identifiers :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> 'id set"
where "identifiers prot \<equiv> {x. identifiersp x prot}"

lemma identifiersp_identifiers_eq [pred_set_conv]: "identifiersp x = (\<lambda>prot. x \<in> identifiers prot)"
by(simp add: identifiers_def)

context begin
local_setup {* Local_Theory.map_background_naming (Name_Space.mandatory_path "identifiers") *}

lemmas intros [intro?] = identifiersp.intros[to_set]
  and Computation        = Computation[to_set]
  and Computation_yes    = Computation_yes[to_set]
  and Computation_no     = Computation_no[to_set]
  and Output             = Output[to_set]
  and Output_succ        = Output_succ[to_set]
  and Input              = Input[to_set]
  and Input_succ         = Input_succ[to_set]
  and Control            = Control[to_set]
  and Control_succ       = Control_succ[to_set]
  and Probabilistic      = Probabilistic[to_set]
  and Probabilistic_succ = Probabilistic_succ[to_set]
  and induct [consumes 1, case_names
    Computation Computation_yes Computation_no
    Output Output_succ
    Input Input_succ
    Control Control_succ
    Probabilistic Probabilistic_succ
    , induct set: identifiers] = identifiersp.induct[to_set]
  and cases [consumes 1, case_names
    Computation Computation_yes Computation_no
    Output Output_succ
    Input Input_succ
    Control Control_succ
    Probabilistic Probabilistic_succ
    , cases set: identifiers] = identifiersp.cases[to_set]
  and simps = identifiersp.simps[to_set]
end

end

lemma identifiers_simps [simp]:
  "identifiers (Computation ident lab data yes no) = insert ident (identifiers yes \<union> identifiers no)"
  "identifiers (Output ident data' succ) = insert ident (identifiers succ)"
  "identifiers (Input ident succ) = insert ident (identifiers succ)"
  "identifiers (Control ident out A succs) = insert ident (\<Union>x\<in>neset A. identifiers (succs x))"
  "identifiers (Probabilistic ident p succs) = insert ident (\<Union>x\<in>set_pmf p. identifiers (succs x))"
unfolding set_eq_iff by(subst identifiers.simps; auto)+


coinductive wf_label :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> backlink \<Rightarrow> bool"
where
  Computation: 
  "\<lbrakk> \<forall>bl\<in>set data. bl < L; wf_label yes (L + 1); wf_label no L \<rbrakk>
  \<Longrightarrow> wf_label (Computation ident lab data yes no) L"
| Output: "\<lbrakk> data < L; wf_label succ L \<rbrakk> \<Longrightarrow> wf_label (Output ident data succ) L"
| Input: "wf_label succ (L + 1) \<Longrightarrow> wf_label (Input ident succ) L"
| Control: "\<lbrakk> \<And>x. x \<in>ne A \<Longrightarrow> wf_label (succs x) L \<rbrakk> \<Longrightarrow> wf_label (Control ident out A succs) L"
| Probabilistic: "(\<And>x. x \<in> set_pmf p \<Longrightarrow> wf_label (succs x) L) \<Longrightarrow> wf_label (Probabilistic ident p succs) L"

inductive_simps wf_label_simps [folded Ball_def, simp]:
  "wf_label (Computation ident lab data yes no) L"
  "wf_label (Output ident data' succ) L"
  "wf_label (Input ident succ) L"
  "wf_label (Control ident out A succs) L"
  "wf_label (Probabilistic ident p succs) L"

text \<open> The predicate @{text wf_ident} ensures that identifiers are unique in the protocol,
  but this predicate should not be needed, as the referencing of arguments is done by offsets.\<close>

coinductive wf_ident :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> bool"
where
  Computation:
  "\<lbrakk> identifiers yes \<inter> identifiers no = {}; ident \<notin> identifiers yes; ident \<notin> identifiers no; 
     wf_ident yes; wf_ident no \<rbrakk>
  \<Longrightarrow> wf_ident (Computation ident lab data yes no)"
| Output:
  "\<lbrakk> ident \<notin> identifiers succ; wf_ident succ \<rbrakk>
  \<Longrightarrow> wf_ident (Output ident data succ)"
| Input:
  "\<lbrakk> ident \<notin> identifiers succ; wf_ident succ \<rbrakk>
  \<Longrightarrow> wf_ident (Input ident succ)"
| Control:
  "\<lbrakk> \<And>x. x \<in> neset A \<Longrightarrow> ident \<notin> identifiers (succs x);
     disjoint_family_on (\<lambda>x. identifiers (succs x)) (neset A);
     \<And>x. x \<in> neset A \<Longrightarrow> wf_ident (succs x) \<rbrakk>
  \<Longrightarrow> wf_ident (Control ident out A succs)"
| "\<lbrakk> \<And>x. x \<in> set_pmf p \<Longrightarrow> ident \<notin> identifiers (succs x);
     disjoint_family_on (\<lambda>x. identifiers (succs x)) (set_pmf p);
     \<And>x. x \<in> set_pmf p \<Longrightarrow> wf_ident (succs x) \<rbrakk>
  \<Longrightarrow> wf_ident (Probabilistic ident p succs)"

inductive_simps wf_ident_simps [folded Ball_def, simp]:
  "wf_ident (Computation ident lab data yes no)"
  "wf_ident (Output ident data' succ)"
  "wf_ident (Input ident succ)"
  "wf_ident (Control ident out A succs)"
  "wf_ident (Probabilistic ident p succs)"


type_synonym ('t, 'id, 'c, 'n, 'd) symb_state = "'t set \<times> ('id, 'c, 'n, 'd) protocol \<times> 't list"

print_translation -- \<open>pretty printing for @{typ "('t, 'id, 'c, 'n, 'd) symb_state"}\<close> \<open>
  let
    fun tr' [Const (@{type_syntax set}, _) $ t1, 
      Const (@{type_syntax prod}, _) $ 
        (Const (@{type_syntax protocol}, _) $ id $ c $ n $ d ) $
        (Const (@{type_syntax list}, _) $ t2)] =
      if t1 = t2 then Syntax.const @{type_syntax symb_state} $ t1 $ id $ c $ n $ d
      else raise Match;
  in [(@{type_syntax "prod"}, K tr')]
  end
\<close>
typ "('t, 'id, 'c, 'n, 'd) symb_state"

abbreviation rel_symb_state
  :: "('t \<Rightarrow> 't' \<Rightarrow> bool) \<Rightarrow> ('id \<Rightarrow> 'id' \<Rightarrow> bool) \<Rightarrow> ('c \<Rightarrow> 'c' \<Rightarrow> bool) \<Rightarrow> ('n \<Rightarrow> 'n' \<Rightarrow> bool) \<Rightarrow> ('d \<Rightarrow> 'd' \<Rightarrow> bool)
  \<Rightarrow> ('t, 'id, 'c, 'n, 'd) symb_state \<Rightarrow> ('t', 'id', 'c', 'n', 'd') symb_state \<Rightarrow> bool"
where "rel_symb_state T ID C N D \<equiv> rel_prod (rel_set T) (rel_prod (rel_protocol ID C N D) (list_all2 T))"

type_synonym ('t, 'id, 'c, 'n, 'd) full_trace = "('t, 'id, 'c, 'n, 'd) symb_state list"

context symbolic_model begin

fun symb_step :: "('t, 'id, 'c, 'n, 'd) symb_state \<Rightarrow> ('t, 'id, 'c, 'n, 'd) symb_state set"
where
  "symb_step (S, Computation ident lab data yes no, f) = 
  (case (case lab of Constructor c \<Rightarrow> filter_option \<T> (Some (\<C> c (map (nth f) data)))
        | Nonce n \<Rightarrow> Some (\<N> n)
        | Destructor d \<Rightarrow> \<D> d (map (nth f) data)) of
      None \<Rightarrow> {(S, no, f)}
    | Some m \<Rightarrow> {(S, yes, f @ [m])})"
| "symb_step (S, Output ident data succ, f) = {(S \<union> {nth f data}, succ, f)}"
| "symb_step (S, Input ident succ, f) = {(S, succ, f @ [m])|m. deduce S m}"
   -- \<open>We assume that @{term deduce} preserves well-typedness, i.e., we need not check @{term \<T>} here.\<close>
| "symb_step (S, Control ident data A succs, f) = {(S, p', f)|p'. p' \<in> succs ` neset A}"
| "symb_step (S, Probabilistic ident p succs, f) = {(S, p', f)|p'. p' \<in> succs ` set_pmf p}"

abbreviation full_trace :: "('t, 'id, 'c, 'n, 'd) symb_state \<Rightarrow> ('t, 'id, 'c, 'n, 'd) symb_state list set"
where "full_trace \<equiv> iterate symb_step"

abbreviation full_traces :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> ('t, 'id, 'c, 'n, 'd) symb_state list set"
where "full_traces prot \<equiv> full_trace ({}, prot, [])"

definition node_traces :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> 'id list set"
where "node_traces prot = map (\<lambda>(_, p, _). identifier p) ` full_traces prot"

end

subsection {* Computational model *}

type_synonym ('id, 'c, 'n, 'd) comp_state 
  = "('id, 'c, 'n, 'd) protocol \<times> bitstring list \<times> ('n \<rightharpoonup> bitstring)"

print_translation -- \<open>pretty printing for @{typ "('id, 'c, 'n, 'd) comp_state"}\<close> \<open>
  let
    fun tr' [Const (@{type_syntax protocol}, _) $ id $ c $ n1 $ d, 
      Const (@{type_syntax prod}, _) $ 
        (Const (@{type_syntax list}, _) $
          (Const (@{type_syntax list}, _) $ (Const (@{type_syntax bool}, _)))) $
        (Const (@{type_syntax fun}, _) $ n2 $
          (Const (@{type_syntax option}, _) $ (Const (@{type_syntax list}, _) $ (Const (@{type_syntax bool}, _)))))] =
      if n1 = n2 then Syntax.const @{type_syntax comp_state} $ id $ c $ n1 $ d
      else raise Match;
  in [(@{type_syntax "prod"}, K tr')]
  end
\<close>
typ "('id, 'c, 'n, 'd) comp_state"


record 's adversary' =
  gen_input :: "'s \<Rightarrow> (bitstring \<times> 's) spmf"
  rec_output :: "'s \<Rightarrow> bitstring \<Rightarrow> ('s) spmf"
  control :: "'s \<Rightarrow> bitstring \<Rightarrow> (bitstring \<times> 's) spmf"

type_synonym 's adversary = "security \<Rightarrow> 's adversary' \<times> 's"

print_translation -- \<open>pretty printing for @{typ "'s adversary"}\<close> \<open>
  let
    fun tr' [Const (@{type_syntax nat}, _), 
      Const (@{type_syntax prod}, _) $ 
        (Const (@{type_syntax adversary'_ext}, _) $ s1 $ Const (@{type_syntax unit}, _)) $
        s2] =
      if s1 = s2 then Syntax.const @{type_syntax adversary} $ s1
      else raise Match;
  in [(@{type_syntax "fun"}, K tr')]
  end
\<close>
typ "'s adversary"

type_synonym bound = "security \<Rightarrow> nat"

locale computational_implementation = symbolic_model +
  constrains \<C> :: "'c \<Rightarrow> 't list \<Rightarrow> 't"
  and \<N> :: "'n \<Rightarrow> 't"
  and \<D> :: "'d \<Rightarrow> 't list \<Rightarrow> 't option"
  and \<T> :: "'t \<Rightarrow> bool"
  and deduce :: "'t set \<Rightarrow> 't \<Rightarrow> bool"
  fixes A_C :: "'c  \<Rightarrow> security \<Rightarrow> bitstring list \<Rightarrow> bitstring option"
  and A_N :: "'n \<Rightarrow> security \<Rightarrow> bitstring pmf"
  and A_D :: "'d \<Rightarrow> security \<Rightarrow> bitstring list \<Rightarrow> bitstring option"
begin

text {*
  In CoSP's intermediate language for protocols (these trees with backlinks), there are just
  two processes on the computational level (Def. 9): the attacker as an interactive turing machine
  and the interpreter of the protocol which contains all interleavings of the protocol (controlled 
  by the attacker via inner control nodes. They also give an encoding from the applied $\pi$-calculus
  to the tree language (\S 4.4, Def. 22 and below):
  At each execution step of the protocol, the whole state of the process is sent to the attacker
  (without the environments for nonces and variables) and the attacker responds
  with an evaluation context that determines the next step of the protocol. Internal communication
  between the two subprocesses is possible if the attacker says so. To restrict his access to
  channels whose names he knows, when the attacker wants to access such a channel, he has to
  provide the name of the channel. By picking a replication, the attacker can spawn a new instance.

  Summary: Attacker is not part of the process language, but part of the environment.
  The protocols themselves call the attacker to decide what to do next. There are always at most
  two agents and the attacker gets a full description of the control state of the participants
  (but not their data).
*}

fun (in -) fuel_left :: "('id, 'c, 'n, 'd) comp_state \<Rightarrow> nat \<Rightarrow> bool"
where "fuel_left (prot, f, nonces) n \<longleftrightarrow> sum_list (map length f) < n"

context
  fixes \<A> :: "'s adversary'"
  and \<eta> :: security
begin

fun comp_step :: "('id, 'c, 'n, 'd) comp_state \<times> 's \<Rightarrow> (('id, 'c, 'n, 'd) comp_state \<times> 's) spmf"
where
  "comp_step ((Computation ident (Nonce n) data yes no, f, nonces), s) =
   (case nonces n of None \<Rightarrow> 
      bind_pmf (A_N n \<eta>) (\<lambda>nonce. return_spmf ((yes, f @ [nonce], nonces(n \<mapsto> nonce)), s))
    | Some nonce \<Rightarrow> return_spmf ((yes, f @ [nonce], nonces), s))"
| "comp_step ((Computation ident (Constructor c) data yes no, f, nonces), s) =
   (case A_C c \<eta> (map (nth f) data) of None \<Rightarrow> return_spmf ((no, f, nonces), s)
    | Some x \<Rightarrow> return_spmf ((yes, f @ [x], nonces), s))"
| "comp_step ((Computation ident (Destructor d) data yes no, f, nonces), s) =
   (case A_D d \<eta> (map (nth f) data) of None \<Rightarrow> return_spmf ((no, f, nonces), s)
    | Some x \<Rightarrow> return_spmf ((yes, f @ [x], nonces), s))"
| "comp_step ((Input ident succ, f, nonces), s) = 
   bind_spmf (gen_input \<A> s) (\<lambda>(x, s'). return_spmf ((succ, f @ [x], nonces), s'))"
| "comp_step ((Output ident data succ, f, nonces), s) =
   bind_spmf (rec_output \<A> s (nth f data)) (\<lambda>s'. return_spmf ((succ, f, nonces), s'))"
| "comp_step ((Control ident out A succs, f, nonces), s) =
   bind_spmf (control \<A> s out)
     (\<lambda>(input, s'). return_spmf ((succs (select input (neset A)), f, nonces), s'))"
| "comp_step ((Probabilistic ident p succs, f, nonces), s) =
   bind_pmf p (\<lambda>x. return_spmf ((succs x, f, nonces), s))"

text {*
  @{const comp_step} must be iterated to get a distribution on traces. 
  In CoSP, the length of the run is bounded by a polynomial. We assume that
  the polynomial is represented as a fuel in the initial state from which every step
  deducts a bit.
*}

context includes applicative_syntax begin

fun comp_step_bounded :: "(('id, 'c, 'n, 'd) comp_state \<times> 's) \<times> nat \<Rightarrow> ((('id, 'c, 'n, 'd) comp_state \<times> 's) \<times> nat) option spmf"
where
  "comp_step_bounded (state, bound) =
   (if fuel_left (fst state) bound then
      pure_spmf (\<lambda>state'. Some (state', bound - 1)) \<diamondop> comp_step state
    else return_spmf None)"

end

text \<open>
  When the adversary does not complete its call, @{const comp_step} returns @{const None}.
  The iteration must not drop the corresponding trace, but should only stop the iteration. 
\<close>

definition full_comp_trace
  :: "nat \<Rightarrow> (('id, 'c, 'n, 'd) comp_state \<times> 's)
  \<Rightarrow> (('id, 'c, 'n, 'd) comp_state \<times> 's) list spmf"
where
  "full_comp_trace bound state = 
   map_spmf (map fst) (iterate_spmf comp_step_bounded (state, bound))"

(* lemma iterate_spmf_comp_step_bounded: "iterate_spmf comp_step_bounded (inv_image less_than snd)"
by unfold_locales(auto simp add: enat_iless_iff  split: split_if_asm)

interpretation iterate_spmf "comp_step_bounded" "inv_image less_than snd"
by(fact iterate_spmf_comp_step_bounded)
 *)

lemma full_comp_trace_unfold:
  "full_comp_trace bound state =
  (if fuel_left (fst state) bound then
    bind_spmf (comp_step state) (\<lambda>state'. map_spmf (op # state') (full_comp_trace (bound - 1) state'))
   else return_spmf [])"
unfolding full_comp_trace_def
apply(subst iterate_spmf.simps)
apply(auto simp add: spmf.map_comp o_def bind_return_pmf map_spmf_conv_ap[symmetric] bind_map_spmf map_spmf_bind_spmf)
done

end

definition comp_trace
  :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> 's adversary \<Rightarrow> bound \<Rightarrow> security \<Rightarrow> 'id list spmf"
where
 "comp_trace prot \<A> bound \<eta> =
  (let (\<A>', s) = \<A> \<eta>
   in map_spmf (map (\<lambda>((prot, _, _), _). identifier prot)) 
        (full_comp_trace \<A>' \<eta> (bound \<eta>) ((prot, [], Map.empty), s)))"

end

subsection {* Computational soundness *}

inductive prefix_closed :: "'a list set \<Rightarrow> bool" for A
where prefix_closed: "(\<And>xs ys. \<lbrakk> xs \<in> A; prefix ys xs \<rbrakk> \<Longrightarrow> ys \<in> A) \<Longrightarrow> prefix_closed A"

type_synonym 'id trace_prop = "'id list \<Rightarrow> bool"

definition trace_prop :: "'id trace_prop \<Rightarrow> bool"
where "trace_prop tprop \<longleftrightarrow> prefix_closed {tr. tprop tr}"
 -- \<open>We miss the efficiency part here, there should be an efficient algorithm to determine
     whether @{term prop} contains a given list\<close>

context symbolic_model begin

definition holds :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> 'id trace_prop \<Rightarrow> bool" ("_ \<Turnstile> _" [120, 120] 65)
where "prot \<Turnstile> tprop \<equiv> node_traces prot \<subseteq> {tr. tprop tr}"

end

context computational_implementation begin

inductive holds_computationally_for
  :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> 'id trace_prop \<Rightarrow> 's adversary
  \<Rightarrow> bound => bool"
  ("_ \<Turnstile> _ '/ _, _" [120, 0, 0, 120] 65)
  for prot tprop \<A> p
where
  holds_computationally_for:
  "negligible (\<lambda>\<eta>. spmf (map_spmf tprop (comp_trace prot \<A> p \<eta>)) False)
  \<Longrightarrow> prot \<Turnstile> tprop / \<A>, p"
lemmas holds_computationally_forI = holds_computationally_for

end

subsection {* Simulator *}

record ('s, 'id, 't, 'call, 'ret) simulator' =
  gen_message :: "'s \<Rightarrow> (('t \<times> 's) option, 'call, 'ret) gpv"
  rec_message :: "'s \<Rightarrow> 't \<Rightarrow> ('s option, 'call, 'ret) gpv"
  control :: "'s \<Rightarrow> bitstring \<Rightarrow> ((bitstring \<times> 's) option, 'call, 'ret) gpv"
  info_sim :: "'s \<Rightarrow> 'id \<Rightarrow> 't option \<Rightarrow> ('s option, 'call, 'ret) gpv"

type_synonym ('s, 'id, 't, 'call, 'ret) simulator
  = "security \<Rightarrow> ('s, 'id, 't, 'call, 'ret) simulator' \<times> 's"

print_translation -- \<open>pretty printing for @{typ "('s, 'id, 't, 'call, 'ret) simulator"}\<close> \<open>
  let
    fun tr' [Const (@{type_syntax nat}, _), 
      Const (@{type_syntax prod}, _) $ 
        (Const (@{type_syntax simulator'_ext}, _) $ s1 $ id $ t $ call $ ret $ Const (@{type_syntax unit}, _)) $
        s2] =
      if s1 = s2 then Syntax.const @{type_syntax simulator} $ s1 $ id $ t $ call $ ret
      else raise Match;
  in [(@{type_syntax "fun"}, K tr')]
  end
\<close>
typ "('a, 'out, 'in, 'call, 'ret) simulator"

subsubsection \<open>Parametricity setup for @{typ "('s, 'id, 't, 'call, 'ret, 'more) simulator'_ext"}\<close>
text \<open>@{typ 'ret} is a dead type variable in @{typ "('a, 'call, 'ret) gpv"}, so we also omit it here\<close>

term simulator'_ext

context includes lifting_syntax begin
definition rel_simulator'_ext
  :: "('s \<Rightarrow> 's' \<Rightarrow> bool) \<Rightarrow> ('id \<Rightarrow> 'id' \<Rightarrow> bool) \<Rightarrow> ('t \<Rightarrow> 't' \<Rightarrow> bool)
  \<Rightarrow> ('call \<Rightarrow> 'call' \<Rightarrow> bool) \<Rightarrow> ('more \<Rightarrow> 'more' \<Rightarrow> bool)
  \<Rightarrow> ('s, 'id, 't, 'call, 'ret, 'more) simulator'_ext 
  \<Rightarrow> ('s', 'id', 't', 'call', 'ret, 'more') simulator'_ext \<Rightarrow> bool"
where
  "rel_simulator'_ext S ID T CALL M sim sim' \<longleftrightarrow> 
  (S ===> rel_gpv (rel_option (rel_prod T S)) CALL) (gen_message sim) (gen_message sim') \<and>
  (S ===> T ===> rel_gpv (rel_option S) CALL) (rec_message sim) (rec_message sim') \<and>
  (S ===> op = ===> rel_gpv (rel_option (rel_prod op = S)) CALL) (control sim) (control sim') \<and>
  (S ===> ID ===> rel_option T ===> rel_gpv (rel_option S) CALL) (info_sim sim) (info_sim sim') \<and>
  M (simulator'.more sim) (simulator'.more sim')"

lemma simulator'_ext_rel_eq [relator_eq]:
  "rel_simulator'_ext op = op = op = op = op = = op ="
by(auto simp add: fun_eq_iff rel_simulator'_ext_def gpv.rel_eq prod.rel_eq option.rel_eq rel_fun_eq intro!: simulator'.equality)

lemma gen_message_parametric [transfer_rule]:
  "(rel_simulator'_ext S ID T CALL M ===> S ===> rel_gpv (rel_option (rel_prod T S)) CALL) gen_message gen_message"
by(rule rel_funI)(simp add: rel_simulator'_ext_def)

lemma rec_message_parametric [transfer_rule]:
  "(rel_simulator'_ext S ID T CALL M ===> S ===> T ===> rel_gpv (rel_option S) CALL) rec_message rec_message"
by(rule rel_funI)(simp add: rel_simulator'_ext_def)

lemma control_parametric [transfer_rule]:
  "(rel_simulator'_ext S ID T CALL M ===> S ===> op = ===> rel_gpv (rel_option (rel_prod op = S)) CALL) control control"
by(rule rel_funI)(simp add: rel_simulator'_ext_def)

lemma info_sim_parametric [transfer_rule]:
  "(rel_simulator'_ext S ID T CALL M ===> S ===> ID ===> rel_option T ===> rel_gpv (rel_option S) CALL) info_sim info_sim"
by(rule rel_funI)(simp add: rel_simulator'_ext_def)

lemma simulator'_ext_more_parametric [transfer_rule]:
  "(rel_simulator'_ext S ID T CALL M ===> M) simulator'.more simulator'.more"
by(rule rel_funI)(simp add: rel_simulator'_ext_def)

lemma simulator'_ext_parametric [transfer_rule]:
  "((S ===> rel_gpv (rel_option (rel_prod T S)) CALL) ===>
    (S ===> T ===> rel_gpv (rel_option S) CALL) ===>
    (S ===> op = ===> rel_gpv (rel_option (rel_prod op = S)) CALL) ===>
    (S ===> ID ===> rel_option T ===> rel_gpv (rel_option S) CALL) ===>
    M ===> rel_simulator'_ext S ID T CALL M) simulator'_ext simulator'_ext"
by(rule rel_funI)+(simp add: rel_simulator'_ext_def)

lemma gen_message_update_parametric [transfer_rule]:
  "(((S ===> rel_gpv (rel_option (rel_prod T S)) CALL) ===> S ===> rel_gpv (rel_option (rel_prod T S)) CALL)
   ===> rel_simulator'_ext S ID T CALL M ===> rel_simulator'_ext S ID T CALL M)
   gen_message_update gen_message_update"
by(rule rel_funI)+(auto simp add: rel_simulator'_ext_def dest: rel_funD)

lemma rec_message_update_parametric [transfer_rule]:
  "(((S ===> T ===> rel_gpv (rel_option S) CALL) ===> S ===> T ===> rel_gpv (rel_option S) CALL)
   ===> rel_simulator'_ext S ID T CALL M ===> rel_simulator'_ext S ID T CALL M)
   rec_message_update rec_message_update"
by(rule rel_funI)+(auto simp add: rel_simulator'_ext_def dest: rel_funD)

lemma control_update_parametric [transfer_rule]:
  "(((S ===> op = ===> rel_gpv (rel_option (rel_prod op = S)) CALL) ===> S ===> op = ===> rel_gpv (rel_option (rel_prod op = S)) CALL)
   ===> rel_simulator'_ext S ID T CALL M ===> rel_simulator'_ext S ID T CALL M)
   control_update control_update"
by(rule rel_funI)+(auto simp add: rel_simulator'_ext_def dest: rel_funD)

lemma info_sim_update_parametric [transfer_rule]:
  "(((S ===> ID ===> rel_option T ===> rel_gpv (rel_option S) CALL) ===> S ===> ID ===> rel_option T ===> rel_gpv (rel_option S) CALL)
   ===> rel_simulator'_ext S ID T CALL M ===> rel_simulator'_ext S ID T CALL M)
   info_sim_update info_sim_update"
by(rule rel_funI)+(auto simp add: rel_simulator'_ext_def dest: rel_funD)

lemma simulator'_ext_more_update_parametric [transfer_rule]:
  "((M ===> M)
   ===> rel_simulator'_ext S ID T CALL M ===> rel_simulator'_ext S ID T CALL M)
   simulator'.more_update simulator'.more_update"
by(rule rel_funI)+(auto simp add: rel_simulator'_ext_def dest: rel_funD)

end

type_synonym ('t, 'id, 'c, 'n, 'd, 's) hybrid_state = "('t, 'id, 'c, 'n, 'd) symb_state \<times> 's"

translations
  (type) "('t, 'id, 'c, 'n, 'd, 's) hybrid_state" \<leftharpoondown> (type) "('t, 'id, 'c, 'n, 'd) symb_state \<times> 's"
typ "('t, 'id, 'c, 'n, 'd, 's) hybrid_state"

abbreviation simulator_state :: "('t, 'id, 'c, 'n, 'd, 's) hybrid_state \<Rightarrow> 's"
where "simulator_state \<equiv> snd"

abbreviation symb_state :: "('t, 'id, 'c, 'n, 'd, 's) hybrid_state \<Rightarrow> ('t, 'id, 'c, 'n, 'd) symb_state"
where "symb_state \<equiv> fst"

inductive WT_sim :: "('call, 'ret) \<I> \<Rightarrow> ('s, 'id, 't, 'call, 'ret) simulator' \<Rightarrow> bool" ("(_) \<turnstile>s/ (_) \<surd>" [100, 0] 99)
  for \<I> sim
where
  WT_sim: 
  "\<lbrakk> \<And>s. \<I> \<turnstile>g gen_message sim s \<surd>;
     \<And>s t. \<I> \<turnstile>g rec_message sim s t \<surd>;
     \<And>s bs. \<I> \<turnstile>g control sim s bs \<surd>;
     \<And>s id t. \<I> \<turnstile>g info_sim sim s id t \<surd> \<rbrakk>
  \<Longrightarrow> \<I> \<turnstile>s sim \<surd>"
lemmas WT_simI = WT_sim
hide_fact (open) WT_sim

lemma WT_simD:
  assumes "\<I> \<turnstile>s sim \<surd>"
  shows WT_sim_gen_messageD: "\<I> \<turnstile>g gen_message sim s \<surd>"
  and WT_sim_rec_messageD: "\<I> \<turnstile>g rec_message sim s t \<surd>"
  and WT_sim_controlD: "\<I> \<turnstile>g control sim s bs \<surd>"
  and WT_sim_info_simD: "\<And>id t. \<I> \<turnstile>g info_sim sim s id t \<surd>"
using assms by(blast elim: WT_sim.cases)+

text {* Termination relation for simulators *}

inductive relation_for_sim'
  :: "('s, 'id, 't, 'call, 'ret) simulator' \<Rightarrow> ('s \<times> 's) set \<Rightarrow> bool"
  for sim R
where
  "(\<And>s s' s'' t to ident x y.
    \<lbrakk> Some (t, s') \<in> results'_gpv (gen_message sim s) \<or> 
      Some s' \<in> results'_gpv (rec_message sim s t) \<or> 
      Some (y, s') \<in> results'_gpv (control sim s x) \<or> s = s';
      Some s'' \<in> results'_gpv (info_sim sim s' ident to) \<rbrakk>
    \<Longrightarrow> (s'', s) \<in> R)
  \<Longrightarrow> relation_for_sim' sim R"

inductive relation_for_sim 
  :: "('s, 'id, 't, 'call, 'ret) simulator \<Rightarrow> (security \<Rightarrow> ('s \<times> 's) set) \<Rightarrow> bool"
  for sim R
where
  "(\<And>\<eta>. relation_for_sim' (fst (sim \<eta>)) (R \<eta>))
  \<Longrightarrow> relation_for_sim sim R"

lemma relation_for_sim'D:
  assumes "relation_for_sim' sim R"
  shows "\<lbrakk> Some s'' \<in> results'_gpv (info_sim sim s' ident to); 
    Some (t, s') \<in> results'_gpv (gen_message sim s) \<or>
    Some s' \<in> results'_gpv (rec_message sim s t) \<or>
    Some (y, s') \<in> results'_gpv (control sim s x) \<or> s = s' \<rbrakk> 
  \<Longrightarrow> (s'', s) \<in> R"
using assms by cases blast

context symbolic_model begin

context 
  fixes sim :: "('s, 'id, 't, 'call, 'ret, 'more) simulator'_scheme"
begin

fun hybrid_step1
  :: "('t, 'id, 'c, 'n, 'd, 's) hybrid_state
  \<Rightarrow> (('t option \<times> ('t, 'id, 'c, 'n, 'd, 's) hybrid_state) option, 'call, 'ret) gpv"
where
  "hybrid_step1 ((S, Computation ident lab data yes no, f), \<sigma>) =
     (case (case lab of Constructor c \<Rightarrow> filter_option \<T> (Some (\<C> c (map (nth f) data)))
         | Nonce n \<Rightarrow> Some (\<N> n)
         | Destructor d \<Rightarrow> \<D> d (map (nth f) data)) of
       None \<Rightarrow> Done (Some (None, (S, no, f), \<sigma>))
     | Some t' \<Rightarrow> Done (Some (Some t', (S, yes, f @ [t']), \<sigma>)))"
| "hybrid_step1 ((S, Output ident data succ, f), \<sigma>) =
     map_gpv (map_option (\<lambda>\<sigma>'. (None, (insert (f ! data) S, succ, f), \<sigma>'))) id (rec_message sim \<sigma> (f ! data))"
  -- \<open>Here, we are handing over control to sim to process the input. In CoSP (Def. 13), we should not. This is unclear.\<close>
  -- \<open>CoSP Def. 13 forgets to add the term to be output to @{term S} here\<close>
| "hybrid_step1 ((S, Input ident succ, f), \<sigma>) =
     map_gpv (\<lambda>t\<sigma>'. Option.bind t\<sigma>' (\<lambda>(t, \<sigma>'). if \<T> t then Some (Some t, (S, succ, f @ [t]), \<sigma>') else None)) id
       (gen_message sim \<sigma>)"
  -- \<open>Here we check that the received term is well-typed. Otherwise, the execution aborts.\<close>
| "hybrid_step1 ((S, Control ident out A succs, f), \<sigma>) =
     map_gpv (\<lambda>res. map_option (\<lambda>(input, \<sigma>'). (None, (S, succs (select input (neset A)), f), \<sigma>')) res) id
       (control sim \<sigma> out)"
| "hybrid_step1 ((S, Probabilistic ident p succs, f), \<sigma>) =
     bind_gpv (lift_spmf (spmf_of_pmf p)) (\<lambda>succ. Done (Some (None, (S, succs succ, f), \<sigma>)))"

lemmas hybrid_step1_cases[case_names Computation Output Input Control Probabilistic] = hybrid_step1.cases

lemma hybrid_step1_conv_case:
  "hybrid_step1 state =
  (case state of ((S, Computation ident lab data yes no, f), \<sigma>) \<Rightarrow>
     (case (case lab of Constructor c \<Rightarrow> filter_option \<T> (Some (\<C> c (map (nth f) data)))
         | Nonce n \<Rightarrow> Some (\<N> n)
         | Destructor d \<Rightarrow> \<D> d (map (nth f) data)) of
       None \<Rightarrow> Done (Some (None, (S, no, f), \<sigma>))
     | Some t' \<Rightarrow> Done (Some (Some t', (S, yes, f @ [t']), \<sigma>)))
  | ((S, Output ident data succ, f), \<sigma>) \<Rightarrow>
     map_gpv (map_option (\<lambda>\<sigma>'. (None, (insert (f ! data) S, succ, f), \<sigma>'))) id (rec_message sim \<sigma> (f ! data))
  | ((S, Input ident succ, f), \<sigma>) \<Rightarrow>
     map_gpv (\<lambda>t\<sigma>'. Option.bind t\<sigma>' (\<lambda>(t, \<sigma>'). if \<T> t then Some (Some t, (S, succ, f @ [t]), \<sigma>') else None)) id (gen_message sim \<sigma>)
  | ((S, Control ident out A succs, f), \<sigma>) \<Rightarrow>
     map_gpv (\<lambda>res. map_option (\<lambda>(input, \<sigma>'). (None, (S, succs (select input (neset A)), f), \<sigma>')) res) id
       (control sim \<sigma> out)
  | ((S, Probabilistic ident p succs, f), \<sigma>) \<Rightarrow>
     bind_gpv (lift_spmf (spmf_of_pmf p)) (\<lambda>succ. Done (Some (None, (S, succs succ, f), \<sigma>))))"
by(cases state rule: hybrid_step1.cases) simp_all

definition hybrid_step
  :: "('t, 'id, 'c, 'n, 'd, 's) hybrid_state
  \<Rightarrow> (('t, 'id, 'c, 'n, 'd, 's) hybrid_state option, 'call, 'ret) gpv"
where
  "hybrid_step state = 
    bind_gpv (hybrid_step1 state)
      (\<lambda>result. case result of None \<Rightarrow> Done None | Some (to, (S, prot, f), \<sigma>) \<Rightarrow>
      map_gpv (map_option (\<lambda>\<sigma>'. ((S, prot, f), \<sigma>'))) id (info_sim sim \<sigma> (identifier (fst (snd (fst state)))) to))"

end

context includes lifting_syntax begin
-- \<open>We cannot get parametricity over terms because we use @{const nth} to lookup terms in the history
  and nth is not parametric as it might go beyond the list's end.\<close>
lemma hybrid_step1_parametric [transfer_rule]:
  fixes T Tlist Toption defines "T \<equiv> op =" and "Tlist \<equiv> op =" and "Toption \<equiv> op =" shows
  "((C ===> Tlist ===> T) ===>
    (N ===> T) ===>
    (D ===> Tlist ===> Toption) ===>
    (T ===> op =) ===>
    rel_simulator'_ext S ID T CALL M ===>
    rel_prod (rel_symb_state T ID C N D) S ===>
    rel_gpv (rel_option (rel_prod Toption (rel_prod (rel_symb_state T ID C N D) S))) CALL)
  symbolic_model.hybrid_step1 symbolic_model.hybrid_step1"
proof((rule rel_funI)+, goal_cases)
  have T: "Tlist = list_all2 op =" "Toption = rel_option op ="
    by(simp_all add: Tlist_def Toption_def list.rel_eq option.rel_eq)

  case *[unfolded T_def T, transfer_rule]: (1 \<C> \<C>' \<N> \<N>' \<D> \<D>' \<T> \<T>' sim sim' state state')
  let ?B = "rel_gpv (rel_option (rel_prod (rel_option op =) (rel_prod (rel_symb_state op = ID C N D) S))) CALL"
  show ?case unfolding symbolic_model.hybrid_step1_conv_case T T_def using _ *(6)
    apply(rule case_prod_transfer[THEN rel_funD, THEN rel_funD, where B="?B"])
  proof(intro rel_funI, goal_cases)
    case ** [transfer_rule]: (1 ss ss' s s')
    show ?case using _ **(1)
      apply(rule case_prod_transfer[THEN rel_funD, THEN rel_funD, where B="?B"])
    proof(intro rel_funI, goal_cases)
      case *** [transfer_rule]: (1 S S' rest rest')
      show ?case using _ ***(2)
        apply(rule case_prod_transfer[THEN rel_funD, THEN rel_funD, where B="?B"])
      proof(intro rel_funI, goal_cases)
        case **** [transfer_rule]: (1 comp comp' f f')
        show ?case using _ _ _ _ _ ****(1)
          apply(rule protocol.case_transfer[THEN rel_funD, THEN rel_funD, THEN rel_funD, THEN rel_funD, THEN rel_funD, THEN rel_funD, where B="?B"])
          unfolding list.rel_eq[symmetric]
          apply transfer_prover+
          done
      qed
    qed
  qed
qed
end

definition run_hybrid_step
  :: "('\<sigma>, 'call, 'ret) oracle' \<Rightarrow> ('s, 'id, 't, 'call, 'ret, 'more) simulator'_scheme
  \<Rightarrow> ('t, 'id, 'c, 'n, 'd, 's) hybrid_state \<times> '\<sigma> \<Rightarrow> (('t, 'id, 'c, 'n, 'd, 's) hybrid_state \<times> '\<sigma>) option spmf"
  where
  "run_hybrid_step oracle sim = 
  (\<lambda>(state, \<sigma>). map_spmf (\<lambda>(so, \<sigma>'). map_option (\<lambda>state'. (state', \<sigma>')) so) (exec_gpv oracle (hybrid_step TYPE(_) sim state) \<sigma>))"

(* sublocale SMC: SMC_syntax "run_hybrid_step oracle sim" for "oracle" sim . *)
  
abbreviation SMC_iterate_spmf (* TODO: better name! *)
where "SMC_iterate_spmf oracle sim \<equiv> iterate_spmf (run_hybrid_step oracle sim)"

lemma SMC_iterate_spmf_induct_set_split [consumes 1, case_names stop step]:
  "\<lbrakk> tr \<in> set_spmf (SMC_iterate_spmf oracle sim (((S, prot, f), s), \<sigma>));
     \<And>S prot f s \<sigma>. None \<in> set_spmf (run_hybrid_step oracle sim (((S, prot, f), s), \<sigma>))
     \<Longrightarrow> P [] S prot f s \<sigma>;
     \<And>tr S' prot' f' s' \<sigma>' S prot f s \<sigma>.
     \<lbrakk> Some (((S', prot', f'), s'), \<sigma>') \<in> set_spmf (run_hybrid_step oracle sim (((S, prot, f), s), \<sigma>));
       tr \<in> set_spmf (SMC_iterate_spmf oracle sim (((S', prot', f'), s'), \<sigma>'));
       P tr S' prot' f' s' \<sigma>' \<rbrakk>
    \<Longrightarrow> P ((((S', prot', f'), s'), \<sigma>') # tr) S prot f s \<sigma> \<rbrakk>
\<Longrightarrow> P tr S prot f s \<sigma>"
by(erule iterate_spmf_induct_set[where K="run_hybrid_step oracle sim", split_format (complete)]) blast+

lemma SMC_iterate_spmf_induct_set_split' [consumes 1, case_names stop step]:
  "\<lbrakk> tr \<in> set_spmf (SMC_iterate_spmf oracle sim ((Sprotf, s), \<sigma>));
     \<And>Sprotf s \<sigma>. None \<in>  set_spmf (run_hybrid_step oracle sim ((Sprotf, s), \<sigma>))
     \<Longrightarrow> P [] Sprotf s \<sigma>;
     \<And>tr Sprotf' s' \<sigma>' Sprotf s \<sigma>.
     \<lbrakk> Some ((Sprotf', s'), \<sigma>') \<in> set_spmf (run_hybrid_step oracle sim ((Sprotf, s), \<sigma>));
       tr \<in> set_spmf (SMC_iterate_spmf oracle sim ((Sprotf', s'), \<sigma>'));
       P tr Sprotf' s' \<sigma>' \<rbrakk>
    \<Longrightarrow> P (((Sprotf', s'), \<sigma>') # tr) Sprotf s \<sigma> \<rbrakk>
\<Longrightarrow> P tr Sprotf s \<sigma>"
apply(cases Sprotf)
apply clarsimp
apply(erule SMC_iterate_spmf_induct_set_split; blast)
done

definition full_hybrid_trace'
  :: "('\<sigma>, 'call, 'ret) oracle' \<Rightarrow> ('s, 'id, 't, 'call, 'ret) simulator'
  \<Rightarrow> ('t, 'id, 'c, 'n, 'd, 's) hybrid_state \<times> '\<sigma>
  \<Rightarrow> (('t, 'id, 'c, 'n, 'd) symb_state \<times> '\<sigma>) list spmf"
where
  "full_hybrid_trace' oracle sim state = map_spmf (map (\<lambda>((state, s), \<sigma>). (state, \<sigma>))) (SMC_iterate_spmf oracle sim state)"

lemma hybrid_step1_S_mono:
  "Some (to, (S', prot', f'), \<sigma>') \<in> results_gpv \<I> (hybrid_step1 tytok sim ((S, prot, f), \<sigma>)) \<Longrightarrow> S \<subseteq> S'"
  by(cases "((S, prot, f), \<sigma>)" rule: hybrid_step1_cases)
    (auto simp add: set_spmf_bind_pmf Some_eq_bind_conv split: option.split_asm if_split_asm)

lemma hybrid_step_S_mono:
  "Some ((S', prot', f'), \<sigma>') \<in> results_gpv \<I> (hybrid_step tytok sim ((S, prot, f), \<sigma>)) \<Longrightarrow> S \<subseteq> S'"
by(auto simp add: hybrid_step_def intro: hybrid_step1_S_mono del: subsetI split: option.split_asm)

end
    


text {*
  In CoSP, it is assumed that the hybrid execution always stops after finitely many steps.
  It is the job of the simulator to stop the execution in the @{const info_sim} interaction
  after each step of the protocol. This is left implicit in the CoSP paper in the construction
  of the simulator. We require that the simulator terminates.
*}

locale simulator' = symbolic_model +
  constrains \<C> :: "'c \<Rightarrow> 't list \<Rightarrow> 't"
  and \<N> :: "'n \<Rightarrow> 't"
  and \<D> :: "'d \<Rightarrow> 't list \<Rightarrow> 't option"
  and \<T> :: "'t \<Rightarrow> bool"
  and deduce :: "'t set \<Rightarrow> 't \<Rightarrow> bool"
  fixes sim :: "('s, 'id, 't, 'call, 'ret) simulator'"
  and R :: "('s \<times> 's) set"
  assumes sim_R: "relation_for_sim' sim R"
  and wf_R: "wf R"
begin

lemma simulator': "simulator' sim R" by unfold_locales

lemma hybrid_step_subset_R:
  "Some state' \<in> results'_gpv (hybrid_step tytok sim state) \<Longrightarrow> (state', state) \<in> inv_image R snd"
apply(clarsimp simp add: hybrid_step_def set_pmf_bind_spmf split: option.split_asm)
apply(rename_tac \<sigma>'' S' prot' f' \<sigma>')
apply(case_tac "state" rule: hybrid_step1.cases; clarsimp simp add: gpv.set_map Some_eq_bind_conv set_spmf_bind_pmf map_option_set_bind Some_in_map_option_set_conv in_set_spmf[symmetric] split: computation.split_asm option.split_asm if_split_asm)
apply(blast intro: relation_for_sim'D[OF sim_R])+
done
(*
lemma WT_gpv_hybrid_step1: "\<I> \<turnstile>g hybrid_step1 sim state \<surd>"
by(cases state rule: hybrid_step1.cases)(simp_all add: WT_simD[OF WT_sim] split: computation.split option.split prod.split)

lemma WT_gpv_hybrid_step: "\<I> \<turnstile>g hybrid_step sim state \<surd>"
by(simp add: hybrid_step_def WT_gpv_hybrid_step1 split_beta WT_simD[OF WT_sim])
*)
lemma run_hybrid_step_subset_R:
  "{(state', state). Some state' \<in> set_spmf (run_hybrid_step oracle sim state)} \<subseteq> inv_image R (snd \<circ> fst)"
apply(clarsimp simp add: run_hybrid_step_def)
apply(drule oi_True.in_set_spmf_exec_gpv_into_results_gpv; simp add: results_gpv_\<I>_full Some_eq_map_option)
apply(drule hybrid_step_subset_R; simp)
done

(* sublocale SMC: SMC_syntax "run_hybrid_step oracle sim" . *)

(* lemma iterate_spmf_run_hybrid_step:
  "iterate_spmf (run_hybrid_step oracle sim) (inv_image R (snd \<circ> fst))"
by(unfold_locales)(rule run_hybrid_step_subset_R, simp add: wf_R)

sublocale SMC: iterate_spmf "run_hybrid_step oracle sim" "inv_image R (snd \<circ> fst)"
by(fact iterate_spmf_run_hybrid_step)
 *)

lemma full_hybrid_trace'_unfold:
  "full_hybrid_trace' oracle sim (state, \<sigma>) =
   bind_spmf (exec_gpv oracle (hybrid_step tytok sim state) \<sigma>) (\<lambda>(ostate, \<sigma>'). case ostate of None \<Rightarrow> return_spmf []
     | Some state' \<Rightarrow> map_spmf (op # (fst state', \<sigma>')) (full_hybrid_trace' oracle sim (state', \<sigma>')))"
unfolding full_hybrid_trace'_def
by(subst iterate_spmf.simps)
  (simp add: run_hybrid_step_def map_spmf_bind_spmf bind_map_spmf option.case_distrib spmf.map_comp o_def split_def cong del: option.case_cong_weak)

(* lemmas full_hybrid_trace'_induct[case_names full_hybrid_trace'] = SMC.iterate_spmf_induct

lemma full_hybrid_trace'_induct2[case_names full_hybrid_trace']:
  assumes "\<And>S prot f s \<sigma>.
     (\<And>S' prot' f' s' \<sigma>'. (((S', prot', f'), s'), \<sigma>') \<in> set_spmf (exec_gpv oracle (hybrid_step tytok sim ((S, prot, f), s)) \<sigma>)
      \<Longrightarrow> P S' prot' f' s' \<sigma>')
     \<Longrightarrow> P S prot f s \<sigma>"
  shows "P S prot f s \<sigma>"
  -- \<open>cannot use @{text "split_format (complete)"} properly inside a locale with assumptions\<close>
using assms
by(rule full_hybrid_trace'_induct[unfolded run_hybrid_step_def, where s="(((_, _, _), _), _)", split_format (complete), OF simulator', simplified])

lemma full_hybrid_trace'_induct1[case_names full_hybrid_trace']:
  "\<And>state. \<lbrakk> \<And>state s \<sigma>.
     (\<And>state' s' \<sigma>'. ((state', s'), \<sigma>') \<in> set_spmf (exec_gpv oracle (hybrid_step tytok sim (state, s)) \<sigma>)
      \<Longrightarrow> P state' s' \<sigma>')
     \<Longrightarrow> P state s \<sigma>\<rbrakk>
  \<Longrightarrow> P state s \<sigma>"
by clarify(erule full_hybrid_trace'_induct2)
 *)

end

context symbolic_model begin

definition full_hybrid_trace
  :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> ('\<sigma>, 'call, 'ret) oracle \<Rightarrow> ('s, 'id, 't, 'call, 'ret) simulator
  \<Rightarrow> security \<Rightarrow> (('t, 'id, 'c, 'n, 'd) symb_state \<times> '\<sigma>) list spmf"
where
  "full_hybrid_trace prot oracle sim \<eta> = 
  (let (sim', s) = sim \<eta>; (oracle', \<sigma>) = oracle \<eta>
   in full_hybrid_trace' oracle' sim' ((({}, prot, []), s), \<sigma>))"

definition hybrid_trace
  :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> ('\<sigma>, 'call, 'ret) oracle \<Rightarrow> ('s, 'id, 't, 'call, 'ret) simulator
  \<Rightarrow> security \<Rightarrow> 'id list spmf"
where
  "hybrid_trace prot oracle sim \<eta> = 
   map_spmf (map (\<lambda>((_, prot, _), _). identifier prot)) (full_hybrid_trace prot oracle sim \<eta>)"

end

context begin
qualified inductive output_tp :: "'t list \<Rightarrow> 't \<Rightarrow> ('id, 'c, 'n, 'd) protocol \<Rightarrow> bool"
  for f t
where "f ! data = t \<Longrightarrow> output_tp f t (Output ident data succ)"

definition output_t :: "'t list \<Rightarrow> ('id, 'c, 'n, 'd) protocol \<Rightarrow> 't set"
where "output_t f prot \<equiv> {t. output_tp f t prot}"

lemma output_tp_output_t_eq [pred_set_conv]: "output_tp f t = (\<lambda>prot. t \<in> output_t f prot)"
by(simp add: output_t_def)

context begin
local_setup {* Local_Theory.map_background_naming (Name_Space.mandatory_path "output_t") *}
lemmas intros [intro!] = output_tp.intros[to_set]
  and induct [consumes 1, case_names output_t, induct set: output_t] = output_tp.induct[to_set]
  and cases [consumes 1, case_names output_T, cases set: output_t] = output_tp.cases[to_set]
  and simps = output_tp.simps[to_set]
end
end

lemma output_t_simps [simp]:
  "output_t f (Output ident data succ) = {f ! data}"
  "output_t f (Computation ident lab data' yes no) = {}"
  "output_t f (Input ident succ) = {}"
  "output_t f (Control ident out A succs) = {}"
  "output_t f (Probabilistic ident p succs) = {}"
unfolding set_eq_iff by(auto simp add: output_t.simps)

inductive_set outputs_t :: "('t, 'id, 'c, 'n, 'd) symb_state list \<Rightarrow> 't set"
  for tr
where outputs_t: "\<lbrakk> (S, prot, f) \<in> set tr; t \<in> output_t f prot \<rbrakk> \<Longrightarrow> t \<in> outputs_t tr"

lemmas outputs_tI = outputs_t
hide_fact (open) outputs_t

lemma outputs_t_conv_UN: "outputs_t tr = (\<Union>(S, prot, f)\<in>set tr. output_t f prot)"
by(fastforce simp add: outputs_t.simps)

lemma outputs_t_simps [simp]:
  shows outputs_t_Nil: "outputs_t [] = {}"
  and outputs_t_Cons: "outputs_t ((S, prot, f) # tr) = output_t f prot \<union> outputs_t tr"
by(simp_all add: outputs_t_conv_UN)

context symbolic_model begin

lemma output_t_subset_S1:
 "Some (to, (S', protf'), s') \<in> results_gpv \<I> (hybrid_step1 tytok sim ((S, prot, f), s))
 \<Longrightarrow> output_t f prot \<subseteq> S'"
by(cases "((S, prot, f), s)" rule: hybrid_step1_cases)(simp_all split: option.split_asm)

lemma output_t_subset_S:
 "Some ((S', protf'), s') \<in> results_gpv \<I> (hybrid_step tytok sim ((S, prot, f), s)) \<Longrightarrow> output_t f prot \<subseteq> S'"
by(clarsimp simp add: hybrid_step_def output_t_subset_S1 del: subsetI split: option.split_asm)

end

context symbolic_model begin

text \<open>
  CoSP defines Dolev-Yao-ness of a trace by requiring deducibility of input terms
  from terms previously output. However, the set @{term S} in the trace already records the
  sent messages (plus the messages in @{term S} in the initial state, but this is @{term "{}"}.
  We require deducibility from @{term S} rather than the outputs to be in line with
  the symbolic definition of trace. We separately prove that the two definitions are
  equivalent for hybrid traces starting in the start state.
\<close>

inductive dy_trace :: "('t, 'id, 'c, 'n, 'd) symb_state list \<Rightarrow> bool"
  for tr
where
  dy_trace:
  "(\<And>n S ident succ f t.
    \<lbrakk> n + 1 < length tr;
      tr ! n = (S, Input ident succ, f); 
      last (snd (snd (tr ! (n + 1)))) = t \<rbrakk>
    \<Longrightarrow> S (*outputs_t (take n tr)*) \<turnstile> t)
  \<Longrightarrow> dy_trace tr"

lemmas dy_traceI = dy_trace (*hide_fact (open) dy_trace*)

lemma dy_traceD:
  assumes "dy_trace tr"
  and "n + 1 < length tr"
  and "tr ! n = (S, Input ident succ, f)"
  and "last (snd (snd (tr ! (n + 1)))) = t"
  shows "S (*outputs_t (take n tr)*) \<turnstile> t"
using assms by - (rule dy_trace.cases)

end

context simulator' begin

lemma dy_trace_symbolic_aux:
  assumes tr: "tr \<in> set_spmf (full_hybrid_trace' oracle sim ((state, s), \<sigma>))"
  and dy: "dy_trace (state # map fst tr)"
  shows "map fst tr \<in> full_trace state"
proof(rule iterate_all_nthI)
  fix n
  assume n: "n < length (map fst tr)"
  
  from tr obtain tr' where tr: "tr = map (\<lambda>((state', s'), \<sigma>'). (state', \<sigma>')) tr'" 
    and tr': "tr' \<in> set_spmf (SMC_iterate_spmf oracle sim ((state, s), \<sigma>))"
    by(clarsimp simp add: full_hybrid_trace'_def)
  from iterate_spmf_subset_iterate[of "run_hybrid_step oracle sim" "((state, s), \<sigma>)"] tr'
  have "tr' \<in> iterate (\<lambda>s. \<Union>x\<in>set_spmf (run_hybrid_step oracle sim s). set_option x) ((state, s), \<sigma>)" ..
  hence "tr' ! n \<in> (\<Union>x\<in>set_spmf (run_hybrid_step oracle sim ((((state, s), \<sigma>) # tr') ! n)). set_option x)"
    using n tr by(elim iterate_nthD)simp
  hence "Some (fst (tr' ! n)) \<in> results'_gpv (hybrid_step TYPE(_) sim (fst ((((state, s), \<sigma>) # tr') ! n)))"
    by(cases "tr' ! n")(auto simp add: run_hybrid_step_def split_def results_gpv_\<I>_full dest: oi_True.in_set_spmf_exec_gpv_into_results_gpv)
  then obtain S prot f s' s'' to
    where *: "Some (to, (S, prot, f), s') \<in> results'_gpv (hybrid_step1 TYPE(_) sim (fst ((((state, s), \<sigma>) # tr') ! n)))"
    and "Some s'' \<in> results'_gpv (info_sim sim s' (identifier (fst (snd (symb_state (fst ((((state, s), \<sigma>) # tr') ! n)))))) to)"
    and tr'n: "fst (tr' ! n) = ((S, prot, f), s'')"
    by(clarsimp simp add: hybrid_step_def gpv.set_map split: option.split_asm)
  hence "fst (fst (tr' ! n)) \<in> symb_step (fst (fst ((((state, s), \<sigma>) # tr') ! n)))"
  proof(cases "fst ((((state, s), \<sigma>) # tr') ! n)" rule: hybrid_step1_cases)
    case (Input S' ident succ f' \<sigma>''')
    from * tr'n Input obtain m
      where "Some (m, s') \<in> results'_gpv (gen_message sim \<sigma>''')" 
      and [simp]: "S' = S" "f = f' @ [m]" "succ = prot"
      and "\<T> m" -- \<open>unused\<close>
      by(clarsimp simp add: gpv.set_map Some_eq_bind_conv split: if_split_asm)
    note dy
    moreover have "n + 1 < length (state # map fst tr)" using n by simp 
    moreover have "(state # map fst tr) ! n = (S, Input ident succ, f')"
      using Input tr n by(auto simp add: nth_Cons' split_beta)
    moreover have "last (snd (snd ((state # map fst tr) ! (n + 1)))) = m"
      using Input tr n tr'n by(auto simp add: nth_Cons' split_beta split: if_split_asm)
    ultimately have "S \<turnstile> m" by- (rule dy_traceD)
    then show ?thesis using Input tr'n by(auto)
  qed(auto simp add: set_spmf_bind_pmf gpv.set_map split: computation.split_asm option.split_asm if_split_asm)
  then show "map fst tr ! n \<in> symb_step ((state # map fst tr) ! n)" using tr n
    by(auto simp add: nth_Cons' split_beta)
qed

lemma S_mono_hybrid_trace'_aux:
  "\<lbrakk> tr' \<in> set_spmf (SMC_iterate_spmf oracle sim (((S, prot, f), s), \<sigma>));
    (((S', protf'), s'), \<sigma>') \<in> set tr' \<rbrakk>
  \<Longrightarrow> S \<subseteq> S'"
proof(induction rule: SMC_iterate_spmf_induct_set_split)
  case (step tr S'' prot'' f'' s'' \<sigma>'' S prot f s \<sigma>)
  from step.prems show ?case unfolding list.set
  proof(cases rule: insertE[consumes 1, case_names insert rest])
    case insert
    with step.hyps show ?thesis
      by(auto simp add: run_hybrid_step_def Some_eq_map_option dest!: oi_True.in_set_spmf_exec_gpv_into_results_gpv hybrid_step_S_mono)
  next
    case rest
    from step.IH[OF this] step.hyps show ?thesis
      by(auto 4 3 simp add: run_hybrid_step_def Some_eq_map_option dest!: oi_True.in_set_spmf_exec_gpv_into_results_gpv hybrid_step_S_mono)
  qed
qed simp

lemma S_mono_hybrid_trace':
  "\<lbrakk> tr \<in> set_spmf (full_hybrid_trace' oracle sim (((S, prot, f), s), \<sigma>));
     ((S', protf'), \<sigma>') \<in> set tr \<rbrakk>
  \<Longrightarrow> S \<subseteq> S'"
by(auto 4 3 simp add: full_hybrid_trace'_def dest: S_mono_hybrid_trace'_aux)

lemma outputs_t_subset_S:
  assumes tr: "tr \<in> set_spmf (full_hybrid_trace' oracle sim ((state, s), \<sigma>))"
  and S'': "((S'', protf), \<sigma>'') \<in> set (drop n ((state, \<sigma>) # tr))"
  shows "outputs_t (map fst (take n ((state, \<sigma>) # tr))) \<subseteq> S''"
proof -
  from tr obtain tr' S prot f where tr: "tr = map (\<lambda>((state, s), \<sigma>). (state, \<sigma>)) tr'"
    and tr': "tr' \<in> set_spmf (SMC_iterate_spmf oracle sim ((state, s), \<sigma>))"
    and state: "state = (S, prot, f)"
    by(cases state; cases n)(auto 4 3 simp add: full_hybrid_trace'_def)
  from S'' tr state obtain s'' 
    where S'': "(((S'', protf), s''), \<sigma>'') \<in> set (drop n (((state, s), \<sigma>) # tr'))"
    by(cases n)(auto simp add: drop_map)
  show ?thesis using tr' S'' unfolding state tr
  proof(induction arbitrary: n rule: SMC_iterate_spmf_induct_set_split)
    case stop
    thus ?case by(cases n) simp_all
  next
    case (step tr S' prot' f' s' \<sigma>' S prot f s \<sigma>)
    show ?case
    proof(cases n)
      case 0
      thus ?thesis by simp
    next
      case (Suc n')
      from step.hyps have "output_t f prot \<subseteq> S'" unfolding state
        by(auto dest!: oi_True.in_set_spmf_exec_gpv_into_results_gpv output_t_subset_S simp add: run_hybrid_step_def Some_eq_map_option)
      also have *: "(((S'', protf), s''), \<sigma>'') \<in> set (drop n' ((((S', prot', f'), s'), \<sigma>') # tr))"
        using step.prems Suc by simp
      hence "(((S'', protf), s''), \<sigma>'') \<in> set ((((S', prot', f'), s'), \<sigma>') # tr)"
        by(auto dest: in_set_dropD)
      with step.hyps(2) have "S' \<subseteq> S''" by(auto elim: S_mono_hybrid_trace'_aux del: subsetI)
      finally show ?thesis using step.IH[OF *] Suc by simp
    qed
  qed
qed

lemma outputs_t_subset_S_nth:
  "\<lbrakk> tr \<in> set_spmf (full_hybrid_trace' oracle sim ((state, s), \<sigma>)); n < length tr \<rbrakk>
  \<Longrightarrow> outputs_t (map fst (take n ((state, \<sigma>) # tr))) \<subseteq> fst (fst (tr ! n))"
by(erule outputs_t_subset_S)(auto 4 3 intro: prod.expand intro!: exI[where x=1] simp add: in_set_conv_nth)

end

locale simulator = symbolic_model +
  constrains \<C> :: "'c \<Rightarrow> 't list \<Rightarrow> 't"
  and \<N> :: "'n \<Rightarrow> 't"
  and \<D> :: "'d \<Rightarrow> 't list \<Rightarrow> 't option"
  and \<T> :: "'t \<Rightarrow> bool"
  and deduce :: "'t set \<Rightarrow> 't \<Rightarrow> bool"
  fixes sim :: "('s, 'id, 't, 'call, 'ret) simulator"
  and R :: "security \<Rightarrow> ('s \<times> 's) set"
  assumes sim_R: "relation_for_sim' (fst (sim \<eta>)) (R \<eta>)"
  and wf_R: "wf (R \<eta>)"
begin

sublocale simulator' \<C> \<N> \<D> \<T> deduce "fst (sim \<eta>)" "R \<eta>" for \<eta>
by unfold_locales(rule sim_R wf_R)+

corollary dy_trace_symbolic:
  "\<lbrakk> tr \<in> set_spmf (full_hybrid_trace prot oracle sim \<eta>);
     dy_trace ((({}, prot, []) # map fst tr)) \<rbrakk>
  \<Longrightarrow> map fst tr \<in> full_traces prot"
by(rule dy_trace_symbolic_aux)(simp_all add: full_hybrid_trace_def split_beta)

end

type_synonym 'a distinguisher = "security \<Rightarrow> 'a \<Rightarrow> bool spmf"

definition distinguish_spmf_advantage :: "'a spmf \<Rightarrow> 'a spmf \<Rightarrow> ('a \<Rightarrow> bool spmf) \<Rightarrow> real"
where "distinguish_spmf_advantage p q d = \<bar>spmf (TRY p \<bind> d ELSE coin_spmf) True - spmf (TRY q \<bind> d ELSE coin_spmf) True\<bar>"

lemma distinguish_spmf_advantage_nonneg: "0 \<le> distinguish_spmf_advantage p q d"
by(simp add: distinguish_spmf_advantage_def)

inductive indistinguishable_spmf :: "(security \<Rightarrow> 'a spmf) \<Rightarrow> (security \<Rightarrow> 'a spmf) \<Rightarrow> 'a distinguisher \<Rightarrow> bool"
  for p q d
where
  indistinguishable_spmf:
  "\<lbrakk> negligible (\<lambda>\<eta>. distinguish_spmf_advantage (p \<eta>) (q \<eta>) (d \<eta>));
     negligible (\<lambda>\<eta>. weight_spmf (p \<eta>) - weight_spmf (q \<eta>)) \<rbrakk>
  \<Longrightarrow> indistinguishable_spmf p q d"

lemmas indistinguishable_spmfI = indistinguishable_spmf
hide_fact (open) indistinguishable_spmf

lemma indistinguishable_spmfD:
  fixes p q :: "security \<Rightarrow> _ spmf" and d :: "security \<Rightarrow> _ \<Rightarrow> _ spmf"
  assumes "indistinguishable_spmf p q d"
  shows indistinguishable_spmfD_distinguishable_spmf_advantage: 
  "negligible (\<lambda>\<eta>. spmf (try_spmf (p \<eta> \<bind> d \<eta>) coin_spmf) True - spmf (try_spmf (q \<eta> \<bind> d \<eta>) coin_spmf) True)"
  and indistinguishable_spmfD_weight: "negligible (\<lambda>\<eta>. weight_spmf (p \<eta>) - weight_spmf (q \<eta>))"
using assms by(auto elim!: indistinguishable_spmf.cases simp add: distinguish_spmf_advantage_def)

lemma spmf_True_conv_False: "spmf p True = weight_spmf p - spmf p False" (* TODO: Generalise in SPMF *)
proof -
  have "spmf p False + spmf p True = measure (measure_spmf p) {False} + measure (measure_spmf p) {True}"
    by(simp add: spmf_conv_measure_spmf)
  also have "\<dots> = measure (measure_spmf p) ({False} \<union> {True})"
    by(subst measure_spmf.finite_measure_Union) simp_all
  also have "{False} \<union> {True} = space (measure_spmf p)" by(auto simp add: space_measure_spmf)
  finally show ?thesis by simp
qed

lemma spmf_False_conv_True: "spmf p False = weight_spmf p - spmf p True" (* TODO: Generalise in SPMF *)
by(simp add: spmf_True_conv_False)

lemma weight_bind_spmf_lossless2: (* Move to SPMF *)
  "(\<And>x. x \<in> set_spmf p \<Longrightarrow> lossless_spmf (f x)) \<Longrightarrow> weight_spmf (p \<bind> f) = weight_spmf p"
unfolding weight_bind_spmf o_def
by(subst integral_cong_AE)(auto simp add: lossless_weight_spmfD)

lemma indistinguishable_spmfI_False:
  assumes negl: "negligible (\<lambda>\<eta>. spmf (try_spmf (p \<eta> \<bind> d \<eta>) coin_spmf) False - spmf (try_spmf (q \<eta> \<bind> d \<eta>) coin_spmf) False)"
  and similar: "negligible (\<lambda>\<eta>. weight_spmf (p \<eta>) - weight_spmf (q \<eta>))"
  shows "indistinguishable_spmf p q d"
proof
  from negl[THEN negligible_uminus[THEN iffD2]]
  show "negligible (\<lambda>\<eta>. distinguish_spmf_advantage (p \<eta>) (q \<eta>) (d \<eta>))"
    by(simp del: negligible_uminus add: distinguish_spmf_advantage_def spmf_False_conv_True lossless_weight_spmfD)
qed(rule similar)

lemma indistinguishable_spmfD_False:
  assumes ind: "indistinguishable_spmf p q d"
  shows "negligible (\<lambda>\<eta>. spmf (try_spmf (p \<eta> \<bind> d \<eta>) coin_spmf) False - spmf (try_spmf (q \<eta> \<bind> d \<eta>) coin_spmf) False)"
proof -
  have "negligible (\<lambda>\<eta>. - (spmf (TRY p \<eta> \<bind> d \<eta> ELSE coin_spmf) True - spmf (TRY q \<eta> \<bind> d \<eta> ELSE coin_spmf) True))"
    using ind[THEN indistinguishable_spmfD_distinguishable_spmf_advantage] by(simp only: negligible_uminus)
  then show ?thesis by(simp add: spmf_False_conv_True lossless_weight_spmfD)
qed

context symbolic_model begin

inductive dy_simulator
  :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> ('\<sigma>, 'call, 'ret) oracle
  \<Rightarrow> ('s, 'id, 't, 'call, 'ret) simulator \<Rightarrow> bool"
  for prot "oracle" sim
where
  dy_simulator:
  "negligible (\<lambda>\<eta>. spmf (map_spmf (\<lambda>tr. dy_trace (({}, prot, []) # map fst tr)) (full_hybrid_trace prot oracle sim \<eta>)) False)
  \<Longrightarrow> dy_simulator prot oracle sim"

lemmas dy_simulatorI = dy_simulator

end

context computational_implementation begin

inductive indistinguishable_simulator
  :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> 's1 adversary \<Rightarrow> ('\<sigma>, 'call, 'ret) oracle
  \<Rightarrow> ('s2, 'id, 't, 'call, 'ret) simulator \<Rightarrow> bound \<Rightarrow> 'id list distinguisher \<Rightarrow> bool"
  for prot adv "oracle" sim bound d
where indistinguishable_simulator:
  "indistinguishable_spmf (comp_trace prot adv bound) (hybrid_trace prot oracle sim) d
  \<Longrightarrow> indistinguishable_simulator prot adv oracle sim bound d"
lemmas indistinguishable_simulatorI = indistinguishable_simulator

inductive good_simulator
  :: "('id, 'c, 'n, 'd) protocol \<Rightarrow> 's1 adversary \<Rightarrow> ('\<sigma>, 'call, 'ret) oracle
  \<Rightarrow> ('s2, 'id, 't, 'call, 'ret) simulator \<Rightarrow> bound \<Rightarrow> 'id list distinguisher \<Rightarrow> bool"
  for prot adv "oracle" sim bound d
where
  good_simulator:
  "\<lbrakk> dy_simulator prot oracle sim; indistinguishable_simulator prot adv oracle sim bound d \<rbrakk>
  \<Longrightarrow> good_simulator prot adv oracle sim bound d"
lemmas good_simulatorI = good_simulator

end


locale computational_simulator = computational_implementation + simulator +
  constrains \<C> :: "'c \<Rightarrow> 't list \<Rightarrow> 't"
  and \<N> :: "'n \<Rightarrow> 't"
  and \<D> :: "'d \<Rightarrow> 't list \<Rightarrow> 't option"
  and \<T> :: "'t \<Rightarrow> bool"
  and deduce :: "'t set \<Rightarrow> 't \<Rightarrow> bool"
  and A_C :: "'c  \<Rightarrow> security \<Rightarrow> bitstring list \<Rightarrow> bitstring option"
  and A_N :: "'n \<Rightarrow> security \<Rightarrow> bitstring pmf"
  and A_D :: "'d \<Rightarrow> security \<Rightarrow> bitstring list \<Rightarrow> bitstring option"
  and sim :: "('s, 'id, 't, 'call, 'ret) simulator"
  and R :: "security \<Rightarrow> ('s \<times> 's) set"
begin

lemma good_simulator_sound:
  fixes tprop :: "'id trace_prop"
  defines "d \<equiv> \<lambda>_ tr. return_spmf (tprop tr)"
  assumes good: "good_simulator prot adv oracle sim bound d"
  and "trace_prop tprop" -- \<open>unused\<close>
  and holds: "prot \<Turnstile> tprop"
  shows "prot \<Turnstile> tprop / adv, bound"
proof
  from good obtain dy: "dy_simulator prot oracle sim"
    and ind_sim: "indistinguishable_simulator prot adv oracle sim bound d" ..
  obtain dy_neg: "negligible (\<lambda>\<eta>. spmf (map_spmf (\<lambda>tr. dy_trace (({}, prot, []) # map fst tr)) (full_hybrid_trace prot oracle sim \<eta>)) False)"
    (is "negligible ?dy_hyb") using dy ..
  then have "negligible (\<lambda>\<eta>. spmf (map_spmf tprop (hybrid_trace prot oracle sim \<eta>)) False)" (is "negligible ?tprop_hyb")
  proof(rule negligible_mono)
    have "ennreal (?tprop_hyb \<eta>) \<le> ?dy_hyb \<eta>" for \<eta>
      unfolding ennreal_spmf_map_conv_nn_integral hybrid_trace_def nn_integral_map_spmf o_def
    proof(rule nn_integral_mono_AE; unfold AE_measure_spmf_iff; intro strip)
      fix tr
      assume tr: "tr \<in> set_spmf (full_hybrid_trace prot oracle sim \<eta>)"
      { assume "\<not> tprop (map (\<lambda>((_, prot, _), _). identifier prot) tr)"
        hence "\<not> dy_trace (({}, prot, []) # map fst tr)"
        proof(rule contrapos_nn)
          assume "dy_trace (({}, prot, []) # map fst tr)"
          with tr have "map fst tr \<in> full_traces prot" by(rule dy_trace_symbolic)
          hence "map (\<lambda>(_, prot, _). identifier prot) (map fst tr) \<in> node_traces prot"
            by(simp add: node_traces_def del: map_map)
          with holds show "tprop (map (\<lambda>((_, prot, _), _). identifier prot) tr)"
            unfolding holds_def map_map o_def split_def by blast
        qed }
      then show "indicator (tprop -` {False}) (map (\<lambda>((_, prot, _), _). identifier prot) tr)
         \<le> (indicator ((\<lambda>tr. dy_trace (({}, prot, []) # map fst tr)) -` {False}) tr :: ennreal)"
        by(simp split: split_indicator)
    qed
    then show "?tprop_hyb \<in> O(?dy_hyb)" by(intro bigoI[where c=1])(simp add: pmf_nonneg)
  qed
  moreover
  have "negligible (\<lambda>\<eta>. spmf (TRY comp_trace prot adv bound \<eta> \<bind> d \<eta> ELSE coin_spmf) False - spmf (TRY hybrid_trace prot oracle sim \<eta> \<bind> d \<eta> ELSE coin_spmf) False)"
  proof(rule indistinguishable_spmfD_False)
    from ind_sim show "indistinguishable_spmf (comp_trace prot adv bound) (hybrid_trace prot oracle sim) d" ..
  qed
  moreover have "negligible (\<lambda>\<eta>. 1 / 2 * (weight_spmf (comp_trace prot adv bound \<eta>) - weight_spmf (hybrid_trace prot oracle sim \<eta>)))"
    using ind_sim by cases (rule indistinguishable_spmfD_weight negligible_cmultI)+
  ultimately have "negligible (\<lambda>\<eta>. spmf (map_spmf tprop (hybrid_trace prot oracle sim \<eta>)) False + (spmf (TRY comp_trace prot adv bound \<eta> \<bind> d \<eta> ELSE coin_spmf) False - spmf (TRY hybrid_trace prot oracle sim \<eta> \<bind> d \<eta> ELSE coin_spmf) False) + \<dots> \<eta>)"
    by(intro negligible_plus)
  also have "\<dots> = (\<lambda>\<eta>. spmf (map_spmf tprop (comp_trace prot adv bound \<eta>)) False)"
    by(simp add: d_def fun_eq_iff map_spmf_conv_bind_spmf[symmetric] spmf_try_spmf pmf_None_eq_weight_spmf spmf_of_set diff_divide_distrib)
  finally show "negligible \<dots>" .
qed

end

end