theory CoSP_PKES imports
  "$AFP/CryptHOL/GPV_Bisim"
  CoSP_PKES_Simulator_Eq
begin

subsection \<open>Move upwards\<close>

text \<open>Experiment with a Hoare-style calculus for the state monad transformer \<close>

definition hoare_state :: "('s \<Rightarrow> bool) \<Rightarrow> ('s, 'm) stateT \<Rightarrow> ('m \<Rightarrow> bool) \<Rightarrow> bool" ("(\<lbrace> _ \<rbrace>)/ _/ (\<lbrace> _ \<rbrace>)" [0, 0, 0] 100)
where "hoare_state P m Q \<longleftrightarrow> (\<forall>s. P s \<longrightarrow> Q (run_state m s))"

lemma hoare_stateI [intro?]: "(\<And>s. P s \<Longrightarrow> Q (run_state m s)) \<Longrightarrow> \<lbrace>P\<rbrace> m \<lbrace>Q\<rbrace>"
by(simp add: hoare_state_def)

lemma hoare_stateD: "\<lbrakk> \<lbrace> P \<rbrace> m \<lbrace> Q \<rbrace>; P s \<rbrakk> \<Longrightarrow> Q (run_state m s)"
by(simp add: hoare_state_def)

lemma hoare_state_returnI: "(\<And>s. P s \<Longrightarrow> Q (return (x, s))) \<Longrightarrow> \<lbrace> P \<rbrace> return_state return x \<lbrace> Q \<rbrace>" for return
by(rule hoare_stateI) simp

lemma hoare_state_bindI:
  fixes return bind
  assumes x: "\<lbrace> P \<rbrace> x \<lbrace> R \<rbrace>"
  and f: "\<And>x. \<lbrace> \<lambda>s. R (return (x, s)) \<rbrace> f x \<lbrace> Q \<rbrace>"
  and bind: "\<And>x f. \<lbrakk> R x; \<And>a. R (return a) \<Longrightarrow> Q (f a) \<rbrakk> \<Longrightarrow> Q (bind x f)"
  shows "\<lbrace> P \<rbrace> bind_state bind x f \<lbrace> Q \<rbrace>"
apply(rule hoare_stateI)
apply simp
apply(rule bind)
 apply(erule hoare_stateD[OF x])
apply clarsimp
apply(erule hoare_stateD[OF f])
done

definition get_state :: "('s \<Rightarrow> ('s, 'm) stateT) \<Rightarrow> ('s, 'm) stateT"
where "get_state f = StateT (\<lambda>s. run_state (f s) s)"

lemma hoare_state_getI:
  assumes "\<And>s. P s \<Longrightarrow> \<lbrace> \<lambda>s'. s' = s \<rbrace> f s \<lbrace> Q \<rbrace>"
  shows "\<lbrace> P \<rbrace> get_state f \<lbrace> Q \<rbrace>"
apply(rule hoare_stateI)
apply(simp add: get_state_def)
apply(drule assms)
apply(erule hoare_stateD)
apply simp
done

definition put_state :: "'s \<Rightarrow> ('s, 'm) stateT \<Rightarrow> ('s, 'm) stateT"
where "put_state s m = StateT (\<lambda>_. run_state m s)"

lemma hoare_state_putI:
  assumes "\<And>s'. P s' \<Longrightarrow> \<lbrace> \<lambda>s'. s' = s \<rbrace> m \<lbrace> Q \<rbrace>"
  shows "\<lbrace> P \<rbrace> put_state s m \<lbrace> Q \<rbrace>"
apply(rule hoare_stateI)
apply(simp add: put_state_def)
apply(erule hoare_stateD[OF assms])
apply simp
done

lemma pred_gpv_bindI:
  "\<lbrakk> pred_gpv Q R x; \<And>a. Q a \<Longrightarrow> pred_gpv P R (f a) \<rbrakk> \<Longrightarrow> pred_gpv P R (bind_gpv x f)"
by(auto simp add: pred_gpv_def outs_bind_gpv)

notepad begin -- \<open>This does not really scale, does it?\<close>
  have "\<lbrace> even \<rbrace>
         bind_state bind_gpv (get_state (return_state Done))
         (\<lambda>n. put_state (n + 1) (return_state Done (n mod 2)))
        \<lbrace> pred_gpv (\<lambda>(x :: nat, s). x = 0 \<and> odd s) bot \<rbrace>"
    apply(rule hoare_state_bindI[where R="pred_gpv (pred_prod even top) bot" and return=Done])
      apply(rule hoare_state_getI)
      apply(rule hoare_state_returnI)
      apply simp
     apply simp
     apply(rule hoare_state_putI)
     apply(rule hoare_state_returnI)
     apply(simp)
    apply(clarsimp simp del: pred_gpv_bind)
    apply(rule pred_gpv_bindI)
     apply assumption
    apply auto
    done

end


end