theory CoSP_PKES_Faking imports
  CoSP_PKES_Simulator2
  "$AFP/CryptHOL/GPV_Bisim"
begin

lemma spmf_bind_spmf_of_set: "spmf (bind_spmf (spmf_of_set A) f) y = (\<Sum>x\<in>A. spmf (f x) y) / card A"
by(simp add: spmf_bind integral_spmf_of_set)

primrec intercept_right :: "('call2, 'ret2, 's2) callee
  \<Rightarrow> 's2 \<Rightarrow> 'call1 + 'call2 \<Rightarrow> (('ret1 + 'ret2) \<times> 's2, 'call1, 'ret1) gpv"
where
  "intercept_right callee2 s2 (Inl out1) = Pause out1 (\<lambda>input. Done (Inl input, s2))"
| "intercept_right callee2 s2 (Inr out2) = lift_spmf (map_spmf (apfst Inr) (callee2 s2 out2))"

lemma intercept_right_parametric [transfer_rule]:
  includes lifting_syntax shows
  "((S2 ===> CALL2 ===> rel_spmf (rel_prod op = S2)) ===> S2 ===> rel_sum CALL1 CALL2
  ===> rel_gpv (rel_prod op = S2) CALL1)
  intercept_right intercept_right"
unfolding intercept_right_def by(fold relator_eq)transfer_prover

lemma exec_gpv_intercept_right [simp]:
  "exec_gpv callee1 (intercept_right callee2 s2 c) s1 = 
   map_spmf (\<lambda>(res, s1', s2'). ((res, s2'), s1')) (parallel_oracle callee1 callee2 (s1, s2) c)"
by(cases c)(simp_all add: split_def map_spmf_conv_bind_spmf[symmetric] spmf.map_comp o_def apfst_def map_prod_def)

lemma exec_gpv_state_iso:
  assumes "\<And>s. g (f s) = s" "\<And>s. f (g s) = s"
  shows
  "exec_gpv (\<lambda>s c. map_spmf (apsnd g) (callee (f s) c)) = (\<lambda>gpv s.
   map_spmf (apsnd g) (exec_gpv callee gpv (f s)))"
using exec_gpv_parametric[of "BNF_Def.Grp UNIV g" "BNF_Def.Grp UNIV id" "op =", unfolded eq_alt]
unfolding prod.rel_Grp option.rel_Grp pmf.rel_Grp
unfolding eq_alt[symmetric] gpv.rel_eq
apply(simp add: rel_fun_def Grp_def apsnd_def)
apply(rule ext)+
apply(erule allE)+
apply(erule impE)
prefer 2
apply(erule allE)+
apply(drule sym)
apply(rule sym)
apply(erule trans)
apply(unfold assms)
apply(rule refl)
apply(unfold assms)
apply simp
  done

lemma run_gpv_iterate:
  fixes callee exec_gpv2 and K :: "'s \<Rightarrow> ('s option, 'out, 'ret) gpv" and s :: "'\<sigma>"
  defines "exec_gpv2 \<equiv> exec_gpv callee"
  shows "run_gpv callee (iterate_gpv K x) s = 
   map_spmf (map fst) (iterate_spmf (\<lambda>(x, s). map_spmf (\<lambda>(xo, s'). map_option (\<lambda>x'. (x', s')) xo) (exec_gpv2 (K x) s)) (x, s))"
  (is "?lhs = ?rhs")
proof(rule spmf.leq_antisym)
  note [[show_variants]]
  note [simp del] = last.simps
    and [cong del] = option.case_cong_weak sum.case_cong_weak
  
  def exec_gpv3 \<equiv> "exec_gpv :: ('\<sigma> \<Rightarrow> 'out \<Rightarrow> ('ret \<times> '\<sigma>) spmf) \<Rightarrow> ('s list, 'out, 'ret) gpv \<Rightarrow> _"
  have "ord_spmf op = ?lhs ?rhs"
    and "exec_gpv callee (map_gpv g id gpv) s = map_spmf (apfst g) (exec_gpv callee gpv s)"
    -- \<open>Stupid! We again have to do yet another induction to show that @{const exec_gpv}
        commutes with bind and map! And we must use @{typ "'s list \<Rightarrow> 's list"} instead of
        @{typ "'s option => 's list"} here, because otherwise we'd have different types for @{const exec_gpv}. :-(\<close>
    and "ord_spmf op = (exec_gpv callee (bind_gpv gpv f) s)
         (bind_spmf (exec_gpv3 callee gpv s) (\<lambda>(x, s') \<Rightarrow> exec_gpv callee (f x) s'))"
    for gpv :: "('s list, 'out, 'ret) gpv" and f :: "'s list \<Rightarrow> ('s list, 'out, 'ret) gpv" and g :: "'s list \<Rightarrow> 's list"
  proof(induction arbitrary: x s gpv f g rule: exec_gpv_fixp_induct_strong2)
    case adm show ?case
      apply(rule cont_intro)+
      subgoal for a b c d by(cases d; clarsimp)
      done
    case bottom case 1 show ?case by simp
    case bottom case 2 show ?case by simp
    case bottom case 3 show ?case by simp
  next
    case exec_gpv: (step exec_gpv') case 1
    have search: "ord_spmf op = (bind_spmf (iterate_search K x) (\<lambda>xo. case xo of Inl x \<Rightarrow> return_spmf x
          | Inr (xs, out, rpv) \<Rightarrow> bind_spmf (callee s out)
                  (\<lambda>(input, s'). map_spmf fst (exec_gpv' (bind_gpv (rpv input) (\<lambda>so. case so of None \<Rightarrow> Done xs | Some s' \<Rightarrow> map_gpv (op @ (xs @ [s'])) id (iterate_gpv K s'))) s'))))
      (map_spmf (map fst) (iterate_spmf (\<lambda>(x, s). map_spmf (\<lambda>(xo, s'). map_option (\<lambda>x'. (x', s')) xo) (exec_gpv callee (K x) s)) (x, s)))"
    proof(induction arbitrary: x s rule: iterate_search_fixp_induct)
      case adm show ?case by simp
      case bottom show ?case by simp
    next
      case search: (step iterate_search')
      show ?case
        apply(simp add: bind_spmf_pmf_assoc)
        apply(subst iterate_spmf.simps)
        apply simp
        apply(subst exec_gpv.simps)
        apply(simp add: bind_map_spmf map_spmf_bind_spmf o_def)
        apply(rule ord_spmf_bind_reflI)
        apply(clarsimp split: generat.split)
        apply(intro conjI strip)
         apply(clarsimp split: option.split)
         apply(clarsimp simp add: bind_map_spmf o_def case_sum_map_sum spmf.map_comp)
         subgoal for x' using search.IH[of x' s, THEN monotoneD[OF monotone_map_spmf], where ?f1="op # x'"]
           apply(simp add: map_spmf_bind_spmf o_def sum.case_distrib[where h="map_spmf _"] split_def)
           apply(subst (asm) spmf.map_comp)
           apply(subst (asm) fst_comp_apfst[symmetric])
           apply(subst (asm) spmf.map_comp[symmetric])
           apply(subst (asm) exec_gpv.IH(2)[symmetric])
           apply(simp add: map_gpv_id_bind_gpv o_def option.case_distrib[where h="map_gpv _ _"] gpv.map_comp id_def[symmetric])
           apply(simp cong: gpv.map_cong)
           apply(erule spmf.leq_trans)
           apply(simp add: spmf.map_comp o_def)
           done
         apply(rule ord_spmf_bind_reflI)
         apply clarsimp
         subgoal for out rpv input \<sigma>'
           using exec_gpv.IH(3)[of "map_gpv list_of_option id (rpv input)" "\<lambda>xs. case xs of [] \<Rightarrow> Done [] | s' # _ \<Rightarrow> map_gpv (op # s') id (iterate_gpv K s')" \<sigma>', THEN monotoneD[OF monotone_map_spmf], where ?f1="fst"]
           apply(simp add: bind_map_gpv o_def id_def[symmetric])
           apply(erule spmf.leq_trans)
           apply(simp add: exec_gpv3_def exec_gpv_map_gpv_id bind_map_spmf map_spmf_bind_spmf o_def)
           apply(rule ord_spmf_bind_reflI)
           apply(clarsimp split!: option.split)
           subgoal for s' using exec_gpv.hyps(2)[of "Done []" s'] by(auto simp add: ord_spmf_return_spmf2)
           apply(subst exec_gpv.IH(2))
           apply(subst spmf.map_comp)
           apply(subst fst_comp_apfst)
           apply(subst spmf.map_comp[symmetric])
           apply(rule spmf.leq_trans)
           apply(rule monotone_map_spmf[THEN monotoneD])
           apply(rule exec_gpv.IH(1))
           apply(simp add: spmf.map_comp o_def exec_gpv2_def)
           done
         done
     qed
      
    show ?case using search
      apply(rewrite iterate_gpv_sel)
      apply(simp add: map_spmf_bind_spmf exec_gpv2_def bind_map_spmf o_def spmf.map_comp
          sum.case_distrib[where h="\<lambda>x. bind_spmf x _"] prod.case_distrib[where h="\<lambda>x. bind_spmf x _"]
          sum.case_distrib[where h="map_spmf _"] prod.case_distrib[where h="map_spmf _"] option.case_distrib[where h="map_spmf _"]
          prod.case_distrib[where h="case_option _ _"]
          cong del: sum.case_cong_weak option.case_cong_weak)
      done
  next
    case step case 3 show ?case unfolding exec_gpv3_def
      apply(rewrite exec_gpv.simps)
      apply(simp del: bind_gpv_sel' add: bind_gpv_sel)
      apply(rule ord_spmf_bind_reflI)
      apply(clarsimp split: generat.split)
      apply(rule ord_spmf_bind_reflI)
      apply clarsimp
      apply(rule spmf.leq_trans)
       apply(rule step.IH)
      apply(simp add: exec_gpv3_def)
      apply(rule ord_spmf_bind_reflI)
      apply clarsimp
      apply(rule step.hyps)
      done
  next
    case step case 2 show ?case
      apply(simp add: gpv.map_sel map_spmf_bind_spmf bind_map_spmf o_def del: id_apply)
      apply(rule bind_spmf_cong[OF refl])
      apply(clarsimp split: generat.split simp add: map_spmf_bind_spmf o_def)
      apply(rule bind_spmf_cong[OF refl])
      apply clarsimp
      apply(rule step.IH)
      done
  qed
  then show "ord_spmf op = ?lhs ?rhs" by -
  
  show "ord_spmf op = ?rhs ?lhs" unfolding exec_gpv2_def
  proof(induction arbitrary: x s rule: iterate_spmf_fixp_induct)
    case adm show ?case by simp
    case bottom show ?case by simp
    case iterate_spmf: (step iterate_spmf')

    show ?case
      apply(simp add: bind_map_spmf map_spmf_bind_spmf)
      apply(subst iterate_gpv_unroll)
      apply(simp add: exec_gpv_bind map_spmf_bind_spmf)
      apply(rule ord_spmf_bind_reflI)
      apply(clarsimp simp add: o_def)
      subgoal for so \<sigma>'
        apply(cases so)
        subgoal by clarsimp
        subgoal for s'
          apply(simp add: exec_gpv_map)
          using iterate_spmf.IH(1)[of s' \<sigma>', THEN monotoneD[OF monotone_map_spmf], where ?f1="op # s'"]
          apply(simp add: spmf.map_comp o_def)
          done
        done
      done
  qed
qed

lemma numeral_mult_abs: 
  fixes x :: "'a :: linordered_idom" shows "numeral n * abs x = abs (numeral n * x)"
  by(simp add: abs_mult)

lemma inj_Not [simp]: "inj_on Not A"
  by(simp add: inj_on_def)

lemma distinguisher_indistinguishable:
  "\<bar>spmf (TRY f True ELSE coin_spmf) True - spmf (TRY f False ELSE coin_spmf) True\<bar> =
  2 * \<bar>spmf (TRY bind_spmf coin_spmf (\<lambda>b. bind_spmf (f b) (\<lambda>b'. return_spmf (b = b'))) ELSE coin_spmf) True - 1 / 2\<bar>"
  (is "?lhs = ?rhs")
proof -
  have "?rhs = \<bar>spmf (TRY f True ELSE coin_spmf) True + spmf (map_spmf Not (TRY f False ELSE coin_spmf)) (\<not> False) - 1\<bar>"
    by(simp add: try_spmf_bind_out[symmetric] spmf_bind integral_spmf_of_set UNIV_bool map_try_spmf insert_commute numeral_mult_abs add_divide_distrib)
      (simp add: map_spmf_conv_bind_spmf add.commute)
  also have "spmf (TRY f True ELSE coin_spmf) True = spmf (f True) True + (1 - weight_spmf (f True)) / 2" (is "_ = ?True")
    by(simp add: spmf_try_spmf pmf_None_eq_weight_spmf spmf_of_set)
  also have "spmf (map_spmf Not (TRY f False ELSE coin_spmf)) (\<not> False) = spmf (f False) False + (1 - weight_spmf (f False)) / 2"
    by(rewrite spmf_map_inj')(simp_all add: spmf_try_spmf pmf_None_eq_weight_spmf spmf_of_set)
  also have "spmf (f False) False = weight_spmf (f False) - spmf (f False) True" by(simp add: spmf_False_conv_True)
  also have "\<bar>?True + (\<dots> + (1 - weight_spmf (f False)) / 2) - 1\<bar> = ?lhs"
    by(simp add: algebra_simps spmf_try_spmf pmf_None_eq_weight_spmf diff_divide_distrib add_divide_distrib spmf_of_set)
  finally show ?thesis by simp
qed

context comp_impl_sims begin

definition faking_reduction
  :: "nonce_typing \<Rightarrow> 'id cosp_protocol \<Rightarrow> 's adversary \<Rightarrow> (nat \<Rightarrow> nat) \<Rightarrow> ('id list \<Rightarrow> bool spmf) 
  \<Rightarrow> (protocol_nonce, bitstring, bitstring, bitstring) ind_cca2.adversary\<^sub>n"
where
  "faking_reduction \<Gamma> prot \<A> bound d \<eta> = (let (\<A>', \<sigma>) = \<A> \<eta> in do {
     (trace, oss) \<leftarrow> inline (intercept_right (suf_cma.oracle\<^sub>n \<eta>))
        (iterate_gpv (hybrid_step_sim2 \<Gamma> \<A>' \<eta>) (({}, prot, []), sim2.initial () \<sigma> (bound \<eta>))) (\<lambda>_. None);
     lift_spmf (d (map (\<lambda>((_, prot, _), _). identifier prot) trace))
   })"
  
lemma faking_advantage:
  "distinguish_spmf_advantage (hybrid_trace2 prot True \<Gamma> \<A> bound \<eta>) (hybrid_trace2 prot False \<Gamma> \<A> bound \<eta>) d =
   2 * ind_cca2.advantage\<^sub>n (faking_reduction \<Gamma> prot \<A> bound d) \<eta>"
  (is "?lhs = ?rhs")
proof -
  obtain \<A>' \<sigma> where \<A>: "\<A> \<eta> = (\<A>', \<sigma>)" by(cases "\<A> \<eta>")
  have "?lhs = \<bar>spmf (TRY hybrid_trace2 prot True \<Gamma> \<A> bound \<eta> \<bind> d ELSE coin_spmf) True - spmf (TRY hybrid_trace2 prot False \<Gamma> \<A> bound \<eta> \<bind> d ELSE coin_spmf) True\<bar>"
    by(simp add: distinguish_spmf_advantage_def)
  also have "\<dots> = 2 * \<bar>spmf (TRY coin_spmf \<bind> (\<lambda>b. (hybrid_trace2 prot b \<Gamma> \<A> bound \<eta> \<bind> d) \<bind> (\<lambda>b'. return_spmf (b \<longleftrightarrow> b'))) ELSE coin_spmf) True - 1 / 2\<bar>"
    by(rule distinguisher_indistinguishable)
  also have "hybrid_trace2 prot b \<Gamma> \<A> bound \<eta> \<bind> d = exec_gpv (ind_cca2.oracle\<^sub>n b \<eta>) (faking_reduction \<Gamma> prot \<A> bound d \<eta>) Map.empty \<bind> (\<lambda>(guess, \<sigma>). return_spmf guess)" for b
  proof -
    have "hybrid_trace2 prot b \<Gamma> \<A> bound \<eta> \<bind> d = bind_spmf (run_gpv (oracle\<^sub>n' b \<eta>) (iterate_gpv (hybrid_step_sim2 \<Gamma> \<A>' \<eta>) (({}, prot, []), sim2.initial () \<sigma> (bound \<eta>))) (empty, empty)) (\<lambda>x. d (map (\<lambda>x. identifier (fst (snd (symb_state x)))) x))"
      by(subst run_gpv_iterate)(simp add: hybrid_trace_def full_hybrid_trace_def sim2.sim_def \<A> oracle_def full_hybrid_trace'_def spmf.map_comp o_def split_beta bind_map_spmf run_hybrid_step_def)
    also have "\<dots> = bind_spmf (exec_gpv (oracle\<^sub>n' b \<eta>) (iterate_gpv (hybrid_step_sim2 \<Gamma> \<A>' \<eta>) (({}, prot, []), sim2.initial () \<sigma> (bound \<eta>))) (empty, empty))
                (\<lambda>x. d (map (\<lambda>((_, prot, _), _). identifier prot) (fst x)))"
      by(simp add: split_def bind_map_spmf o_def)
    also have "\<dots> = exec_gpv (ind_cca2.oracle\<^sub>n b \<eta>) (faking_reduction \<Gamma> prot \<A> bound d \<eta>) Map.empty \<bind> (\<lambda>(guess, \<sigma>). return_spmf guess)"
      apply(subst faking_reduction_def)
      apply(simp add: \<A>)
      apply(subst exec_gpv_bind)
      apply(subst exec_gpv_inline)
      apply simp
      apply(subst spmf.map_comp)
      apply(simp add: o_def)
      apply(simp add: split_beta)
      apply(fold oracle\<^sub>n_def)
      apply(subst (6) split_def)
      apply(subst exec_gpv_state_iso[unfolded apsnd_def map_prod_def split_def id_apply]; simp)
      apply(simp add: spmf.map_comp o_def bind_map_spmf)
      done
    finally show ?thesis .
  qed
  ultimately show "?lhs = ?rhs"
    by(simp add: ind_cca2.advantage\<^sub>n_def ind_cca2.ind_cca2\<^sub>n_def split_def eq_commute)
qed

end

end