theory CoSP_PKES_Key_Safety_Simulator imports
  CoSP_PKES_Key_Safety
  CoSP_PKES_Simulator_Gen
begin

context comp_impl_\<beta>_base begin

lemma  random_usage_\<beta>_rec_mono:
  "\<beta>_rec t t' \<Longrightarrow> random_usage t' \<le> random_usage t"
by(induction t)(auto simp add: \<beta>_rec_E \<beta>_rec_SIG sup.coboundedI1 sup.coboundedI2)

end

context comp_impl_base begin

lemma random_usage_parse_string: "parse_string \<eta> m = Some t \<Longrightarrow> random_usage t = \<bottom>"
by(induction m t rule: parse_string_induct) simp_all

lemma key_safe_parse_string: "parse_string \<eta> m = Some t \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t t : PublicT"
by(induction m t rule: parse_string_induct) simp_all

lemma \<beta>1_pnonces_parse_string [simp]: "parse_string \<eta> m = Some t \<Longrightarrow> \<beta>1_pnonces t = {}"
by(induction m t rule: parse_string_induct)(simp_all)

lemma \<beta>1_anonces_parse_string [simp]: "parse_string \<eta> m = Some t \<Longrightarrow> \<beta>1_anonces t = {}"
by(induction m t rule: parse_string_induct)(simp_all)

end

definition wf_rand_history :: "random_usage \<Rightarrow> history \<Rightarrow> bool"
where "wf_rand_history ru hist \<longleftrightarrow> (\<forall>t\<in>dom hist. random_usage t \<le> ru)"

lemma wf_rand_history_cong:
  "\<lbrakk> ru = ru'; dom hist = dom hist' \<rbrakk> \<Longrightarrow> wf_rand_history ru hist = wf_rand_history ru' hist'"
by(simp add: wf_rand_history_def)

lemma wf_rand_history_mono:
  "\<lbrakk> wf_rand_history ru hist; ru \<le> ru' \<rbrakk> \<Longrightarrow> wf_rand_history ru' hist"
by(auto simp add: wf_rand_history_def dom_def dest: le_funD)

lemma wf_rand_history_mono2:
  "\<lbrakk> wf_rand_history ru hist; dom hist' \<subseteq> insert t (dom hist); random_usage t \<le> ru \<rbrakk>
  \<Longrightarrow> wf_rand_history ru hist'"
by(auto simp add: wf_rand_history_def dest: le_funD iff del: domIff)

lemma wf_rand_historyD:
  "\<lbrakk> wf_rand_history ru hist; t \<in> dom hist \<rbrakk> \<Longrightarrow> random_usage t \<le> ru"
by(simp add: wf_rand_history_def)

lemma wf_rand_historyD':
  "\<lbrakk> wf_rand_history ru hist; hist t = Some m \<rbrakk> \<Longrightarrow> random_usage t \<le> ru"
by(auto simp add: wf_rand_history_def)

lemma wf_rand_history_empty [simp]: "wf_rand_history ru Map.empty"
by(simp add: wf_rand_history_def)

context comp_impl_parse begin

lemma random_usage_\<tau>:
  "\<lbrakk> t \<in> results (\<tau> \<eta> \<sigma> m); wf_rand_history ru (history \<sigma>) \<rbrakk> \<Longrightarrow> random_usage t \<le> ru"
proof(induction rule: \<tau>_results_induct)
  { case history                  then show ?case by(auto simp add: wf_rand_history_def) }
  { case PNonce                   then show ?case by simp }
  { case ANonce                   then show ?case by simp }
  { case Cipher_garbageE_key      then show ?case by auto }
  { case Cipher_garbageE          then show ?case by auto }
  { case Cipher                   then show ?case by auto }
  { case Ekey                     then show ?case by auto }
  { case Signature_garbageSIG_key then show ?case by auto }
  { case Signature_garbageSIG     then show ?case by auto }
  { case Signature                then show ?case by auto }
  { case Vkey                     then show ?case by auto }
  { case (Pair m t1 t2) then show "random_usage (PAIR t1 t2) \<le> ru" (* avoid eta-expasion *) by auto }
  { case String                   then show ?case by(auto dest: random_usage_parse_string) }
  { case garbage                  then show ?case by auto }
qed

end

context comp_impl_sim_gen begin

lemma wf_random_usage_gen_message:
  "\<lbrakk> Some (t, \<sigma>') \<in> results'_gpv (gen_message \<A> \<eta> \<sigma>); 
     wf_random_usage \<Gamma> f; wf_rand_history (random_usages f) (history \<sigma>) \<rbrakk>
  \<Longrightarrow> wf_random_usage \<Gamma> (f @ [t])"
by(auto simp add: gen_message_def split_beta dest!: random_usage_\<tau> elim: wf_random_usage'_mono)

inductive ops_history_dom_same :: "security \<Rightarrow> bool"
  for \<eta>
where
  "\<lbrakk> \<And>n \<sigma> m \<sigma>'. (m, \<sigma>') \<in> results'_gpv (run_state (pnonce \<eta> n) \<sigma>) \<Longrightarrow> dom (history \<sigma>') = dom (history \<sigma>);
     \<And>t m \<sigma> m' \<sigma>'. (m', \<sigma>') \<in> results'_gpv (run_state (record t m) \<sigma>) \<Longrightarrow> dom (history \<sigma>') \<subseteq> insert t (dom (history \<sigma>));
     \<And>k \<sigma> m' \<sigma>'. (m', \<sigma>') \<in> results'_gpv (run_state (encryption_key \<eta> k) \<sigma>) \<Longrightarrow> dom (history \<sigma>') = dom (history \<sigma>);
     \<And>k t m r \<sigma> m' \<sigma>'. (m', \<sigma>') \<in> results'_gpv (run_state (encryption \<eta> k t m r) \<sigma>) \<Longrightarrow> dom (history \<sigma>') = dom (history \<sigma>);
     \<And>n \<sigma> m \<sigma>'. (m, \<sigma>') \<in> results'_gpv (run_state (verification_key \<eta> n) \<sigma>) \<Longrightarrow> dom (history \<sigma>') = dom (history \<sigma>);
     \<And>k t m r \<sigma> m' \<sigma>'. (m', \<sigma>') \<in> results'_gpv (run_state (signature \<eta> k t m r) \<sigma>) \<Longrightarrow> dom (history \<sigma>') = dom (history \<sigma>) \<rbrakk>
  \<Longrightarrow> ops_history_dom_same \<eta>"

lemma ops_history_dom_sameD:
  assumes "ops_history_dom_same \<eta>"
  notes domIff[iff del]
  shows obs_history_sameD_pnonce: "\<And>m n \<sigma> \<sigma>'. (m, \<sigma>') \<in> results'_gpv (run_state (pnonce \<eta> n) \<sigma>) \<Longrightarrow> dom (history \<sigma>') = dom (history \<sigma>)"
  and obs_history_sameD_record: "\<And>m m' t \<sigma> \<sigma>'. (m', \<sigma>') \<in> results'_gpv (run_state (record t m) \<sigma>) \<Longrightarrow> dom (history \<sigma>') \<subseteq> insert t (dom (history \<sigma>))"
  and obs_history_sameD_encryption_key: "\<And>m m' k \<sigma> \<sigma>'. (m', \<sigma>') \<in> results'_gpv (run_state (encryption_key \<eta> k) \<sigma>) \<Longrightarrow> dom (history \<sigma>') = dom (history \<sigma>)"
  and obs_history_sameD_encryption: "\<And>m m' k t r \<sigma> \<sigma>'. (m', \<sigma>') \<in> results'_gpv (run_state (encryption \<eta> k t m r) \<sigma>) \<Longrightarrow> dom (history \<sigma>') = dom (history \<sigma>)"
  and obs_history_sameD_verification_key: "\<And>bs s' s n. (bs, s') \<in> results'_gpv (run_state (verification_key \<eta> n) s) \<Longrightarrow> dom (history s') = dom (history s)"
  and obs_history_sameD_signature: "\<And>bs s' s n' t m2 n. (bs, s') \<in> results'_gpv (run_state (signature \<eta> n' t m2 n) s) \<Longrightarrow> dom (history s') = dom (history s)"
using assms by(rule ops_history_dom_same.cases; blast 33)+

lemma wf_rand_history_encryption_key':
  assumes "ops_history_dom_same \<eta>"
  shows "\<lbrakk> (m', \<sigma>') \<in> results'_gpv (run_state (encryption_key' \<eta> t k) \<sigma>); wf_rand_history ru (history \<sigma>) \<rbrakk>
  \<Longrightarrow> wf_rand_history (ru \<squnion> random_usage t) (history \<sigma>')"
by(auto simp add: encryption_key'_def sup.coboundedI2 dest!: obs_history_sameD_encryption_key[OF assms] obs_history_sameD_record[OF assms] elim!: wf_rand_history_mono2[rotated] del: subsetI intro: wf_rand_history_mono cong: wf_rand_history_cong)

lemma wf_rand_history_verification_key':
  assumes "ops_history_dom_same \<eta>"
  shows "\<lbrakk> (m', \<sigma>') \<in> results'_gpv (run_state (verification_key' \<eta> t k) \<sigma>); wf_rand_history ru (history \<sigma>) \<rbrakk>
  \<Longrightarrow> wf_rand_history (ru \<squnion> random_usage t) (history \<sigma>')"
by(auto simp add: verification_key'_def sup.coboundedI2 dest!: obs_history_sameD_verification_key[OF assms] obs_history_sameD_record[OF assms] elim!: wf_rand_history_mono2[rotated] del: subsetI intro: wf_rand_history_mono cong: wf_rand_history_cong)

lemma wf_rand_history_\<beta>:
  assumes m: "(m, \<sigma>') \<in> results'_gpv (run_state (\<beta> \<eta> t) \<sigma>)"
  and \<sigma>: "wf_rand_history ru (history \<sigma>)"
  and wf: "ops_history_dom_same \<eta>"
  notes [simp del] = \<beta>.simps sup_apply 
    and [dest!] = ops_history_dom_sameD[OF wf]
  shows "wf_rand_history (ru \<squnion> random_usage t) (history \<sigma>')"
using m \<sigma>
apply(induction t m \<sigma> \<sigma>' arbitrary: ru rule: \<beta>_set_induct)
apply(clarsimp, fastforce elim!: wf_rand_history_mono2[rotated] intro: wf_rand_history_mono del: subsetI simp add: sup.coboundedI2)+
done

lemma wf_rand_history_\<beta>':
  assumes m: "(m, \<sigma>') \<in> results'_gpv (run_state (\<beta> \<eta> t) \<sigma>)"
  and \<sigma>: "wf_rand_history ru (history \<sigma>)"
  and wf: "ops_history_dom_same \<eta>"
  and ru: "random_usage t \<le> ru"
  shows "wf_rand_history ru (history \<sigma>')"
using wf_rand_history_\<beta>[OF m \<sigma> wf] by(rule wf_rand_history_mono)(simp add: ru)

lemma wf_rand_history_rec_message:
  "\<lbrakk> Some \<sigma>' \<in> results'_gpv (rec_message \<A> \<eta> \<sigma> t); wf_rand_history ru (history \<sigma>); ops_history_dom_same \<eta> \<rbrakk>
  \<Longrightarrow> wf_rand_history (ru \<squnion> random_usage t) (history \<sigma>')"
by(auto 4 3 simp add: rec_message_def intro: wf_rand_history_\<beta> split: option.split_asm)

lemma wf_rand_history_gen_message:
  "\<lbrakk> Some (t, \<sigma>') \<in> results'_gpv (gen_message \<A> \<eta> \<sigma>); wf_rand_history ru (history \<sigma>) \<rbrakk>
  \<Longrightarrow> wf_rand_history ru (history \<sigma>')"
by(auto simp add: gen_message_def)

lemma wf_rand_history_sim_control:
  "\<lbrakk> Some (m', \<sigma>') \<in> results'_gpv (sim_control \<A> \<eta> \<sigma> m); wf_rand_history ru (history \<sigma>) \<rbrakk>
  \<Longrightarrow> wf_rand_history ru (history \<sigma>')"
by(auto simp add: sim_control_def)

lemma wf_rand_history_sim_info_sim:
  "\<And>id. \<lbrakk> Some \<sigma>' \<in> results'_gpv (sim_info_sim \<Gamma> \<A> \<eta> \<sigma> id to); wf_rand_history ru (history \<sigma>) \<rbrakk>
  \<Longrightarrow> wf_rand_history ru (history \<sigma>')"
by(cases to)(simp_all split: if_split_asm t.split nonce.split option.split add: Let_def)

lemma wf_rand_history_hybrid_step1:
  assumes step: "Some (to, (s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step1' (sim' \<Gamma> \<A> \<eta>) ((s,v,f),\<sigma>))"
  and ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v"
  and wf: "wf_rand_history (random_usages f) (history \<sigma>)"
  and *: "ops_history_dom_same \<eta>"
  shows "wf_rand_history (random_usages f') (history \<sigma>')"
proof(cases v)
  case (Computation id comp c_data yes no)
  with step have "\<sigma>' = \<sigma>" "random_usages f \<le> random_usages f'" by(auto split: option.split_asm)
  with wf show ?thesis by(auto elim: wf_rand_history_mono)
next
  case (Output id out succ)
  with ks have "out < length f" by(auto simp add: key_safe_simps)
  hence "random_usages f \<squnion> random_usage (f ! out) = random_usages f"
    by(auto intro!: antisym member_le_random_usages)
  with Output step wf * show ?thesis by(auto dest!: wf_rand_history_rec_message simp add: gpv.set_map)
next
  { case Input         with step wf show ?thesis
      by(auto split: if_split_asm elim!: wf_rand_history_gen_message wf_rand_history_mono simp add: gpv.set_map Some_eq_bind_conv) }
  { case Control       with step wf show ?thesis by(auto elim: wf_rand_history_sim_control simp add: gpv.set_map) }
  { case Probabilistic with step wf show ?thesis by(auto) }
qed

lemma wf_rand_history_hybrid_step:
  "\<lbrakk> Some ((s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step' (sim' \<Gamma> \<A> \<eta>) ((s,v,f),\<sigma>)); 
     key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v; 
     wf_rand_history (random_usages f) (history \<sigma>);
     ops_history_dom_same \<eta> \<rbrakk>
  \<Longrightarrow> wf_rand_history (random_usages f') (history \<sigma>')"
by(auto 4 3 simp add: hybrid_step_def gpv.set_map dest: wf_rand_history_hybrid_step1 wf_rand_history_sim_info_sim split: option.split_asm)

end

context comp_impl begin

definition key_safe_sim :: "protocol_nonce set \<Rightarrow> security \<Rightarrow> nonce_typing \<Rightarrow> 'more' simulator_state'_scheme \<Rightarrow> bool"
where
  "key_safe_sim PN \<eta> \<Gamma> s \<longleftrightarrow> 
   \<T>_history \<eta> (history s) \<and>
   (\<forall>t\<in>dom (history s). A,\<eta>,\<Gamma> \<turnstile>t t : node_type_of_t \<Gamma> t) \<and>
   (\<forall>n. EK (N (PNonce n)) \<in> dom (history s) \<longrightarrow> \<Gamma> n = Some Encryption) \<and>
   (\<forall>n. VK (N (PNonce n)) \<in> dom (history s) \<longrightarrow> \<Gamma> n = Some Signature) \<and>
   (\<forall>m m'. history s (VK (N (ANonce m))) = Some m' \<longrightarrow> m = m') \<and> 
   (\<forall>t \<in> dom (history s). \<beta>1_anonces t \<inter> dom \<Gamma> \<subseteq> randomness s \<union> PN) \<and>
   (dom (pnonce s) \<inter> dom \<Gamma> \<subseteq> randomness s \<union> PN) \<and>
   (\<forall>m\<in>ran (pnonce s). \<T>_nonce A \<eta> m)"

lemma key_safe_simD:
  assumes "key_safe_sim PN \<eta> \<Gamma> s"
  shows key_safe_sim_\<T>_history: "\<T>_history \<eta> (history s)"
  and key_safe_sim_key_safe_t: "history s t = Some m \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t t : node_type_of_t \<Gamma> t"
  and key_safe_sim_EK: "EK (N (PNonce n)) \<in> dom (history s) \<Longrightarrow> \<Gamma> n = Some Encryption"
  and key_safe_sim_VK_PNonce: "VK (N (PNonce n)) \<in> dom (history s) \<Longrightarrow> \<Gamma> n = Some Signature"
  and key_safe_sim_VK_ANonce: "history s (VK (N (ANonce m))) = Some m' \<Longrightarrow> m = m'"
  and key_safe_sim_history_PN: "\<lbrakk> history s t = Some m; n \<in> \<beta>1_anonces t \<rbrakk> \<Longrightarrow> \<Gamma> n = None \<or> n \<in> randomness s \<or> n \<in> PN"
  and key_safe_sim_pnonce_\<Gamma>: "pnonce s n = Some m \<Longrightarrow> \<Gamma> n = None \<or> n \<in> randomness s \<or> n \<in> PN"
  and key_safe_sim_pnonce_\<T>: "m \<in> ran (pnonce s) \<Longrightarrow> \<T>_nonce A \<eta> m"
using assms apply(auto simp add: key_safe_sim_def)
apply(auto simp add: dom_def)
apply(drule spec) back back back back
apply(erule impE)
apply blast
apply fastforce
done

lemma key_safe_simI [intro?]:
  "\<lbrakk> \<T>_history \<eta> (history s);
     \<And>t m. history s t = Some m \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t t : node_type_of_t \<Gamma> t;
     \<And>n m. history s (EK (N (PNonce n))) = Some m \<Longrightarrow> \<Gamma> n = Some Encryption;
     \<And>n m. history s (VK (N (PNonce n))) = Some m \<Longrightarrow> \<Gamma> n = Some Signature;
     \<And>m m'. history s (VK (N (ANonce m))) = Some m' \<Longrightarrow> m = m';
     \<And>t m n nt. \<lbrakk> history s t = Some m; n \<in> \<beta>1_anonces t; n \<notin> PN; \<Gamma> n = Some nt \<rbrakk> \<Longrightarrow> n \<in> randomness s;
     \<And>n m nt. \<lbrakk> pnonce s n = Some m; n \<notin> PN; \<Gamma> n = Some nt \<rbrakk> \<Longrightarrow> n \<in> randomness s;
     \<And>n m. pnonce s n = Some m \<Longrightarrow> \<T>_nonce A \<eta> m \<rbrakk>
   \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s"
by(auto simp add: key_safe_sim_def ran_def)

lemma key_safe_sim_randomness_updateI:
  "\<lbrakk> key_safe_sim PN \<eta> \<Gamma> s; randomness s \<subseteq> f (randomness s) \<rbrakk> \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> (randomness_update f s)"
by(auto 4 3 intro!: key_safe_simI dest: key_safe_simD simp add: ran_def)

lemma key_safe_sim_fuel_update [simp]: "key_safe_sim PN \<eta> \<Gamma> (fuel_update f s) = key_safe_sim PN \<eta> \<Gamma> s"
by(simp add: key_safe_sim_def)

lemma key_safe_sim_pnonce_updateI:
  "\<lbrakk> key_safe_sim PN \<eta> \<Gamma> s; \<Gamma> r \<noteq> None \<Longrightarrow> r \<in> randomness s \<or> r \<in> PN; \<T>_nonce A \<eta> m \<rbrakk>
   \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> (pnonce_update (\<lambda>pns. pns(r \<mapsto> m)) s)"
by(rule key_safe_simI)(auto 4 3 dest: key_safe_simD split: if_split_asm simp add: ran_def intro: key_safe_sim_pnonce_\<T>)

lemma key_safe_sim_history_updateI:
  assumes safe: "key_safe_sim PN \<eta> \<Gamma> s"
  and \<T>: "\<T>_of_t \<eta> t m"
  and key_safe_t: "A,\<eta>,\<Gamma> \<turnstile>t t : node_type_of_t \<Gamma> t"
  and keys: "\<And>n. t = EK (N (PNonce n)) \<Longrightarrow> \<Gamma> n = Some Encryption"
            "\<And>n. t = VK (N (PNonce n)) \<Longrightarrow> \<Gamma> n = Some Signature"
            "\<And>m'. t = VK (N (ANonce m')) \<Longrightarrow> m = m'"
  and nonces: "\<And>n nt. \<lbrakk> n \<in> \<beta>1_anonces t; \<Gamma> n = Some nt; n \<notin> PN \<rbrakk> \<Longrightarrow> n \<in> randomness s"
  shows "key_safe_sim PN \<eta> \<Gamma> (history_update (\<lambda>hist. hist(t \<mapsto> m)) s)"
apply(rule key_safe_simI)
apply(auto 4 3 split: if_split_asm intro: key_safe_simD[OF safe] \<T>_history_updateI \<T> key_safe_t keys(1-2) nonces dest: key_safe_sim_pnonce_\<Gamma>[OF safe] key_safe_sim_history_PN[OF safe] simp add: ran_def keys(3))
done

lemma key_safe_sim_more_update [simp]:
  "key_safe_sim PN \<eta> \<Gamma> (simulator_state'.more_update f s) = key_safe_sim PN \<eta> \<Gamma> s"
by(simp add: key_safe_sim_def)

lemma key_safe_sim_adv_state_update [simp]:
  "key_safe_sim PN \<eta> \<Gamma> (adv_state_update f s) = key_safe_sim PN \<eta> \<Gamma> s"
by(simp add: key_safe_sim_def)

lemma key_safe_sim_randomness: "key_safe_sim PN \<eta> \<Gamma> s \<Longrightarrow> key_safe_sim (PN - randomness s) \<eta> \<Gamma> s"
by(simp add: key_safe_sim_def)

lemma key_safe_sim_state_cong:
  "\<lbrakk> history s = history s'; randomness s = randomness s'; pnonce s = pnonce s' \<rbrakk>
  \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s = key_safe_sim PN \<eta> \<Gamma> s'"
by(simp add: key_safe_sim_def)

lemma key_safe_sim_PN_mono:
  "\<lbrakk> key_safe_sim PN \<eta> \<Gamma> s; PN \<subseteq> PN' \<rbrakk> \<Longrightarrow> key_safe_sim PN' \<eta> \<Gamma> s"
by(rule key_safe_simI)(auto dest: key_safe_simD intro!: key_safe_sim_pnonce_\<T> simp add: ran_def)

end

locale key_safe_sim_assms = comp_impl A 
  for A +
  fixes pnonce :: "security \<Rightarrow> protocol_nonce \<Rightarrow> (('s, 'more) simulator_state_scheme, (bitstring \<times> ('s, 'more) simulator_state_scheme, 'call, 'ret) gpv) stateT"
  and "record" :: "t \<Rightarrow> bitstring \<Rightarrow> (('s, 'more) simulator_state_scheme, (bitstring \<times> ('s, 'more) simulator_state_scheme, 'call, 'ret) gpv) stateT"
  and encryption_key :: "security \<Rightarrow> protocol_nonce \<Rightarrow> (('s, 'more) simulator_state_scheme, (bitstring \<times> ('s, 'more) simulator_state_scheme, 'call, 'ret) gpv) stateT"
  and encryption :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> bitstring \<Rightarrow> protocol_nonce \<Rightarrow> (('s, 'more) simulator_state_scheme, (bitstring \<times> ('s, 'more) simulator_state_scheme, 'call, 'ret) gpv) stateT"
  and verification_key :: "security \<Rightarrow> protocol_nonce \<Rightarrow> (('s, 'more) simulator_state_scheme, (vkey \<times> ('s, 'more) simulator_state_scheme, 'call, 'ret) gpv) stateT"
  and signature :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> message \<Rightarrow> protocol_nonce \<Rightarrow> (('s, 'more) simulator_state_scheme, (signature \<times> ('s, 'more) simulator_state_scheme, 'call, 'ret) gpv) stateT"
  and PN :: "protocol_nonce set"
  and \<Gamma> :: nonce_typing
  and \<eta> :: security
  assumes pnonce_random: "\<And>r s m s' bl.
    \<lbrakk> (m, s') \<in> results'_gpv (run_state (pnonce \<eta> r) s);
      key_safe_sim PN \<eta> \<Gamma> s; \<Gamma> r = Some (Randomness bl); r \<in> PN \<rbrakk>
    \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s'"
  and pnonce_public: "\<And>n s m' s' msg s''.
    \<lbrakk> (m', s') \<in> results'_gpv (run_state (pnonce \<eta> n) s);
      (msg, s'') \<in> results'_gpv (run_state (record (N (PNonce n)) m') s');
      \<Gamma> n = None; key_safe_sim PN \<eta> \<Gamma> s \<rbrakk>
    \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s''"
  and "record": "\<And>t m s m' s'.
    \<lbrakk> (m', s') \<in> results'_gpv (run_state (record t m) s);
      key_safe_sim PN \<eta> \<Gamma> s;
      \<forall>n. t \<noteq> EK (N (PNonce n)); \<forall>n. t \<noteq> VK (N (PNonce n)); \<forall>m'. t = VK (N (ANonce m')) \<longrightarrow> m = m';
      \<beta>1_anonces t \<inter> dom \<Gamma> \<subseteq> PN;
      \<T>_of_t \<eta> t m; A,\<eta>,\<Gamma> \<turnstile>t t : PublicT \<rbrakk>
    \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s'"
  and encryption_key: "\<And>k s m s' m' s''. 
    \<lbrakk> (m, s') \<in> results'_gpv (run_state (encryption_key \<eta> k) s); 
      (m', s'') \<in> results'_gpv (run_state (record (EK (N (PNonce k))) m) s');
      key_safe_sim PN \<eta> \<Gamma> s; \<Gamma> k = Some Encryption; k \<in> PN \<rbrakk>
    \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s''"
  and encryption: "\<And>k t m r s m' s' m'' s'' bl.
    \<lbrakk> (m', s') \<in> results'_gpv (run_state (encryption \<eta> k t m r) s);
      (m'', s'') \<in> results'_gpv (run_state (record (E (EK (N (PNonce k))) t (N (PNonce r))) m') s');
      key_safe_sim PN \<eta> \<Gamma> s;
      \<Gamma> k = Some Encryption; k \<in> PN; A,\<eta>,\<Gamma> \<turnstile>t t : PublicT;
      \<Gamma> r = Some (Randomness bl); r \<in> PN; \<beta>1_anonces t \<inter> dom \<Gamma> \<subseteq> PN \<rbrakk>
    \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s''"
  and verification_key: "\<And>k s vk' s' vk s''.
    \<lbrakk> (vk', s') \<in> results'_gpv (run_state (verification_key \<eta> k) s);
      (vk, s'') \<in> results'_gpv (run_state (record (VK (N (PNonce k))) vk') s');
      key_safe_sim PN \<eta> \<Gamma> s;
      \<Gamma> k = Some Signature; k \<in> PN \<rbrakk>
    \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s''"
  and signature: "\<And>k t m r s m' s' msg s'' bl.
    \<lbrakk> (m', s') \<in> results'_gpv (run_state (signature \<eta> k t m r) s);
      (msg, s'') \<in> results'_gpv (run_state (record (SIG (SK (N (PNonce k))) t (N (PNonce r))) m') s');
      key_safe_sim PN \<eta> \<Gamma> s;
      \<Gamma> k = Some Signature; k \<in> PN; A,\<eta>,\<Gamma> \<turnstile>t t : PublicT;
      \<Gamma> r = Some (Randomness bl); r \<in> PN; \<beta>1_anonces t \<inter> dom \<Gamma> \<subseteq> PN \<rbrakk>
   \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s''"

context comp_impl_sim_gen begin

context
  fixes PN :: "protocol_nonce set" and \<Gamma> :: nonce_typing and \<eta> :: security
  assumes ops: "key_safe_sim_assms A pnonce record encryption_key encryption verification_key signature PN \<Gamma> \<eta>"
  notes [simp del] = \<beta>.simps
begin

interpretation key_safe_sim_assms A pnonce "record" encryption_key encryption verification_key signature PN \<Gamma> \<eta> by fact

lemma key_safe_sim_\<beta>:
  "\<lbrakk> (m, \<sigma>') \<in> results'_gpv (run_state (\<beta> \<eta> t) \<sigma>); A,\<eta>,\<Gamma> \<turnstile>t t : PublicT; key_safe_sim PN \<eta> \<Gamma> \<sigma>; \<beta>1_anonces t \<inter> dom \<Gamma> \<subseteq> PN \<rbrakk>
  \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> \<sigma>'"
proof(induction rule: \<beta>_set_induct)
  { case E           thus ?case by(auto dest: encryption_key encryption) }
  { case E_key       thus ?case by(auto dest!: pnonce_random "record" intro: \<T>_A_e') }
  { case E_adversary thus ?case by(auto 4 3 dest!: "record") }
  { case EK_PNonce   thus ?case by(auto dest: encryption_key) }
  { case EK_ANonce   thus ?case by(auto dest: "record") }
  { case DK          thus ?case by auto }
  { case SIG         thus ?case by(auto dest: verification_key signature) }
  { case SIG_ANonce  thus ?case by(auto 4 3 dest: "record") }
  { case VK_PNonce   thus ?case by(auto dest: verification_key) }
  { case VK_ANonce   thus ?case by(auto dest: "record") }
  { case SK          thus ?case by auto }
  { case PAIR        thus ?case by(auto dest!: "record") }
  { case EMPTY       thus ?case by(auto dest!: "record" intro: \<T>_A_empty') }
  { case STRING      thus ?case by(auto dest!: "record" intro: \<T>_A_string') }
  { case PNonce      thus ?case by(auto dest: pnonce_public) }
  { case ANonce      thus ?case by(auto dest: "record") }
  { case garbage     thus ?case by(auto dest: "record") }
  { case garbageE    thus ?case by(auto 4 3 dest: "record") }
  { case garbageSIG  thus ?case by(auto 4 3 dest: "record") }
qed

lemma key_safe_sim_rec_message:
  "\<lbrakk> Some s' \<in> results'_gpv (rec_message \<A> \<eta> s t); A,\<eta>,\<Gamma> \<turnstile>t t : PublicT; key_safe_sim PN \<eta> \<Gamma> s; \<beta>1_anonces t \<inter> dom \<Gamma> \<subseteq> PN \<rbrakk>
  \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s'"
by(clarsimp simp add: rec_message_def key_safe_sim_\<beta> split: option.split_asm)

end

lemma key_safe_sim_initial [simp]: "\<And>more. key_safe_sim {} \<eta> \<Gamma> (initial more \<sigma> bound)"
by(simp add: key_safe_sim_def)

lemma key_safe_sim_gen_message:
  "\<lbrakk> Some (t, s') \<in> results'_gpv (gen_message \<A> \<eta> s); key_safe_sim PN \<eta> \<Gamma> s \<rbrakk> \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s'"
by(auto simp add: gen_message_def)

lemma key_safe_sim_control:
  "\<lbrakk> Some (m', s') \<in> results'_gpv (sim_control \<A> \<eta> s m); key_safe_sim PN \<eta> \<Gamma> s \<rbrakk> \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s'"
by(auto simp add: sim_control_def)

lemma key_safe_sim_info_sim:
  "\<And>id. \<lbrakk> Some s' \<in> results'_gpv (sim_info_sim \<Gamma> \<A> \<eta> s id to); key_safe_sim PN \<eta> \<Gamma> s \<rbrakk> \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s'"
by(cases to)(auto split: if_split_asm t.split nonce.split option.split simp add: Let_def intro!: key_safe_sim_randomness_updateI)

lemma key_safe_\<tau>:
  assumes "t \<in> results (\<tau> \<eta> s m)"
  and safe: "key_safe_sim {} \<eta> \<Gamma> s"
  shows "A,\<eta>,\<Gamma> \<turnstile>t t : PublicT"
using assms
proof(induction rule: \<tau>_results_induct)
  case (history m t)
  from safe[THEN key_safe_sim_\<T>_history] history(1) have "\<T>_of_t \<eta> t m" by(rule \<T>_historyD)
  moreover from safe history(1) have "A,\<eta>,\<Gamma> \<turnstile>t t : node_type_of_t \<Gamma> t" by(rule key_safe_sim_key_safe_t)
  ultimately show ?case using history(2) by(cases t)(auto dest: \<T>_disjoint')
next
  case (PNonce m n)
  with key_safe_sim_pnonce_\<Gamma>[OF safe, of n m] show ?case by simp
next
  case (ANonce m)
  then show ?case by simp
next
  case Cipher_garbageE_key then show ?case by simp
next
  case (Cipher_garbageE m key n)
  from \<open>EK (N (PNonce n)) \<in> results (\<tau> \<eta> s key)\<close>
  have "history s (EK (N (PNonce n))) = Some key" by(auto dest: EK_PNonce_in_\<tau>D)
  hence "\<Gamma> n = Some Encryption" using safe by(auto dest: key_safe_sim_EK)
  then show ?case using \<open>\<T>_cipher A \<eta> m\<close> by simp
next
  case Cipher thus ?case by(auto dest!: EK_PNonce_in_\<tau>D key_safe_sim_key_safe_t[OF safe])
next
  case (Signature_garbageSIG m vkey n)
  then show ?case
  proof(cases n)
    case (PNonce n')
    with \<open>VK (N n) \<in> results (\<tau> \<eta> s vkey)\<close>
    have "history s (VK (N n)) = Some vkey" by(auto dest: VK_PNonce_in_\<tau>D)
    hence "\<Gamma> n' = Some Signature" using safe PNonce by(auto dest: key_safe_sim_VK_PNonce)
    then show ?thesis using PNonce \<open>\<T>_signature A \<eta> m\<close> by simp
  next
    case (ANonce n')
    thus ?thesis using Signature_garbageSIG
      by(auto dest!: VK_ANonce_in_\<tau>D \<T>_A_vkof' dest: \<T>_disjoint' key_safe_sim_VK_ANonce)
  qed
next
  case (Signature m vkey n m' plain)
  thus ?case
    by(cases n)(auto dest!: VK_PNonce_in_\<tau>D VK_ANonce_in_\<tau>D \<T>_A_vkof' dest: \<T>_disjoint' key_safe_sim_VK_ANonce[OF safe] intro: key_safe_sim_VK_PNonce[OF safe])
qed(auto dest: key_safe_parse_string)

lemma key_safe_gen_message:
  "\<lbrakk> Some (t, \<sigma>') \<in> results'_gpv (gen_message \<A> \<eta> \<sigma>); key_safe_sim {} \<eta> \<Gamma> \<sigma> \<rbrakk>
  \<Longrightarrow> A,\<eta>,\<Gamma> \<turnstile>t t : PublicT"
by(auto simp add: gen_message_def intro: key_safe_\<tau>)

lemma \<beta>1_anonces_\<tau>:
  assumes t: "t \<in> results (\<tau> \<eta> s m)"
  and ks: "key_safe_sim PN \<eta> \<Gamma> s"
  shows "\<beta>1_anonces t \<inter> dom \<Gamma> \<subseteq> randomness s \<union> PN"
using t
proof(induction rule: \<tau>_results_induct)
  note [simp] = Int_Un_distrib2
  { case (history m t)                   thus ?case by(auto dest: key_safe_sim_history_PN[OF ks]) }
  { case PNonce                          thus ?case by(auto dest: key_safe_sim_pnonce_\<Gamma>[OF ks]) }
  { case ANonce                          thus ?case by simp }
  { case Cipher_garbageE_key             thus ?case by simp }
  { case Cipher_garbageE                 thus ?case by(auto dest!: EK_PNonce_in_\<tau>D key_safe_sim_history_PN[OF ks]) }
  { case Cipher                          thus ?case by(auto dest!: EK_PNonce_in_\<tau>D key_safe_sim_history_PN[OF ks]) }
  { case Ekey                            thus ?case by simp }
  { case Signature_garbageSIG_key        thus ?case by simp }
  { case (Signature_garbageSIG m vkey n) thus ?case by(cases n)(auto dest!: VK_PNonce_in_\<tau>D key_safe_sim_history_PN[OF ks]) }
  { case (Signature m vkey n m plain)    thus ?case by(cases n)(auto dest!: VK_PNonce_in_\<tau>D key_safe_sim_history_PN[OF ks]) }
  { case Vkey                            thus ?case by simp }
  { case Pair                            thus ?case by simp }
  { case String                          thus ?case by simp }
  { case garbage                         thus ?case by simp }
qed

end

inductive randomness_sampled_recorded :: "nonce_typing \<Rightarrow> 'more simulator_state'_scheme \<Rightarrow> bool"
  for \<Gamma> s
where
  "(\<And>n bl. \<lbrakk> \<Gamma> n = Some (Randomness bl); n \<in> dom (pnonce s) \<rbrakk>
    \<Longrightarrow> (\<exists>k t. E k t (N (PNonce n)) \<in> dom (history s)) \<or> (\<exists>k t. SIG k t (N (PNonce n)) \<in> dom (history s)))
  \<Longrightarrow> randomness_sampled_recorded \<Gamma> s"

lemma randomness_sampled_recordedD:
  "\<lbrakk> randomness_sampled_recorded \<Gamma> s; \<Gamma> n = Some (Randomness bl); n \<in> dom (pnonce s) \<rbrakk>
  \<Longrightarrow> (\<exists>k t. E k t (N (PNonce n)) \<in> dom (history s)) \<or> (\<exists>k t. SIG k t (N (PNonce n)) \<in> dom (history s))"
by(blast elim: randomness_sampled_recorded.cases)

lemma randomness_sampled_recorded_cong:
  "\<lbrakk> \<Gamma> = \<Gamma>'; pnonce s = pnonce s'; history s = history s' \<rbrakk>
  \<Longrightarrow> randomness_sampled_recorded \<Gamma> s = randomness_sampled_recorded \<Gamma>' s'"
by(simp add: randomness_sampled_recorded.simps)

lemma [simp]:
  shows randomness_sampled_recorded_fuel_update:
  "\<And>f. randomness_sampled_recorded \<Gamma> (fuel_update f s) = randomness_sampled_recorded \<Gamma> s"
  and randomness_sampled_recorded_randomness_update:
  "\<And>f. randomness_sampled_recorded \<Gamma> (randomness_update f s) = randomness_sampled_recorded \<Gamma> s"
  and randomness_sampled_recorded_more_update:
  "\<And>f. randomness_sampled_recorded \<Gamma> (more_update f s) = randomness_sampled_recorded \<Gamma> s"
  and randomness_sampled_recorded_adv_state_update:
  "\<And>f s s. randomness_sampled_recorded \<Gamma> (adv_state_update f s) = randomness_sampled_recorded \<Gamma> s"
by(simp_all cong: randomness_sampled_recorded_cong)

lemma randomness_sampled_recorded_history_update_mono:
  assumes rsr: "randomness_sampled_recorded \<Gamma> s"
  and mono: "dom (history s) \<subseteq> dom (f (history s))"
  shows "randomness_sampled_recorded \<Gamma> (history_update f s)"
using mono by(auto 4 3 dest!: randomness_sampled_recordedD[OF rsr] iff del: domIff intro!: randomness_sampled_recorded.intros)

lemma randomness_sampled_recorded_pnonce_update:
  assumes rsr: "randomness_sampled_recorded \<Gamma> s"
  and "\<And>n bl. \<lbrakk> \<Gamma> n = Some (Randomness bl); n \<in> dom (pnonce s'); n \<notin> dom (pnonce s) \<rbrakk>
       \<Longrightarrow> (\<exists>k t. E k t (N (PNonce n)) \<in> dom (history s')) \<or> (\<exists>k t. SIG k t (N (PNonce n)) \<in> dom (history s'))"
  and "dom (history s) \<subseteq> dom (history s')"
  shows "randomness_sampled_recorded \<Gamma> s'"
using assms(2-3)
apply(intro randomness_sampled_recorded.intros)
subgoal for n bl
  by(cases "n \<in> dom (pnonce s)")
    (auto 4 4 intro!: randomness_sampled_recorded.intros dest: randomness_sampled_recordedD[OF rsr] iff del: domIff)
done

context comp_impl_sim_gen begin

lemma randomness_sampled_recorded_initial [simp]:
  "\<And>more. randomness_sampled_recorded \<Gamma> (initial more \<sigma> bound)"
by(auto intro!: randomness_sampled_recorded.intros)

inductive randomness_sampled_recorded_assms :: "nonce_typing \<Rightarrow> security \<Rightarrow> bool"
  for \<Gamma> \<eta>
where
  "\<lbrakk> \<And>t m s m' s'.
     \<lbrakk> (m', s') \<in> results'_gpv (run_state (record t m) s); randomness_sampled_recorded \<Gamma> s \<rbrakk>
     \<Longrightarrow> randomness_sampled_recorded \<Gamma> s';
     \<And>n s m s'.
     \<lbrakk> (m, s') \<in> results'_gpv (run_state (pnonce \<eta> n) s);
       randomness_sampled_recorded \<Gamma> s; \<forall>bl. \<Gamma> n \<noteq> Some (Randomness bl)\<rbrakk>
     \<Longrightarrow> randomness_sampled_recorded \<Gamma> s';
     \<And>r s m s' k t m' m'' s'' bl.
     \<lbrakk> (m, s') \<in> results'_gpv (run_state (pnonce \<eta> r) s);
       (m'', s'') \<in> results'_gpv (run_state (record (E k t (N (PNonce r))) m') s');
       randomness_sampled_recorded \<Gamma> s;
       \<Gamma> r = Some (Randomness bl) \<rbrakk>
     \<Longrightarrow> randomness_sampled_recorded \<Gamma> s'';
     \<And>k s m s'. 
     \<lbrakk> (m, s') \<in> results'_gpv (run_state (encryption_key \<eta> k) s); 
       randomness_sampled_recorded \<Gamma> s; \<Gamma> k = Some Encryption \<rbrakk>
     \<Longrightarrow> randomness_sampled_recorded \<Gamma> s';
     \<And>k t m r s m' s' m'' s'' bl.
     \<lbrakk> (m', s') \<in> results'_gpv (run_state (encryption \<eta> k t m r) s);
       (m'', s'') \<in> results'_gpv (run_state (record (E (EK (N (PNonce k))) t (N (PNonce r))) m') s');
       randomness_sampled_recorded \<Gamma> s; \<Gamma> k = Some Encryption; \<Gamma> r = Some (Randomness bl) \<rbrakk>
     \<Longrightarrow> randomness_sampled_recorded \<Gamma> s'';
     \<And>k s m s'.
     \<lbrakk> (m, s') \<in> results'_gpv (run_state (verification_key \<eta> k) s);
       randomness_sampled_recorded \<Gamma> s; \<Gamma> k = Some Signature \<rbrakk>
     \<Longrightarrow> randomness_sampled_recorded \<Gamma> s';
     \<And>k t m r s m' s' m'' s'' bl.
     \<lbrakk> (m', s') \<in> results'_gpv (run_state (signature \<eta> k t m r) s);
       (m'', s'') \<in> results'_gpv (run_state (record (SIG (SK (N (PNonce k))) t (N (PNonce r))) m') s');
       randomness_sampled_recorded \<Gamma> s; \<Gamma> k = Some Signature; \<Gamma> r = Some (Randomness bl) \<rbrakk>
     \<Longrightarrow> randomness_sampled_recorded \<Gamma> s''
  \<rbrakk> \<Longrightarrow> randomness_sampled_recorded_assms \<Gamma> \<eta>"

lemma randomness_sampled_recorded_assmsD:
  assumes "randomness_sampled_recorded_assms \<Gamma> \<eta>"
  shows wf_randomness_sampled_recordedD_record: "\<And>t m s m' s'.
    \<lbrakk> (m', s') \<in> results'_gpv (run_state (record t m) s); randomness_sampled_recorded \<Gamma> s \<rbrakk>
    \<Longrightarrow> randomness_sampled_recorded \<Gamma> s'"
  and wf_randomness_sampled_recordedD_pnonce: "\<And>n s m s'.
    \<lbrakk> (m, s') \<in> results'_gpv (run_state (pnonce \<eta> n) s);
      randomness_sampled_recorded \<Gamma> s; \<forall>bl. \<Gamma> n \<noteq> Some (Randomness bl)\<rbrakk>
    \<Longrightarrow> randomness_sampled_recorded \<Gamma> s'"
  and wf_randomness_sampled_recordedD_pnonce_randomness: "\<And>r s m s' k t m' m'' s'' bl.
    \<lbrakk> (m, s') \<in> results'_gpv (run_state (pnonce \<eta> r) s);
      (m'', s'') \<in> results'_gpv (run_state (record (E k t (N (PNonce r))) m') s');
      randomness_sampled_recorded \<Gamma> s;
      \<Gamma> r = Some (Randomness bl) \<rbrakk>
    \<Longrightarrow> randomness_sampled_recorded \<Gamma> s''"
  and wf_randomness_sampled_recordedD_encryption_key: "\<And>k s m s'. 
    \<lbrakk> (m, s') \<in> results'_gpv (run_state (encryption_key \<eta> k) s); 
      randomness_sampled_recorded \<Gamma> s; \<Gamma> k = Some Encryption \<rbrakk>
    \<Longrightarrow> randomness_sampled_recorded \<Gamma> s'"
  and wf_randomness_sampled_recordedD_encryption: "\<And>k t m r s m' s' m'' s'' bl.
    \<lbrakk> (m', s') \<in> results'_gpv (run_state (encryption \<eta> k t m r) s);
      (m'', s'') \<in> results'_gpv (run_state (record (E (EK (N (PNonce k))) t (N (PNonce r))) m') s');
      randomness_sampled_recorded \<Gamma> s; \<Gamma> k = Some Encryption; \<Gamma> r = Some (Randomness bl) \<rbrakk>
    \<Longrightarrow> randomness_sampled_recorded \<Gamma> s''"
  and wf_randomness_sampled_recordedD_verification_key: "\<And>k s m s'.
    \<lbrakk> (m, s') \<in> results'_gpv (run_state (verification_key \<eta> k) s);
      randomness_sampled_recorded \<Gamma> s; \<Gamma> k = Some Signature \<rbrakk>
    \<Longrightarrow> randomness_sampled_recorded \<Gamma> s'"
  and wf_randomness_sampled_recordedD_signature: "\<And>k t m r s m' s' m'' s'' bl.
    \<lbrakk> (m', s') \<in> results'_gpv (run_state (signature \<eta> k t m r) s);
      (m'', s'') \<in> results'_gpv (run_state (record (SIG (SK (N (PNonce k))) t (N (PNonce r))) m') s');
      randomness_sampled_recorded \<Gamma> s; \<Gamma> k = Some Signature; \<Gamma> r = Some (Randomness bl) \<rbrakk>
    \<Longrightarrow> randomness_sampled_recorded \<Gamma> s''"
using assms by(rule randomness_sampled_recorded_assms.cases; blast 47)+

lemma wf_randomness_sampled_recordedD_encryption_key':
  assumes "randomness_sampled_recorded_assms \<Gamma> \<eta>"
  shows "\<And>t k s m s'.
  \<lbrakk> (m, s') \<in> results'_gpv (run_state (encryption_key' \<eta> t k) s); 
    randomness_sampled_recorded \<Gamma> s; \<Gamma> k = Some Encryption \<rbrakk>
  \<Longrightarrow> randomness_sampled_recorded \<Gamma> s'"
by(auto simp add: encryption_key'_def dest!: wf_randomness_sampled_recordedD_encryption_key[OF assms] wf_randomness_sampled_recordedD_record[OF assms])

lemma wf_randomness_sampled_recordedD_verification_key':
  assumes "randomness_sampled_recorded_assms \<Gamma> \<eta>"
  shows "\<And>t k s m s'.
  \<lbrakk> (m, s') \<in> results'_gpv (run_state (verification_key' \<eta> t k) s); 
    randomness_sampled_recorded \<Gamma> s; \<Gamma> k = Some Signature \<rbrakk>
  \<Longrightarrow> randomness_sampled_recorded \<Gamma> s'"
by(auto simp add: verification_key'_def dest!: wf_randomness_sampled_recordedD_verification_key[OF assms] wf_randomness_sampled_recordedD_record[OF assms])

lemma randomness_sampled_recorded_\<beta>:
  assumes "(msg, s') \<in> results'_gpv (run_state (\<beta> \<eta> t) s)"
  and "A,\<eta>,\<Gamma> \<turnstile>t t : PublicT"
  and "randomness_sampled_recorded \<Gamma> s"
  and *: "randomness_sampled_recorded_assms \<Gamma> \<eta>"
  shows "randomness_sampled_recorded \<Gamma> s'"
using assms(1-3)
proof(induction rule: \<beta>_set_induct)
  note [dest] =
    randomness_sampled_recorded_assmsD[OF *]
    wf_randomness_sampled_recordedD_encryption_key'[OF *]
    wf_randomness_sampled_recordedD_verification_key'[OF *]
  { case E           thus ?case by auto }
  { case E_key       thus ?case by(auto split: t.split_asm) }
  { case E_adversary thus ?case by auto }
  { case EK_PNonce   thus ?case by auto }
  { case EK_ANonce   thus ?case by auto }
  { case DK          thus ?case by auto }
  { case SIG         thus ?case by auto }
  { case SIG_ANonce  thus ?case by auto }
  { case VK_PNonce   thus ?case by auto }
  { case VK_ANonce   thus ?case by auto }
  { case SK          thus ?case by auto }
  { case PAIR        thus ?case by auto }
  { case EMPTY       thus ?case by auto }
  { case STRING      thus ?case by auto }
  { case PNonce      thus ?case by auto }
  { case ANonce      thus ?case by auto }
  { case garbage     thus ?case by auto }
  { case garbageE    thus ?case by auto }
  { case garbageSIG  thus ?case by auto }
qed

lemma randomness_sampled_recorded_rec_message:
  "\<lbrakk> Some s' \<in> results'_gpv (rec_message \<A> \<eta> s t); A,\<eta>,\<Gamma> \<turnstile>t t : PublicT; randomness_sampled_recorded \<Gamma> s; randomness_sampled_recorded_assms \<Gamma> \<eta> \<rbrakk>
  \<Longrightarrow> randomness_sampled_recorded \<Gamma> s'"
by(clarsimp simp add: rec_message_def randomness_sampled_recorded_\<beta> split: option.split_asm)

lemma randomness_sampled_recorded_gen_message:
  "\<lbrakk> Some (t, s') \<in> results'_gpv (gen_message \<A> \<eta> s); randomness_sampled_recorded \<Gamma> s \<rbrakk> \<Longrightarrow> randomness_sampled_recorded \<Gamma> s'"
by(auto simp add: gen_message_def)

lemma randomness_sampled_recorded_sim_control:
  "\<lbrakk> Some (m', s') \<in> results'_gpv (sim_control \<A> \<eta> s m); randomness_sampled_recorded \<Gamma> s \<rbrakk> \<Longrightarrow> randomness_sampled_recorded \<Gamma> s'"
by(auto simp add: sim_control_def)

lemma randomness_sampled_recorded_info_sim:
  "\<And>id. \<lbrakk> Some s' \<in> results'_gpv (sim_info_sim \<Gamma> \<A> \<eta> s id to); randomness_sampled_recorded \<Gamma> s \<rbrakk> \<Longrightarrow> randomness_sampled_recorded \<Gamma> s'"
by(cases to)(auto split: if_split_asm t.split nonce.split option.split simp add: Let_def)

lemma randomness_sampled_recorded_hybrid_step1:
  assumes step: "Some (to, (s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step1' (sim' \<Gamma> \<A> \<eta>) ((s,v,f),\<sigma>))"
  and ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v"
  and kst: "key_safe_trace A \<eta> \<Gamma> f"
  and rsr: "randomness_sampled_recorded \<Gamma> \<sigma>"
  and *: "randomness_sampled_recorded_assms \<Gamma> \<eta>"
  shows "randomness_sampled_recorded \<Gamma> \<sigma>'"
proof(cases v)
  case (Output id out succ)
  with ks have "out < length f" "node_type_of_t \<Gamma> (f ! out) = PublicT" by(auto simp add: key_safe_simps)
  with kst have "A,\<eta>,\<Gamma> \<turnstile>t f ! out : PublicT" by(auto dest: key_safe_traceD)
  with Output step rsr * show ?thesis by(auto dest: randomness_sampled_recorded_rec_message simp add: gpv.set_map)
next
  { case Computation   with step rsr show ?thesis by(simp split: computation.split_asm if_split_asm option.split_asm) }
  { case Input         with step rsr show ?thesis by(auto split: if_split_asm elim: randomness_sampled_recorded_gen_message simp add: gpv.set_map Some_eq_bind_conv) }
  { case Control       with step rsr show ?thesis by(auto elim: randomness_sampled_recorded_sim_control simp add: gpv.set_map) }
  { case Probabilistic with step rsr show ?thesis by(auto) }
qed

lemma randomness_sampled_recorded_hybrid_step:
  "\<lbrakk> Some ((s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step' (sim' \<Gamma> \<A> \<eta>) ((s,v,f),\<sigma>)); 
     key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v; 
     key_safe_trace A \<eta> \<Gamma> f; 
     randomness_sampled_recorded \<Gamma> \<sigma>;
     randomness_sampled_recorded_assms \<Gamma> \<eta> \<rbrakk>
  \<Longrightarrow> randomness_sampled_recorded \<Gamma> \<sigma>'"
by(auto 4 3 simp add: hybrid_step_def gpv.set_map dest: randomness_sampled_recorded_hybrid_step1 randomness_sampled_recorded_info_sim split: option.split_asm)

end

end