theory CoSP_PKES_Parsing imports
  CoSP_PKES_Simulator
begin

subsubsection \<open>Parsing\<close>

context comp_impl_base begin

context fixes \<eta> :: security begin

partial_function (option') parse_string :: "bitstring \<Rightarrow> t option"
where
  "parse_string m =
  (if \<T>_string A \<eta> m then 
    (case A_unstring A \<eta> False m of None \<Rightarrow>
       (case A_unstring A \<eta> True m of None \<Rightarrow> 
          if A_empty A \<eta> = Some m then Some EMPTY else Some (garbage (N (ANonce m)))
        | Some m' \<Rightarrow> map_option (STRING True) (parse_string m'))  (* recursive invocation of \<tau> on m' missing in CoSP paper *)
     | Some m' \<Rightarrow> map_option (STRING False) (parse_string m'))    (* recursive invocation of \<tau> on m' missing in CoSP paper *)
   else Some (garbage (N (ANonce m))))"

lemma parse_string_induct [consumes 1, case_names False True EMPTY garbage \<T>]:
  assumes major: "parse_string m = Some t"
  and False: "\<And>m t m'. \<lbrakk> \<T>_string A \<eta> m; A_unstring A \<eta> False m = Some m'; parse_string m' = Some t; P m' t \<rbrakk> \<Longrightarrow> P m (STRING False t)"
  and True: "\<And>m t m'. \<lbrakk> \<T>_string A \<eta> m; A_unstring A \<eta> False m = None; A_unstring A \<eta> True m = Some m'; parse_string m' = Some t; P m' t \<rbrakk> \<Longrightarrow> P m (STRING True t)"
  and EMPTY: "\<And>m. \<lbrakk> \<T>_string A \<eta> m; A_unstring A \<eta> False m = None; A_unstring A \<eta> True m = None; A_empty A \<eta> = Some m \<rbrakk> \<Longrightarrow> P m EMPTY"
  and garbage: "\<And>m. \<lbrakk> \<T>_string A \<eta> m; A_unstring A \<eta> False m = None; A_unstring A \<eta> True m = None; A_empty A \<eta> \<noteq> Some m \<rbrakk> \<Longrightarrow> P m (garbage (N (ANonce m)))"
  and \<T>: "\<And>m. \<not> \<T>_string A \<eta> m \<Longrightarrow> P m (garbage (N (ANonce m)))"
  shows "P m t"
using major
proof(induction rule: parse_string.raw_induct[rotated, consumes 1])
  case (1 parse_string' m t)
  have *: "parse_string x = Some y" if "parse_string' x = Some y" for x y
    using "1.hyps"(2)[of x] that by(simp add: flat_ord_def)
  show ?case using "1.IH" "1.hyps"(1)
    by(auto split: if_split_asm option.split_asm intro: False True EMPTY garbage \<T> * )
qed

lemma parse_string_neq_EK [simp]: "parse_string m \<noteq> Some (EK t)"
by(subst parse_string.simps)(auto split: option.split)

lemma parse_string_neq_VK [simp]: "parse_string m \<noteq> Some (VK t)"
by(subst parse_string.simps)(auto split: option.split)

end

end

text \<open>Decryption of ciphertexts is deterministic, so we can use resumptions instead of GPVs for parsing\<close>

locale comp_impl_parse_base = comp_impl_base A 
  for A :: implementation
  +
  fixes decryption :: "security \<Rightarrow> protocol_nonce \<Rightarrow> cipher \<Rightarrow> 'more simulator_state'_scheme \<Rightarrow> (plain option, 'call, 'ret) resumption"
begin

context
  fixes \<eta> :: security
  and s :: "'more simulator_state'_scheme"
  notes [[function_internals]]
begin

partial_function (resumption) \<tau> :: "bitstring \<Rightarrow> (t, 'call, 'ret) resumption"
where
  "\<tau> m =
  (if \<T>_nonce A \<eta> m then 
     (if \<exists>n. n \<notin> randomness s \<and> pnonce s n = Some m then
        DONE (N (PNonce (SOME n. n \<notin> (randomness s) \<and> pnonce s n = Some m)))
      else DONE (N (ANonce m)))
   else if \<T>_cipher A \<eta> m then
     (if m \<in> ran (history s) then DONE (SOME t. history s t = Some m)
       (* we do not check that t has the format as required in the CoSP definition, 
          because the check should always succeed by construction of outputs *)
      else (do {
          t1 \<leftarrow> \<tau> (the (A_ekof A \<eta> m)); (* By implementation condition 9, A_ekof is always defined *)
          case t1 of EK (N (PNonce n')) \<Rightarrow> do {
              plain \<leftarrow> decryption \<eta> n' m s;
              (case plain of None \<Rightarrow> DONE (garbageE (EK (N (PNonce n'))) (N (ANonce m)))
              | Some plain' \<Rightarrow> do {
                  t2 \<leftarrow> \<tau> plain';
                  DONE (E (EK (N (PNonce n'))) t2 (N (ANonce m)))
                })
            }
          | _ \<Rightarrow> DONE (garbageE t1 (N (ANonce m)))
      }))
   else if \<T>_ekey A \<eta> m then
     (if m \<in> ran (history s) then
        DONE (SOME t. history s t =  Some m) (* Again, no check for the right format *)
      else DONE (EK (N (ANonce m))))
   else if \<T>_signature A \<eta> m then
     (if m \<in> ran (history s) then
        DONE (SOME t. history s t = Some m) (* Again, no check for the right format *)
      else (do {
          vk \<leftarrow> \<tau> (the (A_vkof A \<eta> m)); (* By implementation condition 10, A_vkof is always defined *)
          (case vk of VK (N n) \<Rightarrow>
             (case A_verify A \<eta> (the (A_vkof A \<eta> m)) m of Some m' \<Rightarrow> do {
                  t2 \<leftarrow> \<tau> m';
                  DONE (SIG (SK (N n)) t2 (N (ANonce m)))
                }
              | None \<Rightarrow> DONE (garbageSIG (VK (N n)) (N (ANonce m))))
           | t \<Rightarrow> DONE (garbageSIG t (N (ANonce m))))
      }))
    else if \<T>_vkey A \<eta> m then
      (if m \<in> ran (history s) then
         DONE (SOME t. history s t = Some m) (* Again, no check for the right format *)
       else DONE (VK (N (ANonce m))))
    else if \<T>_pair A \<eta> m then do {
         t1 \<leftarrow> \<tau> (the (A_fst A \<eta> m));
         t2 \<leftarrow> \<tau> (the (A_snd A \<eta> m));
         DONE (PAIR t1 t2)
       }
    else if \<T>_string A \<eta> m then resumption.Done (parse_string \<eta> m)
    else DONE (garbage (N (ANonce m))))"

lemmas \<tau>_induct [case_names adm bot step]
  = resumption.fixp_induct_strong2_uc[of "\<lambda>x. x" _ "\<lambda>x. x", OF \<tau>.mono \<tau>_def refl]

end

end

lemma (in comp_impl_base) \<tau>_cases:
  obtains (Nonce) "\<T>_nonce A \<eta> m"
  | (Cipher) "\<T>_cipher A \<eta> m" "\<not> \<T>_nonce A \<eta> m"
  | (Ekey) "\<T>_ekey A \<eta> m" "\<not> \<T>_nonce A \<eta> m" "\<not> \<T>_cipher A \<eta> m"
  | (Signature) "\<T>_signature A \<eta> m" "\<not> \<T>_nonce A \<eta> m" "\<not> \<T>_cipher A \<eta> m" "\<not> \<T>_ekey A \<eta> m"
  | (Vkey) "\<T>_vkey A \<eta> m" "\<not> \<T>_nonce A \<eta> m" "\<not> \<T>_cipher A \<eta> m" "\<not> \<T>_ekey A \<eta> m" "\<not> \<T>_signature A \<eta> m"
  | (Pair) "\<T>_pair A \<eta> m" "\<not> \<T>_nonce A \<eta> m" "\<not> \<T>_cipher A \<eta> m" "\<not> \<T>_ekey A \<eta> m" "\<not> \<T>_signature A \<eta> m" "\<not> \<T>_vkey A \<eta> m"
  | (String) "\<T>_string A \<eta> m" "\<not> \<T>_nonce A \<eta> m" "\<not> \<T>_cipher A \<eta> m" "\<not> \<T>_ekey A \<eta> m" "\<not> \<T>_signature A \<eta> m" "\<not> \<T>_vkey A \<eta> m" "\<not> \<T>_pair A \<eta> m"
  | (garbage) "\<not> \<T>_nonce A \<eta> m" "\<not> \<T>_cipher A \<eta> m" "\<not> \<T>_ekey A \<eta> m" "\<not> \<T>_signature A \<eta> m" "\<not> \<T>_vkey A \<eta> m" "\<not> \<T>_pair A \<eta> m" "\<not> \<T>_string A \<eta> m"
by blast

locale comp_impl_parse = comp_impl_parse_base A decryption + comp_impl A
  for A :: implementation
  and decryption :: "security \<Rightarrow> protocol_nonce \<Rightarrow> cipher \<Rightarrow> 'more simulator_state'_scheme \<Rightarrow> (plain option, 'call, 'ret) resumption"
begin
  
lemma \<tau>_simps:
  shows \<tau>_nonce: 
    "\<T>_nonce A \<eta> m \<Longrightarrow> \<tau> \<eta> s m = 
     (if \<exists>n. n \<notin> randomness s \<and> pnonce s n = Some m then
        DONE (N (PNonce (SOME n. n \<notin> randomness s \<and> pnonce s n = Some m)))
      else DONE (N (ANonce m)) )"
  and \<tau>_ekey:
    "\<T>_ekey A \<eta> m \<Longrightarrow> \<tau> \<eta> s m = 
     (if m \<in> ran (history s) then DONE (SOME t. history s t = Some m) else DONE (EK (N (ANonce m))))"
  and \<tau>_vkey:
    "\<T>_vkey A \<eta> m \<Longrightarrow> \<tau> \<eta> s m =
     (if m \<in> ran (history s) then DONE (SOME t. history s t = Some m) else DONE (VK (N (ANonce m))))"
  and \<tau>_cipher:
    "\<T>_cipher A \<eta> m \<Longrightarrow> \<tau> \<eta> s m = 
      (if m \<in> ran (history s) then DONE (SOME t. history s t = Some m)
       else do {
            t1 \<leftarrow> \<tau> \<eta> s (the (A_ekof A \<eta> m));
            case t1 of EK (N (PNonce n')) \<Rightarrow> do {
                plain \<leftarrow> decryption \<eta> n' m s;
                (case plain of None \<Rightarrow> DONE (garbageE (EK (N (PNonce n'))) (N (ANonce m)))
                | Some plain' \<Rightarrow> do {
                    t2 \<leftarrow> \<tau> \<eta> s plain';
                    DONE (E (EK (N (PNonce n'))) t2 (N (ANonce m)))
                  })
              }
            | _ \<Rightarrow> DONE (garbageE t1 (N (ANonce m)))
          })"
  and \<tau>_signature:
    "\<T>_signature A \<eta> m \<Longrightarrow> \<tau> \<eta> s m = 
      (if m \<in> ran (history s) then DONE (SOME t. history s t = Some m)
       else (do {
            vk \<leftarrow> \<tau> \<eta> s (the (A_vkof A \<eta> m));
            (case vk of VK (N n) \<Rightarrow>
               (case A_verify A \<eta> (the (A_vkof A \<eta> m)) m of Some m' \<Rightarrow> do {
                    t2 \<leftarrow> \<tau> \<eta> s m';
                    DONE (SIG (SK (N n)) t2 (N (ANonce m)))
                  }
                | None \<Rightarrow> DONE (garbageSIG (VK (N n)) (N (ANonce m))))
             | t \<Rightarrow> DONE (garbageSIG t (N (ANonce m))))
         }))"
  and \<tau>_pair: 
    "\<T>_pair A \<eta> m \<Longrightarrow> \<tau> \<eta> s m = do {
       t1 \<leftarrow> \<tau> \<eta> s (the (A_fst A \<eta> m));
       t2 \<leftarrow> \<tau> \<eta> s (the (A_snd A \<eta> m));
       DONE (PAIR t1 t2)
     }"
  and \<tau>_string: "\<T>_string A \<eta> m \<Longrightarrow> \<tau> \<eta> s m = resumption.Done (parse_string \<eta> m)"
  and \<tau>_simps_garbage:
    "\<lbrakk> \<not> \<T>_nonce A \<eta> m;  \<not> \<T>_cipher A \<eta> m; \<not> \<T>_ekey A \<eta> m; \<not> \<T>_signature A \<eta> m; \<not> \<T>_vkey A \<eta> m; \<not> \<T>_pair A \<eta> m; \<not> \<T>_string A \<eta> m \<rbrakk>
    \<Longrightarrow> \<tau> \<eta> s m = DONE (garbage (N (ANonce m)))"
apply(subst \<tau>.simps, ((rule if_False_eq, erule (1) \<T>_disjoint')+)?, simp split del: if_split; fail)+
done

lemma \<tau>_results_induct [consumes 1,
  case_names history PNonce ANonce 
    Cipher_garbageE_key Cipher_garbageE Cipher
    Ekey
    Signature_garbageSIG_key Signature_garbageSIG Signature
    Vkey
    Pair String garbage]:
  assumes "t \<in> results (\<tau> \<eta> \<sigma> m)"
  and history: "\<And>m m'. \<lbrakk> history \<sigma> m' = Some m; \<T>_cipher A \<eta> m \<or> \<T>_ekey A \<eta> m \<or> \<T>_signature A \<eta> m \<or> \<T>_vkey A \<eta> m \<rbrakk> \<Longrightarrow> Q m' m"
  and PNonce: "\<And>m n. \<lbrakk> \<T>_nonce A \<eta> m; n \<notin> randomness \<sigma>; pnonce \<sigma> n = Some m \<rbrakk> \<Longrightarrow> Q (N (PNonce n)) m"
  and ANonce: "\<And>m. \<lbrakk> \<T>_nonce A \<eta> m; \<not> (\<exists>n. n \<notin> randomness \<sigma> \<and> pnonce \<sigma> n = Some m) \<rbrakk> \<Longrightarrow> Q (N (ANonce m)) m"

  and cipher_garE_key:
    "\<And>m key key_t. \<lbrakk> \<T>_cipher A \<eta> m; m \<notin> ran (history \<sigma>); A_ekof A \<eta> m = Some key;
        key_t \<in> results (\<tau> \<eta> \<sigma> key); \<not> (\<exists>n'. key_t = EK (N (PNonce n'))); Q key_t key \<rbrakk>
     \<Longrightarrow> Q (garbageE key_t (N (ANonce m))) m"
  and cipher_garE_EK:
    "\<And>m key n. \<lbrakk> \<T>_cipher A \<eta> m; m \<notin> ran (history \<sigma>); A_ekof A \<eta> m = Some key;
        EK (N (PNonce n)) \<in> results (\<tau> \<eta> \<sigma> key); None \<in> results (decryption \<eta> n m \<sigma>) \<rbrakk>
     \<Longrightarrow> Q (garbageE (EK (N (PNonce n))) (N (ANonce m))) m"
  and cipher_E: "\<And>m key n plain m'.
    \<lbrakk> \<T>_cipher A \<eta> m; m \<notin> ran (history \<sigma>); A_ekof A \<eta> m = Some key; EK (N (PNonce n)) \<in> results (\<tau> \<eta> \<sigma> key); 
      Some plain \<in> results (decryption \<eta> n m \<sigma>); m' \<in> results (\<tau> \<eta> \<sigma> plain); Q m' plain \<rbrakk>
    \<Longrightarrow> Q (E (EK (N (PNonce n))) m' (N (ANonce m))) m"

  and ekey_EK: "\<And>m. \<lbrakk> \<T>_ekey A \<eta> m; m \<notin> ran (history \<sigma>) \<rbrakk> \<Longrightarrow> Q (EK (N (ANonce m))) m"

  and sig_garSIG_key:
    "\<And>m vkey vkey_t. \<lbrakk> \<T>_signature A \<eta> m; m \<notin> ran (history \<sigma>); A_vkof A \<eta> m = Some vkey; 
       vkey_t \<in> results (\<tau> \<eta> \<sigma> vkey); \<forall>n'. vkey_t \<noteq> VK (N n'); Q vkey_t vkey \<rbrakk>
     \<Longrightarrow> Q (garbageSIG vkey_t (N (ANonce m))) m"
  and sig_garSIG_SK:
    "\<And>m vkey n. \<lbrakk> \<T>_signature A \<eta> m; m \<notin> ran (history \<sigma>); A_vkof A \<eta> m = Some vkey;
       VK (N n) \<in> results (\<tau> \<eta> \<sigma> vkey); A_verify A \<eta> vkey m = None \<rbrakk>
    \<Longrightarrow> Q (garbageSIG (VK (N n)) (N (ANonce m))) m"
  and sig_SIG:
    "\<And>m vkey n' m' plain. \<lbrakk> \<T>_signature A \<eta> m; m \<notin> ran (history \<sigma>); A_vkof A \<eta> m = Some vkey;
        VK (N n') \<in> results (\<tau> \<eta> \<sigma> vkey); A_verify A \<eta> vkey m = Some plain; m' \<in> results (\<tau> \<eta> \<sigma> plain); Q m' plain \<rbrakk>
     \<Longrightarrow> Q (SIG (SK (N n')) m' (N (ANonce m))) m"

  and vkey_VK: "\<And>m. \<lbrakk> \<T>_vkey A \<eta> m; m \<notin> ran (history \<sigma>) \<rbrakk> \<Longrightarrow> Q (VK (N (ANonce m))) m"
  
  and pair: "\<And>m t1 t2. \<lbrakk> \<T>_pair A \<eta> m; t1 \<in> results (\<tau> \<eta> \<sigma> (the (A_fst A \<eta> m))); 
      t2 \<in> results (\<tau> \<eta> \<sigma> (the (A_snd A \<eta> m))); Q t1 (the (A_fst A \<eta> m)) ; Q t2 (the (A_snd A \<eta> m)) \<rbrakk>
    \<Longrightarrow>  Q (PAIR t1 t2) m"

  and string: "\<And>m t. \<lbrakk> \<T>_string A \<eta> m; parse_string \<eta> m = Some t \<rbrakk> \<Longrightarrow> Q t m"

  and garbage: 
    "\<And>m. \<lbrakk> \<not> \<T>_nonce A \<eta> m;  \<not> \<T>_cipher A \<eta> m; \<not> \<T>_ekey A \<eta> m; \<not> \<T>_signature A \<eta> m; \<not> \<T>_vkey A \<eta> m; \<not> \<T>_pair A \<eta> m; \<not> \<T>_string A \<eta> m \<rbrakk> 
    \<Longrightarrow> Q (garbage (N (ANonce m))) m"

  shows "Q t m"
proof -
  have "results (\<tau> \<eta> \<sigma> m) \<subseteq> {t. Q t m}"
  proof(induction arbitrary: m rule: \<tau>_induct)
    case adm show ?case by(rule admissible_all)(rule ccpo_class.admissible_leI; simp)
  next
    case bot show ?case by simp
  next
    case (step \<tau>' m)
    from step(1) have results: "t \<in> results (\<tau>' m) \<Longrightarrow> t \<in> results (\<tau> \<eta> \<sigma> m)" for t m
      by(auto simp add: fun_ord_def dest: mcont_results[THEN mcont_monoD])
    have IH: "t \<in> results (\<tau>' m) \<Longrightarrow> Q t m" for t m using step.IH[of m] by blast
    show ?case
    proof(cases \<eta> m rule: \<tau>_cases)
      case Nonce
      let ?P = "\<lambda>n. n \<notin> randomness \<sigma> \<and> pnonce \<sigma> n = Some m"
      show ?thesis
      proof(cases "\<exists>n. ?P n")
        case True
        then have "?P (Eps ?P)" by(rule someI_ex)
        then show ?thesis using Nonce True  PNonce[of m "Eps ?P"] by simp
      next 
        case False
        then show ?thesis using Nonce ANonce[of m] False by auto
      qed
    next
      case Cipher
      show ?thesis 
      proof(cases "m \<in> ran (history \<sigma>)")
        case True
        let ?P = "\<lambda>t. history \<sigma> t = Some m"
        from True have "?P (Eps ?P)" by(auto intro: someI simp add: ran_def)
        from history[OF this] True Cipher show ?thesis by simp
      next
        case hist: False
        from Cipher \<T>_cipher_A_ekofD[of \<eta> m] obtain key where key: "A_ekof A \<eta> m = Some key" by auto
        have "results (do {
               t1 \<leftarrow> \<tau>' key;
               case t1 of EK (N (PNonce n')) \<Rightarrow> do {
                 plain \<leftarrow> decryption \<eta> n' m \<sigma>;
                 (case plain of None \<Rightarrow> DONE (garbageE (EK (N (PNonce n'))) (N (ANonce m)))
                  | Some plain' \<Rightarrow> do {
                        t2 \<leftarrow> \<tau>' plain';
                        DONE (E (EK (N (PNonce n'))) t2 (N (ANonce m)))
                      })}
               | _ \<Rightarrow> DONE (garbageE t1 (N (ANonce m)))
             })
           \<subseteq> {t. Q t m}" (is "?lhs \<subseteq> ?rhs")
        proof
          fix t
          assume "t \<in> ?lhs"
          then obtain key_t
            where key_t: "key_t \<in> results (\<tau>' key)"
            and t: "t \<in> results (case key_t of EK (N (PNonce n')) \<Rightarrow> do {
                    plain \<leftarrow> decryption \<eta> n' m \<sigma>;
                    (case plain of None \<Rightarrow> DONE (garbageE (EK (N (PNonce n'))) (N (ANonce m)))
                    | Some plain' \<Rightarrow> do {
                        t2 \<leftarrow> \<tau>' plain';
                        DONE (E (EK (N (PNonce n'))) t2 (N (ANonce m)))
                      })
                  }
                | _ \<Rightarrow> DONE (garbageE key_t (N (ANonce m))))"
            by(auto simp add: results_bind_resumption)
          consider (garbage) "\<And>n'. key_t \<noteq> EK (N (PNonce n'))" | (EK) n' where "key_t = EK (N (PNonce n'))" by blast
          then show "t \<in> ?rhs"
          proof(cases)
            case garbage
            with t have "t = garbageE key_t (N (ANonce m))" by(auto split: t.split_asm nonce.split_asm)
            with IH[of key_t key] key key_t t hist cipher_garE_key[OF Cipher(1) hist key key_t[THEN results]] garbage Cipher
            show ?thesis by(auto)
          next
            case (EK n')
            then obtain plain where plain: "plain \<in> results (decryption \<eta> n' m \<sigma>)"
              and t: "t \<in> results (case plain of None \<Rightarrow> DONE (garbageE (EK (N (PNonce n'))) (N (ANonce m)))
                    | Some plain' \<Rightarrow> do {
                        t2 \<leftarrow> \<tau>' plain';
                        DONE (E (EK (N (PNonce n'))) t2 (N (ANonce m)))
                      })"
              using t by(auto simp add: results_bind_resumption)
            then show ?thesis
              using cipher_garE_EK[OF Cipher(1) hist key, of n'] cipher_E[OF Cipher(1) hist key, of n' "the plain"] 
                key_t key EK IH[of _ "the plain"] Cipher
              by(auto split: option.split_asm simp add: results_bind_resumption intro: results)
          qed
        qed
        thus ?thesis using Cipher hist key by simp
      qed
    next
      case Ekey
      show ?thesis
      proof(cases "m \<in> ran (history \<sigma>)")
        case True
        let ?P = "\<lambda>t. history \<sigma> t = Some m"
        from True have "?P (Eps ?P)" by(auto intro: someI simp add: ran_def)
        from history[OF this] True Ekey show ?thesis by(auto)
      next
        case hist: False
        with ekey_EK[OF Ekey(1) this] Ekey show ?thesis by simp
      qed
    next
      case Signature
      show ?thesis
      proof(cases "m \<in> ran (history \<sigma>)")
        case True
        let ?P = "\<lambda>t. history \<sigma> t = Some m"
        from True have "?P (Eps ?P)" by(auto intro: someI simp add: ran_def)
        from history[OF this] Signature True show ?thesis apply(auto) done
      next
        case hist: False
        from Signature \<T>_signature_A_vkofD[of \<eta> m]
        obtain key where key: "A_vkof A \<eta> m = Some key" by auto
        have "results (do {
              key_t \<leftarrow> \<tau>' key;
              case key_t of VK (N n) \<Rightarrow>
                 (case A_verify A \<eta> key m of Some m' \<Rightarrow> do {
                      t2 \<leftarrow> \<tau>' m';
                      DONE (SIG (SK (N n)) t2 (N (ANonce m)))
                    }
                  | None \<Rightarrow> DONE (garbageSIG (VK (N n)) (N (ANonce m))))
               | t \<Rightarrow> DONE (garbageSIG t (N (ANonce m))) }) 
           \<subseteq> {t. Q t m}" (is "?lhs \<subseteq> ?rhs")
        proof
          fix t
          assume "t \<in> ?lhs"
          then obtain key_t
            where key_t: "key_t \<in> results (\<tau>' key)"
            and t: "t \<in> results (case key_t of VK (N n) \<Rightarrow>
                   (case A_verify A \<eta> key m of Some m' \<Rightarrow> do {
                        t2 \<leftarrow> \<tau>' m';
                        DONE (SIG (SK (N n)) t2 (N (ANonce m)))
                      }
                    | None \<Rightarrow> DONE (garbageSIG (VK (N n)) (N (ANonce m))))
                 | t \<Rightarrow> DONE (garbageSIG t (N (ANonce m))))"
            by(auto simp add: results_bind_resumption)
          consider (garbage) "\<And>n'. key_t \<noteq> VK (N n')" | (VK) n' where "key_t = VK (N n')" by blast
          then show "t \<in> ?rhs"
          proof(cases)
            case garbage
            with t have "t = garbageSIG key_t (N (ANonce m))" by(auto split: t.split_asm nonce.split_asm)
            with IH[of key_t key] key key_t t hist sig_garSIG_key[OF Signature(1) hist key key_t[THEN results]] garbage Signature
            show ?thesis by(auto)
          next
            case (VK n')
            with sig_garSIG_SK[OF Signature(1) hist key] sig_SIG[OF Signature(1) hist key] key_t t key
            show ?thesis by(auto split: option.split_asm simp add: results_bind_resumption intro: results IH)
          qed
        qed
        then show ?thesis using Signature hist key by simp
      qed
    next
      case Vkey
      show ?thesis 
      proof(cases "m \<in> ran (history \<sigma>)")
        case True
        let ?P = "\<lambda>t. history \<sigma> t = Some m"
        from True have "?P (Eps ?P)" by(auto intro: someI simp add: ran_def)
        from history[OF this] True Vkey show ?thesis by(auto)
      next
        case hist: False
        with vkey_VK[OF Vkey(1) this] Vkey show ?thesis by simp
      qed
    next
      case Pair
      from pair[OF Pair(1)] Pair
      show ?thesis by(auto simp add: results_bind_resumption intro: IH results)
    next
      case String
      with string[OF String(1)] show ?thesis by auto
    next
      case *: garbage
      with garbage[OF this] show ?thesis by simp
    qed
  qed
  with assms(1) show ?thesis by auto
qed

end

text \<open>Assumptions needed to guarantee that the parsing function @{text \<tau>} terminates.\<close>

locale comp_impl_parse_termination = comp_impl_parse A decryption
  for A :: implementation
  and decryption :: "security \<Rightarrow> protocol_nonce \<Rightarrow> cipher \<Rightarrow> 'more simulator_state'_scheme \<Rightarrow> (plain option, 'call, 'ret) resumption"
  +
  assumes A_d_shorter: "\<lbrakk> A_d A \<eta> dk m = Some m'; \<T>_dkey A \<eta> dk \<rbrakk> \<Longrightarrow> length m' < length m"
  and A_verify_shorter: "\<lbrakk> A_verify A \<eta> vk m = Some m'; \<T>_vkey A \<eta> vk \<rbrakk> \<Longrightarrow> length m' < length m"
  and A_pair_longer1: "length m1 < length (A_pair A \<eta> m1 m2)"
  and A_pair_longer2: "length m2 < length (A_pair A \<eta> m1 m2)"
  -- \<open>We assume that there is a well-founded relation on bitstrings that ensure termination of the
    @{term A_unstring} operation. Taking the length as measure (as we do for @{term A_d},
    @{term A_verify}, and @{term A_pair}) does not seem sensible, because we would require a decrease
    for every bit. If padding to at least 8 bits is used, this restriction could be too strong.
    On the other hand, it seems sensible to take the length as measure for decryption, signature
    verification and pairing as they only extract proper subterms (signatures contain the
    verification key, encryptions the encryption key, pairs both components)\<close>
  and decryption_shorter: "\<And>s. \<lbrakk> Some m' \<in> results (decryption \<eta> n m s); \<T>_cipher A \<eta> m \<rbrakk> \<Longrightarrow> length m' < length m"
  fixes rel_string :: "security \<Rightarrow> (bitstring \<times> bitstring) set"
  assumes wf_rel_string: "wf (rel_string \<eta>)"
  and A_unstring_shorter: "\<lbrakk> \<T>_string A \<eta> s; A_unstring A \<eta> b s = Some s' \<rbrakk> \<Longrightarrow> (s', s) \<in> rel_string \<eta>"
begin

lemma parse_string_induct:
  "(\<And>m. (\<And>m'. \<lbrakk>\<T>_string A \<eta> m; A_unstring A \<eta> True m = Some m'\<rbrakk> \<Longrightarrow> P m') \<Longrightarrow> P m) \<Longrightarrow> P m"
apply induction_schema
subgoal by pat_completeness
apply(rule wf_rel_string)
apply(blast intro: A_unstring_shorter)
done

context fixes \<eta> :: security begin
text \<open>
  The recursive calls on encryption and verification keys cannot lead to cycles because
  they cannot contain arbitrary other terms. Therefore, we prove termination using the
  lexicographic combination of two measures. The first is 1 iff the message can contain
  terms of arbitrary size. The second is just the length of the string.
\<close>
private definition type_order :: "bitstring \<Rightarrow> nat"
where "type_order m = (if \<T>_cipher A \<eta> m \<or> \<T>_signature A \<eta> m \<or> \<T>_pair A \<eta> m then 1 else 0)"

private lemma type_order_simps:
  "\<T>_nonce A \<eta> m \<Longrightarrow> type_order m = 0"
  "\<T>_cipher A \<eta> m \<Longrightarrow> type_order m = 1"
  "\<T>_ekey A \<eta> m \<Longrightarrow> type_order m = 0"
  "\<T>_dkey A \<eta> m \<Longrightarrow> type_order m = 0"
  "\<T>_signature A \<eta> m \<Longrightarrow> type_order m = 1"
  "\<T>_vkey A \<eta> m \<Longrightarrow> type_order m = 0"
  "\<T>_pair A \<eta> m \<Longrightarrow> type_order m = 1"
  "\<T>_string A \<eta> m \<Longrightarrow> type_order m = 0"
by(auto simp add: type_order_def dest: \<T>_disjoint')

private lemma type_order_neq_1: "type_order m \<noteq> Suc 0 \<longleftrightarrow> type_order m = 0"
  and type_order_neq_0: "type_order m \<noteq> 0 \<longleftrightarrow> type_order m = 1"
by(simp_all add: type_order_def)

lemma \<tau>_induct:
  "(\<And>m. \<lbrakk> \<T>_cipher A \<eta> m \<Longrightarrow> P (the (A_ekof A \<eta> m));
       \<And>n plain. \<lbrakk> \<T>_cipher A \<eta> m; EK (N (PNonce n)) \<in> results (\<tau> \<eta> s (the (A_ekof A \<eta> m)));
         Some plain \<in> results (decryption \<eta> n m s) \<rbrakk> \<Longrightarrow> P plain;
       \<T>_signature A \<eta> m \<Longrightarrow> P (the (A_vkof A \<eta> m));
       \<And>n msg.
          \<lbrakk>\<T>_signature A \<eta> m; VK (N n) \<in> results (\<tau> \<eta> s (the (A_vkof A \<eta> m))); 
           A_verify A \<eta> (the (A_vkof A \<eta> m)) m = Some msg\<rbrakk>
          \<Longrightarrow> P msg;
       \<T>_pair A \<eta> m \<Longrightarrow> P (the (A_fst A \<eta> m));
       \<T>_pair A \<eta> m \<Longrightarrow> P (the (A_snd A \<eta> m)) \<rbrakk>
    \<Longrightarrow> P m)
   \<Longrightarrow> P m"
apply induction_schema
supply type_order_simps[simp] type_order_neq_1[simp]
subgoal by pat_completeness
apply(relation "measures [type_order, length]")
subgoal by simp
subgoal by(auto 4 4 split: if_split_asm dest: \<T>_cipher_A_ekofD A_type_correct)
subgoal
  apply(clarsimp split: bind_split_asm if_split_asm)
  apply(frule \<T>_cipher_A_ekofD, clarsimp)
  apply(drule A_type_correct)+
  apply(drule (1) decryption_shorter)
  apply simp
  done
subgoal by(auto 4 4 split: if_split_asm dest: \<T>_signature_A_vkofD A_type_correct)
subgoal
  apply (simp split: if_split_asm)
  apply(frule \<T>_signature_A_vkofD, clarsimp)
  apply(drule A_type_correct)+
  apply(drule (1) A_verify_shorter)
  apply clarsimp
  done
subgoal
  apply clarsimp
  apply(frule A_pair_exhaust)
  apply(clarsimp simp add: A_fst_pair)
  using A_pair_longer1 apply simp
  done
subgoal
  apply clarsimp
  apply(frule A_pair_exhaust)
  apply(clarsimp simp add: A_snd_pair)
  using A_pair_longer2 apply simp
  done
done
end

end

context comp_impl begin

lemma \<T>_of_t_parse_string: "parse_string \<eta> m = Some t \<Longrightarrow> \<T>_of_t \<eta> t m"
by(subst (asm) parse_string.simps)(auto split: option.split_asm if_split_asm)

end

context comp_impl_parse begin
                   
lemma \<T>_of_t_\<tau>:
  assumes t: "t \<in> results (\<tau> \<eta> s m)"
  and hist: "\<T>_history \<eta> (history s)"
  shows "\<T>_of_t \<eta> t m"
using t
proof(induction rule: \<tau>_results_induct)
  case (history m t)
  from hist history(1) show ?case by(rule \<T>_historyD)
qed(simp_all add: \<T>_of_t_parse_string)

lemma EK_PNonce_in_\<tau>D: -- \<open>The converse does not hold because @{text \<tau>} uses choice\<close>
  "EK (N (PNonce n)) \<in> results (\<tau> \<eta> s msg)
  \<Longrightarrow> history s (EK (N (PNonce n))) = Some msg \<and> (\<T>_cipher A \<eta> msg \<or> \<T>_ekey A \<eta> msg \<or> \<T>_signature A \<eta> msg \<or> \<T>_vkey A \<eta> msg)"
by(induction t\<equiv>"EK (N (PNonce n))" msg rule: \<tau>_results_induct) simp_all

lemma VK_PNonce_in_\<tau>D:
  "VK (N (PNonce n)) \<in> results (\<tau> \<eta> s msg)
  \<Longrightarrow> history s (VK (N (PNonce n))) = Some msg \<and> (\<T>_cipher A \<eta> msg \<or> \<T>_ekey A \<eta> msg \<or> \<T>_signature A \<eta> msg \<or> \<T>_vkey A \<eta> msg)"
by(induction t\<equiv>"VK (N (PNonce n))" msg rule: \<tau>_results_induct) auto

lemma VK_ANonce_in_\<tau>D:
  "VK (N (ANonce m)) \<in> results (\<tau> \<eta> s msg)
  \<Longrightarrow> history s (VK (N (ANonce m))) = Some msg \<and> (\<T>_cipher A \<eta> msg \<or> \<T>_ekey A \<eta> msg \<or> \<T>_signature A \<eta> msg \<or> \<T>_vkey A \<eta> msg) \<or> m = msg"
by(induction t\<equiv>"VK (N (ANonce m))" msg rule: \<tau>_results_induct) simp_all

end

end