theory CoSP_PKES_Simulator imports
  CoSP_PKES_Computational
begin

subsection \<open>Simulator\<close>

subsubsection \<open>Type definitions\<close>

type_synonym pnonce = "protocol_nonce \<rightharpoonup> bitstring"
type_synonym history = "t \<rightharpoonup> bitstring"

record simulator_state' =
  fuel :: nat
  pnonce :: pnonce -- \<open>serialisation of protocol nonces\<close>
  randomness :: "protocol_nonce set" -- \<open>randomness nonces visited so far\<close>
  history :: history -- \<open>history of terms output by @{text \<beta>} and the sampled nonces\<close>

record 's simulator_state = simulator_state' +
  adv_state :: 's -- \<open>simulated adversary state\<close>

type_synonym ('s, 'id) cosp_simulator' = "('s simulator_state, 'id, t, comp_impl.call, comp_impl.ret) simulator'"
type_synonym ('s, 'id) cosp_simulator = "('s simulator_state, 'id, t, comp_impl.call, comp_impl.ret) simulator"

translations
  (type) "('s, 'id) cosp_simulator'" \<leftharpoondown> (type) "('s simulator_state, 'id, t, comp_impl.call, comp_impl.ret) simulator'"
  (type) "('s, 'id) cosp_simulator" \<leftharpoondown> (type) "('s simulator_state, 'id, t, comp_impl.call, comp_impl.ret) simulator"

typ "('s, 'id) cosp_simulator'"
typ "('s, 'id) cosp_simulator"

type_synonym 'id cosp_symb_state = "(t, 'id, constructor, nonce, destructor) symb_state"
type_synonym ('id, 's) cosp_hybrid_state = "(t, 'id, constructor, nonce, destructor, 's simulator_state) hybrid_state"
translations
  (type) "'id cosp_symb_state" \<leftharpoondown> (type) "(t, 'id, constructor, nonce, destructor) symb_state"
  (type) "('id, 's) cosp_hybrid_state" \<leftharpoondown> (type) "(t, 'id, constructor, nonce, destructor, 's simulator_state) hybrid_state"
typ "'id cosp_symb_state"
typ "('id, 's) cosp_hybrid_state"

lemma history_update_id [simp]: "history_update (\<lambda>x. x) = id"
by(simp add: fun_eq_iff)

lemma history_pnonce_update_commute:
  "history_update f (pnonce_update g s) = pnonce_update g (history_update f s)"
by(rule simulator_state'.equality) simp_all

context comp_impl begin

context fixes \<eta> :: security begin

primrec \<T>_of_t :: "t \<Rightarrow> bitstring \<Rightarrow> bool"
where
  "\<T>_of_t (E t1 t2 t3) = \<T>_cipher A \<eta>"
| "\<T>_of_t (EK t) = \<T>_ekey A \<eta>"
| "\<T>_of_t (DK t) = \<T>_dkey A \<eta>"
| "\<T>_of_t (SIG t1 t2 t3) = \<T>_signature A \<eta>"
| "\<T>_of_t (VK t) = \<T>_vkey A \<eta>"
| "\<T>_of_t (SK t) = \<T>_sigkey A \<eta>"
| "\<T>_of_t (PAIR t1 t2) = \<T>_pair A \<eta>"
| "\<T>_of_t EMPTY = \<T>_string A \<eta>"
| "\<T>_of_t (STRING b t) = \<T>_string A \<eta>"
| "\<T>_of_t (N n) = \<T>_nonce A \<eta>"
| "\<T>_of_t (garbage t) = (\<lambda>_. True)" (* FIXME: what do we need here? *)
| "\<T>_of_t (garbageE t1 t2) = \<T>_cipher A \<eta>"
| "\<T>_of_t (garbageSIG t1 t2) = \<T>_signature A \<eta>"

definition \<T>_history :: "history \<Rightarrow> bool"
where "\<T>_history hist \<longleftrightarrow> (\<forall>t m. hist t = Some m \<longrightarrow> \<T>_of_t t m)"

lemma \<T>_historyI: "(\<And>t m. hist t = Some m \<Longrightarrow> \<T>_of_t t m) \<Longrightarrow> \<T>_history hist"
by(simp add: \<T>_history_def)

lemma \<T>_historyD: "\<lbrakk> \<T>_history hist; hist t = Some m \<rbrakk> \<Longrightarrow> \<T>_of_t t m"
by(simp add: \<T>_history_def)

lemma \<T>_history_empty [simp]: "\<T>_history Map.empty"
by(simp add: \<T>_history_def)

lemma \<T>_history_updateI: "\<lbrakk> \<T>_history hist; \<T>_of_t t m \<rbrakk> \<Longrightarrow> \<T>_history (hist(t \<mapsto> m))"
by(simp add: \<T>_history_def)

end

end

end