theory CoSP_PKES_Simulator1 imports
  CoSP_PKES_Key_Safety_Simulator
begin

text \<open>Eager and lazy sampling of nonces\<close>

context comp_impl_base begin

context fixes \<eta> :: security begin

definition sample_pnonce' :: "protocol_nonce \<Rightarrow> pnonce pmf \<Rightarrow> pnonce pmf"
where
  "sample_pnonce' n p = do {
    pns \<leftarrow> p;
    bs \<leftarrow> sample_bitstring (size_nonce A \<eta>);
    return_pmf (pns(n \<mapsto> A_nonce A \<eta> bs))
  }"

lemma comp_fun_commute_sample_pnonce': "comp_fun_commute sample_pnonce'"
proof
  fix n n'
  show "sample_pnonce' n \<circ> sample_pnonce' n' = sample_pnonce' n' \<circ> sample_pnonce' n"
    unfolding sample_pnonce'_def o_def bind_assoc_pmf bind_return_pmf
    by(rewrite in "bind_pmf _ \<hole>" bind_commute_pmf)(cases "n = n'"; simp_all add: fun_upd_twist)
qed

interpretation sample_pnonce: comp_fun_commute sample_pnonce'
by(rule comp_fun_commute_sample_pnonce')

lemma comp_fun_idem_sample_pnonce': "comp_fun_idem sample_pnonce'"
by unfold_locales(simp add: sample_pnonce'_def o_def bind_assoc_pmf bind_return_pmf fun_eq_iff)

interpretation sample_nonce: comp_fun_idem sample_pnonce'
by(rule comp_fun_idem_sample_pnonce')

abbreviation gen_sample_pnonces :: "pnonce pmf \<Rightarrow> protocol_nonce set \<Rightarrow> pnonce pmf"
where "gen_sample_pnonces == Finite_Set.fold sample_pnonce'"

definition sample_pnonces :: "protocol_nonce set \<Rightarrow> pnonce \<Rightarrow> pnonce pmf"
where "sample_pnonces X s = gen_sample_pnonces (return_pmf s) (X - dom s)"

definition sample_pnonce :: "protocol_nonce \<Rightarrow> pnonce \<Rightarrow> pnonce pmf"
where "sample_pnonce n = sample_pnonces {n}"

lemma sample_pnonce_alt_def:
  "sample_pnonce' n p = map_pmf (\<lambda>(pns, bs). pns(n \<mapsto> A_nonce A \<eta> bs)) (pair_pmf p (sample_bitstring (size_nonce A \<eta>)))"
by(simp add: sample_pnonce'_def map_pmf_def pair_pmf_def bind_assoc_pmf bind_return_pmf)

lemma sample_pnonce'_bind_return:
  "sample_pnonce' n p = bind_pmf p (sample_pnonce' n \<circ> return_pmf)"
by(simp add: sample_pnonce'_def o_def bind_return_pmf bind_assoc_pmf)

lemma gen_sample_pnonces_bind_return:
  "gen_sample_pnonces p X = bind_pmf p (\<lambda>s. gen_sample_pnonces (return_pmf s) X)"
proof(cases "finite X")
  case True then show ?thesis
  proof(induction)
    case (insert x X)
    show ?case using insert.hyps
      apply simp
      apply(subst insert.IH)
      apply(subst (1 2) sample_pnonce'_bind_return)
      apply(simp add: o_def bind_assoc_pmf)
      done
   qed(simp add: bind_return_pmf')
qed(simp add: bind_return_pmf')

lemma sample_pnonce'_bind:
  "sample_pnonce' n (p \<bind> f) = bind_pmf p (\<lambda>x. f x \<bind> sample_pnonce' n \<circ> return_pmf)"
by(simp add: sample_pnonce'_def o_def bind_return_pmf bind_assoc_pmf)

lemma gen_sample_pnonces_bind:
  "gen_sample_pnonces (p \<bind> f) X = bind_pmf p (\<lambda>x. f x \<bind> (\<lambda>s. gen_sample_pnonces (return_pmf s) X))"
proof(cases "finite X")
  case True then show ?thesis
  proof(induction arbitrary: p f rule: finite_induct)
    case insert
    then show ?case
      apply(simp add: bind_assoc_pmf)
      apply(subst (1 3) sample_pnonce'_bind_return)
      apply(simp add: o_def bind_assoc_pmf)
      done
  qed(simp add: bind_return_pmf')
qed(simp add: bind_return_pmf')

lemma sample_pnonce'_dom: "rel_pmf (\<lambda>s s'. dom s' = insert n (dom s)) p (sample_pnonce' n p)"
proof -
  have "rel_pmf (rel_prod (\<lambda>s s'. insert n (dom s') = insert n (dom s)) top)
    (pair_pmf p (return_pmf ())) (pair_pmf p (sample_bitstring (size_nonce A \<eta>)))"
    by(subst rel_pmf_rel_prod)(simp add: pmf.rel_refl)
  thus ?thesis
    by(simp add: pair_return_pmf2 sample_pnonce_alt_def pmf.rel_map pmf.map_comp o_def split_def rel_prod_sel[abs_def])
qed

lemma sample_pnonce'_domD:
  "s' \<in> set_pmf (sample_pnonce' n (return_pmf s)) \<Longrightarrow> dom s' = insert n (dom s)"
using sample_pnonce'_dom[of n "return_pmf s"] by(simp add: rel_pmf_return_pmf1)

lemma gen_sample_pnonces_dom:
  "finite X \<Longrightarrow> rel_pmf (\<lambda>s s'. dom s' = X \<union> dom s) p (gen_sample_pnonces p X)" (is "_ \<Longrightarrow> rel_pmf (?R X) _ _")
proof(induction rule: finite_induct)
  case (insert x X)
  have "rel_pmf (\<lambda>s s'. dom s' = insert x (dom s)) (gen_sample_pnonces p X) (sample_pnonce' x (gen_sample_pnonces p X))"
    (is "rel_pmf ?R2 _ ?p2") by(rule sample_pnonce'_dom)
  with insert.IH have "rel_pmf (?R X OO ?R2) p ?p2" unfolding pmf.rel_compp by(rule relcomppI)
  moreover have "?R X OO ?R2 \<le> ?R (insert x X)" by blast
  ultimately show ?case unfolding sample_pnonce.fold_insert[OF \<open>finite X\<close> \<open>x \<notin> X\<close>] by(rule pmf_rel_mono')
qed(simp add: pmf.rel_refl)

lemma gen_sample_pnonces_domD:
  "\<lbrakk> s' \<in> set_pmf (gen_sample_pnonces (return_pmf s) X); finite X \<rbrakk> \<Longrightarrow> dom s' = X \<union> dom s"
using gen_sample_pnonces_dom[of X "return_pmf s"] by(simp add: rel_pmf_return_pmf1)

corollary sample_pnonces_dom:
  "\<lbrakk> s' \<in> set_pmf (sample_pnonces X s); finite X \<rbrakk> \<Longrightarrow> dom s' = X \<union> dom s"
using gen_sample_pnonces_domD[of s' s "X - dom s"] by(simp add: sample_pnonces_def)

lemma sample_pnonce'_mono:
  assumes "\<forall>s\<in>set_pmf p. n \<notin> dom s"
  shows "rel_pmf (op \<subseteq>\<^sub>m) p (sample_pnonce' n p)"
proof -
  have "rel_pmf (rel_prod op = top)
    (pair_pmf p (return_pmf ())) (pair_pmf p (sample_bitstring (size_nonce A \<eta>)))"
    by(subst rel_pmf_rel_prod)(simp add: pmf.rel_refl)
  then show ?thesis
    apply(simp add: pair_return_pmf2 sample_pnonce_alt_def pmf.rel_map pmf.map_comp o_def split_def rel_prod_sel[abs_def])
    apply(erule pmf.rel_mono_strong)
    apply(auto dest!: assms[rule_format] intro: map_le_fun_upd2)
    done
qed

lemma gen_sample_pnonces_mono:
  assumes "\<forall>s\<in>set_pmf p. dom s \<inter> X = {}"
  shows "rel_pmf (op \<subseteq>\<^sub>m) p (gen_sample_pnonces p X)"
proof(cases "finite X")
  case True
  then show ?thesis using assms
  proof(induction rule: finite_induct)
    case (insert x X)
    from insert.prems have "\<forall>s\<in>set_pmf p. dom s \<inter> X = {}" by auto
    hence "rel_pmf op \<subseteq>\<^sub>m p (gen_sample_pnonces p X)" (is "rel_pmf _ _ ?p") by(rule insert.IH)
    moreover have "rel_pmf (\<lambda>s s'. dom s' = X \<union> dom s) p ?p" (is "rel_pmf ?R _ _")
      using \<open>finite X\<close> by(rule gen_sample_pnonces_dom)
    with insert.prems[folded pred_pmf_def[THEN meta_eq_to_obj_eq, THEN fun_cong, THEN fun_cong]]
    have "pred_pmf (Imagep ?R (\<lambda>s. dom s \<inter> insert x X = {})) ?p" by(rule pred_rel_pmf)
    hence "\<forall>s\<in>set_pmf ?p. x \<notin> dom s" using \<open>x \<notin> X\<close> by(auto 4 3 simp add: pred_pmf_def Imagep.simps)
    then have "rel_pmf op \<subseteq>\<^sub>m ?p (gen_sample_pnonces p (insert x X))"
      unfolding sample_pnonce.fold_insert[OF \<open>finite X\<close> \<open>x \<notin> X\<close>] by(rule sample_pnonce'_mono)
    ultimately have "rel_pmf (op \<subseteq>\<^sub>m OO op \<subseteq>\<^sub>m) p \<dots>" unfolding pmf.rel_compp by blast
    moreover have "op \<subseteq>\<^sub>m OO op \<subseteq>\<^sub>m \<le> op \<subseteq>\<^sub>m" by(auto intro: map_le_trans)
    ultimately show ?case by(rule pmf_rel_mono')
  qed(simp add: pmf.rel_refl)
qed(simp add: pmf.rel_refl)

corollary sample_pnonces_mono: "s' \<in> set_pmf (sample_pnonces X s) \<Longrightarrow> s \<subseteq>\<^sub>m s'"
using gen_sample_pnonces_mono[of "return_pmf s" "X - dom s"]
by(simp add: rel_pmf_return_pmf1 sample_pnonces_def)

lemma sample_pnonces_union:
  assumes "finite X" "finite Y"
  shows "sample_pnonces (X \<union> Y) s = sample_pnonces X s \<bind> sample_pnonces Y"
proof -
  have "X \<union> Y - dom s = (X - dom s) \<union> (Y - (X \<union> dom s))" by auto
  with sample_nonce.fold_set_union[of "X - dom s" "Y - (X \<union> dom s)" "return_pmf s"]
  show ?thesis using assms unfolding sample_pnonces_def[abs_def]
    apply simp
    apply(rewrite at "\<hole> = _" gen_sample_pnonces_bind_return)
    apply(auto intro!: bind_pmf_cong[OF refl] dest: gen_sample_pnonces_domD)
    done
qed

lemma sample_pnonces_insert:
  "finite X \<Longrightarrow> sample_pnonces (insert x X) s = sample_pnonce x s \<bind> sample_pnonces X"
unfolding sample_pnonce_def using sample_pnonces_union[of "{x}" X s] by simp

lemma sample_pnonces_idem:
  assumes "X \<subseteq> dom s"
  shows "sample_pnonces X s = return_pmf s"
proof -
  from assms have "X - dom s = {}" by auto
  thus ?thesis by(simp add: sample_pnonces_def del: Diff_eq_empty_iff)
qed

lemma sample_pnonces_empty [simp]: "sample_pnonces {} s = return_pmf s"
by(simp add: sample_pnonces_def)

lemma sample_pnonce_idem [simp]:
  "n \<in> dom s \<Longrightarrow> sample_pnonce n s = return_pmf s"
by(simp add: sample_pnonce_def sample_pnonces_idem)

lemma sample_pnonce_None:
  assumes "s n = None"
  shows "sample_pnonce n s = map_pmf (\<lambda>bs. s(n \<mapsto> A_nonce A \<eta> bs)) (pmf_of_set (nlists UNIV (size_nonce A \<eta>)))"
proof -
  from assms have "{n} - dom s = {n}" by auto
  thus ?thesis
    by(simp add: sample_pnonce_def sample_pnonces_def sample_pnonce_alt_def pmf.map_comp o_def pair_return_pmf1)
qed

lemma bind_pmf_sample_pnonces:
  "\<lbrakk> finite X; finite Y \<rbrakk>
  \<Longrightarrow> bind_pmf (sample_pnonces X pns) (\<lambda>pns'. bind_pmf (sample_pnonces Y pns') g) =
   bind_pmf (sample_pnonces (X \<union> Y) pns) g"
by(subst sample_pnonces_union)(simp_all add: bind_assoc_pmf)

end

end

subsubsection \<open>First simulator\<close>

text \<open> For the first simulator, we abstract over the monad and the sampling of nonces. \<close>
locale comp_impl_sim1_base = comp_impl_base A
  for A :: implementation
  +
  fixes return :: "(bitstring, 'm) return"
  and bind :: "(bitstring, 'm) bind"
  and fail :: "'m fail"
  and pnonce_sim :: "security \<Rightarrow> protocol_nonce \<Rightarrow> 'm"
begin

adhoc_overloading Monad_Syntax.bind "bind"
adhoc_overloading Monad_Syntax.bind "monad.bind_option fail"

definition encryption_key
  :: "security \<Rightarrow> protocol_nonce \<Rightarrow> 'm"
where
  "encryption_key \<eta> n = do {
     r_n \<leftarrow> pnonce_sim \<eta> n;
     msg \<leftarrow> A_ek A \<eta> r_n;
     return msg
   }"

definition encryption
  :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> plain \<Rightarrow> protocol_nonce \<Rightarrow> 'm"
where
  "encryption \<eta> n _ msg n' = do {
     ek \<leftarrow> encryption_key \<eta> n;
     r \<leftarrow> pnonce_sim \<eta> n';
     msg \<leftarrow> A_e A \<eta> ek msg r;
     return msg
  }"

definition verification_key
  :: "security \<Rightarrow> protocol_nonce \<Rightarrow> 'm"
where
  "verification_key \<eta> n = do {
     r_n \<leftarrow> pnonce_sim \<eta> n;
     msg \<leftarrow> A_vk A \<eta> r_n;
     return msg
  }"

definition signature
  :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> message \<Rightarrow> protocol_nonce \<Rightarrow> 'm"
where
  "signature \<eta> n t2 msg n' = do {
     r_n \<leftarrow> pnonce_sim \<eta> n;
     r_n' \<leftarrow> pnonce_sim \<eta> n';
     sk \<leftarrow> A_sk A \<eta> r_n;
     msg2 \<leftarrow> A_sig A \<eta> sk msg r_n';
     return msg2
   }"

end

context includes lifting_syntax begin

term comp_impl_sim1_base.encryption_key

lemma encryption_key_parametric [transfer_rule]:
  "(op = 
  ===> (op = ===> M) ===> (M ===> (op = ===> M) ===> M) ===> M
  ===> (op = ===> op = ===> M) ===> op = ===> op = ===> M)
   comp_impl_sim1_base.encryption_key comp_impl_sim1_base.encryption_key"
unfolding comp_impl_sim1_base.encryption_key_def[abs_def] list.rel_eq[symmetric]
by transfer_prover

lemma encryption_parametric [transfer_rule]:
  "(op = 
  ===> (op = ===> M) ===> (M ===> (op = ===> M) ===> M) ===> M
  ===> (op = ===> op = ===> M) ===> op = ===> op = ===> op =
  ===> op = ===> op = ===> M)
  comp_impl_sim1_base.encryption comp_impl_sim1_base.encryption"
unfolding comp_impl_sim1_base.encryption_def[abs_def] list.rel_eq[symmetric]
by transfer_prover

lemma verification_key_parametric [transfer_rule]:
  "(op =
  ===> (op = ===> M) ===> (M ===> (op = ===> M) ===> M) ===> M
  ===> (op = ===> op = ===> M) ===> op = ===> op = ===> M)
  comp_impl_sim1_base.verification_key comp_impl_sim1_base.verification_key"
unfolding comp_impl_sim1_base.verification_key_def[abs_def] list.rel_eq[symmetric]
by transfer_prover

lemma signature_parametric [transfer_rule]:
  "(op =
  ===> (op = ===> M) ===> (M ===> (op = ===> M) ===> M) ===> M
  ===> (op = ===> op = ===> M) ===> op = ===> op = ===> op = ===> op = ===> op = ===> M)
  comp_impl_sim1_base.signature comp_impl_sim1_base.signature"
unfolding comp_impl_sim1_base.signature_def[abs_def] prod.rel_eq[symmetric] list.rel_eq[symmetric]
by transfer_prover

end

context comp_impl_base begin

definition decryption_option :: "security \<Rightarrow> protocol_nonce \<Rightarrow> cipher \<Rightarrow> 'more' simulator_state'_scheme \<Rightarrow> plain option"
where
  "decryption_option \<eta> n m s = (do {
      r_n \<leftarrow> pnonce s n;
      dk \<leftarrow> A_dk A \<eta> r_n;
      A_d A \<eta> dk m
  })"

end

context comp_impl_parse_termination begin

lemma decryption_option_shorter:
  "\<lbrakk> decryption_option \<eta> n m s = Some m'; \<T>_cipher A \<eta> m\<rbrakk>
  \<Longrightarrow> length m' < length m"
by(auto simp add: decryption_option_def Some_eq_bind_conv bind_eq_Some_conv dest: A_type_correct A_d_shorter)

end

context comp_impl_sims begin

abbreviation decryption_sim :: "security \<Rightarrow> protocol_nonce \<Rightarrow> cipher \<Rightarrow> 'more' simulator_state'_scheme \<Rightarrow> (plain option, 'call, 'ret) resumption"
where "decryption_sim \<eta> n m s \<equiv> DONE (decryption_option \<eta> n m s)"

sublocale sim1: comp_impl_sim1_base
  A
  "return_state Done" "bind_state bind_gpv" "fail_state Fail"
  pnonce_sim
.

sublocale sim1: comp_impl_sim_gen
  A pnonce_sim "record_sim :: _ \<Rightarrow> _ \<Rightarrow> (('s, 'more) simulator_state_scheme, _) stateT"
  sim1.encryption_key sim1.encryption
  sim1.verification_key sim1.signature
  decryption_sim
by unfold_locales

(* FIXME: abbreviations are not contracted because the expanded right-hand side contains
an occurrence of a bind_gpv, which adhoc_overloading replaces with bind before the folding triggers,
so the folding rule does not match. Therefore, we disable adhoc_overloading in the uncheck phase. *)
declare [[show_variants]]

abbreviation sim1 :: "nonce_typing \<Rightarrow> 's adversary \<Rightarrow> (security \<Rightarrow> nat)
  \<Rightarrow> ('s, 'id) cosp_simulator"
where "sim1 \<equiv> sim1.sim ()"

abbreviation sim1_initial :: "'s \<Rightarrow> nat \<Rightarrow> 's simulator_state"
where "sim1_initial \<equiv> sim1.initial ()"

abbreviation sim1' :: "nonce_typing \<Rightarrow> 's adversary' \<Rightarrow> security \<Rightarrow> ('s, 'id) cosp_simulator'"
where "sim1' \<equiv> sim1.sim'"

abbreviation sim1_\<beta>
  :: "security \<Rightarrow> t \<Rightarrow> (('s, 'more') simulator_state_scheme, 
    (bitstring \<times> ('s, 'more') simulator_state_scheme, comp_impl.call, comp_impl.ret) gpv) stateT"
where "sim1_\<beta> \<equiv> sim1.\<beta>"

abbreviation sim1_\<tau> 
  :: "security \<Rightarrow> ('s, 'more') simulator_state_scheme \<Rightarrow> bool list
  \<Rightarrow> (t, comp_impl.call, comp_impl.ret) resumption"
where "sim1_\<tau> \<equiv> sim1.\<tau> TYPE('s) TYPE('more')"

lemma fuel_encryption_key:
  "(m, s') \<in> results'_gpv (run_state (sim1.encryption_key \<eta> n) s) \<Longrightarrow> fuel s' = fuel s"
by(auto simp add: sim1.encryption_key_def dest: fuel_pnonce_sim)

lemma fuel_encryption:
  "(bs, s') \<in> results'_gpv (run_state (sim1.encryption \<eta> n' t m n) s) \<Longrightarrow> fuel s' = fuel s"
by(auto simp add: sim1.encryption_def dest!: fuel_pnonce_sim fuel_encryption_key)

lemma fuel_verification_key:
  "(m, s') \<in> results'_gpv (run_state (sim1.verification_key \<eta> n) s) \<Longrightarrow> fuel s' = fuel s"
by(auto simp add: sim1.verification_key_def dest: fuel_pnonce_sim)

lemma fuel_signature:
  "(bs, s') \<in> results'_gpv (run_state (sim1.signature \<eta> n' t m n) s) \<Longrightarrow> fuel s' = fuel s"
by(auto simp add: sim1.signature_def dest!: fuel_pnonce_sim split: if_split_asm)

lemma fuel_\<beta>1_same: "(bs, s') \<in> results'_gpv (run_state (sim1.\<beta> \<eta> t) s) \<Longrightarrow> fuel s' = fuel s"
by(erule sim1.fuel_\<beta>_same)(auto dest: fuel_pnonce_sim fuel_encryption_key fuel_encryption fuel_verification_key fuel_signature split: if_split_asm)

lemma randomness_encryption_key:
  "(m, s') \<in> results'_gpv (run_state (sim1.encryption_key \<eta> n) s) \<Longrightarrow> randomness s' = randomness s"
by(auto simp add: sim1.encryption_key_def dest: randomness_pnonce_sim del: equalityI)

lemma randomness_encryption:
  "(bs, s') \<in> results'_gpv (run_state (sim1.encryption \<eta> n' t m n) s) \<Longrightarrow> randomness s' = randomness s"
by(auto simp add: sim1.encryption_def dest!: randomness_pnonce_sim randomness_encryption_key del: equalityI)

lemma randomness_verification_key:
  "(m, s') \<in> results'_gpv (run_state (sim1.verification_key \<eta> n) s) \<Longrightarrow> randomness s' = randomness s"
by(auto simp add: sim1.verification_key_def dest: randomness_pnonce_sim del: equalityI)

lemma randomness_signature:
  "(bs, s') \<in> results'_gpv (run_state (sim1.signature \<eta> n' t m n) s) \<Longrightarrow> randomness s' = randomness s"
by(auto simp add: sim1.signature_def dest!: randomness_pnonce_sim split: if_split_asm del: equalityI)

lemma randomness_\<beta>1_same: "(bs, s') \<in> results'_gpv (run_state (sim1.\<beta> \<eta> t) s) \<Longrightarrow> randomness s' = randomness s"
by(erule sim1.randomness_\<beta>_same)(auto dest: randomness_pnonce_sim randomness_encryption_key randomness_encryption randomness_verification_key randomness_signature del: equalityI)

lemma randomness_rec_message1_same: "Some s' \<in> results'_gpv (sim1.rec_message \<A> \<eta> s t) \<Longrightarrow> randomness s' = randomness s"
by(auto simp add: sim1.rec_message_def dest!: randomness_\<beta>1_same del: equalityI split: option.split_asm)

lemma relation_for_sim'1: "relation_for_sim' (sim1' \<Gamma> \<A> \<eta>) (Wellfounded.measure fuel)"
proof
  fix s s' s'' t to and ident :: 'id and x y
  assume *: "Some s'' \<in> results'_gpv (info_sim (sim1' \<Gamma> \<A> \<eta>) s' ident to)"
  assume "Some (t, s') \<in> results'_gpv (gen_message (sim1' \<Gamma> \<A> \<eta>) s)
    \<or> Some s' \<in> results'_gpv (rec_message (sim1' \<Gamma> \<A> \<eta>) s t)
    \<or> Some (y, s') \<in> results'_gpv (control (sim1' \<Gamma> \<A> \<eta>) s x)
    \<or> s = s'"
    (is "?gen_message \<or> ?rec_message \<or> ?control \<or> ?same")
  hence "fuel s' = fuel s"
    by(auto simp add: sim1.gen_message_def sim1.rec_message_def sim_control_def dest!: fuel_\<beta>1_same split: option.split_asm)
  thus "(s'', s) \<in> Wellfounded.measure fuel" using *
    by(cases to)(auto simp add: Let_def split: if_split_asm t.split_asm nonce.split_asm option.split_asm node_type.split_asm)
qed

lemma simulator'_sim1: "simulator' (sim1' \<Gamma> \<A> \<eta>) (Wellfounded.measure fuel)"
by unfold_locales(simp_all add: relation_for_sim'1)

sublocale sim1: simulator' \<C> N \<D> \<T> deduce "sim1' \<Gamma> \<A> \<eta>" "Wellfounded.measure fuel" for \<Gamma> \<A> \<eta> R
by(fact simulator'_sim1)

lemma key_safe_sim_pnonce_sim:
  "\<lbrakk> (m, s') \<in> results'_gpv (run_state (pnonce_sim \<eta> r) s); key_safe_sim PN \<eta> \<Gamma> s; \<Gamma> r \<noteq> None \<Longrightarrow> r \<in> randomness s \<or> r \<in> PN \<rbrakk>
  \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s'"
by(clarsimp simp add: pnonce_sim_def Let_def in_nlists_UNIV \<T>_A_nonce key_safe_sim_pnonce_updateI simp del: not_None_eq split: option.split_asm if_split_asm)

lemma key_safe_sim_encryption_key:
  "\<lbrakk> (m, s') \<in> results'_gpv (run_state (sim1.encryption_key \<eta> k) s); key_safe_sim PN \<eta> \<Gamma> s; \<Gamma> k = Some Encryption; k \<in> randomness s \<or> k \<in> PN \<rbrakk>
   \<Longrightarrow> key_safe_sim PN \<eta> \<Gamma> s'"
by(auto simp add: sim1.encryption_key_def dest: key_safe_sim_pnonce_sim)

lemma key_safe_sim_assms_sim1:
  "key_safe_sim_assms A pnonce_sim record_sim sim1.encryption_key sim1.encryption sim1.verification_key sim1.signature PN \<Gamma> \<eta>"
apply unfold_locales
subgoal by(rule key_safe_sim_pnonce_sim) auto
subgoal by(auto 4 3 simp add: pnonce_sim_def Let_def in_nlists_UNIV \<T>_A_nonce ran_def split: option.split_asm intro!: key_safe_sim_history_updateI key_safe_sim_pnonce_updateI dest: key_safe_sim_pnonce_\<T>)
subgoal by(auto intro!: key_safe_sim_history_updateI simp add: key_safe_t_nodetype)
subgoal by(auto simp add: sim1.encryption_key_def intro!: key_safe_sim_history_updateI dest: key_safe_sim_pnonce_sim intro: \<T>_A_ek')
subgoal by(auto 4 3 simp add: sim1.encryption_def intro!: key_safe_sim_history_updateI dest: key_safe_sim_encryption_key key_safe_sim_pnonce_sim \<beta>1_pnonces_into_anonces intro: \<T>_A_e')
subgoal by(auto simp add: sim1.verification_key_def intro!: key_safe_sim_history_updateI dest: key_safe_sim_pnonce_sim intro: \<T>_A_vk')
subgoal by(auto 4 3 simp add: sim1.signature_def intro!: key_safe_sim_history_updateI dest!: key_safe_sim_pnonce_sim \<beta>1_pnonces_into_anonces intro: \<T>_A_sig')
done

lemma key_safe_trace_hybrid_step_sim1:
  assumes step: "Some ((s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step' (sim1' \<Gamma> \<A> \<eta>) ((s,v,f),\<sigma>))"
  and ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v"
  and kst: "key_safe_trace A \<eta> \<Gamma> f"
  and safe: "key_safe_sim {} \<eta> \<Gamma> \<sigma>"
  shows "key_safe_trace A \<eta> \<Gamma> f'"
using step ks kst
by(rule key_safe_trace_hybrid_step)(auto dest: sim1.key_safe_gen_message[OF _ safe] simp add: key_safe_t_nodetype)

lemma sim1_pnonce_random:
  assumes step: "Some ((s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step' (sim1' \<Gamma> \<A> \<eta>) ((s,v,f),\<sigma>))"
  and ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v"
  and rand: "\<forall>t\<in>set f. \<beta>1_anonces t \<inter> dom \<Gamma> \<subseteq> randomness \<sigma>"
  and key_safe_sim: "key_safe_sim {} \<eta> \<Gamma> \<sigma>"
  shows "\<forall>t \<in> set f'. \<beta>1_anonces t \<inter> dom \<Gamma> \<subseteq> randomness \<sigma>'"
proof(cases v)
  case (Computation id comp c_data yes no)
  from step obtain to \<sigma>''
    where to: "Some (to, (s', v', f'), \<sigma>'') \<in> results'_gpv (hybrid_step1' (sim1' \<Gamma> \<A> \<eta>) ((s, v, f), \<sigma>))"
    and \<sigma>': "Some \<sigma>' \<in> results'_gpv (sim_info_sim \<Gamma> \<A> \<eta> \<sigma>'' (identifier v) to :: (_, comp_impl.call, comp_impl.ret) gpv)"
    by(auto simp add: hybrid_step_def split: option.split_asm)
  show ?thesis
  proof(cases to)
    case None
    with to Computation have "f' = f" "\<sigma>'' = \<sigma>"
      by(simp_all split: computation.split_asm option.split_asm)
    moreover from \<sigma>' have "randomness \<sigma>' = randomness \<sigma>''" using None by(simp split: if_split_asm)
    ultimately show ?thesis using rand by auto
  next
    case (Some t')
    note [dest!] = bspec[OF rand] and [simp del] = nth_mem
    from \<sigma>' Some
    have \<sigma>'': "randomness \<sigma>' = (if \<exists>n. t' = N (PNonce n) \<and> \<Gamma> n \<noteq> None then insert (the_PNonce (the_N t')) (randomness \<sigma>'') else randomness \<sigma>'')"
      by(auto simp add: Let_def split: if_split_asm t.split_asm option.split_asm nonce.split_asm)
    from to ks have "f' = f @ [t'] \<and> \<beta>1_anonces t' \<inter> dom \<Gamma> \<subseteq> randomness \<sigma>' \<and> \<sigma>'' = \<sigma>" unfolding Some
    proof(cases rule: key_safe_hybrid_step1_cases)
      case (Nonce n)
      thus ?thesis using \<sigma>' Some by(auto split: if_split_asm simp add: Let_def)
    next
      { case (E_C id no bl1 bl2 bl3 n k)   thus ?thesis using \<sigma>'' by(cases k)(auto dest!: \<beta>1_pnonces_into_anonces) }
      { case EK_C                          thus ?thesis using \<sigma>'' by auto }
      { case DK_C                          thus ?thesis using \<sigma>'' by auto }
      { case (SIG_C id no bl1 bl2 bl3 n k) thus ?thesis using \<sigma>'' by(cases k)(auto dest: \<beta>1_pnonces_into_anonces) }      
      { case VK_C                          thus ?thesis using \<sigma>'' by auto }
      { case SK_C                          thus ?thesis using \<sigma>'' by auto }
      { case PAIR_C                        thus ?thesis using \<sigma>'' by(auto dest: \<beta>1_pnonces_into_anonces) }
      { case EMPTY_C                       thus ?thesis using \<sigma>'' by auto }
      { case STRING_C                      thus ?thesis using \<sigma>'' by(auto dest: \<beta>1_pnonces_into_anonces) }
      { case D                             thus ?thesis using \<sigma>'' by(auto dest!: \<beta>1_pnonces_into_anonces) }
      { case isenc                         thus ?thesis using \<sigma>'' by auto }
      { case isek                          thus ?thesis using \<sigma>'' by auto }
      { case ekof                          thus ?thesis using \<sigma>'' by(auto dest!: \<beta>1_pnonces_into_anonces) }
      { case verify                        thus ?thesis using \<sigma>'' by(auto dest!: \<beta>1_pnonces_into_anonces) }
      { case issig                         thus ?thesis using \<sigma>'' by auto }
      { case isvk                          thus ?thesis using \<sigma>'' by auto }
      { case vkof                          thus ?thesis using \<sigma>'' by(auto dest!: \<beta>1_pnonces_into_anonces) }
      { case FST                           thus ?thesis using \<sigma>'' by(auto dest!: \<beta>1_pnonces_into_anonces) }
      { case SND                           thus ?thesis using \<sigma>'' by(auto dest!: \<beta>1_pnonces_into_anonces) }
      { case unstring                      thus ?thesis using \<sigma>'' by(auto dest!: \<beta>1_pnonces_into_anonces) }
      { case equals                        thus ?thesis using \<sigma>'' by(auto dest!: \<beta>1_pnonces_into_anonces) }
      { case Input                         thus ?thesis using Computation by simp }
    qed
    then show ?thesis using rand by(auto simp add: \<sigma>'')
  qed
next
  case Output
  with step have "f' = f" "randomness \<sigma> = randomness \<sigma>'"
    by(auto 4 3 simp add: hybrid_step_def split: if_split_asm option.split_asm dest: randomness_rec_message1_same del: equalityI)
  then show ?thesis using rand by simp
next
  case Input
  with step obtain t' \<sigma>'' where f: "f' = f @ [t']" 
    and t': "Some (t', \<sigma>'') \<in> results'_gpv (sim1.gen_message \<A> \<eta> \<sigma> :: (_, comp_impl.call, comp_impl.ret) gpv)"
    and \<sigma>': "\<sigma>' = fuel_update (\<lambda>x. x - Suc (len_of \<eta> t'))
             ((case t' of N (PNonce n) \<Rightarrow> (case \<Gamma> n of None \<Rightarrow> id | Some x \<Rightarrow> randomness_update (insert n)) | _ \<Rightarrow> id) \<sigma>'')"
    by(auto simp add: hybrid_step_def Let_def bind_eq_Some_conv split: if_split_asm option.split_asm)
  from t' obtain msg where t': "t' \<in> results (sim1_\<tau> \<eta> \<sigma> msg)" and \<sigma>'': "randomness \<sigma>'' = randomness \<sigma>"
    by(clarsimp simp add: sim1.gen_message_def)
  from sim1.\<beta>1_anonces_\<tau>[OF t' key_safe_sim] have "\<beta>1_anonces t' \<inter> dom \<Gamma> \<subseteq> randomness \<sigma>" by simp
  with \<sigma>'' rand f show ?thesis by(auto simp add: \<sigma>' split: t.split nonce.split option.split)
next
  case Control
  with step have "f' = f" "randomness \<sigma> = randomness \<sigma>'"
    by(auto 4 4 simp add: hybrid_step_def split: if_split_asm option.split_asm dest: randomness_sim_control_same del: equalityI)
  then show ?thesis using rand by simp
next
  case Probabilistic
  with step have "f' = f" "randomness \<sigma> = randomness \<sigma>'"
    by(auto simp add: hybrid_step_def split: if_split_asm)
  then show ?thesis using rand by simp
qed

lemma key_safe_sim_hybrid_step:
  assumes step: "Some ((s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step' (sim1' \<Gamma> \<A> \<eta>) ((s,v,f),\<sigma>))"
  and ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v"
  and ksim: "key_safe_sim {} \<eta> \<Gamma> \<sigma>"
  and kst: "key_safe_trace A \<eta> \<Gamma> f"
  and anonces: "\<forall>t\<in>set f. \<beta>1_anonces t \<inter> dom \<Gamma> \<subseteq> randomness \<sigma>"
  shows "key_safe_sim {} \<eta> \<Gamma> \<sigma>'"
proof -
  from step obtain to \<sigma>''
    where to: "Some (to, (s', v', f'), \<sigma>'') \<in> results'_gpv (hybrid_step1' (sim1' \<Gamma> \<A> \<eta>) ((s, v, f), \<sigma>))"
    and \<sigma>': "Some \<sigma>' \<in> results'_gpv (sim_info_sim \<Gamma> \<A> \<eta> \<sigma>'' (identifier v) to :: (_, comp_impl.call, comp_impl.ret) gpv)"
    by(auto simp add: hybrid_step_def gpv.set_map split: option.split_asm)
  show ?thesis
  proof(cases v)
    case (Computation id comp c_data yes no)
    with to have "\<sigma>'' = \<sigma>" by(simp split: computation.split_asm option.split_asm if_split_asm)
    with \<sigma>' ksim show ?thesis
      by(cases to)(auto simp add: Let_def intro: key_safe_sim_randomness_updateI split: if_split_asm t.split_asm nonce.split_asm option.split_asm)
  next
    case (Output id bl succ)
    with ks have bl: "bl < length f" and nt: "node_type_of_t \<Gamma> (f ! bl) = PublicT"
      by(auto simp add: key_safe_simps)
    from bl have bl': "f ! bl \<in> set f" by simp
    with kst nt have ksbl: "A,\<eta>,\<Gamma> \<turnstile>t f ! bl : PublicT" by(auto dest: key_safe_traceD')
    have \<sigma>'': "Some \<sigma>'' \<in> results'_gpv (sim1.rec_message \<A> \<eta> \<sigma> (f ! bl) :: (_, comp_impl.call, comp_impl.ret) gpv)"
      using to Output by(simp add: gpv.set_map)
    from ksim have ksim': "key_safe_sim (randomness \<sigma>) \<eta> \<Gamma> \<sigma>" by(rule key_safe_sim_PN_mono) simp
    from anonces bl' have "\<beta>1_anonces (f ! bl) \<inter> dom \<Gamma> \<subseteq> randomness \<sigma>" by blast
    with key_safe_sim_assms_sim1 \<sigma>'' ksbl ksim' 
    have "key_safe_sim (randomness \<sigma>) \<eta> \<Gamma> \<sigma>''" by(rule sim1.key_safe_sim_rec_message)
    moreover from \<sigma>'' have "randomness \<sigma> = randomness \<sigma>''"
      by(blast dest: randomness_rec_message1_same)
    ultimately have "key_safe_sim {} \<eta> \<Gamma> \<sigma>''" by(auto dest: key_safe_sim_randomness)
    with \<sigma>' show ?thesis using Output to by(simp add: gpv.set_map split: if_split_asm)
  next
    case (Input id succ)
    with to obtain t where [simp]: "to = Some t"
      and \<sigma>'': "Some (t, \<sigma>'') \<in> results'_gpv (sim1.gen_message \<A> \<eta> \<sigma> :: (_, comp_impl.call, comp_impl.ret) gpv)" 
      by(auto simp add: gpv.set_map Some_eq_bind_conv split: if_split_asm)
    from \<sigma>'' ksim have "key_safe_sim {} \<eta> \<Gamma> \<sigma>''" by(rule sim1.key_safe_sim_gen_message)
    with \<sigma>' show ?thesis
      by(auto simp add: Let_def split: if_split_asm t.split_asm nonce.split_asm option.split_asm intro: key_safe_sim_randomness_updateI)
  next
    case (Control id out labels succs)
    with to obtain succ
      where "Some (succ, \<sigma>'') \<in> results'_gpv (sim_control \<A> \<eta> \<sigma> out :: (_, comp_impl.call, comp_impl.ret) gpv)"
      and [simp]: "to = None" by(clarsimp simp add: gpv.set_map)
    from this(1) ksim have "key_safe_sim {} \<eta> \<Gamma> \<sigma>''" by(rule sim1.key_safe_sim_control)
    with \<sigma>' show ?thesis by(simp split: if_split_asm)
  next
    case (Probabilistic id p succs)
    with to have "\<sigma>'' = \<sigma>" "to = None" by simp_all
    with \<sigma>' ksim show ?thesis by(simp split: if_split_asm)
  qed
qed

lemma randomness_sampled_recorded_assms_sim1:
  "sim1.randomness_sampled_recorded_assms tytok1 tytok2 tytok3 tytok4 \<Gamma> \<eta>"
supply Let_def[simp] pnonce_sim_sel[simp] if_split_asm[split]
apply(rule sim1.randomness_sampled_recorded_assms.intros)
subgoal by(auto elim: randomness_sampled_recorded_history_update_mono)
subgoal by(auto elim: randomness_sampled_recorded_pnonce_update)
subgoal by(auto elim!: randomness_sampled_recorded_pnonce_update)
subgoal by(auto simp add: sim1.encryption_key_def elim: randomness_sampled_recorded_pnonce_update)
subgoal by(auto simp add: sim1.encryption_def sim1.encryption_key_def set_pmf_spmf_of_set elim!: randomness_sampled_recorded_pnonce_update)
subgoal by(auto simp add: sim1.verification_key_def elim: randomness_sampled_recorded_pnonce_update)
subgoal by(auto simp add: sim1.signature_def set_pmf_spmf_of_set elim: randomness_sampled_recorded_pnonce_update)
done

lemma ops_history_dom_same_sim1: "sim1.ops_history_dom_same tytok1 tytok2 tytok3 tytok4 \<eta>"
apply(rule sim1.ops_history_dom_same.intros)
apply(auto simp add: Let_def pnonce_sim_sel sim1.encryption_key_def sim1.encryption_def sim1.verification_key_def sim1.signature_def split: if_split_asm)
done

inductive key_safe_sim1_invar
  :: "security \<Rightarrow> nonce_typing \<Rightarrow> t list \<Rightarrow> 'more' simulator_state'_scheme \<Rightarrow> bool"
  for \<eta> \<Gamma> f \<sigma>
where
  key_safe_sim1_invarI:
  "\<lbrakk> key_safe_sim {} \<eta> \<Gamma> \<sigma>;
     key_safe_trace A \<eta> \<Gamma> f;
     \<forall>t\<in>set f. \<beta>1_anonces t \<inter> dom \<Gamma> \<subseteq> randomness \<sigma>;
     randomness_sampled_recorded \<Gamma> \<sigma>;
     wf_rand_history (random_usages f) (history \<sigma>) \<rbrakk>
  \<Longrightarrow> key_safe_sim1_invar \<eta> \<Gamma> f \<sigma>"

lemmas key_safe_sim1_invar_cases =
  key_safe_sim1_invar.cases[consumes 1, case_names key_safe_sim1_invar, cases pred]

lemma key_safe_sim1_invar_initial [simp]:
  "\<And>more. key_safe_sim1_invar \<eta> \<Gamma> [] (sim1.initial more \<sigma> bound)"
by(auto intro!: key_safe_sim1_invar.intros)

lemma key_safe_sim1_invar_adv_state_update [simp]:
  "key_safe_sim1_invar \<eta> \<Gamma> f (adv_state_update h \<sigma>) = key_safe_sim1_invar \<eta> \<Gamma> f \<sigma>"
by(simp add: key_safe_sim1_invar.simps)

lemma key_safe_sim1_invar_hybrid_step:
  assumes step: "Some ((s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step' (sim1' \<Gamma> \<A> \<eta>) ((s, v, f), \<sigma>))"
  and ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v"
  and inv: "key_safe_sim1_invar \<eta> \<Gamma> f \<sigma>"
  shows "key_safe_sim1_invar \<eta> \<Gamma> f' \<sigma>'"
using inv
proof cases
  case key_safe_sim1_invar
  from step ks key_safe_sim1_invar(1-3) have *: "key_safe_sim {} \<eta> \<Gamma> \<sigma>'" by(rule key_safe_sim_hybrid_step)
  moreover from step ks key_safe_sim1_invar(2,1) have "key_safe_trace A \<eta> \<Gamma> f'"
    by(rule key_safe_trace_hybrid_step_sim1)
  moreover from step ks key_safe_sim1_invar(3,1)
  have "\<forall>t\<in>set f'. \<beta>1_anonces t \<inter> dom \<Gamma> \<subseteq> randomness \<sigma>'" by(rule sim1_pnonce_random)
  moreover have "randomness_sampled_recorded \<Gamma> \<sigma>'"
    using step ks key_safe_sim1_invar(2,4) randomness_sampled_recorded_assms_sim1
    by(rule sim1.randomness_sampled_recorded_hybrid_step)
  moreover have "wf_rand_history (random_usages f') (history \<sigma>')"
    using step ks key_safe_sim1_invar(5) ops_history_dom_same_sim1 
    by(rule sim1.wf_rand_history_hybrid_step)
  ultimately show ?thesis ..
qed

lemma key_safe_sim1_hybrid_step:
  assumes step: "Some ((s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step' (sim1' \<Gamma> \<A> \<eta>) ((s, v, f), \<sigma>))"
  and ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v"
  and inv: "key_safe_sim1_invar \<eta> \<Gamma> f \<sigma>"
  shows "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f') v'"
using step ks
proof(rule key_safe_hybrid_step)
  from inv show "key_safe_trace A \<eta> \<Gamma> f" by(auto elim: key_safe_sim1_invar_cases)
  from inv have "key_safe_sim {} \<eta> \<Gamma> \<sigma>" by(auto elim: key_safe_sim1_invar_cases)
  then show "node_type_of_t \<Gamma> t = PublicT" if "Some (t, \<sigma>') \<in> results'_gpv (gen_message (sim1' \<Gamma> \<A> \<eta>) \<sigma>)" for t \<sigma>'
    using that by(auto dest!: sim1.key_safe_gen_message simp add: key_safe_t_nodetype)
qed

lemma wf_random_usage_sim1_hybrid_step:
  "\<lbrakk> Some ((s', v', f'), \<sigma>') \<in> results'_gpv (hybrid_step' (sim1' \<Gamma> \<A> \<eta>) ((s, v, f), \<sigma>)); 
     wf_random_usage \<Gamma> f; key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f) v; key_safe_sim1_invar \<eta> \<Gamma> f \<sigma> \<rbrakk>
  \<Longrightarrow> wf_random_usage \<Gamma> f'"
by(rule wf_random_usage_hybrid_step)(auto dest!: sim1.wf_random_usage_gen_message elim: key_safe_sim1_invar.cases)

abbreviation hybrid_trace1
  :: "'id cosp_protocol \<Rightarrow> bool \<Rightarrow> nonce_typing \<Rightarrow> 's adversary \<Rightarrow> bound
  \<Rightarrow> security \<Rightarrow> 'id list spmf"
where "hybrid_trace1 prot b \<Gamma> \<A> bound \<equiv> hybrid_trace prot (oracle b) (sim1 \<Gamma> \<A> bound)"

abbreviation hybrid_step1_sim1
  :: "nonce_typing \<Rightarrow> 's adversary' \<Rightarrow> security \<Rightarrow> ('id, 's) cosp_hybrid_state
  \<Rightarrow> ((t option \<times> ('id, 's) cosp_hybrid_state) option, call, comp_impl.ret) gpv"
where "hybrid_step1_sim1 \<Gamma> \<A> bound \<equiv> hybrid_step1 TYPE(_) (sim1' \<Gamma> \<A> bound)"

abbreviation hybrid_step_sim1
  :: "nonce_typing \<Rightarrow> 's adversary' \<Rightarrow> security \<Rightarrow> ('id, 's) cosp_hybrid_state 
  \<Rightarrow> (('id, 's) cosp_hybrid_state option, call, ret) gpv"
where "hybrid_step_sim1 \<Gamma> \<A> bound \<equiv> hybrid_step TYPE(_) (sim1' \<Gamma> \<A> bound)"

subsection \<open>Eager sampling of nonces\<close>
text \<open>
  The first simulator can be run in the spmf monad instead of gpv. Here, we omit the recording.
\<close>

definition record_spmf :: "t \<Rightarrow> bitstring \<Rightarrow> (pnonce, (history, ((bitstring \<times> pnonce) \<times> history) spmf) stateT) stateT"
where "record_spmf t bs = StateT (\<lambda>pns. StateT (\<lambda>hist. return_spmf ((bs, pns), hist(t \<mapsto> bs))))"

abbreviation decryption_spmf
  :: "security \<Rightarrow> protocol_nonce \<Rightarrow> cipher \<Rightarrow> 'more simulator_state'_scheme \<Rightarrow> plain option spmf"
where "decryption_spmf \<eta> n m s \<equiv> return_spmf (decryption_option \<eta> n m s)"

sublocale spmf: comp_impl_sim1_base
  A
  "return_state (return_state return_spmf)" "bind_state (bind_state bind_spmf)" "fail_state (fail_state (return_pmf None))"
  "pnonce_spmf"
by unfold_locales

sublocale spmf: comp_impl_\<beta>_base
  A 
  "return_state (return_state return_spmf)" "bind_state (bind_state bind_spmf)" "fail_state (fail_state (return_pmf None))"
  pnonce_spmf record_spmf spmf.encryption_key spmf.encryption spmf.verification_key spmf.signature
by unfold_locales


text \<open>Running @{text \<beta>} in the option monad should give the same result as running @{text \<beta>} in the GPV monad on a term whose nonces have been sampled\<close>

definition pnonce_option :: "security \<Rightarrow> protocol_nonce \<Rightarrow> (pnonce, (history, (bitstring \<times> history) option) stateT) envT"
where "pnonce_option \<eta> n = EnvT (\<lambda>pns. StateT (\<lambda>hist. map_option (\<lambda>bs. (bs, hist)) (pns n)))"

definition record_option :: "t \<Rightarrow> bitstring \<Rightarrow> (pnonce, (history, (bitstring \<times> history) option) stateT) envT"
where "record_option t bs = EnvT (\<lambda>_. (StateT (\<lambda>hist. Some (bs, hist(t \<mapsto> bs)))))"

sublocale option: comp_impl_sim1_base
  A
  "return_env (return_state Some)" "bind_env (bind_state Option.bind)" "fail_env (fail_state None)"
  pnonce_option
.

sublocale option: comp_impl_\<beta>_base
  A
  "return_env (return_state Some)" "bind_env (bind_state Option.bind)" "fail_env (fail_state None)"
  pnonce_option record_option option.encryption_key option.encryption option.verification_key option.signature
  for s
by unfold_locales

lemma pnonce_option_eq_Some_iff [simp]:
  "run_state (run_env (pnonce_option \<eta> n) pns) hist = Some (m, hist') \<longleftrightarrow> hist' = hist \<and> pns n = Some m"
by(auto simp add: pnonce_option_def)

lemma \<beta>1_N_history: "pnonce s n = Some m \<Longrightarrow> run_state (sim1.\<beta> \<eta> (N (PNonce n))) s = run_state (record_sim (N (PNonce n)) m) s"
by(simp add: pnonce_sim_sel)

text \<open>Running the serialisation function in the option monad without recording and sampling\<close>

definition pnonce_pure :: "security \<Rightarrow> protocol_nonce \<Rightarrow> (pnonce, bitstring option) envT"
where "pnonce_pure \<eta> n = EnvT (\<lambda>pns. pns n)"

sublocale pure: comp_impl_sim1_base
  A
  "return_env Some" "bind_env Option.bind" "fail_env None"
  pnonce_pure
.

sublocale pure: comp_impl_\<beta>_base
  A
  "return_env Some" "bind_env Option.bind" "fail_env None"
  pnonce_pure "\<lambda>_. return_env Some" pure.encryption_key pure.encryption pure.verification_key pure.signature
by unfold_locales

lemma record_sim_idem: "history s t = Some m \<Longrightarrow> run_state (record_sim t m) s = Done (m, s)"
apply(simp add: record_sim_def)
by(rule simulator_state'.equality)(simp_all add: fun_upd_idem)

end

abbreviation sampled1 :: "t \<Rightarrow> (protocol_nonce \<rightharpoonup> bitstring) \<Rightarrow> bool"
where "sampled1 t s == \<beta>1_pnonces t \<subseteq> dom s"
  -- \<open>Do not use @{term "sampled \<beta>_nonce1"} on right-hand side,
      because the abbreviation for @{term \<beta>1_nonces} interferes\<close>

context comp_impl_sims begin

text \<open>The pure versions do not care about additional sampling provided that enough has already been sampled before\<close>

lemma pnonce_pure_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns \<rbrakk> \<Longrightarrow> run_env (pnonce_pure \<eta> n) pns = run_env (pnonce_pure \<eta> n) pns'"
by(clarsimp simp add: pnonce_pure_def fun_eq_iff map_le_def)

lemma encryption_key_pure_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns \<rbrakk>
  \<Longrightarrow> run_env (pure.encryption_key \<eta> n) pns = run_env (pure.encryption_key \<eta> n) pns'"
by(simp add: pure.encryption_key_def pnonce_pure_incr)

lemma encryption_key'_pure_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns \<rbrakk>
  \<Longrightarrow> run_env (pure.encryption_key' \<eta> t n) pns = run_env (pure.encryption_key' \<eta> t n) pns'"
by(simp add: pure.encryption_key'_def encryption_key_pure_incr)

lemma encryption_pure_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns; r \<in> dom pns \<rbrakk>
  \<Longrightarrow> run_env (pure.encryption \<eta> n t m r) pns = run_env (pure.encryption \<eta> n t m r) pns'"
by(simp add: pure.encryption_def encryption_key_pure_incr pnonce_pure_incr)

lemma verification_key_pure_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns \<rbrakk>
  \<Longrightarrow> run_env (pure.verification_key \<eta> n) pns = run_env (pure.verification_key \<eta> n) pns'"
by(simp add: pure.verification_key_def pnonce_pure_incr)

lemma verification_key'_pure_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns \<rbrakk>
  \<Longrightarrow> run_env (pure.verification_key' \<eta> t n) pns = run_env (pure.verification_key' \<eta> t n) pns'"
by(simp add: pure.verification_key'_def verification_key_pure_incr)

lemma signature_pure_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns; r \<in> dom pns \<rbrakk>
  \<Longrightarrow> run_env (pure.signature \<eta> n t m r) pns = run_env (pure.signature \<eta> n t m r) pns'"
by(simp add: pure.signature_def pnonce_pure_incr verification_key_pure_incr)

lemma \<beta>_pure_incr:
  assumes incr: "pns \<subseteq>\<^sub>m pns'"
  and "sampled1 t pns"
  shows "run_env (pure.\<beta> \<eta> t) pns = run_env (pure.\<beta> \<eta> t) pns'"
proof -
  note [iff del] = domIff
    and [split] = t.split nonce.split
    and [OF incr, simp] =
      pnonce_pure_incr encryption_key'_pure_incr encryption_pure_incr
      verification_key_pure_incr verification_key'_pure_incr signature_pure_incr
  let ?P = "\<lambda>t. sampled1 t pns \<longrightarrow> run_env (pure.\<beta> \<eta> t) pns = run_env (pure.\<beta> \<eta> t) pns'"
  have "?P t"
  proof(induction rule: t.induct[where P="?P"])
    { case E          thus ?case by auto }
    { case EK         thus ?case by auto }
    { case DK         thus ?case by auto }
    { case SIG        thus ?case by auto }
    { case VK         thus ?case by auto }
    { case SK         thus ?case by auto }
    { case PAIR       thus ?case by auto }
    { case STRING     thus ?case by auto }
    { case EMPTY      thus ?case by auto }
    { case N          thus ?case by auto }
    { case garbage    thus ?case by auto }
    { case garbageE   thus ?case by auto }
    { case garbageSIG thus ?case by auto }
  qed
  with \<open>sampled1 t pns\<close> show ?thesis by simp
qed

text \<open> If not all sampling has been done before, then we still get monotonicity. \<close>

context notes le_option_bind_mono[simp, intro!] begin

lemma pnonce_pure_incr':
  "pns \<subseteq>\<^sub>m pns' \<Longrightarrow> le_option (run_env (pnonce_pure \<eta> n) pns) (run_env (pnonce_pure \<eta> n) pns')"
by(cases "pns n")(auto simp add: pnonce_pure_def dest: map_le_Some1D)

lemma encryption_key_pure_incr':
  "pns \<subseteq>\<^sub>m pns' \<Longrightarrow> le_option (run_env (pure.encryption_key \<eta> n) pns) (run_env (pure.encryption_key \<eta> n) pns')"
by(simp add: pure.encryption_key_def pnonce_pure_incr')

lemma encryption_key'_pure_incr':
  "pns \<subseteq>\<^sub>m pns' \<Longrightarrow> le_option (run_env (pure.encryption_key' \<eta> t n) pns) (run_env (pure.encryption_key' \<eta> t n) pns')"
by(simp add: pure.encryption_key'_def encryption_key_pure_incr')

lemma encryption_pure_incr':
  "pns \<subseteq>\<^sub>m pns'
  \<Longrightarrow> le_option (run_env (pure.encryption \<eta> n t m r) pns) (run_env (pure.encryption \<eta> n t m r) pns')"
by(simp add: pure.encryption_def encryption_key_pure_incr' pnonce_pure_incr')

lemma verification_key_pure_incr':
  "pns \<subseteq>\<^sub>m pns'
  \<Longrightarrow> le_option (run_env (pure.verification_key \<eta> n) pns) (run_env (pure.verification_key \<eta> n) pns')"
by(simp add: pure.verification_key_def pnonce_pure_incr')

lemma verification_key'_pure_incr':
  "pns \<subseteq>\<^sub>m pns'
  \<Longrightarrow> le_option (run_env (pure.verification_key' \<eta> t n) pns) (run_env (pure.verification_key' \<eta> t n) pns')"
by(simp add: pure.verification_key'_def verification_key_pure_incr')

lemma signature_pure_incr':
  "pns \<subseteq>\<^sub>m pns'
  \<Longrightarrow> le_option (run_env (pure.signature \<eta> n t m r) pns) (run_env (pure.signature \<eta> n t m r) pns')"
by(simp add: pure.signature_def pnonce_pure_incr' verification_key_pure_incr')

lemma \<beta>_pure_incr':
  assumes incr: "pns \<subseteq>\<^sub>m pns'"
  shows "le_option (run_env (pure.\<beta> \<eta> t) pns) (run_env (pure.\<beta> \<eta> t) pns')"
proof(induction t rule: t.induct)
  note [iff del] = domIff
    and [split] = t.split nonce.split
    and [OF incr, simp] =
      pnonce_pure_incr' encryption_key'_pure_incr' encryption_pure_incr'
      verification_key_pure_incr' verification_key'_pure_incr' signature_pure_incr'
  { case E          thus ?case by auto }
  { case EK         thus ?case by auto }
  { case DK         thus ?case by auto }
  { case SIG        thus ?case by auto }
  { case VK         thus ?case by auto }
  { case SK         thus ?case by auto }
  { case PAIR       thus ?case by auto }
  { case STRING     thus ?case by auto }
  { case EMPTY      thus ?case by auto }
  { case N          thus ?case by auto }
  { case garbage    thus ?case by auto }
  { case garbageE   thus ?case by auto }
  { case garbageSIG thus ?case by auto }
qed

end

text \<open> The recording can be done afterwards \<close>

definition record_all :: "security \<Rightarrow> t \<Rightarrow> pnonce \<Rightarrow> history"
where
  "record_all \<eta> t pns = (\<lambda>t'. if \<beta>_rec t t' then run_env (pure.\<beta> \<eta> t') pns else None)"

lemma record_all_apply: "record_all \<eta> t pns t' = (if \<beta>_rec t t' then run_env (pure.\<beta> \<eta> t') pns else None)"
by(simp add: record_all_def)

lemma dom_record_all: "dom (record_all \<eta> t pns) = {t'. \<beta>_rec t t' \<and> run_env (pure.\<beta> \<eta> t') pns \<noteq> None}"
by(auto simp add: domIff record_all_apply split: if_split_asm)

lemma pnonce_option_conv_pure:
  "run_state (run_env (pnonce_option \<eta> n) pns) hist = map_option (\<lambda>m. (m, hist)) (run_env (pnonce_pure \<eta> n) pns)"
by(simp add: pnonce_pure_def pnonce_option_def)

lemma encryption_key_option_conv_pure:
  "run_state (run_env (option.encryption_key \<eta> n) pns) hist = map_option (\<lambda>m. (m, hist)) (run_env (pure.encryption_key \<eta> n) pns)"
by(simp add: option.encryption_key_def pure.encryption_key_def pnonce_option_conv_pure map_conv_bind_option o_def)

lemma encryption_key'_option_conv_pure:
  "run_state (run_env (option.encryption_key' \<eta> t n) pns) hist = map_option (\<lambda>m. (m, hist(t \<mapsto> m))) (run_env (pure.encryption_key' \<eta> t n) pns)"
by(simp add: option.encryption_key'_def pure.encryption_key'_def encryption_key_option_conv_pure bind_map_option o_def record_option_def map_conv_bind_option)

lemma encryption_option_conv_pure:
  "run_state (run_env (option.encryption \<eta> n t m r) pns) hist = map_option (\<lambda>m. (m, hist)) (run_env (pure.encryption \<eta> n t m r) pns)"
by(simp add: option.encryption_def pure.encryption_def encryption_key_option_conv_pure pnonce_option_conv_pure o_def map_conv_bind_option)

lemma verification_key_option_conv_pure:
  "run_state (run_env (option.verification_key \<eta> n) pns) hist = map_option (\<lambda>m. (m, hist)) (run_env (pure.verification_key \<eta> n) pns)"
by(simp add: option.verification_key_def pure.verification_key_def pnonce_option_conv_pure map_conv_bind_option o_def)

lemma verification_key'_option_conv_pure:
  "run_state (run_env (option.verification_key' \<eta> t n) pns) hist = map_option (\<lambda>m. (m, hist(t \<mapsto> m))) (run_env (pure.verification_key' \<eta> t n) pns)"
by(simp add: option.verification_key'_def pure.verification_key'_def verification_key_option_conv_pure bind_map_option o_def record_option_def map_conv_bind_option)

lemma signature_option_conv_pure:
  "run_state (run_env (option.signature \<eta> n t m r) pns) hist = map_option (\<lambda>m. (m, hist)) (run_env (pure.signature \<eta> n t m r) pns)"
by(simp add: option.signature_def pure.signature_def verification_key_option_conv_pure map_conv_bind_option pnonce_option_conv_pure o_def)

lemma \<beta>_option_conv_pure:
  "run_state (run_env (option.\<beta> \<eta> t) pns) hist = map_option (\<lambda>m. (m, hist ++ record_all \<eta> t pns)) (run_env (pure.\<beta> \<eta> t) pns)"
proof -
  note [cong del] = if_weak_cong
    and [simp] = pnonce_option_conv_pure encryption_key_option_conv_pure
      encryption_key'_option_conv_pure encryption_option_conv_pure
      verification_key_option_conv_pure verification_key'_option_conv_pure signature_option_conv_pure
      record_option_def o_def record_all_def fun_eq_iff map_conv_bind_option
      map_add_apply sup_option_ai
    and [split] = t.split nonce.split
    and [intro!] = bind_option_cong
  let ?P = "\<lambda>t. \<forall>hist. run_state (run_env (option.\<beta> \<eta> t) pns) hist = map_option (\<lambda>m. (m, hist ++ record_all \<eta> t pns)) (run_env (pure.\<beta> \<eta> t) pns)"
  have "?P t"
  proof(induction rule: t.induct[where P="?P"])
    { case X: E          show ?case by(auto simp add: X \<beta>_rec_E) }
    { case X: EK         show ?case by(auto) }
    { case X: DK         show ?case by(auto) }
    { case X: SIG        show ?case by(auto simp add: X \<beta>_rec_SIG) }
    { case X: VK         show ?case by(auto) }
    { case X: SK         show ?case by(auto) }
    { case X: PAIR       show ?case by(auto simp add: X Let_def) }
    { case X: EMPTY      show ?case by(auto) }
    { case X: STRING     show ?case by(auto simp add: X) }
    { case X: N          show ?case by(auto) }
    { case X: garbage    show ?case by(auto) }
    { case X: garbageE   show ?case by(auto) }
    { case X: garbageSIG show ?case by(auto) }
  qed
  thus ?thesis by simp
qed

lemmas option_convs_pure =
  pnonce_option_conv_pure
  encryption_key_option_conv_pure
  encryption_key'_option_conv_pure
  encryption_option_conv_pure
  verification_key_option_conv_pure
  verification_key'_option_conv_pure
  signature_option_conv_pure
  \<beta>_option_conv_pure

text \<open>Transfer the theorems about increasing the sampling cache to the option version\<close>

lemma record_all_incr:
  assumes "sampled1 t pns" "pns \<subseteq>\<^sub>m pns'"
  shows "record_all \<eta> t pns = record_all \<eta> t pns'"
unfolding record_all_def fun_eq_iff
proof(rule allI if_cong refl)+
  fix t'
  assume "\<beta>_rec t t'"
  with \<open>sampled1 t pns\<close> have "sampled1 t' pns" by(rule sampled_\<beta>_recD)
  with \<open>pns \<subseteq>\<^sub>m pns'\<close> show "run_env (pure.\<beta> \<eta> t') pns = run_env (pure.\<beta> \<eta> t') pns'" by(rule \<beta>_pure_incr)
qed simp

context notes [iff del] = domIff and [simp] = fun_eq_iff and [intro!] = stateT.expand begin

lemma pnonce_option_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns \<rbrakk> \<Longrightarrow> run_env (pnonce_option \<eta> n) pns = run_env (pnonce_option \<eta> n) pns'"
by(auto simp add: pnonce_option_conv_pure pnonce_pure_incr)

lemma record_option_incr: "pns \<subseteq>\<^sub>m pns' \<Longrightarrow> run_env (record_option t m) pns = run_env (record_option t m) pns'" 
by(simp add: record_option_def)

lemma encryption_key_option_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns \<rbrakk>
  \<Longrightarrow> run_env (option.encryption_key \<eta> n) pns = run_env (option.encryption_key \<eta> n) pns'"
by(auto simp add: encryption_key_option_conv_pure encryption_key_pure_incr)

lemma encryption_key'_option_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns \<rbrakk>
  \<Longrightarrow> run_env (option.encryption_key' \<eta> t n) pns = run_env (option.encryption_key' \<eta> t n) pns'"
by(auto simp add: encryption_key'_option_conv_pure encryption_key'_pure_incr)

lemma encryption_option_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns; r \<in> dom pns \<rbrakk>
  \<Longrightarrow> run_env (option.encryption \<eta> n t m r) pns = run_env (option.encryption \<eta> n t m r) pns'"
by(auto simp add: encryption_option_conv_pure encryption_pure_incr)

lemma verification_key_option_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns \<rbrakk>
  \<Longrightarrow> run_env (option.verification_key \<eta> n) pns = run_env (option.verification_key \<eta> n) pns'"
by(auto simp add: verification_key_option_conv_pure verification_key_pure_incr)

lemma verification_key'_option_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns \<rbrakk>
  \<Longrightarrow> run_env (option.verification_key' \<eta> t n) pns = run_env (option.verification_key' \<eta> t n) pns'"
by(auto simp add: verification_key'_option_conv_pure verification_key'_pure_incr)

lemma signature_option_incr:
  "\<lbrakk> pns \<subseteq>\<^sub>m pns'; n \<in> dom pns; r \<in> dom pns \<rbrakk>
  \<Longrightarrow> run_env (option.signature \<eta> n t m r) pns = run_env (option.signature \<eta> n t m r) pns'"
by(auto simp add: signature_option_conv_pure signature_pure_incr)

lemma \<beta>_option_incr:
  "\<lbrakk> sampled1 t pns; pns \<subseteq>\<^sub>m pns' \<rbrakk> \<Longrightarrow> run_env (option.\<beta> \<eta> t) pns = run_env (option.\<beta> \<eta> t) pns'"
by(auto simp add: \<beta>_option_conv_pure \<beta>_pure_incr record_all_incr)

end

text \<open> The sampling can be done up-front \<close>

abbreviation embed_sampling :: "pnonce pmf \<Rightarrow> history \<Rightarrow> (pnonce, (history, (bitstring \<times> history) option) stateT) envT \<Rightarrow> ((bitstring \<times> pnonce) \<times> history) spmf"
where "embed_sampling p hist f == map_pmf (\<lambda>pns. map_option (\<lambda>(m, hist'). ((m, pns), hist')) (run_state (run_env f pns) hist)) p"

text \<open>Proof strategy: Pull the sampling to the front using @{thm [source] bind_spmf_pmf_commute}
  and convert everything into bind operations on pmf and spmf. As we use the same functions, this
  results in the same structure.\<close>

lemma pnonce_spmf_conv_option:
  "run_state (run_state (pnonce_spmf \<eta> n) pns) hist = embed_sampling (sample_pnonces \<eta> {n} pns) hist (pnonce_option \<eta> n)"
by(auto simp add: bind_spmf_of_set bind_return_pmf sample_pnonce_None pnonce_spmf_def pnonce_option_def Let_def sample_pnonces_insert map_bind_pmf bind_map_pmf domIff split: option.split)

lemma record_spmf_conv_option:
  "run_state (run_state (record_spmf t m) pns) hist =
   return_pmf (map_option (\<lambda>(m, hist). ((m, pns), hist)) (run_state (run_env (record_option t m) pns) hist))"
by(simp add: record_spmf_def record_option_def)

lemma encryption_key_spmf_conv_option:
  "run_state (run_state (spmf.encryption_key \<eta> n) pns) hist =
   embed_sampling (sample_pnonces \<eta> {n} pns) hist (option.encryption_key \<eta> n)"
by(simp add: spmf.encryption_key_def option.encryption_key_def pnonce_spmf_conv_option o_def map_conv_bind_option map_pmf_def bind_spmf_pmf_assoc return_pmf_bind_option split_def)

lemma encryption_key'_spmf_conv_option:
  "run_state (run_state (spmf.encryption_key' \<eta> t n) pns) hist =
   embed_sampling (sample_pnonces \<eta> {n} pns) hist (option.encryption_key' \<eta> t n)"
by(simp add: spmf.encryption_key'_def option.encryption_key'_def encryption_key_spmf_conv_option map_pmf_def bind_spmf_pmf_assoc split_def map_conv_bind_option o_def return_pmf_bind_option record_spmf_conv_option)

lemma encryption_spmf_conv_option:
  "run_state (run_state (spmf.encryption \<eta> n t m r) pns) hist =
   embed_sampling (sample_pnonces \<eta> {n, r} pns) hist (option.encryption \<eta> n t m r)"
  (is "?lhs = ?rhs")
proof -
  let ?rest = "\<lambda>pns' pns''. bind_spmf (return_pmf (run_state (run_env (option.encryption_key \<eta> n) pns') hist))
    (\<lambda>(ek, hist'). bind_spmf (return_pmf (run_state (run_env (pnonce_option \<eta> r) pns'') hist'))
    (\<lambda>(r', hist''). bind_spmf (return_pmf (A_e A \<eta> ek m r')) (\<lambda>x. return_spmf ((x, pns''), hist''))))"
  have "?lhs = bind_pmf (sample_pnonces \<eta> {n} pns) (\<lambda>pns'. bind_pmf (sample_pnonces \<eta> {r} pns') (?rest pns'))"
    by(simp add: spmf.encryption_def encryption_key_spmf_conv_option map_pmf_def bind_spmf_pmf_assoc return_pmf_map_option_conv_bind return_pmf_bind_option o_def pnonce_spmf_conv_option bind_spmf_pmf_commute split_def)
  also have "\<dots> = bind_pmf (sample_pnonces \<eta> {n} pns) (\<lambda>pns'. bind_pmf (sample_pnonces \<eta> {r} pns') (\<lambda>pns''. ?rest pns'' pns''))"
  proof(rule bind_pmf_cong[OF refl])+
    fix pns' pns''
    assume "pns'' \<in> set_pmf (sample_pnonces \<eta> {r} pns')"
    hence "pns' \<subseteq>\<^sub>m pns''" by(rule sample_pnonces_mono)
    moreover
    assume "pns' \<in> set_pmf (sample_pnonces \<eta> {n} pns)"
    hence "n \<in> dom pns'" by(auto dest: sample_pnonces_dom)
    ultimately show "?rest pns' pns'' = ?rest pns'' pns''"
      by(simp add: encryption_key_option_incr)
  qed
  also have "\<dots> = bind_pmf (sample_pnonces \<eta> {n, r} pns) (\<lambda>pns''. ?rest pns'' pns'')"
    by(simp add: sample_pnonces_insert bind_assoc_pmf bind_return_pmf)
  also have "\<dots> = ?rhs"
    by(simp add: option.encryption_def map_pmf_def return_pmf_map_option_conv_bind return_pmf_bind_option o_def split_def)
  finally show ?thesis .
qed

lemma verification_key_spmf_conv_option:
  "run_state (run_state (spmf.verification_key \<eta> n) pns) hist =
   embed_sampling (sample_pnonces \<eta> {n} pns) hist (option.verification_key \<eta> n)"
by(simp add: spmf.verification_key_def option.verification_key_def pnonce_spmf_conv_option map_pmf_def bind_spmf_pmf_assoc return_pmf_map_option_conv_bind return_pmf_bind_option o_def split_def)

lemma verification_key'_spmf_conv_option:
  "run_state (run_state (spmf.verification_key' \<eta> t n) pns) hist =
   embed_sampling (sample_pnonces \<eta> {n} pns) hist (option.verification_key' \<eta> t n)"
by(simp add: spmf.verification_key'_def option.verification_key'_def verification_key_spmf_conv_option map_pmf_def bind_spmf_pmf_assoc split_def map_conv_bind_option o_def return_pmf_bind_option record_spmf_conv_option)

lemma signature_spmf_conv_option:
  "run_state (run_state (spmf.signature \<eta> n t m r) pns) hist =
   embed_sampling (sample_pnonces \<eta> {n, r} pns) hist (option.signature \<eta> n t m r)"
   (is "?lhs = ?rhs")
proof -
  let ?rest = "\<lambda>pns' pns''. bind_spmf (return_pmf (run_state (run_env (pnonce_option \<eta> n) pns') hist))
    (\<lambda>(n', hist'). bind_spmf (return_pmf (run_state (run_env (pnonce_option \<eta> r) pns'') hist')) 
    (\<lambda>(r', hist''). bind_spmf (return_pmf (A_sk A \<eta> n'))
    (\<lambda>sk. bind_spmf (return_pmf (A_sig A \<eta> sk m r')) (\<lambda>m. return_spmf ((m, pns''), hist'')))))"
  have "?lhs = bind_pmf (sample_pnonces \<eta> {n} pns) (\<lambda>pns'. bind_pmf (sample_pnonces \<eta> {r} pns') (\<lambda>pns''. ?rest pns' pns''))"
    by(simp add: spmf.signature_def pnonce_spmf_conv_option map_pmf_def bind_spmf_pmf_assoc return_pmf_map_option_conv_bind bind_spmf_pmf_commute split_def)
  also have "\<dots> = bind_pmf (sample_pnonces \<eta> {n} pns) (\<lambda>pns'. bind_pmf (sample_pnonces \<eta> {r} pns') (\<lambda>pns''. ?rest pns'' pns''))"
  proof(rule bind_pmf_cong[OF refl])+
    fix pns' pns''
    assume "pns' \<in> set_pmf (comp_impl_base.sample_pnonces A \<eta> {n} pns)"
      and "pns'' \<in> set_pmf (comp_impl_base.sample_pnonces A \<eta> {r} pns')"
    hence "pns' \<subseteq>\<^sub>m pns''" "n \<in> dom pns'" by(auto dest: sample_pnonces_mono sample_pnonces_dom)
    then show "?rest pns' pns'' = ?rest pns'' pns''" by(simp add: pnonce_option_incr)
  qed
  also have "\<dots> = ?rhs"
    by(auto simp add: option.signature_def sample_pnonces_insert map_pmf_def bind_assoc_pmf bind_return_pmf return_pmf_map_option_conv_bind return_pmf_bind_option intro!: bind_pmf_cong[OF refl] bind_spmf_cong[OF refl])
  finally show ?thesis .
qed

lemma \<beta>_spmf_conv_option:
  "run_state (run_state (spmf.\<beta> \<eta> t) pns) hist = embed_sampling (sample_pnonces \<eta> (\<beta>1_pnonces t) pns) hist (option.\<beta> \<eta> t)"
proof -
  note [split] = t.split nonce.split
    and [simp] = pnonce_spmf_conv_option record_spmf_conv_option 
      encryption_key_spmf_conv_option encryption_key'_spmf_conv_option
      encryption_spmf_conv_option
      verification_key_spmf_conv_option verification_key'_spmf_conv_option
      signature_spmf_conv_option
    and [simp] = bind_spmf_pmf_assoc map_pmf_def map_option_bind return_pmf_bind_option 
      return_pmf_map_option_conv_bind o_def bind_return_pmf bind_spmf_pmf_commute split_def
    and [iff del] = domIff
  let ?lhs = "\<lambda>t pns hist. run_state (run_state (spmf.\<beta> \<eta> t) pns) hist"
  and ?rhs = "\<lambda>t pns hist. embed_sampling (sample_pnonces \<eta> (\<beta>1_pnonces t) pns) hist (option.\<beta> \<eta> t)"
  let ?P = "\<lambda>t. \<forall>pns hist. ?lhs t pns hist = ?rhs t pns hist"
  have ?thesis
  proof(induction t arbitrary: pns hist rule: t.induct)
    case (E t1 t2 t3)
    { fix n r
      assume t1: "t1 = EK (N (PNonce n))"
        and t3: "t3 = N (PNonce r)"
      let ?rest = "\<lambda>pns' pns'' pns'''. bind_spmf (return_pmf (run_state (run_env (option.encryption_key' \<eta> t1 n) pns') hist))
        (\<lambda>(ek, hist'). bind_spmf (return_pmf (run_state (run_env (option.\<beta> \<eta> t2) pns'') hist'))
        (\<lambda>(m, hist''). bind_spmf (return_pmf (run_state (run_env (option.encryption \<eta> n t2 m r) pns''') hist''))
        (\<lambda>(bs, hist'''). run_state (run_state (record_spmf (E t1 t2 t3) bs) pns''') hist''')))"
      have "?lhs (E t1 t2 t3) pns hist = bind_pmf (sample_pnonces \<eta> {n} pns)
        (\<lambda>pns'. bind_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t2) pns')
        (\<lambda>pns''. bind_pmf (sample_pnonces \<eta> {n, r} pns'') (\<lambda>pns'''. ?rest pns' pns'' pns''')))"
        by(simp add: E t1 t3)
      also have "\<dots> = bind_pmf (sample_pnonces \<eta> {n} pns)
        (\<lambda>pns'. bind_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t2) pns')
        (\<lambda>pns''. bind_pmf (sample_pnonces \<eta> {n, r} pns'') (\<lambda>pns'''. ?rest pns''' pns''' pns''')))"
      proof(rule bind_pmf_cong[OF refl])+
        fix pns' pns'' pns'''
        assume "pns' \<in> set_pmf (sample_pnonces \<eta> {n} pns)"
          and "pns'' \<in> set_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t2) pns')"
          and "pns''' \<in> set_pmf (sample_pnonces \<eta> {n, r} pns'')"
        then have "pns' \<subseteq>\<^sub>m pns'''" "pns'' \<subseteq>\<^sub>m pns'''" "n \<in> dom pns'" "r \<in> dom pns'''" "n \<in> dom pns'''" "sampled1 t2 pns''"
          by(auto dest: sample_pnonces_mono sample_pnonces_dom intro: map_le_trans)
        then show "?rest pns' pns'' pns''' = ?rest pns''' pns''' pns'''"
          by(simp add: \<beta>_option_incr encryption_key'_option_incr)
      qed
      also have "\<dots> = bind_pmf (sample_pnonces \<eta> (insert n (insert r (\<beta>1_pnonces t2))) pns) (\<lambda>pns'''. ?rest pns''' pns''' pns''')"
        by(simp add: bind_pmf_sample_pnonces insert_commute)
      also have "\<dots> = ?rhs (E t1 t2 t3) pns hist" by(simp add: t1 t3)
      also note calculation }
    moreover {
      fix n r
      assume t1: "t1 = EK (N (ANonce n))"
        and t3: "t3 = N (PNonce r)"
      let ?rest = "\<lambda>pns pns' pns''. bind_spmf (return_pmf (run_state (run_env (record_option t1 n) pns) hist))
        (\<lambda>(ek, hist'). bind_spmf (return_pmf (run_state (run_env (option.\<beta> \<eta> t2) pns') hist'))
        (\<lambda>(m, hist''). bind_spmf (return_pmf (run_state (run_env (pnonce_option \<eta> r) pns'') hist''))
        (\<lambda>(r', hist'''). bind_spmf (return_pmf (A_e A \<eta> ek m r')) (\<lambda>m. run_state (run_state (record_spmf (E t1 t2 t3) m) pns'') hist'''))))"
      have "?lhs (E t1 t2 t3) pns hist = bind_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t2) pns) (\<lambda>pns'. bind_pmf (sample_pnonces \<eta> {r} pns') (\<lambda>pns''. ?rest pns pns' pns''))"
        by(simp add: E t1 t3)
      also have "\<dots> = bind_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t2) pns) (\<lambda>pns'. bind_pmf (sample_pnonces \<eta> {r} pns') (\<lambda>pns''. ?rest pns'' pns'' pns''))"
      proof(rule bind_pmf_cong[OF refl])+
        fix pns' pns''
        assume "pns' \<in> set_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t2) pns)"
          and "pns'' \<in> set_pmf (sample_pnonces \<eta> {r} pns')"
        hence "sampled1 t2 pns'" "pns' \<subseteq>\<^sub>m pns''" "pns \<subseteq>\<^sub>m pns''"
          by(auto dest: sample_pnonces_mono sample_pnonces_dom intro: map_le_trans)
        then show "?rest pns pns' pns'' = ?rest pns'' pns'' pns''"
          by(simp add: \<beta>_option_incr record_option_incr)
      qed
      also have "\<dots> = bind_pmf (sample_pnonces \<eta> (insert r (\<beta>1_pnonces t2)) pns) (\<lambda>pns''. ?rest pns'' pns'' pns'')"
        by(simp add: bind_pmf_sample_pnonces)
      also have "\<dots> = ?rhs (E t1 t2 t3) pns hist" by(simp add: t1 t3)
      also note calculation }
    ultimately show ?case by simp
  next
    case (SIG t1 t2 t3)
    { fix n r
      assume t1: "t1 = SK (N (PNonce n))" and t3: "t3 = N (PNonce r)"
      let ?rest = "\<lambda>pns' pns'' pns'''. bind_spmf (return_pmf (run_state (run_env (option.verification_key' \<eta> (VK (N (PNonce n))) n) pns') hist))
        (\<lambda>(vk, hist'). bind_spmf (return_pmf (run_state (run_env (option.\<beta> \<eta> t2) pns'') hist'))
        (\<lambda>(m, hist''). bind_spmf (return_pmf (run_state (run_env (option.signature \<eta> n t2 m r) pns''') hist''))
        (\<lambda>(bs, hist'''). run_state (run_state (record_spmf (SIG t1 t2 t3) bs) pns''') hist''')))"
      have "?lhs (SIG t1 t2 t3) pns hist = bind_pmf (sample_pnonces \<eta> {n} pns)
        (\<lambda>pns'. bind_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t2) pns')
        (\<lambda>pns''. bind_pmf (sample_pnonces \<eta> {n, r} pns'') (\<lambda>pns'''. ?rest pns' pns'' pns''')))"
        by(simp add: SIG t1 t3)
      also have "\<dots> = bind_pmf (sample_pnonces \<eta> {n} pns)
        (\<lambda>pns'. bind_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t2) pns')
        (\<lambda>pns''. bind_pmf (sample_pnonces \<eta> {n, r} pns'') (\<lambda>pns'''. ?rest pns''' pns''' pns''')))"
      proof(rule bind_pmf_cong[OF refl])+
        fix pns' pns'' pns'''
        assume "pns' \<in> set_pmf (sample_pnonces \<eta> {n} pns)"
          and "pns'' \<in> set_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t2) pns')"
          and "pns''' \<in> set_pmf (sample_pnonces \<eta> {n, r} pns'')"
        then have "pns' \<subseteq>\<^sub>m pns'''" "pns'' \<subseteq>\<^sub>m pns'''" "n \<in> dom pns'" "sampled1 t2 pns''"
          by(auto dest: sample_pnonces_mono sample_pnonces_dom intro: map_le_trans)
        then show "?rest pns' pns'' pns''' = ?rest pns''' pns''' pns'''"
          by(simp add: verification_key'_option_incr \<beta>_option_incr record_option_incr)
      qed
      also have "\<dots> = bind_pmf (sample_pnonces \<eta> (insert n (insert r (\<beta>1_pnonces t2))) pns) (\<lambda>pns'''. ?rest pns''' pns''' pns''')"
        by(simp add: bind_pmf_sample_pnonces insert_commute)
      also have "\<dots> = ?rhs (SIG t1 t2 t3) pns hist" by(simp add: t1 t3)
      also note calculation }
    then show ?case by simp
  next
    case (PAIR t1 t2)
    let ?rest = "\<lambda>pns' pns''.  bind_spmf (return_pmf (run_state (run_env (option.\<beta> \<eta> t1) pns') hist))
      (\<lambda>(m1, hist'). bind_spmf (return_pmf (run_state (run_env (option.\<beta> \<eta> t2) pns'') hist'))
      (\<lambda>(m2, hist''). run_state (run_state (record_spmf (PAIR t1 t2) (A_pair A \<eta> m1 m2)) pns'') hist''))"
    have "?lhs (PAIR t1 t2) pns hist = bind_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t1) pns)
      (\<lambda>pns'. bind_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t2) pns') (\<lambda>pns''. ?rest pns' pns''))"
      by(simp add: PAIR)
    also have "\<dots> = bind_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t1) pns)
      (\<lambda>pns'. bind_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t2) pns') (\<lambda>pns''. ?rest pns'' pns''))"
    proof(rule bind_pmf_cong[OF refl])+
      fix pns' pns''
      assume "pns' \<in> set_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t1) pns)"
        and "pns'' \<in> set_pmf (sample_pnonces \<eta> (\<beta>1_pnonces t2) pns')"
      hence "sampled1 t1 pns'" "pns' \<subseteq>\<^sub>m pns''"
        by(auto dest: sample_pnonces_mono sample_pnonces_dom)
      then show "?rest pns' pns'' = ?rest pns'' pns''" by(simp add: \<beta>_option_incr)
    qed
    also have "\<dots> = ?rhs (PAIR t1 t2) pns hist" by(simp add: bind_pmf_sample_pnonces)
    finally show ?case by(simp)
  next
    { case EK         show ?case by(auto) }
    { case DK         show ?case by(auto) }
    { case VK         show ?case by(auto) }
    { case SK         show ?case by(auto) }
    { case EMPTY      show ?case by(auto) }
    { case STRING     show ?case by(auto simp add: STRING) }
    { case N          show ?case by(auto) }
    { case garbage    show ?case by(auto) }
    { case garbageE   show ?case by(auto) }
    { case garbageSIG show ?case by(auto) }
  qed
  then show ?thesis by simp
qed

lemmas spmf_convs_option =
  pnonce_spmf_conv_option
  record_spmf_conv_option
  encryption_key_spmf_conv_option
  encryption_key'_spmf_conv_option
  encryption_spmf_conv_option
  verification_key_spmf_conv_option
  verification_key'_spmf_conv_option
  signature_spmf_conv_option
  \<beta>_spmf_conv_option

end

fun spmf2sim' :: "nat \<times> protocol_nonce set \<times> 'more \<Rightarrow> pnonce \<times> history \<Rightarrow> 'more simulator_state'_scheme"
where "\<And>more. spmf2sim' (fuel', randomness', more) (pns, hist) = \<lparr>fuel = fuel', pnonce = pns, randomness = randomness', history = hist, \<dots> = more\<rparr>"

lemma spmf2sim'_conv_upd:
  "spmf2sim' (fuel s, randomness s, simulator_state'.more s) = (\<lambda>(pns, hist). s\<lparr>pnonce := pns, history := hist\<rparr>)"
by(simp add: fun_eq_iff)

fun spmf2sim :: "(nat \<times> protocol_nonce set \<times> 'more) \<Rightarrow> pnonce \<times> history \<Rightarrow> 'more simulator_state'_scheme \<Rightarrow> bool"
where "\<And>more. spmf2sim (fuel', randomness', more) (pns, hist) s \<longleftrightarrow> pns = pnonce s \<and> hist = history s \<and> fuel' = fuel s \<and> randomness' = randomness s \<and> more = simulator_state'.more s"

lemma spmf2sim_alt_def: "spmf2sim frm = BNF_Def.Grp UNIV (spmf2sim' frm)"
by(cases frm)(auto simp add: fun_eq_iff Grp_def)

lemma right_unique_spmf2sim: "right_unique (spmf2sim frm)"
by(simp add: spmf2sim_alt_def)

lemma left_total_spmf2sim: "left_total (spmf2sim frm)"
by(simp add: spmf2sim_alt_def)

(* context includes lifting_syntax begin
lemma rel_monad_extI [intro?]:
  "\<lbrakk> (M ===> (A ===> M) ===> M) (bind m1) (bind m2); 
     (A ===> M) (return m1) (return m2);
     X (monad.more m1) (monad.more m2) \<rbrakk>
   \<Longrightarrow> rel_monad_ext A M X m1 m2"
unfolding rel_monad_ext_def
by(simp add: vimage2p_def monad.select_defs iso_tuple_snd_def tuple_iso_tuple_def iso_tuple_fst_def monad_ext_Tuple_Iso_def rel_prod_sel)

lemma rel_monad_extD:
  "rel_monad_ext A M X m1 m2 \<Longrightarrow> 
  (M ===> (A ===> M) ===> M) (bind m1) (bind m2) \<and> (A ===> M) (return m1) (return m2) \<and> X (monad.more m1) (monad.more m2)"
unfolding rel_monad_ext_def
by(simp add: vimage2p_def monad.select_defs iso_tuple_snd_def tuple_iso_tuple_def iso_tuple_fst_def monad_ext_Tuple_Iso_def rel_prod_sel)

lemma ST_fail_ext_parametric [transfer_rule]:
  "((rel_monad_fail_scheme (rel_prod A S) M X ===> X') ===> rel_monad_fail_scheme (rel_prod A S) M X ===> rel_monad_fail_ext (rel_ST S M) X')
  ST_fail_ext ST_fail_ext"
unfolding ST_fail_ext_def[abs_def] by transfer_prover

lemma spmf_monad_parametric [transfer_rule]:
  "rel_monad_fail A (rel_spmf A) spmf_monad spmf_monad"
unfolding spmf_monad_def monad_fail_ext.rel_eq[symmetric]
apply(rule monad_ext_parametric[THEN rel_funD, THEN rel_funD, THEN rel_funD])
apply transfer_prover
apply transfer_prover
apply(rule monad_fail_ext_parametric[THEN rel_funD, THEN rel_funD])
apply transfer_prover+
done

end *)

(* context spmf_to_gpv begin
lemma spmf_to_gpv_monad_transfer [transfer_rule]: 
  "rel_monad_fail A (pcr_spmf_gpv A) spmf_monad gpv_monad"
unfolding spmf_monad_def gpv_monad_def by transfer_prover
end
 *)

definition cr_stateT2_reassoc :: "('a \<Rightarrow> 'd \<Rightarrow> bool) \<Rightarrow> ('b \<times> 'c \<Rightarrow> 'e \<times> 'f \<Rightarrow> bool)
   \<Rightarrow> ('a \<times> 'b) \<times> 'c \<Rightarrow> 'd \<times> 'e \<times> 'f \<Rightarrow> bool"
where "cr_stateT2_reassoc A BC = BNF_Def.vimage2p (\<lambda>((x, y), z). (x, y, z)) (\<lambda>xyz. xyz) (rel_prod A BC)"

lemma cr_stateT2_reassoc_apply:
  "cr_stateT2_reassoc A S ((x, y), z) (x', y', z') \<longleftrightarrow> A x x' \<and> S (y, z) (y', z')"
by(simp add: cr_stateT2_reassoc_def vimage2p_def)

lemma cr_stateT2_reassoc_Grp:
  "cr_stateT2_reassoc (BNF_Def.Grp A f) (BNF_Def.Grp B g) = 
   BNF_Def.Grp {((x, y), z). x \<in> A \<and> (y, z) \<in> B} (\<lambda>((x, y), z). (f x, g (y, z)))"
unfolding cr_stateT2_reassoc_def prod.rel_Grp
apply(subst type_copy_vimage2p_Grp_Rep[OF type_copy_id'])
apply(simp add: o_def split_def prod_set_defs)
done

lemma type_copy_run_state2: "type_definition StateT run_state UNIV"
by(unfold_locales) auto

lemma type_copy_run_state: "type_definition run_state StateT UNIV"
by(unfold_locales) auto

context includes lifting_syntax begin

definition cr_stateT2 :: "('s1 \<times> 's1' \<Rightarrow> 's2 \<times> 's2' \<Rightarrow> bool) \<Rightarrow> ('m1 \<Rightarrow> 'm2 \<Rightarrow> bool)
  \<Rightarrow> ('s1, ('s1', 'm1) stateT) stateT \<Rightarrow> ('s2 \<times> 's2', 'm2) stateT \<Rightarrow> bool"
where "cr_stateT2 S M = BNF_Def.vimage2p (\<lambda>x (s1, s2). run_state (run_state x s1) s2) run_state (S ===> M)"

lemma cr_stateT2_apply:
  "cr_stateT2 S M m1 m2 \<longleftrightarrow> (S ===> M) (\<lambda>(s1, y). run_state (run_state m1 s1) y) (run_state m2)"
by(simp add: cr_stateT2_def vimage2p_def)

lemma cr_stateT2_Grp:
  assumes "type_definition Rep f S"
  shows "cr_stateT2 (BNF_Def.Grp S f) (BNF_Def.Grp M g) = 
  BNF_Def.Grp {x. (\<lambda>(s1, y). run_state (run_state x s1) y) ` S \<subseteq> M}
     (\<lambda>x. StateT ((Rep ---> g) (\<lambda>(s1, s2). run_state (run_state x s1) s2)))"
unfolding cr_stateT2_def
apply(subst rel_fun_Grp_copy_Abs[OF assms])
apply(subst type_copy_vimage2p_Grp_Rep[OF type_copy_run_state])
apply(simp add: o_def)
done

lemma bind_stateT2_transfer: 
  "((M ===> (cr_stateT2_reassoc A S ===> M) ===> M)
  ===> cr_stateT2 S M ===> (A ===> cr_stateT2 S M) ===> cr_stateT2 S M)
   (\<lambda>bind. bind_state (bind_state bind)) bind_state"
apply(unfold cr_stateT2_def vimage2p_def)
apply(rule rel_funI)+
apply clarify
apply(frule (1) rel_funD)
apply simp
apply(drule (1) rel_funD)
apply(erule rel_funD)
apply(rule rel_funI)
apply(clarsimp simp add: cr_stateT2_reassoc_apply)
apply(drule (1) rel_funD)
apply(drule (1) rel_funD; simp; drule (1) rel_funD; simp; fail)
done

lemma return_stateT2_transfer: 
  "((cr_stateT2_reassoc A S ===> M) ===> A ===> cr_stateT2 S M)
   (\<lambda>return. return_state (return_state return)) return_state"
apply(rule rel_funI)+
apply(simp add: cr_stateT2_apply)
apply(rule rel_funI)
apply clarsimp
apply(erule rel_funD)
apply(simp add: cr_stateT2_reassoc_apply)
done

lemma fail_stateT2_transfer:
  "(M ===> cr_stateT2 S M) (\<lambda>fail. fail_state (fail_state fail)) fail_state"
by(simp add: cr_stateT2_apply split_def rel_fun_def)

(* definition cr_stateT2_more
where "cr_stateT2_more AS M X X' = 
  BNF_Def.vimage2p (\<lambda>(more1, more2) m. more1 (ST_scheme more2 m)) (\<lambda>m. m) (rel_monad_ext AS M X ===> X')"
 *)
(* definition map_monad_ext :: "('b \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> ('more1 \<Rightarrow> 'more2) \<Rightarrow> ('b, 'm, 'more1) monad_ext \<Rightarrow> ('a, 'm, 'more2) monad_ext"
where "\<And>more. map_monad_ext f1 f2 more m = \<lparr>bind = \<lambda>x g. bind m x (\<lambda>y. g (f1 y)), return = \<lambda>x. return m (f2 x), \<dots> = more (monad.more m)\<rparr>"
 *)
(*
definition monad_more_update' :: "('more \<Rightarrow> 'more') \<Rightarrow> ('a, 'm, 'more) monad_ext \<Rightarrow> ('a, 'm, 'more') monad_ext"
where "monad_more_update' f x = \<lparr>bind = bind x, return = return x, \<dots> = f (monad.more x)\<rparr>"
*)

(*
definition ST2_more2 :: "(('a \<times> 'b \<times> 'c, 'm, 'more) monad_ext \<Rightarrow> 'more') \<Rightarrow> (('a \<times> 'b) \<times> 'c, 'm, 'more) monad_ext \<Rightarrow> 'more'"
where "\<And>more. ST2_more2 more m = more (map_monad_ext (\<lambda>((a, b), c). (a, b, c)) (\<lambda>(a, b, c). ((a, b), c)) id m)"
*)

(* definition ST2_more1 :: "(('a \<times> 'b, 'm, 'more) monad_ext \<Rightarrow> 'more') \<Rightarrow> ('a, ('b, 'c) ST, 'more) monad_ext \<Rightarrow> 'more"
where "ST2_more1 = undefined"
 *)
(* lemma ST_ext_cr_stateT2_transfer:
  "(cr_stateT2_more (cr_stateT2_reassoc A S) M X X' ===> rel_monad_ext (cr_stateT2_reassoc A S) M X ===> rel_monad_ext A (cr_stateT2 S M) X')
  (\<lambda>(MORE1, MORE2) m. ST_ext MORE1 (ST_ext MORE2 m)) ST_ext"
apply(rule rel_funI)+
apply(rule rel_monad_extI; clarsimp simp add: bind_ST return_ST)
  apply(erule bind_state_ST2_transfer[THEN rel_funD])
 apply(erule return_state_ST2_transfer[THEN rel_funD])
apply(simp add: cr_stateT2_more_def vimage2p_def)
apply(erule (1) rel_funD)
done
 *)
(* lemma ST_fail_ext_cr_stateT2_aux:
  "cr_stateT2_more (cr_stateT2_reassoc op = op =) (rel_spmf (cr_stateT2_reassoc op = op =)) (rel_monad_fail_ext (rel_spmf (cr_stateT2_reassoc op = op =)) op =)
     (rel_monad_fail_ext (cr_stateT2 op = (rel_spmf (cr_stateT2_reassoc op = op =))) op =)
     (ST_fail_ext monad_fail.more, ST_fail_ext monad_fail.more) (ST_fail_ext monad_fail.more)"
apply(simp add: cr_stateT2_more_def vimage2p_def)
apply(rule rel_funI)
apply(simp add: rel_monad_fail_ext_def vimage2p_def ST_fail_ext_def monad_fail_ext_def iso_tuple_cons_def monad_fail_ext_Tuple_Iso_def Abs_monad_fail_ext_inverse)
apply(drule rel_monad_extD)
apply(rule conjI)
 apply(simp add: cr_stateT2_def vimage2p_def)
 apply(rule rel_funI)
 apply clarsimp
 apply(simp add: monad.select_defs(3) iso_tuple_snd_def monad_ext_Tuple_Iso_def iso_tuple_fst_def monad_fail_ext_Tuple_Iso_def monad_fail.select_defs)
 apply(rule fst_transfer[THEN rel_funD])
 apply simp
apply(simp add: monad.select_defs(3) iso_tuple_snd_def monad_ext_Tuple_Iso_def iso_tuple_fst_def monad_fail_ext_Tuple_Iso_def monad_fail.select_defs)
apply clarsimp
apply(rule snd_transfer[THEN rel_funD])
apply simp
done
 *)
(*
lemma ST_fail_ext_cr_stateT2_transfer:
  "(P ===> cr_stateT2_more A M X X')
  (map_prod ST_fail_ext ST_fail_ext) ST_fail_ext"
apply(rule rel_funI)
apply(clarsimp simp add: cr_stateT2_more_def vimage2p_def ST_sel)
apply(rule rel_funI)
oops
*)

lemma rel_stateT_alt_def: "rel_stateT S M = BNF_Def.vimage2p run_state run_state (S ===> M)"
apply(rule ext)+
subgoal for x y
  apply(cases x; cases y)
  apply(simp add: rel_stateT_def vimage2p_def)
  done
done

lemma rel_STD:
  "\<lbrakk> rel_stateT S M x y; S s1 s2 \<rbrakk> \<Longrightarrow> M (run_state x s1) (run_state y s2)"
by(auto simp add: rel_stateT_def dest: rel_funD)

lemma rel_ST_Grp:
  assumes "type_definition Rep f S"
  shows "rel_stateT (BNF_Def.Grp S f) (BNF_Def.Grp A g) = 
    BNF_Def.Grp {x. run_state x ` S \<subseteq> A} (StateT \<circ> (Rep ---> g) \<circ> run_state)"
by(simp add: rel_stateT_alt_def type_copy_vimage2p_Grp_Rep[OF type_copy_run_state] rel_fun_Grp_copy_Abs[OF assms])

end

context comp_impl_sims begin

context includes lifting_syntax begin interpretation spmf_to_gpv .

lemma pnonce_spmf2sim_transfer: "(spmf2sim frm ===> op =) fst pnonce"
by(cases frm)(simp add: rel_fun_def)

lemma history_spmf2sim_transfer: "(spmf2sim frm ===> op =) snd history"
by(cases frm)(simp add: rel_fun_def)

lemma pnonce_update_spmf2sim_transfer:
  "(op = ===> spmf2sim frm ===> spmf2sim frm) (\<lambda>f s. (f (fst s), snd s)) pnonce_update"
by(cases frm)(simp add: rel_fun_def)

lemma history_update_spmf2sim_transfer:
  "(op = ===> spmf2sim frm ===> spmf2sim frm) (\<lambda>f s. (fst s, f (snd s))) history_update"
by(cases frm)(simp add: rel_fun_def)

lemma cr_spmf_gpv_Grp: "cr_spmf_gpv = BNF_Def.Grp UNIV lift_spmf"
by(simp add: fun_eq_iff cr_spmf_gpv_def Grp_def)

lemma pcr_spmf_gpv_Grp:
  "\<And>A. pcr_spmf_gpv (BNF_Def.Grp A f) = BNF_Def.Grp {p. set_spmf p \<subseteq> A} (lift_spmf \<circ> map_spmf f)"
apply(simp add: pcr_spmf_gpv_def cr_spmf_gpv_Grp)
apply(rewrite in "\<hole> = _" eq_alt)
apply(rewrite gpv.rel_Grp)
apply(fastforce simp add: BNF_Def.Grp_def fun_eq_iff map_lift_spmf)
done

lemma lazy_convs_spmf:
  fixes \<Gamma> \<eta> t k msg r n embed bs and s :: "('s, 'more) simulator_state_scheme"
  defines [simp]: "embed \<equiv> \<lambda>monad. lift_spmf (map_spmf (\<lambda>((m, pns), hist). (m, s\<lparr>pnonce := pns, history := hist\<rparr>)) (run_state (run_state monad (pnonce s)) (history s)))"
  -- \<open>We use \<open>defines\<close> to avoid type annotations for preventing excessive polymorphism.\<close>
  and "\<beta>1 \<equiv> sim1.\<beta> \<eta> t :: (_, (_, 'call, 'ret) gpv) stateT"
  and "pnonce1 \<equiv> pnonce_sim \<eta> n :: (_, (_, 'call, 'ret) gpv) stateT"
  and "record1 \<equiv> record_sim t bs :: (_, (_, 'call, 'ret) gpv) stateT"
  and "encryption_key1 \<equiv> sim1.encryption_key \<eta> k :: (_, (_, 'call, 'ret) gpv) stateT"
  and "encryption1 \<equiv> sim1.encryption \<eta> k t msg r :: (_, (_, 'call, 'ret) gpv) stateT"
  and "encryption_key'1 \<equiv> sim1.encryption_key' \<eta> t k :: (_, (_, 'call, 'ret) gpv) stateT"
  and "verification_key1 \<equiv> sim1.verification_key \<eta> k :: (_, (_, 'call, 'ret) gpv) stateT"
  and "verification_key'1 \<equiv> sim1.verification_key' \<eta> t k :: (_, (_, 'call, 'ret) gpv) stateT"
  and "signature1 \<equiv> sim1.signature \<eta> k t msg r :: (_, (_, 'call, 'ret) gpv) stateT"
  shows pnonce_sim_conv_spmf: "run_state pnonce1 s = embed (pnonce_spmf \<eta> n)" (is "?pnonce")
  and record_sim_conv_spmf: "run_state record1 s = embed (record_spmf t bs)" (is "?record")
  and encryption_key1_conv_spmf: "run_state encryption_key1 s = embed (spmf.encryption_key \<eta> k)" (is "?encryption_key")
  and encryption_key'1_conv_spmf: "run_state encryption_key'1 s = embed (spmf.encryption_key' \<eta> t k)" (is "?encryption_key'")
  and encryption1_conv_spmf: "run_state encryption1 s = embed (spmf.encryption \<eta> k t msg r)" (is "?encryption")
  and verification_key1_conv_spmf: "run_state verification_key1 s = embed (spmf.verification_key \<eta> k)" (is "?verification_key")
  and verification_key'1_conv_spmf: "run_state verification_key'1 s = embed (spmf.verification_key' \<eta> t k)" (is "?verification_key'")
  and signature1_conv_spmf: "run_state signature1 s = embed (spmf.signature \<eta> k t msg r)" (is "?signature")
  and \<beta>1_conv_spmf: "run_state \<beta>1 s = embed (spmf.\<beta> \<eta> t)" (is "?\<beta>")
proof -
  define return1 :: "(bitstring, (pnonce \<times> history, _) stateT) return" 
    where "return1 \<equiv> return_state return_spmf"
  define bind1 :: "(bitstring, (pnonce \<times> history, _) stateT) bind" 
    where "bind1 \<equiv> bind_state bind_spmf"
  define fail1 :: "(pnonce \<times> history, (bitstring \<times> (pnonce \<times> history)) spmf) stateT fail" 
    where "fail1 \<equiv> fail_state (return_pmf None)"
  note [transfer_rule] = pnonce_spmf2sim_transfer history_spmf2sim_transfer pnonce_update_spmf2sim_transfer history_update_spmf2sim_transfer

  def pnonce' \<equiv> "\<lambda>\<eta> n. StateT (\<lambda>(pns, hist). map_spmf (\<lambda>((m, pns'), hist'). (m, pns', hist')) (run_state (run_state (pnonce_spmf \<eta> n) pns) hist))"
  def record' \<equiv> "\<lambda>t m. StateT (\<lambda>(pns, hist). map_spmf (\<lambda>((m, pns'), hist'). (m, pns', hist')) (run_state (run_state (record_spmf t m) pns) hist))"
  interpret spmf': comp_impl_sim1_base A
    return1 bind1 fail1
    pnonce'
    .
  interpret spmf': comp_impl_\<beta>_base A
    return1 bind1 fail1
    pnonce' record' spmf'.encryption_key spmf'.encryption spmf'.verification_key spmf'.signature
    .
  have pnonce' [transfer_rule]:
    "(op = ===> op = ===> rel_stateT (spmf2sim frm) (pcr_spmf_gpv (rel_prod op = (spmf2sim frm)))) pnonce' pnonce_sim"
    for frm :: "(nat \<times> protocol_nonce set \<times> 'more')"
    unfolding pnonce'_def pnonce_sim_def[abs_def] pnonce_spmf_def list.rel_eq[symmetric]
    apply(unfold map_spmf_bind_spmf o_def Let_def map_return_pmf option.map prod.simps split_def fst_conv snd_conv prod.collapse stateT.sel option.case_distrib[where h="map_spmf _"])
    apply transfer_prover
    done
  have [transfer_rule]:
    "(op = ===> op = ===> rel_stateT (spmf2sim frm) (pcr_spmf_gpv (rel_prod op = (spmf2sim frm)))) record' record_sim"
    for frm :: "(nat \<times> protocol_nonce set \<times> 'more')"
    unfolding record'_def record_sim_def[abs_def] record_spmf_def map_spmf_bind_spmf o_def Let_def map_return_pmf option.map prod.simps split_def fst_conv snd_conv prod.collapse stateT.sel
    by transfer_prover

  let ?cr = "cr_stateT2 op = (rel_spmf (cr_stateT2_reassoc op = op =))"
  let ?g = "\<lambda>monad. map_spmf (\<lambda>((m, pns), history). (m, pns, history)) (run_state (run_state monad (pnonce s)) (history s))"
  have [transfer_rule]: "(op = ===> op = ===> ?cr) pnonce_spmf pnonce'"
    by(auto simp add: rel_fun_def pnonce'_def cr_stateT2_def vimage2p_def spmf_rel_map split_beta cr_stateT2_reassoc_def intro!: rel_spmf_reflI)
  have [transfer_rule]: "(op = ===> op = ===> ?cr) record_spmf record'"
    by(auto simp add: rel_fun_def record'_def cr_stateT2_def vimage2p_def spmf_rel_map split_beta cr_stateT2_reassoc_def intro!: rel_spmf_reflI)
  have [transfer_rule]: "(op = ===> ?cr) (return_state (return_state return_spmf)) return1"
    unfolding return1_def by(rule return_stateT2_transfer[THEN rel_funD]) transfer_prover
  have [transfer_rule]: "(?cr ===> (op = ===> ?cr) ===> ?cr) (bind_state (bind_state bind_spmf)) bind1"
    unfolding bind1_def by(rule bind_stateT2_transfer[THEN rel_funD]) transfer_prover
  have [transfer_rule]: "?cr (fail_state (fail_state (return_pmf None))) fail1"
    unfolding fail1_def by(rule fail_stateT2_transfer[THEN rel_funD]) transfer_prover

  def frm \<equiv> "(fuel s, randomness s, simulator_state'.more s)"
  have frm: "spmf2sim frm (pnonce s, history s) s" unfolding frm_def by simp

  let ?f = "\<lambda>m. lift_spmf (map_spmf (map_prod id (\<lambda>(pns, hist). s\<lparr>pnonce := pns, history := hist\<rparr>)) m)"
  let ?rel1 = "pcr_spmf_gpv (rel_prod op = (spmf2sim frm))"
  let ?rel = "rel_stateT (spmf2sim frm) ?rel1"

  -- \<open>@{const pnonce_sim}\<close>
  have "?rel (pnonce' \<eta> n) pnonce1" unfolding pnonce1_def by transfer_prover
  hence [unfolded eq_alt]: "?rel1 (run_state (pnonce' \<eta> n) (pnonce s, history s)) (run_state pnonce1 s)"
    using frm by(rule rel_STD)
  then have "run_state pnonce1 s = ?f (run_state (pnonce' \<eta> n) (pnonce s, history s))"
    by(auto simp add: spmf2sim_alt_def prod.rel_Grp pcr_spmf_gpv_Grp rel_ST_Grp spmf2sim'_conv_upd frm_def elim!: GrpE)
  also have [unfolded eq_alt]: "?cr (pnonce_spmf \<eta> n) (pnonce' \<eta> n)" by transfer_prover
  hence "run_state (pnonce' \<eta> n) (pnonce s, history s) = ?g (pnonce_spmf \<eta> n)"
    by(auto simp add: cr_stateT2_reassoc_Grp option.rel_Grp pmf.rel_Grp cr_stateT2_Grp[OF type_copy_id] elim!: GrpE)
  finally show ?pnonce by(simp add: spmf.map_comp o_def split_def)

  -- \<open>@{const record_sim}\<close>
  have "?rel (record' t bs) record1" unfolding list.rel_eq[symmetric] record1_def by transfer_prover
  hence [unfolded eq_alt]: "?rel1 (run_state (record' t bs) (pnonce s, history s)) (run_state record1 s)"
    using frm by(rule rel_STD)
  then have "run_state record1 s = ?f (run_state (record' t bs) (pnonce s, history s))"
    by(auto simp add: spmf2sim_alt_def prod.rel_Grp pcr_spmf_gpv_Grp rel_ST_Grp spmf2sim'_conv_upd frm_def elim!: GrpE)
  also have [unfolded eq_alt]: "?cr (record_spmf t bs) (record' t bs)"
    unfolding list.rel_eq[symmetric] by transfer_prover
  hence "run_state (record' t bs) (pnonce s, history s) = ?g (record_spmf t bs)"
    by(auto simp add: cr_stateT2_reassoc_Grp option.rel_Grp pmf.rel_Grp cr_stateT2_Grp[OF type_copy_id] elim!: GrpE)
  finally show ?record by(simp add: spmf.map_comp o_def split_def)

  -- \<open>@{const sim1.encryption_key}\<close>
  have "?rel (spmf'.encryption_key \<eta> k) encryption_key1"
    unfolding list.rel_eq[symmetric] encryption_key1_def return1_def bind1_def fail1_def by transfer_prover
  hence [unfolded eq_alt]: "?rel1 (run_state (spmf'.encryption_key \<eta> k) (pnonce s, history s)) (run_state encryption_key1 s)"
    using frm by(rule rel_STD)
  then have "run_state encryption_key1 s = ?f (run_state (spmf'.encryption_key \<eta> k) (pnonce s, history s))"
    by(auto simp add: spmf2sim_alt_def prod.rel_Grp pcr_spmf_gpv_Grp rel_ST_Grp spmf2sim'_conv_upd frm_def elim!: GrpE)
  also have [unfolded eq_alt]: "?cr (spmf.encryption_key \<eta> k) (spmf'.encryption_key \<eta> k)"
    by(fold relator_eq)  transfer_prover
  hence "run_state (spmf'.encryption_key \<eta> k) (pnonce s, history s) = ?g (spmf.encryption_key \<eta> k)"
    by(auto simp add: cr_stateT2_reassoc_Grp option.rel_Grp pmf.rel_Grp cr_stateT2_Grp[OF type_copy_id] elim!: GrpE)
  finally show ?encryption_key by(simp add: spmf.map_comp o_def split_def)

  -- \<open>@{const sim1.encryption_key'}\<close>
  have "?rel (spmf'.encryption_key' \<eta> t k) encryption_key'1"
    unfolding list.rel_eq[symmetric] encryption_key'1_def return1_def bind1_def fail1_def by transfer_prover
  hence [unfolded eq_alt]: "?rel1 (run_state (spmf'.encryption_key' \<eta> t k) (pnonce s, history s)) (run_state encryption_key'1 s)"
    using frm by(rule rel_STD)
  then have "run_state encryption_key'1 s = ?f (run_state (spmf'.encryption_key' \<eta> t k) (pnonce s, history s))"
    by(auto simp add: spmf2sim_alt_def prod.rel_Grp pcr_spmf_gpv_Grp rel_ST_Grp spmf2sim'_conv_upd frm_def elim!: GrpE)
  also have [unfolded eq_alt]: "?cr (spmf.encryption_key' \<eta> t k) (spmf'.encryption_key' \<eta> t k)"
    by(fold relator_eq) transfer_prover
  hence "run_state (spmf'.encryption_key' \<eta> t k) (pnonce s, history s) = ?g (spmf.encryption_key' \<eta> t k)"
    by(auto simp add: cr_stateT2_reassoc_Grp option.rel_Grp pmf.rel_Grp cr_stateT2_Grp[OF type_copy_id] elim!: GrpE)
  finally show ?encryption_key' by(simp add: spmf.map_comp o_def split_def)

  -- \<open>@{const sim1.encryption}\<close>
  have "?rel (spmf'.encryption \<eta> k t msg r) encryption1"
    unfolding list.rel_eq[symmetric] encryption1_def return1_def bind1_def fail1_def by transfer_prover
  hence [unfolded eq_alt]: "?rel1 (run_state (spmf'.encryption \<eta> k t msg r) (pnonce s, history s)) (run_state encryption1 s)"
    using frm by(rule rel_STD)
  then have "run_state encryption1 s = ?f (run_state (spmf'.encryption \<eta> k t msg r) (pnonce s, history s))"
    by(auto simp add: spmf2sim_alt_def prod.rel_Grp pcr_spmf_gpv_Grp rel_ST_Grp spmf2sim'_conv_upd frm_def elim!: GrpE)
  also have [unfolded eq_alt]: "?cr (spmf.encryption \<eta> k t msg r) (spmf'.encryption \<eta> k t msg r)"
    by(fold relator_eq) transfer_prover
  hence "run_state (spmf'.encryption \<eta> k t msg r) (pnonce s, history s) = ?g (spmf.encryption \<eta> k t msg r)"
    by(auto simp add: cr_stateT2_reassoc_Grp option.rel_Grp pmf.rel_Grp cr_stateT2_Grp[OF type_copy_id] elim!: GrpE)
  finally show ?encryption by(simp add: spmf.map_comp o_def split_def)

  -- \<open>@{const sim1.verification_key}\<close>
  have "?rel (spmf'.verification_key \<eta> k) verification_key1"
    unfolding list.rel_eq[symmetric] verification_key1_def return1_def bind1_def fail1_def by transfer_prover
  hence [unfolded eq_alt]: "?rel1 (run_state (spmf'.verification_key \<eta> k) (pnonce s, history s)) (run_state verification_key1 s)"
    using frm by(rule rel_STD)
  then have "run_state verification_key1 s = ?f (run_state (spmf'.verification_key \<eta> k) (pnonce s, history s))"
    by(auto simp add: spmf2sim_alt_def prod.rel_Grp pcr_spmf_gpv_Grp rel_ST_Grp spmf2sim'_conv_upd frm_def elim!: GrpE)
  also have [unfolded eq_alt]: "?cr (spmf.verification_key \<eta> k) (spmf'.verification_key \<eta> k)"
    by(fold relator_eq) transfer_prover
  hence "run_state (spmf'.verification_key \<eta> k) (pnonce s, history s) = ?g (spmf.verification_key \<eta> k)"
    by(auto simp add: cr_stateT2_reassoc_Grp option.rel_Grp pmf.rel_Grp cr_stateT2_Grp[OF type_copy_id] elim!: GrpE)
  finally show ?verification_key by(simp add: spmf.map_comp o_def split_def)

  -- \<open>@{const sim1.verification_key'}\<close>
  have "?rel (spmf'.verification_key' \<eta> t k) verification_key'1"
    unfolding list.rel_eq[symmetric] verification_key'1_def return1_def bind1_def fail1_def by transfer_prover
  hence [unfolded eq_alt]: "?rel1 (run_state (spmf'.verification_key' \<eta> t k) (pnonce s, history s)) (run_state verification_key'1 s)"
    using frm by(rule rel_STD)
  then have "run_state verification_key'1 s = ?f (run_state (spmf'.verification_key' \<eta> t k) (pnonce s, history s))"
    by(auto simp add: spmf2sim_alt_def prod.rel_Grp pcr_spmf_gpv_Grp rel_ST_Grp spmf2sim'_conv_upd frm_def elim!: GrpE)
  also have [unfolded eq_alt]: "?cr (spmf.verification_key' \<eta> t k) (spmf'.verification_key' \<eta> t k)"
    by(fold relator_eq) transfer_prover
  hence "run_state (spmf'.verification_key' \<eta> t k) (pnonce s, history s) = ?g (spmf.verification_key' \<eta> t k)"
    by(auto simp add: cr_stateT2_reassoc_Grp option.rel_Grp pmf.rel_Grp cr_stateT2_Grp[OF type_copy_id] elim!: GrpE)
  finally show ?verification_key' by(simp add: spmf.map_comp o_def split_def)

  -- \<open>@{const sim1.signature}\<close>
  have "?rel (spmf'.signature \<eta> k t msg r) signature1"
    unfolding list.rel_eq[symmetric] signature1_def return1_def bind1_def fail1_def by transfer_prover
  hence [unfolded eq_alt]: "?rel1 (run_state (spmf'.signature \<eta> k t msg r) (pnonce s, history s)) (run_state signature1 s)"
    using frm by(rule rel_STD)
  then have "run_state signature1 s = ?f (run_state (spmf'.signature \<eta> k t msg r) (pnonce s, history s))"
    by(auto simp add: spmf2sim_alt_def prod.rel_Grp pcr_spmf_gpv_Grp rel_ST_Grp spmf2sim'_conv_upd frm_def elim!: GrpE)
  also have [unfolded eq_alt]: "?cr (spmf.signature \<eta> k t msg r) (spmf'.signature \<eta> k t msg r)"
    by(fold relator_eq) transfer_prover
  hence "run_state (spmf'.signature \<eta> k t msg r) (pnonce s, history s) = ?g (spmf.signature \<eta> k t msg r)"
    by(auto simp add: cr_stateT2_reassoc_Grp option.rel_Grp pmf.rel_Grp cr_stateT2_Grp[OF type_copy_id] elim!: GrpE)
  finally show ?signature by(simp add: spmf.map_comp o_def split_def)

  -- \<open>@{const sim1.\<beta>}\<close>
  def \<beta>' \<equiv> "spmf'.\<beta> \<eta> t"
  have "?rel \<beta>' \<beta>1" unfolding \<beta>1_def \<beta>'_def list.rel_eq[symmetric] return1_def bind1_def fail1_def by transfer_prover
  then have [unfolded eq_alt]: "pcr_spmf_gpv (rel_prod op = (spmf2sim frm)) (run_state \<beta>' (pnonce s, history s)) (run_state \<beta>1 s)"
    using frm by(rule rel_STD)
  hence "run_state \<beta>1 s = ?f (run_state \<beta>' (pnonce s, history s))"
    by(auto simp add: spmf2sim_alt_def prod.rel_Grp pcr_spmf_gpv_Grp spmf2sim'_conv_upd frm_def elim!: GrpE)
  also have [unfolded eq_alt]: "?cr (spmf.\<beta> \<eta> t) \<beta>'" unfolding \<beta>'_def by(fold relator_eq) transfer_prover
  hence "run_state \<beta>' (pnonce s, history s) = ?g (spmf.\<beta> \<eta> t)"
    by(auto simp add: cr_stateT2_reassoc_Grp option.rel_Grp pmf.rel_Grp cr_stateT2_Grp[OF type_copy_id] elim!: GrpE)
  finally show ?\<beta> by(simp add: spmf.map_comp o_def split_def)
qed

lemmas lazy_convs_pure =
  lazy_convs_spmf
  spmf_convs_option
  option_convs_pure

lemma encryption_key1_sampled: fixes s1 :: "('s, 'more) simulator_state_scheme" shows
  "(m, s1') \<in> results'_gpv (run_state (sim1.encryption_key \<eta> k) s1)
  \<Longrightarrow> \<exists>bs. pnonce s1' k = Some bs \<and> A_ek A \<eta> bs = Some m"
by(auto simp add: lazy_convs_pure pure.encryption_key_def bind_eq_Some_conv pnonce_pure_def dest!: sample_pnonces_dom)

end

inductive wf_history1 :: "security \<Rightarrow> pnonce \<Rightarrow> history \<Rightarrow> bool"
  for \<eta> pns hist
where 
  wf_history1I:
  "\<lbrakk> \<And>t m. hist t = Some m \<Longrightarrow> run_env (pure.\<beta> \<eta> t) pns = Some m;
     \<And>t m. hist t = Some m \<Longrightarrow> sampled1 t pns \<rbrakk>
   \<Longrightarrow> wf_history1 \<eta> pns hist"

lemmas wf_history1_cases [consumes 1, case_names wf_history1, cases pred] = wf_history1.cases

lemma wf_history1D:
  assumes "wf_history1 \<eta> pns hist"
  shows wf_history1D_\<beta>: "\<And>t m. hist t = Some m \<Longrightarrow> run_env (pure.\<beta> \<eta> t) pns = Some m"
  and wf_history1D_sampled1: "\<And>t m. hist t = Some m \<Longrightarrow> sampled1 t pns"
using assms by(auto elim!: wf_history1.cases del: subsetI)

lemma wf_history1_pnonce_mono:
  "\<lbrakk> wf_history1 \<eta> pns hist; pns \<subseteq>\<^sub>m pns' \<rbrakk>
  \<Longrightarrow> wf_history1 \<eta> pns' hist"
by(auto 4 4 intro!: wf_history1I elim!: wf_history1_cases del: subsetI dest: \<beta>_pure_incr[where \<eta>=\<eta>] map_le_implies_dom_le)

lemma wf_history1_history_updateI:
  "\<lbrakk> wf_history1 \<eta> pns hist; run_env (pure.\<beta> \<eta> t) pns = Some m; sampled1 t pns \<rbrakk>
  \<Longrightarrow> wf_history1 \<eta> pns (hist(t \<mapsto> m))"
by(auto intro!: wf_history1I split: if_split_asm dest: wf_history1D del: subsetI)

lemma wf_history1_empty [simp]: "wf_history1 \<eta> pns Map.empty"
by(rule wf_history1I) auto

lemma wf_history1_record_all:
  fixes N
  assumes wfh: "wf_history1 \<eta> pns hist"
  and pns: "pns' \<in> set_pmf (sample_pnonces \<eta> N pns)"
  and hist: "hist' = hist ++ record_all \<eta> t pns'"
  and N: "\<beta>1_pnonces t \<subseteq> N" "finite N"
  shows "wf_history1 \<eta> pns' hist'"
using wfh
proof cases
  case wf_history1
  from pns have incr: "pns \<subseteq>\<^sub>m pns'" by(rule sample_pnonces_mono)
  show ?thesis 
  proof
    fix t' m
    assume tm: "hist' t' = Some m"
    show "run_env (pure.\<beta> \<eta> t') pns' = Some m"
    proof(cases "t' \<in> dom (record_all \<eta> t pns')")
      case True
      thus ?thesis using tm hist
        by(simp add: map_add_dom_app_simps)(simp add: record_all_def domIff split: if_split_asm)
    next
      case False
      with tm hist have "hist t' = Some m" by(simp add: map_add_dom_app_simps)
      hence "run_env (pure.\<beta> \<eta> t') pns = Some m" by(rule wf_history1)
      with \<beta>_pure_incr'[OF incr, of \<eta> t'] show ?thesis by simp
    qed
    show "sampled1 t' pns'"
    proof(cases "t' \<in> dom (record_all \<eta> t pns')")
      case False
      with tm hist have "hist t' = Some m" by(simp add: map_add_dom_app_simps)
      hence "sampled1 t' pns" by(rule wf_history1)
      with incr show ?thesis by(auto 4 3 dest: map_le_Some1D)
    next
      case True
      hence "\<beta>_rec t t'" by(clarsimp simp add: record_all_def split: if_split_asm)
      with pns[THEN sample_pnonces_dom, simplified] hist N show ?thesis
        by(auto 4 3 simp add: \<beta>_pnonces_def intro: rtranclp_trans)
    qed
  qed
qed

lemma wf_history1_encryption_key: fixes s' :: "('s, 'more) simulator_state_scheme" shows
  "\<lbrakk> (m, s') \<in> results'_gpv (run_state (sim1.encryption_key \<eta> k) s);
     wf_history1 \<eta> (pnonce s) (history s) \<rbrakk>
  \<Longrightarrow> wf_history1 \<eta> (pnonce s') (history s')"
by(auto simp add: lazy_convs_pure elim!: wf_history1_pnonce_mono intro: sample_pnonces_mono)

lemma wf_history1_encryption_key': fixes s' :: "('s, 'more) simulator_state_scheme" shows
  "\<lbrakk> (m, s') \<in> results'_gpv (run_state (sim1.encryption_key' \<eta> (EK (N (PNonce k))) k) s);
     wf_history1 \<eta> (pnonce s) (history s) \<rbrakk>
  \<Longrightarrow> wf_history1 \<eta> (pnonce s') (history s')"
by(auto simp add: map_add_dom_app_simps record_all_apply dom_record_all lazy_convs_pure elim!: wf_history1_record_all[where t="EK (N (PNonce k))"] intro!: map_eqI split: if_split_asm)

lemma wf_history1_verification_key: fixes s' :: "('s, 'more) simulator_state_scheme" shows
  "\<lbrakk> (m, s') \<in> results'_gpv (run_state (sim1.verification_key \<eta> k) s);
     wf_history1 \<eta> (pnonce s) (history s) \<rbrakk>
  \<Longrightarrow> wf_history1 \<eta> (pnonce s') (history s')"
by(auto simp add: lazy_convs_pure elim!: wf_history1_pnonce_mono intro: sample_pnonces_mono)

lemma wf_history1_verification_key': fixes s' :: "('s, 'more) simulator_state_scheme" shows
  "\<lbrakk> (m, s') \<in> results'_gpv (run_state (sim1.verification_key' \<eta> (VK (N (PNonce k))) k) s);
     wf_history1 \<eta> (pnonce s) (history s) \<rbrakk>
  \<Longrightarrow> wf_history1 \<eta> (pnonce s') (history s')"
by(auto simp add: map_add_dom_app_simps record_all_apply dom_record_all lazy_convs_pure elim!: wf_history1_record_all[where t="VK (N (PNonce k))"] intro!: map_eqI split: if_split_asm)

lemma wf_history1_\<beta>1:
  "\<lbrakk> (m, s') \<in> results'_gpv (run_state (sim1_\<beta> \<eta> t) s); wf_history1 \<eta> (pnonce s) (history s) \<rbrakk>
  \<Longrightarrow> wf_history1 \<eta> (pnonce s') (history s')"
by(auto simp add: lazy_convs_pure elim: wf_history1_record_all)

lemma wf_history1_rec_message:
  "\<lbrakk> Some \<sigma>' \<in> results'_gpv (sim1.rec_message \<A> \<eta> \<sigma> t :: (_, comp_impl.call, comp_impl.ret) gpv);
     wf_history1 \<eta> (pnonce \<sigma>) (history \<sigma>) \<rbrakk>
   \<Longrightarrow> wf_history1 \<eta> (pnonce \<sigma>') (history \<sigma>')"
by(auto simp add: sim1.rec_message_def dest!: wf_history1_\<beta>1 split: option.split_asm)

lemma wf_history1_gen_message:
  "\<lbrakk> Some (t, \<sigma>') \<in> results'_gpv (sim1.gen_message \<A> \<eta> \<sigma>); wf_history1 \<eta> (pnonce \<sigma>) (history \<sigma>) \<rbrakk>
  \<Longrightarrow> wf_history1 \<eta> (pnonce \<sigma>') (history \<sigma>')"
by(auto simp add: sim1.gen_message_def)

lemma wf_history1_sim_control:
  "\<lbrakk> Some (msg, \<sigma>') \<in> results'_gpv (sim_control \<A> \<eta> \<sigma> out); wf_history1 \<eta> (pnonce \<sigma>) (history \<sigma>) \<rbrakk>
  \<Longrightarrow> wf_history1 \<eta> (pnonce \<sigma>') (history \<sigma>')"
by(auto simp add: sim_control_def)

lemma wf_history1_hybrid_step1:
  "\<lbrakk> Some (to, sprotf', \<sigma>') \<in> results'_gpv (hybrid_step1_sim1 \<Gamma> \<A> \<eta> (sprotf, \<sigma>)); wf_history1 \<eta> (pnonce \<sigma>) (history \<sigma>) \<rbrakk>
  \<Longrightarrow> wf_history1 \<eta> (pnonce \<sigma>') (history \<sigma>')"
by(cases "(sprotf, \<sigma>)" rule: hybrid_step1.cases)(auto split: option.split_asm if_split_asm dest: wf_history1_rec_message wf_history1_gen_message wf_history1_sim_control simp add: gpv.set_map Some_eq_bind_conv)

lemma wf_history1_sim_info_sim:
  "\<lbrakk> Some \<sigma>' \<in> results'_gpv (sim_info_sim \<Gamma> \<A> \<eta> \<sigma> ident to); wf_history1 \<eta> (pnonce \<sigma>) (history \<sigma>) \<rbrakk>
  \<Longrightarrow> wf_history1 \<eta> (pnonce \<sigma>') (history \<sigma>')"
by(cases to)(auto split: if_split_asm simp add: Let_def split: t.split_asm nonce.split_asm option.split_asm)

lemma wf_history1_hybrid_step:
  "\<lbrakk> Some (sprotf', \<sigma>') \<in> results'_gpv (hybrid_step_sim1 \<Gamma> \<A> \<eta> (sprotf, \<sigma>)); wf_history1 \<eta> (pnonce \<sigma>) (history \<sigma>) \<rbrakk>
  \<Longrightarrow> wf_history1 \<eta> (pnonce \<sigma>') (history \<sigma>')"
by(auto simp add: hybrid_step_def split: option.split_asm elim!: wf_history1_sim_info_sim wf_history1_hybrid_step1)

end

end