theory CoSP_PKES_Simulator2 imports
  CoSP_PKES_Key_Safety_Simulator
begin

subsubsection \<open>Second simulator\<close>

context comp_impl_sims begin

definition gen_ekey :: "protocol_nonce \<Rightarrow> (ekey, call, ret) gpv"
where "gen_ekey n = Pause (Inl (n, Inl ())) (Done \<circ> projl \<circ> projl)"

definition gen_vkey :: "protocol_nonce \<Rightarrow> (vkey, call, ret) gpv"
where "gen_vkey n = Pause (Inr (n, Inl ())) (Done \<circ> projl \<circ> projr)"

definition gen_sig :: "protocol_nonce \<Rightarrow> message \<Rightarrow> (signature, call, ret) gpv"
where "gen_sig n m = Pause (Inr (n, Inr m)) (Done \<circ> projr \<circ> projr)"

definition gen_decrypt :: "protocol_nonce \<Rightarrow> cipher \<Rightarrow> (plain option, call, ret) resumption"
where
  "gen_decrypt n cipher = 
   resumption.Pause (Inl (n, (Inr (Inl cipher)))) 
     (\<lambda>input. DONE (case input of Inl (Inr (Inl (Some plain))) \<Rightarrow> Some plain | _ \<Rightarrow> None))"

definition encryption_key_sim2
  :: "security \<Rightarrow> protocol_nonce \<Rightarrow> ('more simulator_state'_scheme, (ekey \<times> 'more simulator_state'_scheme, call, ret) gpv) stateT"
where
  "encryption_key_sim2 \<eta> n = StateT (\<lambda>s. do {
     m \<leftarrow> gen_ekey n;
     Done (m, s)
   })"

definition gen_encrypt :: "protocol_nonce \<Rightarrow> plain \<Rightarrow> plain \<Rightarrow> (cipher, call, ret) gpv"
where
  "gen_encrypt n plain1 plain2 =
   Pause (Inl (n, (Inr (Inr (plain1, plain2)))))
     (\<lambda>input. case input of Inl (Inr (Inr cipher)) \<Rightarrow> Done cipher | _ \<Rightarrow> Fail)"

text \<open>The second simulator sends two challenge plaintexts to the encryption oracle:
  the original message and a sequence of zeros. The oracle then consistenly picks one of them for the encryption.
  If the oracle always uses the first message (i.e., the real one), then we get something equivalent
  to the original simulator, and if it always uses the second message, we get something equivalent
  to the faking simulator.\<close>

definition encryption_sim2
  :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> bitstring \<Rightarrow> protocol_nonce \<Rightarrow>
  ('more simulator_state'_scheme, (bitstring \<times> 'more simulator_state'_scheme, call, ret) gpv) stateT"
where
  "encryption_sim2 \<eta> n t m n' = StateT (\<lambda>s.
   (case history s (E (EK (N (PNonce n))) t (N (PNonce n'))) of Some m \<Rightarrow> Done (m, s)
    | _ \<Rightarrow> do {
       (ek, s') \<leftarrow> run_state (encryption_key_sim2 \<eta> n) s;
       let m' = replicate (length m) False; (* We could compute length m from t by length regularity *)
       cipher \<leftarrow> gen_encrypt n m m';
       Done (cipher, s')
   }))"

definition verification_key_sim2
  :: "security \<Rightarrow> protocol_nonce
  \<Rightarrow> ('more simulator_state'_scheme, (vkey \<times> 'more simulator_state'_scheme, call, ret) gpv) stateT"
where
  "verification_key_sim2 \<eta> n = StateT (\<lambda>s. do {
     m \<leftarrow> gen_vkey n;
     Done (m, s)
  })"

definition signature_sim2
  :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> message \<Rightarrow> protocol_nonce
  \<Rightarrow> ('more simulator_state'_scheme, (signature \<times> 'more simulator_state'_scheme, call, ret) gpv) stateT"
where
  "signature_sim2 \<eta> n t m n' = StateT (\<lambda>s.
  (case history s (SIG (SK (N (PNonce n))) t (N (PNonce n'))) of Some m' \<Rightarrow> Done (m', s)
   | _ \<Rightarrow> do {
     (* We initialise the oracle despite \<beta> triggering the key generation before such
        that sim1 and sim2 behave similar *)
     (vk, s') \<leftarrow> run_state (verification_key_sim2 \<eta> n) s;
     sig \<leftarrow> gen_sig n m;
     Done (sig, s')
   }))"

abbreviation (input) decryption_sim2 
  :: "security \<Rightarrow> protocol_nonce \<Rightarrow> cipher \<Rightarrow> 'more simulator_state'_scheme
  \<Rightarrow> (plain option, call, ret) resumption"
where "decryption_sim2 \<eta> n c s \<equiv> gen_decrypt n c"

(* lemma decryption_sim2_shorter:
  "Some m' \<in> results (decryption_sim2 \<eta> n m s) \<Longrightarrow> length m' < length m"
by(clarsimp simp add: gen_decrypt_def)(simp split: sum.split_asm option.split_asm split_if_asm)
*)

sublocale sim2: comp_impl_sim_gen
  A pnonce_sim record_sim 
  encryption_key_sim2 encryption_sim2
  verification_key_sim2 signature_sim2
  decryption_sim2
by(unfold_locales)

abbreviation sim2 :: "nonce_typing \<Rightarrow> 's adversary \<Rightarrow> (security \<Rightarrow> nat)
  \<Rightarrow> ('s, 'id) cosp_simulator"
where "sim2 \<equiv> sim2.sim ()"

abbreviation sim2_initial :: "'s \<Rightarrow> nat \<Rightarrow> 's simulator_state"
where "sim2_initial \<equiv> sim2.initial ()"

abbreviation sim2' :: "nonce_typing \<Rightarrow> 's adversary' \<Rightarrow> security \<Rightarrow> ('s, 'id) cosp_simulator'"
where "sim2' \<equiv> sim2.sim'"

abbreviation sim2_\<beta> :: "security \<Rightarrow> t \<Rightarrow>
  (('s, 'more') simulator_state_scheme, (bitstring \<times> ('s, 'more') simulator_state_scheme, call, comp_impl.ret) gpv) stateT"
 where "sim2_\<beta> \<equiv> sim2.\<beta>"

abbreviation sim2_\<tau> :: "security \<Rightarrow> ('s, 'more') simulator_state_scheme \<Rightarrow> bool list
  \<Rightarrow> (t, comp_impl.call, comp_impl.ret) resumption"
where "sim2_\<tau> \<equiv> sim2.\<tau> TYPE('s) TYPE('more')"

lemma fuel_encryption_key_sim2:
  "(m, s') \<in> results'_gpv (run_state (encryption_key_sim2 \<eta> n) s) \<Longrightarrow> fuel s' = fuel s"
by(auto simp add: encryption_key_sim2_def)

lemma fuel_encryption_sim2:
  "(bs, s') \<in> results'_gpv (run_state (encryption_sim2 \<eta> n' t m n) s) \<Longrightarrow> fuel s' = fuel s"
by(auto simp add: encryption_sim2_def dest!: fuel_encryption_key_sim2 split: option.split_asm)

lemma fuel_verification_key_sim2:
  "(m, s') \<in> results'_gpv (run_state (verification_key_sim2 \<eta> n) s) \<Longrightarrow> fuel s' = fuel s"
by(auto simp add: verification_key_sim2_def)

lemma fuel_signature_sim2:
  "(bs, s') \<in> results'_gpv (run_state (signature_sim2 \<eta> n' t m n) s) \<Longrightarrow> fuel s' = fuel s"
by(auto simp add: signature_sim2_def dest!: fuel_verification_key_sim2 split: if_split_asm option.split_asm)

lemma fuel_\<beta>2_same: "(bs, s') \<in> results'_gpv (run_state (sim2.\<beta> \<eta> t) s) \<Longrightarrow> fuel s' = fuel s"
by(erule sim2.fuel_\<beta>_same)(auto dest: fuel_pnonce_sim fuel_encryption_key_sim2 fuel_encryption_sim2 fuel_verification_key_sim2 fuel_signature_sim2 split: if_split_asm)

lemma relation_for_sim'2: "relation_for_sim' (sim2' \<Gamma> \<A> \<eta>) (Wellfounded.measure fuel)"
proof
  fix s s' s'' t to and ident :: 'id and x y
  assume *: "Some s'' \<in> results'_gpv (info_sim (sim2' \<Gamma> \<A> \<eta>) s' ident to)"
  assume "Some (t, s') \<in> results'_gpv (gen_message (sim2' \<Gamma> \<A> \<eta>) s)
    \<or> Some s' \<in> results'_gpv (rec_message (sim2' \<Gamma> \<A> \<eta>) s t)
    \<or> Some (y, s') \<in> results'_gpv (control (sim2' \<Gamma> \<A> \<eta>) s x)
    \<or> s = s'"
    (is "?gen_message \<or> ?rec_message \<or> ?control \<or> ?same")
  hence "fuel s' = fuel s"
    by(auto simp add: sim2.gen_message_def sim2.rec_message_def sim_control_def dest!: fuel_\<beta>2_same split: option.split_asm)
  thus "(s'', s) \<in> Wellfounded.measure fuel" using *
    by(cases to)(auto simp add: Let_def split: if_split_asm t.split_asm nonce.split_asm option.split_asm node_type.split_asm)
qed

lemma simulator'_sim2: "simulator' (sim2' \<Gamma> \<A> \<eta>) (Wellfounded.measure fuel)"
by unfold_locales(simp_all add: relation_for_sim'2)

sublocale sim2: simulator' \<C> N \<D> \<T> deduce "sim2' \<Gamma> \<A> \<eta>" "Wellfounded.measure fuel" for \<Gamma> \<A> \<eta> R
by(fact simulator'_sim2)

inductive is_decrypt :: "comp_impl.call \<Rightarrow> bool"
where "is_decrypt (Inl (n, Inr (Inl x)))"

inductive_simps is_decrypt_Inl [simp]: "is_decrypt (Inl x)"
inductive_simps is_decrypt_Inr [simp]: "is_decrypt (Inr x)"

lemma outputs_gen_decrypt: "query \<in> outputs (gen_decrypt k bs) \<Longrightarrow> is_decrypt query"
by(simp add: gen_decrypt_def)

lemma outputs_gen_decrypt': "outputs (gen_decrypt k bs) \<subseteq> {query. is_decrypt query}"
by(auto dest: outputs_gen_decrypt)

lemma outputs_\<tau>2_decrypt:
  assumes "query \<in> outputs (sim2_\<tau> \<eta> s bs)"
  shows "is_decrypt query"
proof -
  have "outputs (sim2_\<tau> \<eta> s bs) \<subseteq> {query. is_decrypt query}"
  proof(induction arbitrary: bs rule: sim2.\<tau>_induct)
    note [cont_intro] = admissible_leI[OF complete_lattice_ccpo']
    case adm thus ?case by simp
  next
    case bot thus ?case by simp
  next
    case (step \<tau>)
    note [split] = if_split_asm t.split nonce.split option.split
      and [simp] = UN_subset_iff  
      and [rule del] = subsetI
    show ?case
    proof(cases \<eta> bs rule: \<tau>_cases, simp_all split del: if_split, goal_cases)
      { case Nonce: 1      show ?case by simp }
      { case Cipher: 2     show ?case using step.IH by(simp add: outputs_gen_decrypt') }
      { case Ekey: 3       show ?case by simp }
      { case Signature: 4  show ?case using step.IH by simp }
      { case Vkey: 5       show ?case by simp }
      { case Pair: 6       show ?case using step.IH by simp }
    qed
  qed
  with assms show ?thesis by blast
qed

lemma \<tau>2_only_decrypt_queries: "pred_resumption (\<lambda>_. True) is_decrypt (sim2_\<tau> \<eta> s2 bs)"
by(auto simp add: pred_resumption_def dest: outputs_\<tau>2_decrypt)

abbreviation hybrid_trace2
  :: "'id cosp_protocol \<Rightarrow> bool \<Rightarrow> nonce_typing \<Rightarrow> 's adversary \<Rightarrow> bound
  \<Rightarrow> security \<Rightarrow> 'id list spmf"
where "hybrid_trace2 prot b \<Gamma> \<A> bound \<equiv> hybrid_trace prot (oracle b) (sim2 \<Gamma> \<A> bound)"

abbreviation hybrid_step1_sim2
  :: "nonce_typing \<Rightarrow> 's adversary' \<Rightarrow> security \<Rightarrow> ('id, 's) cosp_hybrid_state
  \<Rightarrow> ((t option \<times> ('id, 's) cosp_hybrid_state) option, call, comp_impl.ret) gpv"
where "hybrid_step1_sim2 \<Gamma> \<A> bound \<equiv> hybrid_step1 TYPE(_) (sim2' \<Gamma> \<A> bound)"

abbreviation hybrid_step_sim2
  :: "nonce_typing \<Rightarrow> 's adversary' \<Rightarrow> security \<Rightarrow> ('id, 's) cosp_hybrid_state 
  \<Rightarrow> (('id, 's) cosp_hybrid_state option, call, ret) gpv"
where "hybrid_step_sim2 \<Gamma> \<A> bound \<equiv> hybrid_step TYPE(_) (sim2' \<Gamma> \<A> bound)"

end

end