theory CoSP_PKES_Simulator_Eq imports
  CoSP_PKES_Simulator1
  CoSP_PKES_Simulator2
  CoSP_PKES_Key_Safety_Simulator
begin

lemma bind_pmf_spmf_of_set: (* Move to SPMF *)
  "bind_pmf (spmf_of_set A) f = (if finite A \<and> A \<noteq> {} then bind_pmf (pmf_of_set A) (f \<circ> Some) else f None)"
  by(simp add: spmf_of_set_def spmf_of_pmf_def bind_map_pmf o_def bind_return_pmf del: spmf_of_pmf_pmf_of_set)

lemma catch_gpv_bind_option:
  "catch_gpv (monad.bind_option Fail x f) = (case x of None \<Rightarrow> Done None | Some x' \<Rightarrow> catch_gpv (f x'))"
  by(cases x) simp_all
  

locale comp_impl_sim12 = comp_impl_sims A
  for A :: implementation
begin

context fixes \<Gamma> :: nonce_typing and \<eta> :: security begin 

definition pnonce_enc :: "pnonce \<Rightarrow> (protocol_nonce \<Rightarrow> (bitstring, bitstring, bitstring) ind_cca2.state_oracle) \<Rightarrow> bool"
where "pnonce_enc pns ose = 
  (\<forall>n. \<Gamma> n = Some Encryption \<longrightarrow>
   rel_option (\<lambda>bs (ek, dk, ciphers). A_ek A \<eta> bs = Some ek \<and> A_dk A \<eta> bs = Some dk) (pns n) (ose n))"

lemma pnonce_encD: 
  "\<lbrakk> pnonce_enc pns ose; \<Gamma> n = Some Encryption \<rbrakk> \<Longrightarrow> 
  rel_option (\<lambda>bs (ek, dk, ciphers). A_ek A \<eta> bs = Some ek \<and> A_dk A \<eta> bs = Some dk) (pns n) (ose n)"
by(simp add: pnonce_enc_def)

lemma pnonce_enc_updI:
  "\<lbrakk> pnonce_enc pns ose; A_ek A \<eta> bs = Some ek; A_dk A \<eta> bs = Some dk \<rbrakk>
  \<Longrightarrow> pnonce_enc (pns(k \<mapsto> bs)) (ose(k \<mapsto> (ek, dk, ciphers)))"
by(simp add: pnonce_enc_def)

lemma pnonce_enc_upd_other: 
  "\<lbrakk> pnonce_enc pns ose; \<Gamma> n \<noteq> Some Encryption \<rbrakk> \<Longrightarrow> pnonce_enc (pns(n \<mapsto> bs)) ose"
by(auto simp add: pnonce_enc_def)

lemma pnonce_enc_upd_ciphers: 
  "\<lbrakk> pnonce_enc pns ose; \<exists>ciphers. ose n = Some (ek, dk, ciphers) \<rbrakk>
  \<Longrightarrow> pnonce_enc pns (ose(n \<mapsto> (ek, dk, ciphers')))"
unfolding pnonce_enc_def by(auto simp add: option_rel_Some2)

definition pnonce_sig :: "pnonce \<Rightarrow> (protocol_nonce \<Rightarrow> (bitstring \<times> bitstring \<times> (bitstring \<times> bitstring) list) option) \<Rightarrow> bool"
where "pnonce_sig pns oss =
  (\<forall>n. \<Gamma> n = Some Signature \<longrightarrow> 
    rel_option (\<lambda>bs (vk, sk, sigs). A_vk A \<eta> bs = Some vk \<and> A_sk A \<eta> bs = Some sk) (pns n) (oss n))"

lemma pnonce_sigI: 
  "(\<And>n. \<Gamma> n = Some Signature
    \<Longrightarrow> rel_option (\<lambda>bs (vk, sk, sigs). A_vk A \<eta> bs = Some vk \<and> A_sk A \<eta> bs = Some sk) (pns n) (oss n))
  \<Longrightarrow> pnonce_sig pns oss"
by(auto simp add: pnonce_sig_def)

lemma pnonce_sigD:
  "\<lbrakk> pnonce_sig pns oss; \<Gamma> n = Some Signature \<rbrakk>
   \<Longrightarrow> rel_option (\<lambda>bs (vk, sk, sigs). A_vk A \<eta> bs = Some vk \<and> A_sk A \<eta> bs = Some sk) (pns n) (oss n)"
by(simp add: pnonce_sig_def)

lemma pnonce_sig_updI:
  "\<lbrakk> pnonce_sig pns oss; A_vk A \<eta> bs = Some vk; A_sk A \<eta> bs = Some sk \<rbrakk>
  \<Longrightarrow> pnonce_sig (pns(k \<mapsto> bs)) (oss(k \<mapsto> (vk, sk, sigs)))"
by(simp add: pnonce_sig_def)

lemma pnonce_sig_upd_other: 
  "\<lbrakk> pnonce_sig pns oss; \<Gamma> n \<noteq> Some Signature \<rbrakk> \<Longrightarrow> pnonce_sig (pns(n \<mapsto> bs)) oss"
by(auto simp add: pnonce_sig_def)

lemma pnonce_sig_upd_sigs: 
  "\<lbrakk> pnonce_sig pns oss; oss k = Some (vk, sk, sigs) \<rbrakk> \<Longrightarrow> pnonce_sig pns (oss(k \<mapsto> (vk, sk, sigs')))"
by(auto intro!: pnonce_sigI dest!: pnonce_sigD simp add: option_rel_Some2)

definition rand_sim :: "random_usage \<Rightarrow> pnonce \<Rightarrow> pnonce \<Rightarrow> bool" where
  "rand_sim ru pns1 pns2 =
   (\<forall>n bl. \<Gamma> n = Some (Randomness bl)
   \<longrightarrow> (\<forall>k t. RUEncryption (EK (N (PNonce k))) t \<notin> ru n)
   \<longrightarrow> (\<forall>k t. RUSignature (SK (N (PNonce k))) t \<notin> ru n)
   \<longrightarrow> pns1 n = pns2 n)"

lemma rand_simD:
  "\<lbrakk> rand_sim ru pns1 pns2; \<Gamma> n = Some (Randomness bl);
     \<forall>k t. RUEncryption (EK (N (PNonce k))) t \<notin> ru n;
     \<forall>k t. RUSignature (SK (N (PNonce k))) t \<notin> ru n \<rbrakk>
  \<Longrightarrow> pns1 n = pns2 n"
by(auto simp add: rand_sim_def)

lemma rand_simI [intro?]:
   "(\<And>n bl. \<lbrakk> \<Gamma> n = Some (Randomness bl); \<And>k t. RUEncryption (EK (N (PNonce k))) t \<notin> ru n;
        \<And>k t. RUSignature (SK (N (PNonce k))) t \<notin> ru n \<rbrakk> \<Longrightarrow> pns1 n = pns2 n)
   \<Longrightarrow> rand_sim ru pns1 pns2"
by(auto simp add: rand_sim_def)

lemma rand_sim_upd1_other: 
  "\<lbrakk> rand_sim ru pns1 pns2;
     \<forall>bl. \<Gamma> n = Some (Randomness bl) \<longrightarrow> (\<exists>k t. RUEncryption (EK (N (PNonce k))) t \<in> ru n) \<or> (\<exists>k t. RUSignature (SK (N (PNonce k))) t \<in> ru n) \<rbrakk>
  \<Longrightarrow> rand_sim ru (pns1(n \<mapsto> bs)) pns2"
by(auto intro!: rand_simI dest: rand_simD)

lemma rand_sim_updI: "rand_sim ru pns1 pns2 \<Longrightarrow> rand_sim ru (pns1(n \<mapsto> bs)) (pns2(n \<mapsto> bs))"
by(simp add: rand_sim_def)

lemma rand_sim_mono: "\<lbrakk> rand_sim ru pns1 pns2; ru \<le> ru' \<rbrakk> \<Longrightarrow> rand_sim ru' pns1 pns2"
by(intro rand_simI)(blast dest: rand_simD le_funD)

definition pnonce_public :: "pnonce \<Rightarrow> pnonce \<Rightarrow> bool"
where "pnonce_public pns1 pns2 = (\<forall>n. \<Gamma> n = None \<longrightarrow> pns1 n = pns2 n)"

lemma pnonce_publicI: "(\<And>n. \<Gamma> n = None \<Longrightarrow> pns1 n = pns2 n) \<Longrightarrow> pnonce_public pns1 pns2"
by(auto simp add: pnonce_public_def)

lemma pnonce_publicD: "\<lbrakk> pnonce_public pns1 pns2; \<Gamma> n = None \<rbrakk> \<Longrightarrow> pns1 n = pns2 n"
by(auto simp add: pnonce_public_def)

lemma pnonce_public_updI: "pnonce_public pns1 pns2 \<Longrightarrow> pnonce_public (pns1(n := bs)) (pns2(n := bs))"
by(simp add: pnonce_public_def)

lemma assumes "pnonce_public pns1 pns2" "\<Gamma> n \<noteq> None"
  shows pnonce_public_upd1: "pnonce_public (pns1(n := bs)) pns2"
  and pnonce_public_upd2: "pnonce_public pns1 (pns2(n := bs))"
using assms by(auto simp add: pnonce_public_def)

definition pnonce_dom :: "pnonce \<Rightarrow> pnonce \<Rightarrow> bool"
where "pnonce_dom pns1 pns2 \<longleftrightarrow> dom pns2 \<subseteq> dom pns1"

lemma pnonce_domI: "dom pns2 \<subseteq> dom pns1 \<Longrightarrow> pnonce_dom pns1 pns2"
by(simp add: pnonce_dom_def)

lemma pnonce_domD: "pnonce_dom pns1 pns2 \<Longrightarrow> dom pns2 \<subseteq> dom pns1"
by(simp add: pnonce_dom_def)

lemma pnonce_dom_updI: "pnonce_dom pns1 pns2 \<Longrightarrow> pnonce_dom (pns1(n \<mapsto> bs)) (pns2(n \<mapsto> bs'))"
by(auto simp add: pnonce_dom_def)

lemma pnonce_dom_upd1: "pnonce_dom pns1 pns2 \<Longrightarrow> pnonce_dom (pns1(n \<mapsto> bs)) pns2"
by(auto simp add: pnonce_dom_def)

definition history_ciphers :: "history \<Rightarrow> (protocol_nonce \<Rightarrow> (bitstring, bitstring, bitstring) ind_cca2.state_oracle) \<Rightarrow> bool"
where "history_ciphers hist1 os2e \<longleftrightarrow> ind_cca2.cipher_queries os2e \<subseteq> ran hist1"

lemma history_ciphersI: "ind_cca2.cipher_queries os2e \<subseteq> ran hist \<Longrightarrow> history_ciphers hist os2e"
by(simp add: history_ciphers_def)

lemma history_ciphersD: "history_ciphers hist os2e \<Longrightarrow> ind_cca2.cipher_queries os2e \<subseteq> ran hist"
by(simp add: history_ciphers_def)

lemma history_ciphers_upd1: 
  "\<lbrakk> history_ciphers hist os2e; wf_history1 \<eta> pns (hist(t \<mapsto> bs)); wf_history1 \<eta> pns hist \<rbrakk>
  \<Longrightarrow> history_ciphers (hist(t \<mapsto> bs)) os2e"
apply(auto intro!: history_ciphersI dest!: history_ciphersD subsetD simp add: ran_def)
subgoal for \<dots> t' by(cases "t' = t")(auto dest!: wf_history1D_\<beta>)
done

lemma history_ciphers_upd2: 
  "\<lbrakk> history_ciphers hist os2e; set ciphers \<subseteq> ran hist \<union> ind_cca2.cipher_queries os2e \<rbrakk>
  \<Longrightarrow> history_ciphers hist (os2e(n \<mapsto> (ek, dk, ciphers)))"
by(auto intro!: history_ciphersI dest: history_ciphersD elim!: ind_cca2.cipher_queries_updE)

lemma history_ciphers_add: 
  "\<lbrakk> history_ciphers hist os2e; t \<notin> dom hist; os2e n = Some (ek', dk', ciphers) \<rbrakk>
  \<Longrightarrow> history_ciphers (hist(t \<mapsto> bs)) (os2e(n \<mapsto> (ek, dk, bs # ciphers)))"
by(rule history_ciphersI)(auto 4 3 dest!: history_ciphersD elim!: ind_cca2.cipher_queries_updE intro: ind_cca2.cipher_queriesI)

definition invar_sim1 :: "('id, 's) cosp_hybrid_state \<times> state_oracle \<Rightarrow> bool"
where
  "invar_sim1 = pred_prod (\<lambda>((S1, prot1, f1), s1).
      key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f1) prot1 \<and>
      key_safe_sim1_invar \<eta> \<Gamma> f1 s1 \<and>
      wf_random_usage \<Gamma> f1 \<and>
      wf_history1 \<eta> (pnonce s1) (history s1)
    ) (\<lambda>_. True)"

lemma invar_sim1I [intro?]: 
  "\<lbrakk> key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f1) prot1;
     key_safe_sim1_invar \<eta> \<Gamma> f1 s1;
     wf_random_usage \<Gamma> f1;
     wf_history1 \<eta> (pnonce s1) (history s1) \<rbrakk>
   \<Longrightarrow> invar_sim1 (((S1, prot1, f1), s1), (os1e, os1s))"
by(simp add: invar_sim1_def)

lemma invar_sim1D:
  assumes "invar_sim1 (((S1, prot1, f1), s1), (os1e, os1s))"
  shows invar_sim1_ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f1) prot1"
  and invar_sim1_ks1: "key_safe_sim1_invar \<eta> \<Gamma> f1 s1"
  and invar_sim1_ru: "wf_random_usage \<Gamma> f1"
  and invar_sim1_wfh: "wf_history1 \<eta> (pnonce s1) (history s1)"
using assms by(auto simp add: invar_sim1_def)

definition bisim12
  :: "('id, 's) cosp_hybrid_state \<times> state_oracle \<Rightarrow> ('id, 's) cosp_hybrid_state \<times> state_oracle \<Rightarrow> bool"
where "bisim12 = (\<lambda>(((S1, prot1, f1), s1), (os1e, os1s)) (((S2, prot2, f2), s2), (os2e, os2s)).
    prot1 = prot2 \<and> S1 = S2 \<and> f1 = f2 \<and> adv_state s1 = adv_state s2 \<and>
    history s1 = history s2 \<and> fuel s1 = fuel s2 \<and> randomness s1 = randomness s2 \<and>
    pnonce_enc (pnonce s1) os2e \<and>
    pnonce_sig (pnonce s1) os2s \<and>
    rand_sim (random_usages f1) (pnonce s1) (pnonce s2) \<and>
    pnonce_public (pnonce s1) (pnonce s2) \<and>
    pnonce_dom (pnonce s1) (pnonce s2) \<and>
    history_ciphers (history s1) os2e
  )"

lemma bisim12I [intro?]:
  "\<lbrakk> prot1 = prot2; S1 = S2; f1 = f2; adv_state s1 = adv_state s2; history s1 = history s2;
     fuel s1 = fuel s2; randomness s1 = randomness s2;
     pnonce_enc (pnonce s1) os2e; pnonce_sig (pnonce s1) os2s;
     rand_sim (random_usages f1) (pnonce s1) (pnonce s2);
     pnonce_public (pnonce s1) (pnonce s2); pnonce_dom (pnonce s1) (pnonce s2);
     history_ciphers (history s1) os2e \<rbrakk>
  \<Longrightarrow> bisim12 (((S1, prot1, f1), s1), (os1e, os1s)) (((S2, prot2, f2), s2), (os2e, os2s))"
by(simp add: bisim12_def)

lemma bisim12D:
  assumes "bisim12 (((S1, prot1, f1), s1), (os1e, os1s)) (((S2, prot2, f2), s2), (os2e, os2s))"
  shows bisim12_pnonce_enc: "pnonce_enc (pnonce s1) os2e"
  and bisim12_pnonce_sig: "pnonce_sig (pnonce s1) os2s"
  and bisim12_rand_sim: "rand_sim (random_usages f1) (pnonce s1) (pnonce s2)"
  and bisim12_pnonce_public: "pnonce_public (pnonce s1) (pnonce s2)"
  and bisim12_pnonce_dom: "pnonce_dom (pnonce s1) (pnonce s2)"
  and bisim12_history_ciphers: "history_ciphers (history s1) os2e"
using assms by(auto simp add: bisim12_def)

thm iterate_spmf_parametric[THEN rel_funD, THEN rel_funD]

lemma bisim_run_hybrid_trace_sim12:
  fixes prot :: "'id cosp_protocol" and \<A>' :: "'s adversary'"
  assumes "\<Gamma>, [] \<turnstile> prot \<surd>"
  shows "rel_spmf (list_all2 (bisim12 \<upharpoonleft> invar_sim1 \<otimes> (\<lambda>_. True)))
    (SMC_iterate_spmf (oracle\<^sub>n' b \<eta>) (sim1' \<Gamma> \<A>' \<eta>) ((({}, prot, []), sim2.initial () \<sigma> bound), (\<lambda>_. None, \<lambda>_. None)))
    (SMC_iterate_spmf (oracle\<^sub>n' True \<eta>) (sim2' \<Gamma> \<A>' \<eta>) ((({}, prot, []), sim2.initial () \<sigma> bound), (\<lambda>_. None, \<lambda>_. None)))"
    (is "rel_spmf (list_all2 ?Q) (SMC_iterate_spmf _ _ ?init1) (SMC_iterate_spmf _ _ ?init2)")
proof -
  define bisim12' :: "('id, 's) cosp_hybrid_state option \<times> state_oracle \<Rightarrow> ('id, 's) cosp_hybrid_state option \<times> state_oracle \<Rightarrow> bool"
    where "bisim12' \<equiv> \<lambda>(state1, os1) (state2, os2). case (state1, state2) of (None, None) \<Rightarrow> True | (Some s1, Some s2) \<Rightarrow> bisim12 (s1, os1) (s2, os2) | _ \<Rightarrow> False"
  have bisim12' [simp]:
    "bisim12' (Some s1, os1) (Some s2, os2) \<longleftrightarrow> bisim12 (s1, os1) (s2, os2)"
    "bisim12' (None, os1) (None, os2) \<longleftrightarrow> True"
    for s1 os1 s2 os2 by(simp_all add: bisim12'_def)
  
  define invar_sim1' :: "('id, 's) cosp_hybrid_state option \<times> state_oracle \<Rightarrow> bool" where
    "invar_sim1' \<equiv> \<lambda>(state, os). case state of None \<Rightarrow> True | Some state \<Rightarrow> invar_sim1 (state, os)"
  have invar_sim1' [simp]:
    "invar_sim1' (None, os) \<longleftrightarrow> True"
    "invar_sim1' (Some s, os) \<longleftrightarrow> invar_sim1 (s, os)"
    for s os by(simp_all add: invar_sim1'_def)
  {
    fix hs1 os1 hs2 os2
    assume bisim12: "?Q (hs1, os1) (hs2, os2)"
    let ?Q' = "\<lambda>((to1, hs1), os1) ((to2, hs2), os2). to1 = to2 \<and> ?Q (hs1, os1) (hs2, os2)"

    obtain S1 prot1 f1 s1 where hs1 [simp]: "hs1 = ((S1, prot1, f1), s1)" by(cases hs1) auto
    obtain S2 prot2 f2 s2 where hs2 [simp]: "hs2 = ((S2, prot2, f2), s2)" by(cases hs2) auto
    obtain os1e os1s where os1 [simp]: "os1 = (os1e, os1s)" by(cases os1)
    obtain os2e os2s where os2 [simp]: "os2 = (os2e, os2s)" by(cases os2)

    note [iff del] = domIff

    from bisim12 have prot [simp]: "prot2 = prot1" and S [simp]: "S2 = S1" and f [simp]: "f2 = f1" 
      and adv_state [simp]: "adv_state s2 = adv_state s1" 
      and history [simp]: "history s2 = history s1"
      and fuel [simp]: "fuel s1 = fuel s2"
      and randomness: "randomness s1 = randomness s2"
      by(simp_all add: bisim12_def)
    from bisim12 have pnonce_enc: "pnonce_enc (pnonce s1) os2e" by(clarsimp simp add: bisim12_pnonce_enc)
    from bisim12 have pnonce_sig: "pnonce_sig (pnonce s1) os2s" by(clarsimp simp add: bisim12_pnonce_sig)
    from bisim12 have rand_sim: "rand_sim (random_usages f1) (pnonce s1) (pnonce s2)" by(clarsimp simp add: bisim12_rand_sim)
    from bisim12 have pnonce_public: "pnonce_public (pnonce s1) (pnonce s2)" by(clarsimp simp add: bisim12_pnonce_public)
    from bisim12 have pnonce_dom: "pnonce_dom (pnonce s1) (pnonce s2)" by(clarsimp simp add: bisim12_pnonce_dom)
    from bisim12 have history_ciphers: "history_ciphers (history s1) os2e" by(clarsimp simp add: bisim12_history_ciphers)
    from bisim12 have ks: "key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f1) prot1" by(clarsimp simp add: invar_sim1_ks)
    from bisim12 have ks1: "key_safe_sim1_invar \<eta> \<Gamma> f1 s1" by(auto simp add: invar_sim1_ks1)
    hence kssim: "key_safe_sim {} \<eta> \<Gamma> s1" and kstr: "key_safe_trace A \<eta> \<Gamma> f1"
      and anonces: "\<forall>t\<in>set f1. \<beta>1_anonces t \<inter> dom \<Gamma> \<subseteq> randomness s1" 
      and rsr: "randomness_sampled_recorded \<Gamma> s1"
      and wrh: "wf_rand_history (random_usages f1) (history s1)"
      and kst: "key_safe_trace A \<eta> \<Gamma> f1"
      by(auto elim!: key_safe_sim1_invar.cases)
    from bisim12 have ru: "wf_random_usage \<Gamma> f1" by(clarsimp simp add: invar_sim1_ru)
    from bisim12 have wfh: "wf_history1 \<eta> (pnonce s1) (history s1)" by(clarsimp simp add: invar_sim1_wfh)

    have "rel_spmf bisim12'
      (exec_gpv (oracle\<^sub>n' b \<eta>) (hybrid_step_sim1 \<Gamma> \<A>' \<eta> hs1) os1)
      (exec_gpv (oracle\<^sub>n' True \<eta>) (hybrid_step_sim2 \<Gamma> \<A>' \<eta> hs2) os2)"
    proof(cases prot1)
      case (Output id out succ)
      def t \<equiv> "f1 ! out"
      def t' \<equiv> t
      with Output ks have out: "out < length f1" and nt: "node_type_of_t \<Gamma> t = PublicT"
        and t': "t' \<in> set f1" by(auto simp add: key_safe_simps t_def)
      with kstr have kst: "A,\<eta>,\<Gamma> \<turnstile>t t : PublicT" by(auto dest: key_safe_traceD simp add: t_def)
      have tt': "\<beta>_rec t' t" by(simp add: t'_def)

      def ru \<equiv> "random_usages f1"
      define I1' :: "(bitstring \<times> 's simulator_state) \<times> state_oracle \<Rightarrow> bool" where
        "I1' = pred_prod (pred_prod (\<lambda>_. True) (\<lambda>s1.
          randomness_sampled_recorded \<Gamma> s1 \<and>
          wf_rand_history ru (history s1) \<and>
          wf_history1 \<eta> (pnonce s1) (history s1)
        )) (\<lambda>_. True)"
      have I1'I: "I1' ((m1, s1), (os1e, os1s))"
        if "randomness_sampled_recorded \<Gamma> s1"
        and "wf_rand_history ru (history s1)"
        and "wf_history1 \<eta> (pnonce s1) (history s1)"
        for m1 s1 os1e os1s using that by(simp add: I1'_def)
      have I1'D_rsr: "randomness_sampled_recorded \<Gamma> s1"
        and I1'D_wrh: "wf_rand_history ru (history s1)"
        and I1'D_wfh: "wf_history1 \<eta> (pnonce s1) (history s1)"
        if "I1' ((m1, s1), (os1e, os1s))"
        for m1 s1 os1e os1s using that by(simp_all add: I1'_def)
      have I1'_history_updateI:
        "I1' ((m1, history_update (\<lambda>hist. hist(t \<mapsto> bs)) s1), (os1e, os1s))"
        if "I1' ((m1', s1), (os1e, os1s))"
        and "run_env (pure.\<beta> \<eta> t) (pnonce s1) = Some bs"
        and "sampled1 t (pnonce s1)"
        and "random_usage t \<le> ru"
        for m1 s1 os1e os1s m1' t bs using that
        by(auto simp add: I1'_def elim: wf_rand_history_mono2 randomness_sampled_recorded_history_update_mono wf_history1_history_updateI)

      def rel \<equiv> "(\<lambda>M ((m1, s1), (os1e, os1s)) ((m2, s2), (os2e, os2s)).
        M m1 m2 \<and>
        pnonce_enc (pnonce s1) os2e \<and>
        pnonce_sig (pnonce s1) os2s \<and>
        rand_sim ru (pnonce s1) (pnonce s2) \<and>
        pnonce_public (pnonce s1) (pnonce s2) \<and>
        pnonce_dom (pnonce s1) (pnonce s2) \<and>
        history_ciphers (history s1) os2e \<and>
        history s1 = history s2 \<and> fuel s1 = fuel s2 \<and> randomness s1 = randomness s2 \<and> adv_state s1 = adv_state s2
        ) :: (bitstring \<Rightarrow> bitstring \<Rightarrow> bool) \<Rightarrow> (bitstring \<times> 's simulator_state) \<times> state_oracle \<Rightarrow> (bitstring \<times> 's simulator_state) \<times> state_oracle \<Rightarrow> bool"
      have relI: "rel M ((m1, s1), (os1e, os1s)) ((m2, s2), (os2e, os2s))"
        if "M m1 m2" "pnonce_enc (pnonce s1) os2e" "pnonce_sig (pnonce s1) os2s"
           "rand_sim ru (pnonce s1) (pnonce s2)"
           "pnonce_public (pnonce s1) (pnonce s2)"
           "pnonce_dom (pnonce s1) (pnonce s2)"
           "history_ciphers (history s1) os2e"
           "history s1 = history s2" "fuel s1 = fuel s2" "randomness s1 = randomness s2" "adv_state s1 = adv_state s2"
        for M m1 s1 os1e os1s m2 s2 os2e os2s using that by(auto simp add: rel_def)
      have relE [elim?]: thesis
        if "rel M ((m1, s1), (os1e, os1s)) ((m2, s2), (os2e, os2s))"
        and "\<lbrakk> M m1 m2; pnonce_enc (pnonce s1) os2e; pnonce_sig (pnonce s1) os2s;
               rand_sim ru (pnonce s1) (pnonce s2); pnonce_public (pnonce s1) (pnonce s2);
               pnonce_dom (pnonce s1) (pnonce s2); history_ciphers (history s1) os2e;
               history s1 = history s2; fuel s1 = fuel s2; randomness s1 = randomness s2; adv_state s1 = adv_state s2  \<rbrakk>
             \<Longrightarrow> thesis"
        for M m1 s1 os1e os1s m2 s2 os2e os2s thesis using that by(auto simp add: rel_def)
      have relD_m: "M m1 m2"
        and relD_pnonce_enc: "pnonce_enc (pnonce s1) os2e"
        and relD_pnonce_sig: "pnonce_sig (pnonce s1) os2s"
        and relD_rand_sim: "rand_sim ru (pnonce s1) (pnonce s2)"
        and relD_pnonce_public: "pnonce_public (pnonce s1) (pnonce s2)"
        and relD_pnonce_dom: "pnonce_dom (pnonce s1) (pnonce s2)"
        and relD_history_ciphers: "history_ciphers (history s1) os2e"
        and relD_history: "history s1 = history s2"
        and relD_fuel: "fuel s1 = fuel s2"
        and relD_randomness: "randomness s1 = randomness s2"
        and relD_adv_state: "adv_state s1 = adv_state s2"
        if "rel M ((m1, s1), (os1e, os1s)) ((m2, s2), (os2e, os2s))"
        for M m1 s1 os1e os1s m2 s2 os2e os2s using that by(auto simp add: rel_def)
      let ?rel = "\<lambda>M. rel M \<upharpoonleft> I1' \<otimes> (\<lambda>_. True)"
      
      define I1'' :: "(bitstring \<times> 's simulator_state) option \<times> state_oracle \<Rightarrow> bool" where
        "I1'' = pred_prod (pred_option (pred_prod (\<lambda>_. True) (\<lambda>s1.
          randomness_sampled_recorded \<Gamma> s1 \<and>
          wf_rand_history ru (history s1) \<and>
          wf_history1 \<eta> (pnonce s1) (history s1)
        ))) (\<lambda>_. True)"
      have I1'' [simp]: "I1'' (None, os)" and "I1'' (Some s, os) \<longleftrightarrow> I1' (s, os)" for s os
        by(simp_all add: I1''_def I1'_def)
      define rel' where "rel' M \<equiv> \<lambda>(l, os1) (r, os2). case (l, r) of (None, None) \<Rightarrow> True | (Some s1, Some s2) \<Rightarrow> rel M (s1, os1) (s2, os2) | _ \<Rightarrow> False" for M
      have rel'_simps [simp]: 
        "rel' M (Some s1, os1) (Some s2, os2) \<longleftrightarrow> rel M (s1, os1) (s2, os2)"
        "rel' M (None, os1) (None, os2)"
        "\<not> rel' M (Some s1, os1) (None, os2)"
        "\<not> rel' M (None, os1) (Some s2, os2)"
        for M s1 os1 s2 os2 by(simp_all add: rel'_def)
      let ?rel' = "\<lambda>M. rel' M \<upharpoonleft> I1'' \<otimes> (\<lambda>_. True)"

      have rel_history_E_NoneD: "pnonce s1 r = None"
        if rel: "rel M ((m1, s1), (os1e, os1s)) ((m2, s2), (os2e, os2s))"
        and I1': "I1' ((m1, s1), (os1e, os1s))"
        and hist: "history s2 (E (EK (N (PNonce k))) t (N (PNonce r))) = None"
        and r: "\<Gamma> r = Some (Randomness bl)"
        and \<beta>_rec: "\<beta>_rec t' (E (EK (N (PNonce k))) t (N (PNonce r)))"
        for M m1 s1 os1e os1s m2 s2 os2e os2s k t r bl
      proof(rule ccontr)
        assume "pnonce s1 r \<noteq> None"
        hence r_dom: "r \<in> dom (pnonce s1)" by(auto simp add: domIff)
        from \<beta>_rec[THEN sim1.random_usage_\<beta>_rec_mono] member_le_random_usages[OF t']
        have "single_usage r (RUEncryption (EK (N (PNonce k))) t) \<le> ru" by(auto simp add: ru_def)
        with randomness_sampled_recordedD[OF I1'D_rsr[OF I1'] r r_dom] relD_history[OF rel, symmetric]
        have "E (EK (N (PNonce k))) t (N (PNonce r)) \<in> dom (history s2)"
          by(auto 4 3 dest!: single_usage_leD dest: wf_rand_historyD[OF I1'D_wrh[OF I1']] dest: wf_random_usage'D1[OF ru] simp add: ru_def)
        with hist show False by(auto simp add: domIff)
      qed
      have rel_history_SIG_NoneD: "pnonce s1 r = None"
        if rel: "rel M ((m1, s1), (os1e, os1s)) ((m2, s2), (os2e, os2s))"
        and I1': "I1' ((m1, s1), (os1e, os1s))"
        and hist: "history s2 (SIG (SK (N (PNonce k))) t (N (PNonce r))) = None"
        and r: "\<Gamma> r = Some (Randomness bl)"
        and \<beta>_rec: "\<beta>_rec t' (SIG (SK (N (PNonce k))) t (N (PNonce r)))"
        for M m1 s1 os1e os1s m2 s2 os2e os2s k t r bl
      proof(rule ccontr)
        assume "pnonce s1 r \<noteq> None"
        hence r_dom: "r \<in> dom (pnonce s1)" by(auto simp add: domIff)
        from \<beta>_rec[THEN sim1.random_usage_\<beta>_rec_mono] member_le_random_usages[OF t']
        have "single_usage r (RUSignature (SK (N (PNonce k))) t) \<le> ru" by(auto simp add: ru_def)
        with randomness_sampled_recordedD[OF I1'D_rsr[OF I1'] r r_dom] relD_history[OF rel, symmetric]
        have "SIG (SK (N (PNonce k))) t (N (PNonce r)) \<in> dom (history s2)"
          by(auto 4 3 dest!: single_usage_leD dest: wf_rand_historyD[OF I1'D_wrh[OF I1']] dest: wf_random_usage'D1[OF ru] simp add: ru_def)
        with hist show False by(auto simp add: domIff)
      qed
      have rel_history_updateI:
        "rel M ((m1, history_update (\<lambda>hist. hist(t \<mapsto> bs)) s1), (os1e, os1s)) ((m2, history_update (\<lambda>hist. hist(t \<mapsto> bs)) s2), (os2e, os2s))"
        if "rel M' ((m1', s1), (os1e, os1s)) ((m2', s2), (os2e, os2s))"
        and "M m1 m2"
        and "wf_history1 \<eta> pns (history s1)"
        and "wf_history1 \<eta> pns (history s1(t \<mapsto> bs))"
        for M M' m1 s1 os1e os1s m2 s2 os2e os2s m1' m2' t bs pns using that
        by(auto elim!: relE intro!: relI history_ciphers_upd1)

      have I1'_encryption_key: "pred_spmf I1'' (exec_gpv (oracle\<^sub>n' b \<eta>) (catch_gpv (run_state (sim1.encryption_key \<eta> k) s1)) (os1e, os1s))"
        if I1': "I1' ((m1, s1), (os1e, os1s))"
        and k: "\<Gamma> k = Some Encryption"
        for b k s1 os1e os1s m1 using that unfolding I1''_def
        by(intro pred_spmf_exec_gpv[where C="\<lambda>_. True"])(auto 4 3 simp add: pred_gpv_def I1'_def elim!: results'_gpv_catch_gpvE sim1.wf_randomness_sampled_recordedD_encryption_key[OF randomness_sampled_recorded_assms_sim1] wf_history1_encryption_key dest: sim1.obs_history_sameD_encryption_key[OF ops_history_dom_same_sim1] cong: wf_rand_history_cong)
      have rel_encryption_key: "rel_spmf (rel' op =)
        (exec_gpv (oracle\<^sub>n' b \<eta>) (catch_gpv (run_state (sim1.encryption_key \<eta> k) s1)) (os1e, os1s))
        (exec_gpv (oracle\<^sub>n' True \<eta>) (catch_gpv (run_state (encryption_key_sim2 \<eta> k) s2)) (os2e, os2s))"
        if rel: "rel M ((m1, s1), (os1e, os1s)) ((m2, s2), (os2e, os2s))"
        and k: "\<Gamma> k = Some Encryption"
        for M b k s1 os1e os1s s2 os2e os2s m1 m2
        using pnonce_encD[OF relD_pnonce_enc[OF rel] k]
        apply cases
        subgoal 
          apply(simp add: sim1.encryption_key_def encryption_key_sim2_def bind_map_pmf gen_ekey_def exec_gpv_bind catch_gpv_bind_gpv key_gen_pke_def pnonce_sim_sel Let_def bind_map_spmf bind_spmf_spmf_of_set bind_spmf_pmf_assoc)
          apply(clarsimp simp add: catch_gpv_bind_option bind_spmf_map_pmf return_pmf_bind_option bind_pmf_spmf_of_set intro!: rel_pmf_bind_reflI rel_spmf_bind_reflI split: option.split)
          (* this is not going to work. *) oops end end end
          apply(rule rel_spmf_bindI2[rotated])
           subgoal by(auto dest: A_ek_SomeD_dk)
          apply(simp)
          apply(rule relI)
          using rel k
          apply(auto 4 4 intro: pnonce_enc_upd_other pnonce_sig_upd_other pnonce_enc_updI map_le_fun_upd2 rand_sim_upd1_other pnonce_public_upd1 pnonce_dom_upd1 history_ciphers_upd2 elim!: relE)
          done
        subgoal using rel
          by(auto 4 3 simp add: sim1.encryption_key_def encryption_key_sim2_def gen_ekey_def pnonce_sim_sel fun_upd_idem elim: relE intro: relI)
        done
      have rel'_encryption_key: "rel_spmf (?rel' op =)
        (exec_gpv (oracle\<^sub>n' b \<eta>) (catch_gpv (run_state (sim1.encryption_key \<eta> k) s1)) (os1e, os1s))
        (exec_gpv (oracle\<^sub>n' True \<eta>) (catch_gpv (run_state (encryption_key_sim2 \<eta> k) s2)) (os2e, os2s))"
        if rel: "rel M ((m1, s1), (os1e, os1s)) ((m2, s2), (os2e, os2s))"
        and I1': "I1' ((m1, s1), (os1e, os1s))"
        and k: "\<Gamma> k = Some Encryption"
        for M b k s1 os1e os1s s2 os2e os2s m1 m2
        apply(rule rel_spmf_restrict_relpI rel_encryption_key[OF rel k] I1'_encryption_key[OF I1' k])+
        apply(simp)
        done

      have I1'_encryption_key': "pred_spmf I1'' (exec_gpv (oracle\<^sub>n' b \<eta>) (catch_gpv (run_state (sim1.encryption_key' \<eta> (EK (N (PNonce k))) k) s1)) (os1e, os1s))"
        if I1': "I1' ((m1, s1), (os1e, os1s))"
        and k: "\<Gamma> k = Some Encryption"
        for b k s1 os1e os1s m1 using that unfolding I1''_def
        by(intro pred_spmf_exec_gpv[where C="\<lambda>_. True"])
          (auto 4 3 simp add: pred_gpv_def I1'_def elim!: results'_gpv_catch_gpvE sim1.wf_randomness_sampled_recordedD_encryption_key'[OF randomness_sampled_recorded_assms_sim1] wf_history1_encryption_key' dest: sim1.wf_rand_history_encryption_key'[OF ops_history_dom_same_sim1])
      have rel'_encryption_key': "rel_spmf (?rel' op =)
        (exec_gpv (oracle\<^sub>n' b \<eta>) (catch_gpv (run_state (sim1.encryption_key' \<eta> (EK (N (PNonce k))) k) s1)) (os1e, os1s))
        (exec_gpv (oracle\<^sub>n' True \<eta>) (catch_gpv (run_state (sim2.encryption_key' \<eta> (EK (N (PNonce k))) k) s2)) (os2e, os2s))"
        if rel: "rel M ((m1, s1), (os1e, os1s)) ((m2, s2), (os2e, os2s))"
        and I1': "I1' ((m1, s1), (os1e, os1s))"
        and k: "\<Gamma> k = Some Encryption"
        for M b k s1 os1e os1s s2 os2e os2s m1 m2
        using _ I1'_encryption_key'[OF I1' k]
        apply(rule rel_spmf_restrict_relpI')
         apply(simp add: sim1.encryption_key'_def sim2.encryption_key'_def exec_gpv_bind catch_gpv_bind_gpv)
         apply(rule rel_spmf_bindI[where R="?rel' op ="])
          subgoal using rel I1' k by(rule rel'_encryption_key)
         apply(auto simp add: record_sim_def elim!: relE intro!: relI elim!: history_ciphers_upd1 dest!: I1'D_wfh)
        done

      text \<open>@{const sim1.encryption} does not preserve @{term "?rel op ="} because of @{term history_ciphers}, which holds only after updating the history as done by @{term "sim1_\<beta>"}.\<close>

      have I1'_verification_key: "pred_spmf I1'
           (exec_gpv (oracle\<^sub>n' b \<eta>) (run_state (sim1.verification_key \<eta> k) s1) (os1e, os1s))"
        if I1': "I1' ((m1, s1), os1e, os1s)"
        and k: "\<Gamma> k = Some Signature"
        for b k m1 s1 os1e os1s using that unfolding I1'_def
        by(intro pred_spmf_exec_gpv[where C="\<lambda>_. True"])(auto 4 3 simp add: pred_gpv_def elim!: sim1.wf_randomness_sampled_recordedD_verification_key[OF randomness_sampled_recorded_assms_sim1] wf_history1_verification_key dest: sim1.obs_history_sameD_verification_key[OF ops_history_dom_same_sim1] cong: wf_rand_history_cong)
      have rel_verification_key: "rel_spmf (rel op =)
           (exec_gpv (oracle\<^sub>n' b \<eta>) (run_state (sim1.verification_key \<eta> k) s1) (os1e, os1s))
           (exec_gpv (oracle\<^sub>n' True \<eta>) (run_state (verification_key_sim2 \<eta> k) s2) (os2e, os2s))"
        if rel: "rel M ((m1, s1), os1e, os1s) ((m2, s2), os2e, os2s)"
        and k: "\<Gamma> k = Some Signature"
        for M m1 s1 os1e os1s m2 s2 os2e os2s k b
        using pnonce_sigD[OF relD_pnonce_sig[OF rel] k]
        apply cases
        subgoal
          apply(simp add: sim1.verification_key_def verification_key_sim2_def gen_vkey_def exec_gpv_bind key_gen_sig_def pnonce_sim_sel Let_def bind_map_spmf bind_spmf_spmf_of_set bind_spmf_pmf_assoc)
          apply(clarsimp simp add: bind_spmf_map_pmf return_pmf_bind_option intro!: rel_pmf_bind_reflI rel_spmf_bind_reflI)
          apply(rule rel_spmf_bindI2[rotated])
           subgoal by(auto dest: A_vk_SomeD_sk)
          apply(simp)
          apply(rule relI)
          using rel k
          apply(auto 4 4 intro: pnonce_enc_upd_other pnonce_sig_upd_other pnonce_sig_updI map_le_fun_upd2 rand_sim_upd1_other pnonce_public_upd1 pnonce_dom_upd1 elim!: relE)
          done
        subgoal using rel
          by(auto 4 3 simp add: sim1.verification_key_def verification_key_sim2_def gen_vkey_def pnonce_sim_sel fun_upd_idem elim: relE intro: relI)
        done
      have rel'_verification_key: "rel_spmf (?rel op =)
        (exec_gpv (oracle\<^sub>n' b \<eta>) (run_state (sim1.verification_key \<eta> k) s1) (os1e, os1s))
        (exec_gpv (oracle\<^sub>n' True \<eta>) (run_state (verification_key_sim2 \<eta> k) s2) (os2e, os2s))"
        if rel: "rel M ((m1, s1), (os1e, os1s)) ((m2, s2), (os2e, os2s))"
        and I1': "I1' ((m1, s1), (os1e, os1s))"
        and k: "\<Gamma> k = Some Signature"
        for M b k s1 os1e os1s s2 os2e os2s m1 m2
        by(rule rel_spmf_restrict_relpI rel_verification_key[OF rel k] I1'_verification_key[OF I1' k])+ simp

      have I1'_verification_key': "pred_spmf I1'
        (exec_gpv (oracle\<^sub>n' b \<eta>) (run_state (sim1.verification_key' \<eta> (VK (N (PNonce k))) k) s1) (os1e, os1s))"
        if I1': "I1' ((m1, s1), (os1e, os1s))"
        and k: "\<Gamma> k = Some Signature"
        for b k m1 s1 os1e os1s using that unfolding I1'_def
        by(intro pred_spmf_exec_gpv[where C="\<lambda>_. True"])
          (auto 4 3 simp add: pred_gpv_def elim!: sim1.wf_randomness_sampled_recordedD_verification_key'[OF randomness_sampled_recorded_assms_sim1] wf_history1_verification_key' dest: sim1.wf_rand_history_verification_key'[OF ops_history_dom_same_sim1])
      have rel'_verification_key': "rel_spmf (?rel op =)
        (exec_gpv (oracle\<^sub>n' b \<eta>) (run_state (sim1.verification_key' \<eta> (VK (N (PNonce k))) k) s1) (os1e, os1s))
        (exec_gpv (oracle\<^sub>n' True \<eta>) (run_state (sim2.verification_key' \<eta> (VK (N (PNonce k))) k) s2) (os2e, os2s))"
        if rel: "rel M ((m1, s1), (os1e, os1s)) ((m2, s2), (os2e, os2s))"
        and I1': "I1' ((m1, s1), (os1e, os1s))"
        and k: "\<Gamma> k = Some Signature"
        for M b k s1 os1e os1s s2 os2e os2s m1 m2
        apply(intro rel_spmf_restrict_relpI' I1'_verification_key'[OF I1' k])
         apply(simp add: sim1.verification_key'_def sim2.verification_key'_def exec_gpv_bind)
         apply(rule rel_spmf_bindI[where R="?rel op ="])
          subgoal using rel I1' k by(rule rel'_verification_key)
         apply(auto simp add: record_sim_def elim!: relE randomness_sampled_recorded_pnonce_update wf_rand_history_mono2 split: if_split_asm intro!: relI history_ciphers_upd1 dest!: I1'D_wfh)
         done
      have rel_verification_key': "rel_spmf (rel op =)
        (exec_gpv (oracle\<^sub>n' b \<eta>) (run_state (sim1.verification_key' \<eta> (VK (N (PNonce k))) k) s1) (os1e, os1s))
        (exec_gpv (oracle\<^sub>n' True \<eta>) (run_state (sim2.verification_key' \<eta> (VK (N (PNonce k))) k) s2) (os2e, os2s))"
        if rel: "rel M ((m1, s1), (os1e, os1s)) ((m2, s2), (os2e, os2s))"
        and I1': "I1' ((m1, s1), (os1e, os1s))"
        and k: "\<Gamma> k = Some Signature"
        for M b k s1 os1e os1s s2 os2e os2s m1 m2
        using rel'_verification_key'[OF rel I1' k]
        by(rule rel_spmf_mono) simp

      have I1'_\<beta>: "pred_spmf I1'' (exec_gpv (oracle\<^sub>n' b \<eta>) (catch_gpv (run_state (sim1_\<beta> \<eta> t) s1)) (os1e, os1s))"
        if "I1' ((m1, s1), (os1e, os1s))"
        and "A,\<eta>,\<Gamma> \<turnstile>t t : PublicT"
        and "random_usage t \<le> ru"
        for t m1 s1 os1e os1s unfolding I1''_def
      proof(rule pred_spmf_exec_gpv)
        show "pred_gpv (pred_option (pred_prod (\<lambda>_. True) (\<lambda>s1. randomness_sampled_recorded \<Gamma> s1 \<and> wf_rand_history ru (history s1) \<and> wf_history1 \<eta> (pnonce s1) (history s1)))) (\<lambda>_. True) (catch_gpv (run_state (sim1_\<beta> \<eta> t) s1))"
          using that 
          by(auto simp add: pred_gpv_def I1'_def elim!: results'_gpv_catch_gpvE elim: sim1.randomness_sampled_recorded_\<beta>[OF _ _ _ randomness_sampled_recorded_assms_sim1] wf_history1_\<beta>1 sim1.wf_rand_history_\<beta>'[OF _ _ ops_history_dom_same_sim1])
      qed simp_all
          
      define m1 :: bitstring where "m1 \<equiv> []" define m2 :: bitstring where "m2 \<equiv> []"
      have "?rel op = ((m1, s1), os1) ((m2, s2), os2)"
        using pnonce_enc pnonce_sig rand_sim pnonce_public pnonce_dom history_ciphers rsr wrh wfh randomness
        unfolding os1 os2 m1_def m2_def by(auto intro!: relI I1'I simp add: ru_def)
      with kst tt' have "rel_spmf (?rel' op =)
        (exec_gpv (oracle\<^sub>n' b \<eta>) (catch_gpv (run_state (sim1_\<beta> \<eta> t) s1)) os1)
        (exec_gpv (oracle\<^sub>n' True \<eta>) (catch_gpv (run_state (sim2_\<beta> \<eta> t) s2)) os2)"
        unfolding os2 os1
      proof(induction t nt\<equiv>"PublicT" arbitrary: m1 m2 s1 s2 os1e os1s os2e os2s)
        case (E t1 t2 t3 bl)
        have rur: "random_usage (E t1 t2 t3) \<le> ru"
          using member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF E.prems(1)]
          by(auto simp add: ru_def)
        consider (PP) k r where "t1 = EK (N (PNonce k))" "t3 = N (PNonce r)"
          | (AP) k r where "t1 = EK (N (ANonce k))" "t3 = N (PNonce r)"
          | (PA) k r where "t1 = EK (N (PNonce k))" "t3 = N (ANonce r)"
          | (other) "(\<forall>k. t1 \<noteq> EK (N (PNonce k))) \<or> (\<forall>r. t3 \<noteq> N (PNonce r))"
                    "(\<forall>k. t1 \<noteq> EK (N (ANonce k))) \<or> (\<forall>r. t3 \<noteq> N (PNonce r))"
                    "(\<forall>k. t1 \<noteq> EK (N (PNonce k))) \<or> (\<forall>r. t3 \<noteq> N (ANonce r))"
          by blast
        then show ?case
        proof cases
          case other thus ?thesis by(auto split!: t.split nonce.split)
        next
          case (PP k r)
          with E.hyps(1,5) have k: "\<Gamma> k = Some Encryption" and r: "\<Gamma> r = Some (Randomness bl)" by auto
          hence [simp]: "k \<noteq> r" "r \<noteq> k" by auto
          from member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF E.prems(1)] PP
          have "single_usage r (RUEncryption (EK (N (PNonce k))) t2) \<le> ru" by(auto simp add: ru_def)
          hence rur': "RUEncryption (EK (N (PNonce k))) t2 \<in> ru r" by(rule single_usage_leD)
          let ?t = "E (EK (N (PNonce k))) t2 (N (PNonce r))"
          show ?thesis
            apply(rule rel_spmf_restrict_relpI[rotated])
            subgoal using E.hyps(1,3,5) E.prems(2) rur by(intro I1'_\<beta>)(auto elim!: rel_spmf_restrict_relpE)
            subgoal by simp
            apply(simp add: exec_gpv_bind PP catch_gpv_bind_gpv)
            apply(rule rel_spmf_bindI[where R="?rel' op ="])
             subgoal using E.prems(2) k by clarsimp(erule (2) rel'_encryption_key')
            apply clarsimp
            subgoal premises prems for ek1 s1' os1e' os1s' ek2 s2' os2e' os2s' thm prems
              apply(simp add: bind_gpv_assoc exec_gpv_bind spmf.map_comp bind_map_spmf)
              apply(rule rel_spmf_bindI'[where A="?rel op ="])
               subgoal using \<open>rel _ ((ek1, s1'), _) _\<close> \<open>I1' ((ek1, s1'), os1e', os1s') \<close>k E.prems(1) PP E.hyps(3) rur
                 by(intro rel_spmf_restrict_relpI E.hyps(4) I1'_\<beta>)(auto elim!: rtranclp_trans)
              apply(clarsimp)
              subgoal premises prems' for msg1 s1'' os1e'' os1s'' msg2 s2'' os2e'' os2s'' thm prems'
              proof(cases "history s2'' ?t")
                case None
                note rel = \<open>rel _ ((_, s1''), os1e'', os1s'') _\<close> and I1' = \<open>I1' ((msg1, s1''), os1e'', os1s'')\<close>
                have domh: "?t \<notin> dom (history s1'')"
                  using None relD_history[OF rel] by(simp add: domIff)
                show ?thesis using None
                  apply(simp add: exec_gpv_bind sim1.encryption_def encryption_sim2_def)
                  apply(rule rel_spmf_bindI'[where A="?rel op ="])
                   subgoal using rel I1' k by(rule rel'_encryption_key)
                  apply(clarsimp simp add: exec_gpv_bind pnonce_sim_sel bind_map_spmf Let_def gen_encrypt_def)
                  apply(frule relD_m, clarsimp simp add: relD_m[OF rel])
                  apply(drule in_set_spmf_exec_gpv_into_results'_gpv)
                  apply(frule relD_history)
                  apply(frule sim1.obs_history_sameD_encryption_key[OF ops_history_dom_same_sim1])
                  apply(frule (1) rel_history_E_NoneD[OF _ _ _ r E.prems(1)[unfolded PP]])
                   subgoal using relD_history[OF rel] by(clarsimp simp add: eq_None_iff_not_dom[where f="history _"])
                  apply(clarsimp simp add: record_sim_def intro!: rel_spmf_bind_reflI)
                  apply(frule encryption_key1_sampled)
                  apply(frule relD_pnonce_enc[THEN pnonce_encD, OF _ k])
                  apply(clarsimp simp add: option_rel_Some1 encrypt_def bind_spmf_map_pmf bind_spmf_pmf_assoc bind_spmf_spmf_of_set)
                  apply(rule rel_pmf_bind_reflI)
                  apply(rule rel_spmf_bind_reflI)
                  apply clarsimp
                  apply(rule relI)
                  prefer 7 -- \<open>@{term history_ciphers}\<close>
                  subgoal using relD_history[OF rel] domh by(auto elim!: relE intro!: history_ciphers_add)
                  using k r sim1.random_usage_\<beta>_rec_mono[OF E.prems(1)] PP member_le_random_usages[OF t', folded ru_def] rur'
                  apply(auto 4 3 simp add: record_sim_def elim!: relE wf_rand_history_mono2 pnonce_public_upd1 intro!: relI intro: pnonce_enc_upd_other pnonce_enc_upd_ciphers pnonce_sig_upd_other split: if_split_asm intro: rand_sim_upd1_other pnonce_dom_upd1)
                  done
              next
                case (Some bs_e)
                note rel = \<open>rel _ ((_, s1''), os1e'', os1s'') _\<close> and I1' = \<open>I1' ((msg1, s1''), os1e'', os1s'')\<close>
                from I1'[THEN I1'D_wfh] Some
                have run: "run_env (pure.\<beta> \<eta> ?t) (pnonce s1'') = Some bs_e"
                  and sampled': "sampled1 ?t (pnonce s1'')"
                  unfolding relD_history[OF rel, symmetric] by(rule wf_history1D)+
                from sampled' have "{k, r} - dom (pnonce s1'') = {}" by auto
                then have "sample_pnonces \<eta> {k, r} (pnonce s1'') = return_pmf (pnonce s1'')"
                  by(simp add: sample_pnonces_def)
                moreover have "history_update (\<lambda>hist. hist(?t \<mapsto> bs_e)) s1'' = s1''"
                  and "history_update (\<lambda>hist. hist(?t \<mapsto> bs_e)) s2'' = s2''"
                  by(simp_all add: fun_upd_idem Some relD_history[OF rel])
                moreover have "rel op = ((bs_e, s1''), os1e'', os1s'') ((bs_e, s2''), os2e'', os2s'')"
                  using rel by(auto simp add: rel_def)
                moreover have "run_env (pure.\<beta> \<eta> t2) (pnonce s1'') = Some msg1"
                  using \<open>((msg1, s1''), os1e'', os1s'') \<in> _\<close> by(clarsimp simp add: lazy_convs_pure)
                ultimately show ?thesis using run Some rel
                  by(simp add: \<beta>1_conv_spmf encryption1_conv_spmf encryption_spmf_conv_option encryption_option_conv_pure exec_gpv_bind bind_map_spmf option.map_comp encryption_sim2_def bind_eq_Some_conv record_sim_def)
              qed
              done
            done
        next
          case (AP k r)
          with E.hyps(5) have r: "\<Gamma> r = Some (Randomness bl)" by auto

          from AP rur have "single_usage r (RUEncryption (EK (N (ANonce k))) t2) \<le> ru" by(auto simp add: ru_def)
          hence rur': "RUEncryption (EK (N (ANonce k))) t2 \<in> ru r" by(rule single_usage_leD)
          have pns_r: "pnonce s1 r = pnonce s2 r" if "rel M ((m1, s1), (os1e, os1s)) ((m2, s2), (os2e, os2s))"
            for M m1 s1 os1e os1s m2 s2 os2e os2s using that[THEN relD_rand_sim] r
            by(rule rand_simD)(auto dest: wf_random_usage'D1[OF ru, folded ru_def, OF rur'])

          have I1': "I1' ((m, history_update (\<lambda>hist. hist(EK (N (ANonce k)) \<mapsto> k)) s1), os1e, os1s)" for m
            using E.prems(2) by(intro I1'_history_updateI) auto

          let ?s1' = "history_update (\<lambda>hist. hist(EK (N (ANonce k)) \<mapsto> k)) s1"
            and ?s2' = "history_update (\<lambda>hist. hist(EK (N (ANonce k)) \<mapsto> k)) s2"
          have IH: "rel_spmf (?rel op =)
            (exec_gpv (oracle\<^sub>n' b \<eta>) (run_state (sim1_\<beta> \<eta> t2) ?s1') (os1e, os1s))
            (exec_gpv (oracle\<^sub>n' True \<eta>) (run_state (sim2_\<beta> \<eta> t2) ?s2') (os2e, os2s))"
          proof(rule E.hyps)
            show "\<beta>_rec t' t2" using AP E.prems(1) by(auto elim!: rtranclp_trans)
            fix m
            have "I1' ((m, ?s1'), os1e, os1s)"
              using E.prems(2) rur by(auto intro!: I1'_history_updateI)
            then show "?rel op = ((m, ?s1'), os1e, os1s) ((m, ?s2'), os2e, os2s)" using E.prems(2)
              by(auto intro!: rel_history_updateI dest: I1'D_wfh)
          qed
          show ?thesis
            apply(rule rel_spmf_restrict_relpI'[rotated])
            subgoal using E.hyps(1,3,5) E.prems(2) rur by(intro I1'_\<beta>) auto
            subgoal by simp
            apply(simp add: record_sim_def exec_gpv_bind AP)
            apply(rule rel_spmf_bindI[OF IH])
            apply(clarsimp simp add: exec_gpv_bind)
            subgoal premises prems for m1' s1' os1e' os1s' m2' s2' os2e' os2s'
            proof(cases "pnonce s1' r")
              case None
              with \<open>rel _ ((m1', s1'), _) _\<close>[THEN pns_r] have [simp]: "pnonce s2' r = None" by simp
              show ?thesis using \<open>rel _ ((m1', s1'), _) _\<close> \<open>I1' ((m1', s1'), _)\<close> None r AP
                apply(auto simp add: pnonce_sim_def exec_gpv_bind Let_def intro!: rel_spmf_bind_reflI elim!: relE intro!: relI)
                prefer 6 -- \<open>@{term history_ciphers}\<close>
                subgoal
                  apply(erule history_ciphers_upd1)
                   apply(fastforce dest: I1'D_wfh)
                  apply(fastforce intro: wf_history1_pnonce_mono dest: I1'D_wfh intro!: map_le_map_upd2)+
                  done
                apply(auto 4 3 intro: pnonce_enc_upd_other pnonce_sig_upd_other rand_sim_updI pnonce_public_upd1 pnonce_public_upd2 pnonce_dom_updI)
                done
            next
              case (Some bs_r)
              then show ?thesis
                using \<open>rel _ ((m1', s1'), _) _\<close>[THEN relD_m] \<open>rel _ ((m1', s1'), _) _\<close>[THEN pns_r] rur AP \<open>I1' ((m1', s1'), _)\<close>[THEN I1'D_wfh]
                by(auto simp add: pnonce_sim_def split: option.split intro!: rel_spmf_bind_reflI rel_history_updateI[OF \<open>rel _ ((m1', s1'), _) _\<close>] dest!: I1'D_wfh)
            qed
            done
        next
          case (PA k r)
          then have False using E.hyps(1,3,5) by auto
          then show ?thesis ..
        qed
      next
        case (E_ANonce n t2 m)
        show ?case using E_ANonce.prems(2)
          apply(intro rel_spmf_restrict_relpI'[rotated])
          subgoal using E_ANonce.hyps(1,2,4) E_ANonce.prems(2) member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF E_ANonce.prems(1)]
            by(intro I1'_\<beta>)(auto simp add: ru_def)
          subgoal by simp
          subgoal by(auto simp add: record_sim_def intro!: rel_history_updateI dest: I1'D_wfh)
          done
      next
        case (EK t)
        show ?case using EK.prems(2) EK.hyps 
          by(auto split: t.split nonce.split intro: rel'_encryption_key')
      next
        case (EK_ANonce n)
        show ?case using EK_ANonce.prems(2)
          apply(intro rel_spmf_restrict_relpI'[rotated])
          subgoal using EK_ANonce.hyps EK_ANonce.prems(2) member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF EK_ANonce.prems(1)]
            by(intro I1'_\<beta>)(auto simp add: ru_def)
          subgoal by simp
          subgoal by(auto simp add: record_sim_def intro!: rel_history_updateI dest: I1'D_wfh)
          done
      next
        case (VK t)
        show ?case using VK.prems(2) VK.hyps
          by(auto split: t.split nonce.split intro: rel'_verification_key')
      next
        case (VK_ANonce n)
        show ?case using VK_ANonce.prems(2)
          apply(intro rel_spmf_restrict_relpI'[rotated])
          subgoal using VK_ANonce.hyps VK_ANonce.prems(2) member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF VK_ANonce.prems(1)]
            by(intro I1'_\<beta>)(auto simp add: ru_def)
          subgoal by simp
          subgoal by(auto simp add: record_sim_def intro!: rel_history_updateI dest: I1'D_wfh)
          done
      next
        case (SIG t1 t2 t3 bl)
        have rur: "random_usage (SIG t1 t2 t3) \<le> ru"
          using member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF SIG.prems(1)]
          by(auto simp add: ru_def)

        from SIG.hyps(1,4)
        obtain k r where t1: "t1 = SK (N (PNonce k))" and k: "\<Gamma> k = Some Signature"
          and t3: "t3 = N (PNonce r)" and r: "\<Gamma> r = Some (Randomness bl)"
          by(auto elim!: key_safe_t.cases)
        from k r have [simp]: "k \<noteq> r" "r \<noteq> k" by auto
        let ?t = "SIG (SK (N (PNonce k))) t2 (N (PNonce r))"
        show ?case
          apply(intro rel_spmf_restrict_relpI'[rotated])
          subgoal using SIG.hyps(1,2,4) SIG.prems(2) rur t1 t3 by(intro I1'_\<beta>)(auto elim!: rel_spmf_restrict_relpE)
          subgoal by simp
          apply(simp add: exec_gpv_bind t1 t3)
          apply(rule rel_spmf_bindI'[where A="?rel op ="])
           subgoal using SIG.prems(2) k by(intro rel'_verification_key') auto
          apply(clarsimp simp add: exec_gpv_bind)
          apply(rule rel_spmf_bindI'[where A="?rel op ="])
           subgoal using SIG.prems(1) rur t3 t1 by(intro SIG.hyps)(auto elim!: rtranclp_trans)
          apply clarsimp
          subgoal premises prems for bs_sk1 s1' os1e' os1s' bs_sk2 s2' os2e' os2s' msg1 s1'' os1e'' os1s'' msg2 s2'' os2e'' os2s'' thm prems
          proof(cases "history s2'' (SIG (SK (N (PNonce k))) t2 (N (PNonce r)))")
            case None
            note rel = \<open>rel _ ((msg1, s1''), _) _\<close> and I1' = \<open>I1' ((msg1, s1''), _)\<close>
            from member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF SIG.prems(1)] t1 t3
            have "single_usage r (RUSignature (SK (N (PNonce k))) t2) \<le> ru" by(auto simp add: ru_def)
            hence rur: "RUSignature (SK (N (PNonce k))) t2 \<in> ru r" by(rule single_usage_leD)
            def verification_key'' \<equiv> "bind_gpv (run_state (pnonce_sim \<eta> k) s1'') (\<lambda>(r_n, s1'). monad.bind_option gpv_monad 
                    (A_vk A \<eta> r_n) (\<lambda>_. Done (r_n, s1'))) :: (_, comp_impl.call, comp_impl.ret) gpv"
            have eq1:
              "bind_gpv (run_state (sim1.signature \<eta> k t2 msg1 r) s1'') (\<lambda>(msg, s1'). run_state (record_sim ?t msg) s1') = 
               bind_gpv verification_key'' (\<lambda>(r_n, s1'). bind_gpv
                 (run_state (pnonce_sim \<eta> r) s1') (\<lambda>(r_n', s1''). monad.bind_option gpv_monad
                 (A_sk A \<eta> r_n) (\<lambda>bs_sk. monad.bind_option gpv_monad
                 (A_sig A \<eta> bs_sk msg1 r_n') (\<lambda>msg.
                 run_state (record_sim ?t msg) s1''))))"
              apply(clarsimp simp add: sim1.signature_def bind_gpv_assoc verification_key''_def split_def bind_gpv_bind_option_assoc intro!: bind_gpv_cong[OF refl])
              apply(rewrite in "_ = \<hole>" monad.bind_option_K)
               subgoal by(clarsimp simp add: A_sk_vk_None[symmetric] pnonce_sim_def Let_def bind_gpv_assoc bind_gpv_eq_Fail split: option.split)
              apply(simp)
              done
            have "rel_spmf (rel \<top>)
              (exec_gpv (oracle\<^sub>n' b \<eta>) verification_key'' (os1e'', os1s''))
              (exec_gpv (oracle\<^sub>n' True \<eta>) (run_state (verification_key_sim2 \<eta> k) s2'') (os2e'', os2s''))"
              using relD_pnonce_sig[OF rel, THEN pnonce_sigD, OF k]
            proof cases
              case None
              thus ?thesis using rel k
                apply(clarsimp simp add: verification_key''_def pnonce_sim_def Let_def verification_key_sim2_def gen_vkey_def key_gen_sig_def bind_spmf_map_pmf bind_map_spmf exec_gpv_bind bind_spmf_spmf_of_set bind_spmf_pmf_assoc bind_map_pmf bind_assoc_pmf return_pmf_bind_option return_pmf_map_option_conv_bind bind_pmf_spmf_assoc rel_pmf_return_pmf1 bind_return_pmf intro!: rel_pmf_bind_reflI rel_spmf_bind_reflI)
                apply(frule A_vk_SomeD_sk)
                apply(auto 4 4 elim!: relE intro!: relI intro: wf_history1_pnonce_mono pnonce_enc_upd_other pnonce_sig_updI pnonce_public_upd1 pnonce_dom_upd1 intro!: rand_sim_upd1_other elim!: randomness_sampled_recorded_pnonce_update intro!: map_le_map_upd2)
                done
            next
              case Some
              thus ?thesis using rel k
                by(auto simp add: verification_key''_def pnonce_sim_def Let_def verification_key_sim2_def gen_vkey_def key_gen_sig_def rel_spmf_return_spmf1 bind_return_pmf fun_upd_idem elim!: relE intro!: relI)
            qed
            hence rel'_verification_key'': "rel_spmf (?rel \<top>)
              (exec_gpv (oracle\<^sub>n' b \<eta>) verification_key'' (os1e'', os1s''))
              (exec_gpv (oracle\<^sub>n' True \<eta>) (run_state (verification_key_sim2 \<eta> k) s2'') (os2e'', os2s''))"
              apply(rule rel_spmf_restrict_relpI pred_spmf_exec_gpv)
              using I1' k apply(auto 4 5 simp add: I1'_def Let_def verification_key''_def exec_gpv_bind pred_spmf_def pnonce_sim_sel split: if_split_asm intro: wf_history1_pnonce_mono elim!: randomness_sampled_recorded_pnonce_update intro!: map_le_map_upd2)
              done
            show ?thesis
              apply(simp add: exec_gpv_bind eq1 signature_sim2_def None)
              apply(rule rel_spmf_bindI'[OF rel'_verification_key''])
              apply(clarsimp simp add: exec_gpv_bind gen_sig_def bind_map_spmf)
              subgoal premises prems for bs_k s1''' os1e''' os1s''' bs_vk s2''' os2e''' os2s'''
              proof - thm prems
                note rel' = \<open>rel _ ((bs_k, s1'''), _) _\<close>
                from \<open>((bs_k, s1'''), _) \<in> _\<close> 
                have s1'k: "pnonce s1''' k = Some bs_k"
                  and dom: "dom (history s1''') = dom (history s1'')"
                  and pns: "pnonce s1'' \<subseteq>\<^sub>m pnonce s1'''"
                  by(auto simp add: verification_key''_def pnonce_sim_conv_spmf exec_gpv_bind pnonce_spmf_def Let_def
                          split: option.split_asm if_split_asm intro!: map_le_map_upd2)
                with rel'[THEN relD_pnonce_sig, THEN pnonce_sigD, OF k]
                obtain bs_vk bs_sk sigs where os2s'k: "os2s''' k = Some (bs_vk, bs_sk, sigs)"
                  and vk: "A_vk A \<eta> bs_k = Some bs_vk" and sk: "A_sk A \<eta> bs_k = Some bs_sk"
                  by(auto simp add: option_rel_Some1)
                have s1'r: "pnonce s1''' r = None" using dom None relD_history[OF rel] relD_history[OF rel']
                  by(intro rel_history_SIG_NoneD[OF rel' \<open>I1' ((bs_k, s1'''), os1e''', os1s''')\<close> _ r SIG.prems(1)[unfolded t1 t3]])
                    (clarsimp simp add: eq_None_iff_not_dom[where f="history _"])
                show ?thesis using s1'k s1'r os2s'k vk sk dom rel'
                  apply(clarsimp simp add: sign_def pnonce_sim_def Let_def exec_gpv_bind bind_spmf_map_pmf bind_spmf_spmf_of_set bind_spmf_pmf_assoc relD_m[OF rel] record_sim_def intro!: rel_pmf_bind_reflI rel_spmf_bind_reflI)
                  using k r sim1.random_usage_\<beta>_rec_mono[OF SIG.prems(1)] member_le_random_usages[OF t', folded ru_def] rur t1 t3 \<open>I1' ((_, s1'''), _)\<close>
                  apply(auto intro!: relI elim!: relE pnonce_enc_upd_other pnonce_public_upd1 pnonce_dom_upd1 intro: pnonce_sig_upd_other pnonce_sig_upd_sigs rand_sim_upd1_other)
                  apply(fastforce intro!: relI elim!: relE pnonce_enc_upd_other pnonce_public_upd1 pnonce_dom_upd1 intro: pnonce_sig_upd_other pnonce_sig_upd_sigs rand_sim_upd1_other)
                  apply(fastforce intro!: relI elim!: relE pnonce_enc_upd_other pnonce_public_upd1 pnonce_dom_upd1 intro: pnonce_sig_upd_other pnonce_sig_upd_sigs rand_sim_upd1_other)
                  apply(fastforce intro!: history_ciphers_upd1 dest: I1'D_wfh elim!: wf_history1_pnonce_mono intro!: map_le_map_upd2)
                  done
              qed
              done
          next
            case (Some bs_sig)
            note rel = \<open>rel _ ((msg1, s1''), _) _\<close> and I1' = \<open>I1' ((msg1, s1''), _)\<close>
            from I1'[THEN I1'D_wfh] Some
            have run: "run_env (pure.\<beta> \<eta> ?t) (pnonce s1'') = Some bs_sig"
              and sampled': "sampled1 ?t (pnonce s1'')"
              unfolding relD_history[OF rel, symmetric] by(rule wf_history1D)+
            from sampled' have "{k, r} - dom (pnonce s1'') = {}" by auto
            then have "sample_pnonces \<eta> {k, r} (pnonce s1'') = return_pmf (pnonce s1'')"
              by(simp add: sample_pnonces_def)
            moreover have "history_update (\<lambda>hist. hist(?t \<mapsto> bs_sig)) s1'' = s1''"
              and "history_update (\<lambda>hist. hist(?t \<mapsto> bs_sig)) s2'' = s2''"
              by(simp_all add: fun_upd_idem Some relD_history[OF rel])
            moreover have "rel op = ((bs_sig, s1''), os1e'', os1s'') ((bs_sig, s2''), os2e'', os2s'')"
              using rel by(auto simp add: rel_def)
            moreover have "run_env (pure.\<beta> \<eta> t2) (pnonce s1'') = Some msg1"
              using \<open>((msg1, s1''), os1e'', os1s'') \<in> _\<close> by(clarsimp simp add: lazy_convs_pure)
            ultimately show ?thesis using run Some rel
              by(simp add: \<beta>1_conv_spmf signature1_conv_spmf signature_spmf_conv_option signature_option_conv_pure exec_gpv_bind bind_map_spmf option.map_comp signature_sim2_def bind_eq_Some_conv record_sim_def)
          qed
          done
      next
        case (SIG_ANonce n t m)
        show ?case using SIG_ANonce.prems(2)
          apply(intro rel_spmf_restrict_relpI'[rotated])
          subgoal using SIG_ANonce.hyps(1,2,4) SIG_ANonce.prems(2) member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF SIG_ANonce.prems(1)]
            by(intro I1'_\<beta>)(auto simp add: ru_def)
          subgoal by simp
          subgoal by(auto simp add: record_sim_def intro!: rel_history_updateI dest: I1'D_wfh)
          done
      next
        case (PAIR t1 t2)
        have rur: "random_usage (PAIR t1 t2) \<le> ru"
          using member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF PAIR.prems(1)]
          by(auto simp add: ru_def)

        show ?case
          apply(intro rel_spmf_restrict_relpI'[rotated])
          subgoal using PAIR.hyps(1,3) PAIR.prems(2) member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF PAIR.prems(1)]
            by(intro I1'_\<beta>)(auto simp add: ru_def)
          subgoal by simp
          apply(simp add: exec_gpv_bind)
          apply(rule rel_spmf_bindI'[where A="?rel op ="])
           subgoal using PAIR.prems rur by(intro rel_spmf_restrict_relpI PAIR.hyps I1'_\<beta>)(auto elim: rtranclp_trans)
          apply(clarsimp simp add: exec_gpv_bind)
          apply(frule relD_m)
          apply(rule rel_spmf_bindI'[where A="?rel op ="])
           subgoal using PAIR.prems rur by(intro rel_spmf_restrict_relpI PAIR.hyps I1'_\<beta>)(auto elim: rtranclp_trans)
          apply(clarsimp simp add: record_sim_def)
          apply(rotate_tac -3, frule relD_m)
          apply(fastforce intro!: rel_history_updateI dest: I1'D_wfh)
          done
      next
        case EMPTY thus ?case
          apply(intro rel_spmf_restrict_relpI'[rotated])
          subgoal using EMPTY.prems(2) member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF EMPTY.prems(1)]
            by(intro I1'_\<beta>)(auto simp add: ru_def)
          subgoal by simp
          by(auto intro!: rel_spmf_bind_reflI rel_history_updateI simp add: record_sim_def dest!: I1'D_wfh)
      next
        case (STRING t bit)
        have rur: "random_usage (STRING bit t) \<le> ru"
          using member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF STRING.prems(1)]
          by(auto simp add: ru_def)

        show ?case
          apply(intro rel_spmf_restrict_relpI'[rotated])
          subgoal using STRING.hyps(1) STRING.prems(2) member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF STRING.prems(1)]
            by(intro I1'_\<beta>)(auto simp add: ru_def)
          subgoal by simp
          apply(clarsimp simp add: exec_gpv_bind)
          apply(rule rel_spmf_bindI[where R="?rel op ="])
           subgoal using STRING.prems rur by(intro rel_spmf_restrict_relpI STRING.hyps I1'_\<beta>)(auto elim!: rtranclp_trans)
          apply(clarsimp simp add: record_sim_def)
          apply(frule relD_m)
          apply(auto intro!: rel_history_updateI intro!: rel_spmf_bind_reflI dest!: I1'D_wfh)
          done
      next
        case (Nonce_public n)
        from Nonce_public.prems(2) \<open>\<Gamma> n = None\<close>
        have "pnonce s1 n = pnonce s2 n" by(auto dest: relD_pnonce_public pnonce_publicD)
        then show ?case using Nonce_public.prems(2) Nonce_public.hyps
          apply(intro rel_spmf_restrict_relpI'[rotated])
          subgoal using Nonce_public.hyps(1) Nonce_public.prems(2) member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF Nonce_public.prems(1)]
            by(intro I1'_\<beta>)(auto simp add: ru_def)
          subgoal by simp
          apply(clarsimp simp add: exec_gpv_bind)
          apply(cases "pnonce s2 n")
          subgoal
            apply(clarsimp simp add: pnonce_sim_def exec_gpv_bind Let_def record_sim_def intro!: rel_spmf_bind_reflI)
            apply(auto 4 3 simp add: pnonce_pure_def elim!: relE pnonce_enc_upd_other pnonce_sig_upd_other pnonce_public_updI pnonce_dom_updI rand_sim_updI intro!: relI history_ciphers_upd1 map_le_map_upd2 dest!: I1'D_wfh elim: wf_history1_pnonce_mono)
            done
          subgoal
            by(auto simp add: pnonce_sim_def record_sim_def pnonce_pure_def intro!: rel_history_updateI dest: I1'D_wfh)
          done
      next
        case (ANonce m)
        thus ?case
          apply(intro rel_spmf_restrict_relpI'[rotated])
          subgoal using member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF ANonce.prems(1)]
            by(intro I1'_\<beta>)(auto simp add: ru_def)
          subgoal by simp
          subgoal by(auto simp add: record_sim_def intro!: rel_history_updateI dest!: I1'D_wfh)
          done
      next
        case (garbage t)
        thus ?case
          apply(intro rel_spmf_restrict_relpI'[rotated])
          subgoal using member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF garbage.prems(1)]
            by(intro I1'_\<beta>)(auto simp add: ru_def)
          subgoal by simp
          by(auto split: t.split nonce.split simp add: record_sim_def intro!: rel_history_updateI dest!: I1'D_wfh)
      next
        case (garbageE t m)
        show ?case using garbageE.prems(2)
          apply(intro rel_spmf_restrict_relpI'[rotated])
          subgoal using garbageE.hyps(1,3) member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF garbageE.prems(1)]
            by(intro I1'_\<beta>)(auto simp add: ru_def)
          subgoal by simp
          subgoal by(auto simp add: record_sim_def intro!: rel_history_updateI dest: I1'D_wfh)
          done
      next
        case (garbageSIG t m)
        show ?case using garbageSIG.prems(2)
          apply(intro rel_spmf_restrict_relpI'[rotated])
          subgoal using garbageSIG.hyps(1,3) member_le_random_usages[OF t'] sim1.random_usage_\<beta>_rec_mono[OF garbageSIG.prems(1)]
            by(intro I1'_\<beta>)(auto simp add: ru_def)
          subgoal by simp
          subgoal by(auto simp add: record_sim_def intro!: rel_history_updateI dest: I1'D_wfh)
          done
      qed
      then show ?thesis using Output
        apply(simp add: hybrid_step_def split_def map_spmf_conv_bind_spmf[symmetric] spmf_rel_map_restrict_relp sim_control_def sim1.rec_message_def sim2.rec_message_def exec_gpv_bind spmf.map_comp bind_map_spmf map_gpv_id_bind_gpv)
        apply(rule rel_spmf_bindI[where R="?rel op ="])
         subgoal by(simp add: t_def)
        apply(auto elim!: relE simp add: ru_def intro!: bisim12I rel_spmf_bind_reflI)
        done
    next
      case (Control id out labs succs)
      then show ?thesis using randomness pnonce_enc pnonce_sig rand_sim pnonce_public pnonce_dom history_ciphers
        by(auto simp add: hybrid_step_def split_def exec_gpv_bind spmf_rel_map sim_control_def map_gpv_id_bind_gpv intro!: rel_spmf_bind_reflI bisim12I)
    next
      case (Probabilistic id p succs)
      then show ?thesis using randomness pnonce_enc pnonce_sig rand_sim pnonce_public pnonce_dom history_ciphers
        by(auto simp add: hybrid_step_def split_def exec_gpv_bind spmf_rel_map sim_control_def bind_spmf_pmf_assoc intro!: rel_spmf_bind_reflI bisim12I rel_pmf_bind_reflI)
    next
      case (Computation id comp c_data yes no)
      then show ?thesis using randomness pnonce_enc pnonce_sig rand_sim pnonce_public pnonce_dom history_ciphers
        by(auto simp add: hybrid_step_def Let_def bisim12_def split: computation.split t.split nonce.split option.split elim!: rand_sim_mono)
    next
      case (Input id succ)

      let ?rel = "rel_prod op = (\<lambda>_ _. True) :: t \<times> state_oracle \<Rightarrow> t \<times> state_oracle \<Rightarrow> bool"

      def I1'_t \<equiv> "\<lambda>t. A,\<eta>,\<Gamma> \<turnstile>t t : PublicT"
      have I1'I: "I1'_t t"
        if "A,\<eta>,\<Gamma> \<turnstile>t t : PublicT" for t using that by(simp add: I1'_t_def)
      have I1'E: "thesis" if "I1'_t t"
        and "A,\<eta>,\<Gamma> \<turnstile>t t : PublicT \<Longrightarrow> thesis" for thesis t using that by(auto simp add: I1'_t_def)
      have I1'_tD_key_safe_t: "A,\<eta>,\<Gamma> \<turnstile>t t : PublicT"
        if "I1'_t t" for t using that by(auto simp add: I1'_t_def)

      def I2 \<equiv> "\<lambda>(ose, oss). 
        pnonce_enc (pnonce s1) ose \<and> 
        pnonce_sig (pnonce s1) oss \<and>
        history_ciphers (history s1) ose"
      have I2I: "I2 (os2e, os2s)"
        if "pnonce_enc (pnonce s1) os2e"
        and "pnonce_sig (pnonce s1) os2s" 
        and "history_ciphers (history s1) os2e"
        for os2e os2s using that by(simp add: I2_def)
      have I2D_pnonce_enc: "pnonce_enc (pnonce s1) os2e"
        and I2D_pnonce_sig: "pnonce_sig (pnonce s1) os2s"
        and I2D_history_ciphers: "history_ciphers (history s1) os2e"
        if "I2 (os2e, os2s)"
        for os2e os2s using that by(simp_all add: I2_def)

      have I2_oracle: "pred_fun I2 (pred_fun is_decrypt (pred_spmf (pred_prod (\<lambda>_. True) I2))) (oracle\<^sub>n' b \<eta>)" for b
        apply(clarsimp intro!: pred_funI elim!: is_decrypt.cases)
        subgoal for os2e os2s n bs by(cases "os2e n")(auto simp add: fun_upd_idem)
        done

      let ?I1' = "pred_prod I1'_t (\<lambda>_. True)"
      let ?I2' = "pred_prod (\<lambda>_. True) I2"
      let ?rel' = "?rel \<upharpoonleft> ?I1' \<otimes> ?I2'"
      have I1'_\<tau>: "pred_resumption I1'_t (\<lambda>_. True) (sim1_\<tau> \<eta> s1 bs)" for bs using kssim
        by(auto simp add: pred_resumption_def I1'_t_def elim!: sim1.key_safe_\<tau>)

      have I2_start: "I2 (os2e, os2s)"
        using pnonce_enc pnonce_sig history_ciphers by(simp add: I2I)
      then have "rel_spmf ?rel
         (exec_gpv (oracle\<^sub>n' b \<eta>) (lift_resumption (sim1_\<tau> \<eta> s1 bs)) (os1e, os1s))
         (exec_gpv (oracle\<^sub>n' True \<eta>) (lift_resumption (sim2_\<tau> \<eta> s2 bs)) (os2e, os2s))" 
        for bs unfolding exec_gpv_lift_resumption
      proof(induction arbitrary: bs os1e os1s os2e os2s rule: parallel_fixp_induct_strong_1_1[OF resumption_partial_function_definition resumption_partial_function_definition sim1.\<tau>.mono sim2.\<tau>.mono sim1.\<tau>_def sim2.\<tau>_def, case_names adm bot step])
        case adm show ?case by(simp add: resumption.mcont2mcont[OF mcont_call])
      next
        case bot show ?case by(simp add: resumption_lub_empty)
      next
        case (step \<tau>1 \<tau>2)

        have I1'_\<tau>: "pred_resumption I1'_t (\<lambda>_. True) (\<tau>1 bs)" for bs
          using I1'_\<tau> step.hyps(1) by(rule pred_resumption_antimono)
        from \<tau>2_only_decrypt_queries have decrypt: "pred_resumption (\<lambda>_. True) is_decrypt (\<tau>2 bs)" for bs
          using step.hyps(2) by(rule pred_resumption_antimono)

        show ?case
        proof(cases \<eta> bs rule: \<tau>_cases, simp_all split del: if_split, goal_cases) (* Use the same trick above for the induction rule about tau and set *)
          case Nonce: 1
          let ?P1 = "\<lambda>n. n \<notin> randomness s1 \<and> pnonce s1 n = Some bs"
          let ?P2 = "\<lambda>n. n \<notin> randomness s2 \<and> pnonce s2 n = Some bs"
          have eq: "?P1 = ?P2"
          proof(rule ext)
            show "?P1 n = ?P2 n" for n  
              using pnonce_public[THEN pnonce_publicD, of n] key_safe_sim_pnonce_\<Gamma>[OF kssim, of n] pnonce_dom[THEN pnonce_domD] randomness
              by(cases "pnonce s1 n")(auto iff: domIff)
          qed
          then show ?case using step.prems by simp
        next
          case Cipher: 2
          show ?case
          proof(cases "bs \<in> ran (history s1)")
            case True
            thus ?thesis apply simp done
          next
            case False
            thus ?thesis
              apply simp
              apply(rule rel_spmf_bindI[where R="?rel'"])
               subgoal using step.prems I2_oracle
                 by(intro rel_spmf_restrict_relpI step.IH)
                   (simp_all add: pred_spmf_execp_resumption[OF I1'_\<tau>] pred_spmf_execp_resumption[OF decrypt])
              apply(clarsimp split: t.split nonce.split)
              apply(simp add: decryption_option_def gen_decrypt_def bind_map_spmf)
              apply(frule I1'_tD_key_safe_t)
              apply simp
              apply(frule I2D_pnonce_enc)
              apply(drule (1) pnonce_encD)
              apply(erule option.rel_cases)
              subgoal by simp
              apply clarsimp
              apply(intro conjI strip)
              subgoal by(blast dest!: I2D_history_ciphers history_ciphersD intro!: ind_cca2.cipher_queriesI)
              apply(clarsimp split: option.split simp add: fun_upd_idem)
              apply(rule rel_spmf_bindI[where R="?rel'"])
               subgoal using I2_oracle 
                 by(intro rel_spmf_restrict_relpI step.IH)
                   (simp_all add: pred_spmf_execp_resumption[OF I1'_\<tau>] pred_spmf_execp_resumption[OF decrypt])
              apply clarsimp
              done
          qed
        next
          case Ekey: 3
          show ?case by simp
        next
          case Signature: 4
          show ?case
          proof(cases "bs \<in> ran (history s1)")
            case True
            thus ?thesis by simp
          next
            case False
            thus ?thesis
              apply simp
              apply(rule rel_spmf_bindI[where R="?rel'"])
               subgoal using step.prems I2_oracle
                 by(intro rel_spmf_restrict_relpI step.IH)
                   (simp_all add: pred_spmf_execp_resumption[OF I1'_\<tau>] pred_spmf_execp_resumption[OF decrypt])
              apply(clarsimp split: t.split option.split)
              apply(rule rel_spmf_bindI[where R="?rel'"])
               subgoal using step.prems I2_oracle
                 by(intro rel_spmf_restrict_relpI step.IH)
                   (simp_all add: pred_spmf_execp_resumption[OF I1'_\<tau>] pred_spmf_execp_resumption[OF decrypt])
              apply clarsimp
              done
          qed
        next
          case Vkey: 5
          show ?case by simp
        next
          case Pair: 6
          show ?case
            apply(rule rel_spmf_bindI[where R="?rel'"])
             subgoal using step.prems I2_oracle
               by(intro rel_spmf_restrict_relpI step.IH)
                 (simp_all add: pred_spmf_execp_resumption[OF I1'_\<tau>] pred_spmf_execp_resumption[OF decrypt])
            apply clarsimp
            apply(rule rel_spmf_bindI[where R="?rel'"])
             subgoal using step.prems I2_oracle
               by(intro rel_spmf_restrict_relpI step.IH)
                 (simp_all add: pred_spmf_execp_resumption[OF I1'_\<tau>] pred_spmf_execp_resumption[OF decrypt])
            apply clarsimp
            done
        next
          case String: 7
          show ?case by(simp add: option.rel_map)
        qed
      qed
      hence "rel_spmf ?rel'
        (exec_gpv (oracle\<^sub>n' b \<eta>) (lift_resumption (sim1_\<tau> \<eta> s1 bs)) (os1e, os1s))
        (exec_gpv (oracle\<^sub>n' True \<eta>) (lift_resumption (sim2_\<tau> \<eta> s2 bs)) (os2e, os2s))" for bs
        unfolding exec_gpv_lift_resumption
        by(intro rel_spmf_restrict_relpI pred_spmf_execp_resumption[OF I1'_\<tau>] pred_spmf_execp_resumption[OF _ I2_oracle] \<tau>2_only_decrypt_queries I2_start)+
          simp_all
      then show ?thesis using Input randomness rand_sim pnonce_public pnonce_dom
        apply(simp add: hybrid_step_def exec_gpv_bind sim1.gen_message_def sim2.gen_message_def map_gpv_id_bind_gpv)
        apply(rule rel_spmf_bind_reflI)
        apply(clarsimp simp add: exec_gpv_bind map_gpv_id_bind_gpv)
        apply(rule rel_spmf_bindI)
         apply assumption
        apply(clarsimp simp add: Let_def)
        apply(auto simp add: bisim12_def I2_def del: subsetI equalityI split: t.split nonce.split option.split elim!: rand_sim_mono)
        done
    qed
    moreover
    have invar_sim1'_alt: "invar_sim1' = pred_prod (pred_option (\<lambda>((S1, prot1, f1), s1).
        key_safe \<Gamma> (map (node_type_of_t \<Gamma>) f1) prot1 \<and>
        key_safe_sim1_invar \<eta> \<Gamma> f1 s1 \<and>
        wf_random_usage \<Gamma> f1 \<and>
        wf_history1 \<eta> (pnonce s1) (history s1)
      )) (\<lambda>_. True)"
      by(simp add: invar_sim1'_def fun_eq_iff invar_sim1_def split: option.split)
    have "pred_spmf invar_sim1' (exec_gpv (oracle\<^sub>n' b \<eta>) (hybrid_step_sim1 \<Gamma> \<A>' \<eta> hs1) os1)"
      unfolding invar_sim1'_alt using ks ks1 ru wfh
      by(intro pred_spmf_exec_gpv[where C="\<lambda>_. True"])
        (auto simp add: pred_gpv_def pred_option_def elim!: key_safe_sim1_hybrid_step key_safe_sim1_invar_hybrid_step wf_random_usage_sim1_hybrid_step wf_history1_hybrid_step)
    ultimately have "rel_spmf (bisim12' \<upharpoonleft> invar_sim1' \<otimes> (\<lambda>_. True)) (exec_gpv (oracle\<^sub>n' b \<eta>) (hybrid_step_sim1 \<Gamma> \<A>' \<eta> hs1) os1) (exec_gpv (oracle\<^sub>n' True \<eta>) (hybrid_step_sim2 \<Gamma> \<A>' \<eta> hs2) os2)"
      by(rule rel_spmf_restrict_relpI) simp }
  moreover from assms have "?Q ?init1 ?init2"
    by(auto intro!: bisim12I invar_sim1I pnonce_sigI rand_simI pnonce_publicI pnonce_domI history_ciphersI simp add: pnonce_enc_def)
  ultimately show ?thesis
    apply(intro iterate_spmf_parametric[THEN rel_funD, THEN rel_funD] rel_funI)
     apply(auto simp add: run_hybrid_step_def spmf_rel_map)
    apply(erule meta_allE)+
    apply(erule meta_impE)
    prefer 2
     apply(erule rel_spmf_mono)
     apply(auto simp add: invar_sim1'_def bisim12'_def split: option.splits)
    apply(case_tac af; case_tac ah)
       apply auto
    done
qed

lemma sim1_sim2_hybrid_trace:
  fixes prot :: "'id cosp_protocol" and \<A> :: "'s adversary"
  assumes "\<Gamma>, [] \<turnstile> prot \<surd>"
  shows "hybrid_trace1 prot b \<Gamma> \<A> bound \<eta> = hybrid_trace2 prot True \<Gamma> \<A> bound \<eta>"
proof -
  obtain \<A>' \<sigma> where \<A>: "\<A> \<eta> = (\<A>', \<sigma>)" by(cases "\<A> \<eta>")

  let ?I2 = "\<lambda>_ :: ('id, 's) cosp_hybrid_state \<times> state_oracle. True"
  let ?Q = "bisim12 \<upharpoonleft> invar_sim1 \<otimes> ?I2"
  let ?init1 = "((({}, prot, []), sim2.initial () \<sigma> (bound \<eta>)), (\<lambda>_. None, \<lambda>_. None))"
    and ?init2 = "((({}, prot, []), sim2.initial () \<sigma> (bound \<eta>)), (\<lambda>_. None, \<lambda>_. None))"

  from bisim_run_hybrid_trace_sim12[OF assms, of b \<A>' \<sigma> "bound \<eta>"]
  have "rel_spmf (list_all2 (rel_prod op = (\<lambda>_ _. True)))
    (full_hybrid_trace' (oracle\<^sub>n' b \<eta>) (sim1' \<Gamma> \<A>' \<eta>) ?init1)
    (full_hybrid_trace' (oracle\<^sub>n' True \<eta>) (sim2' \<Gamma> \<A>' \<eta>) ?init2)"
    (is "rel_spmf ?Q' _ _")
    unfolding full_hybrid_trace'_def
    by(auto simp add: spmf_rel_map list.rel_map[abs_def] bisim12_def elim!: spmf_rel_mono_strong list.rel_mono_strong)
  then have "rel_spmf ?Q'
      (full_hybrid_trace prot (oracle b) (sim1 \<Gamma> \<A> bound) \<eta>)
      (full_hybrid_trace prot (oracle True) (sim2 \<Gamma> \<A> bound) \<eta>)"
    using \<A> by(simp add: full_hybrid_trace_def sim1.sim_def sim2.sim_def oracle_def)
  thus ?thesis
    by(fold spmf_rel_eq list.rel_eq)
      (auto simp add: hybrid_trace_def spmf_rel_map list.rel_map[abs_def] elim!: spmf_rel_mono_strong list.rel_mono_strong)
qed

end

end

end