theory CoSP_PKES_Simulator_Gen imports
  CoSP_PKES_Serialisation
  CoSP_PKES_Parsing
  "$AFP/CryptHOL/GPV_Bisim"
begin

subsubsection \<open>Generic simulator definition\<close>

context comp_impl begin

definition record_sim :: "t \<Rightarrow> bitstring \<Rightarrow> ('more' simulator_state'_scheme, (bitstring \<times> 'more' simulator_state'_scheme, 'call, 'ret) gpv) stateT"
where "record_sim t bs = StateT (\<lambda>s. Done (bs, history_update (\<lambda>hist. hist(t \<mapsto> bs)) s))"

lemma record_sim_sel [simp]:
  "the_gpv (run_state (record_sim t bs) s) = return_spmf (Pure (bs, history_update (\<lambda>f. f(t \<mapsto> bs)) s))"
by(simp add: record_sim_def)

lemma results'_gpv_record_sim [simp]:
  "results'_gpv (run_state (record_sim t bs) s) = {(bs, history_update (\<lambda>f. f(t \<mapsto> bs)) s)}"
by(simp add: record_sim_def)

definition pnonce_spmf :: "security \<Rightarrow> protocol_nonce \<Rightarrow> (pnonce, (history, ((bitstring \<times> pnonce) \<times> history) spmf) stateT) stateT"
where
  "pnonce_spmf \<eta> n = StateT (\<lambda>pns. StateT (\<lambda>hist.
  (case pns n of None \<Rightarrow> do {
      bs \<leftarrow> spmf_of_pmf (sample_bitstring (size_nonce A \<eta>));
      let r = A_nonce A \<eta> bs;
      return_spmf ((r, pns(n \<mapsto> r)), hist)
    }
  | Some bs \<Rightarrow> return_spmf ((bs, pns), hist))))"

lemma pnonce_spmf_sample:
  "pns n = None \<Longrightarrow> 
  run_state (run_state (pnonce_spmf \<eta> n) pns) hist = 
  map_spmf (\<lambda>bs. let r = A_nonce A \<eta> bs in ((r, pns(n \<mapsto> r)), hist))
    (spmf_of_pmf (sample_bitstring (size_nonce A \<eta>)))"
by(clarsimp simp add: pnonce_spmf_def map_spmf_conv_bind_spmf Let_def)

lemma pnonce_spmf_found:
  "pns n = Some r \<Longrightarrow>
  run_state (run_state (pnonce_spmf \<eta> n) pns) hist = return_spmf ((r, pns), hist)"
by(simp add: pnonce_spmf_def)

lemma pnonce_spmf_sel:
  "run_state (run_state (pnonce_spmf \<eta> n) pns) hist =
   (if pns n = None then 
      map_spmf (\<lambda>bs. let r = A_nonce A \<eta> bs in ((r, pns(n \<mapsto> r)), hist))
            (spmf_of_pmf (sample_bitstring (size_nonce A \<eta>)))
     else return_spmf ((the (pns n), pns), hist))"
by(auto simp add: pnonce_spmf_def pnonce_spmf_sample pnonce_spmf_found Let_def o_def map_spmf_conv_bind_spmf[symmetric] id_def)

lemma in_set_pnonce_spmf:
  "((m, pns'), hist') \<in> set_spmf (run_state (run_state (pnonce_spmf \<eta> n) pns) hist) \<longleftrightarrow> 
   (if pns n = None then
     \<exists>r \<in> nlists UNIV (size_nonce A \<eta>). m = A_nonce A \<eta> r \<and> pns' = pns(n \<mapsto> m) \<and> hist' = hist
    else pns n = Some m \<and> pns' = pns \<and> hist' = hist)"
by(auto simp add: pnonce_spmf_sel Let_def set_pmf_spmf_of_set intro: rev_bexI)

text \<open> We update the set of randomness nonces whenever we sample a randomness nonce such
  that the store @{const pnonce} and the set @{const randomness} are always consistent. \<close>
definition pnonce_sim :: "security \<Rightarrow> protocol_nonce \<Rightarrow> ('more' simulator_state'_scheme, (bitstring \<times> 'more' simulator_state'_scheme, 'call, 'ret) gpv) stateT"
where "pnonce_sim \<eta> n = StateT (\<lambda>s.
  (case pnonce s n of None \<Rightarrow> do {
      bs \<leftarrow> lift_spmf (spmf_of_pmf (sample_bitstring (size_nonce A \<eta>)));
      let r = A_nonce A \<eta> bs;
      Done (r, pnonce_update (\<lambda>pns. pns(n \<mapsto> r)) s)
    }
  | Some bs \<Rightarrow> Done (bs, s)))"

lemma pnonce_sim_sel:
  "run_state (pnonce_sim \<eta> n) s =
   (if pnonce s n = None then 
      lift_spmf (map_spmf (\<lambda>bs. let r = A_nonce A \<eta> bs in (r, pnonce_update (\<lambda>pnonce. pnonce(n \<mapsto> r)) s))
            (spmf_of_pmf (sample_bitstring (size_nonce A \<eta>))))
     else Done (the (pnonce s n), s))"
by(auto simp add: pnonce_sim_def Let_def lift_map_spmf lift_bind_spmf o_def map_gpv_conv_bind id_def record_sim_def)

lemma pnonce_sim_history: "pnonce s n = Some m \<Longrightarrow> run_state (pnonce_sim \<eta> n) s = Done (m, s)"
by(simp add: pnonce_sim_def)

lemma fuel_pnonce_sim: "(bs, s') \<in> results'_gpv (run_state (pnonce_sim \<eta> n) s) \<Longrightarrow> fuel s' = fuel s"
by(auto simp add: pnonce_sim_def Let_def split: option.split_asm)

lemma randomness_pnonce_sim: "(bs, s') \<in> results'_gpv (run_state (pnonce_sim \<eta> n) s) \<Longrightarrow> randomness s' = randomness s"
by(auto simp add: pnonce_sim_def Let_def split: option.split_asm)

definition sim_control
  :: "'s adversary' \<Rightarrow> security \<Rightarrow> ('s, 'more') simulator_state_scheme \<Rightarrow> bitstring
  \<Rightarrow> ((bitstring \<times> ('s, 'more') simulator_state_scheme) option, 'call, 'ret) gpv"
where
  "sim_control \<A> \<eta> s m = do {
     (m', \<sigma>') \<leftarrow> lift_spmf (adversary'.control \<A> (adv_state s) m);
     Done (Some (m', s\<lparr>adv_state := \<sigma>'\<rparr>))
  }"

fun sim_info_sim
  :: "nonce_typing \<Rightarrow> 's adversary' \<Rightarrow> security \<Rightarrow> ('s, 'more') simulator_state_scheme \<Rightarrow> 'id
  \<Rightarrow> t option \<Rightarrow> (('s, 'more') simulator_state_scheme option, 'call, 'ret) gpv"
where
  "sim_info_sim \<Gamma> \<A> \<eta> s ident None = 
  (if 0 < fuel s then Done (Some (fuel_update (\<lambda>x. x - 1) s)) else Done None)"
| "sim_info_sim \<Gamma> \<A> \<eta> s ident (Some t) =
  (let l = len_of \<eta> t
   in if l < fuel s then Done (Some (
     (fuel_update (\<lambda>x. x - l - 1) \<circ> 
     (case t of N (PNonce n) \<Rightarrow>
         (* Is this a randomness node according to typing? --
            typing over-approximates the path-based classification, but only if the nonce
            is not used, as the node types are exclusive *)
         (* Is this update needed at all? *)
         (case \<Gamma> n of Some _ \<Rightarrow> randomness_update (insert n) | _ \<Rightarrow> id)
      | _ \<Rightarrow> id))
      s))
   else Done None)"

lemma randomness_sim_control_same:
  "Some (msg', \<sigma>') \<in> results'_gpv (sim_control \<A> \<eta> \<sigma> msg) \<Longrightarrow> randomness \<sigma>' = randomness \<sigma>"
by(clarsimp simp add: sim_control_def)

end


locale comp_impl_sim_gen = 
  -- \<open>Instantiate the type variable @{typ "'s"} from @{const comp_impl_\<beta>} with @{typ "('s, 'more) simulator_state_scheme"}
      and @{typ "'more"} from @{const comp_impl_parse} with @{typ "('s, 'more) simulator_state_ext"} and specialise
      the monad to @{typ "(bitstring \<times> ('s, 'more) simulator_state_scheme, _, _) gpv"}\<close>
  (* FIXME: cannot instantiate monad operations of comp_impl_\<beta> directly because the monad locale
     defines the map parameter, so we must do it separately with command sublocale below *)
  comp_impl A
  + comp_impl_parse A decryption
  for A :: implementation
  and pnonce :: "security \<Rightarrow> protocol_nonce \<Rightarrow> (('s, 'more) simulator_state_scheme, (bitstring \<times> ('s, 'more) simulator_state_scheme, 'call, 'ret) gpv) stateT"
  and "record" :: "t \<Rightarrow> bitstring \<Rightarrow> (('s, 'more) simulator_state_scheme, (bitstring \<times> ('s, 'more) simulator_state_scheme, 'call, 'ret) gpv) stateT"
  and encryption_key :: "security \<Rightarrow> protocol_nonce \<Rightarrow> (('s, 'more) simulator_state_scheme, (bitstring \<times> ('s, 'more) simulator_state_scheme, 'call, 'ret) gpv) stateT"
  and encryption :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> bitstring \<Rightarrow> protocol_nonce \<Rightarrow> (('s, 'more) simulator_state_scheme, (bitstring \<times> ('s, 'more) simulator_state_scheme, 'call, 'ret) gpv) stateT"
  and verification_key :: "security \<Rightarrow> protocol_nonce \<Rightarrow> (('s, 'more) simulator_state_scheme, (vkey \<times> ('s, 'more) simulator_state_scheme, 'call, 'ret) gpv) stateT"
  and signature :: "security \<Rightarrow> protocol_nonce \<Rightarrow> t \<Rightarrow> message \<Rightarrow> protocol_nonce \<Rightarrow> (('s, 'more) simulator_state_scheme, (signature \<times> ('s, 'more) simulator_state_scheme, 'call, 'ret) gpv) stateT"
  and decryption :: "nat \<Rightarrow> nat \<Rightarrow> bool list \<Rightarrow> ('s, 'more) simulator_state_scheme \<Rightarrow> (bool list option, 'call, 'ret) resumption"
begin

sublocale comp_impl_\<beta> A Done bind_gpv Fail results'_gpv pnonce "record" encryption_key encryption verification_key signature
by unfold_locales auto

(* FIXME: Adapt definition of simulator such that state is the last parameter *)
definition rec_message
  :: "'s adversary' \<Rightarrow> security \<Rightarrow> ('s, 'more) simulator_state_scheme \<Rightarrow> t
  \<Rightarrow> (('s, 'more) simulator_state_scheme option, 'call, 'ret) gpv"
where
  "rec_message \<A> \<eta> s t = do {
     x \<leftarrow> catch_gpv (run_state (\<beta> \<eta> t) s);
     case x of None \<Rightarrow> Done None | Some (m, s') \<Rightarrow> do {
       \<sigma>' \<leftarrow> lift_spmf (rec_output \<A> (adv_state s') m);
       Done (Some (s'\<lparr>adv_state := \<sigma>'\<rparr>))
     }
  }"

definition gen_message
  :: "'s adversary' \<Rightarrow> security \<Rightarrow> ('s, 'more) simulator_state_scheme \<Rightarrow> ((t \<times> ('s, 'more) simulator_state_scheme) option, 'call, 'ret) gpv"
where
  "gen_message \<A> \<eta> s = do {
    (bs, \<sigma>') \<leftarrow> lift_spmf (gen_input \<A> (adv_state s));
    map_gpv (map_option (\<lambda>t. (t, s\<lparr>adv_state := \<sigma>'\<rparr>))) id (catch_gpv (lift_resumption (\<tau> \<eta> s bs)))
  }"

definition sim' :: "nonce_typing \<Rightarrow> 's adversary' \<Rightarrow> security \<Rightarrow> (('s, 'more) simulator_state_scheme, 'id, t, 'call, 'ret) simulator'"
where "sim' \<Gamma> \<A> \<eta> = 
  \<lparr> simulator'.gen_message = gen_message \<A> \<eta>,
    simulator'.rec_message = rec_message \<A> \<eta>, 
    control = sim_control \<A> \<eta>, 
    info_sim = sim_info_sim \<Gamma> \<A> \<eta>
  \<rparr>"

lemma sim'_sel [simp]:
  shows gen_message_sim': "simulator'.gen_message (sim' \<Gamma> \<A> \<eta>) = gen_message \<A> \<eta>"
  and rec_message_sim': "simulator'.rec_message (sim' \<Gamma> \<A> \<eta>) = rec_message \<A> \<eta>"
  and control_sim': "control (sim' \<Gamma> \<A> \<eta>) = sim_control \<A> \<eta>"
  and info_sim_sim': "info_sim (sim' \<Gamma> \<A> \<eta>) = sim_info_sim \<Gamma> \<A> \<eta>"
by(simp_all add: sim'_def)

definition initial :: "'more \<Rightarrow> 's \<Rightarrow> nat \<Rightarrow> ('s, 'more) simulator_state_scheme"
where
  "\<And>more. initial more \<sigma> p =
  \<lparr>fuel = p, pnonce = Map.empty, randomness = {}, history = Map.empty, adv_state = \<sigma>, \<dots> = more\<rparr>"

lemma initial_sel [simp]:
  shows fuel_initial: "\<And>more. fuel (initial more \<sigma> p) = p"
  and pnonce_initial: "\<And>more. simulator_state'.pnonce (initial more \<sigma> p) = Map.empty"
  and randomness_initial: "\<And>more. randomness (initial more \<sigma> p) = {}"
  and history_initial: "\<And>more. history (initial more \<sigma> p) = Map.empty"
  and adv_state_initial: "\<And>more. adv_state (initial more \<sigma> p) = \<sigma>"
  and more_initial: "\<And>more. simulator_state.more (initial more \<sigma> p) = more"
by(simp_all add: initial_def)

definition sim
  :: "'more \<Rightarrow> nonce_typing \<Rightarrow> 's adversary \<Rightarrow> (security \<Rightarrow> nat)
  \<Rightarrow> (('s, 'more) simulator_state_scheme, 'id, t, 'call, 'ret) simulator"
where
  "\<And>more. sim more \<Gamma> \<A> bound \<eta> =
  (let (\<A>', \<sigma>) = \<A> \<eta> in (sim' \<Gamma> \<A>' \<eta>, initial more \<sigma> (bound \<eta>)))"

lemma fuel_\<beta>_same:
  assumes "(bs, s') \<in> results'_gpv (run_state (\<beta> \<eta> t) s)"
  and [dest!]: "\<And>bs s' s n. (bs, s') \<in> results'_gpv (run_state (pnonce \<eta> n) s) \<Longrightarrow> fuel s' = fuel s"
  and [dest!]: "\<And>t1 t2 t3 bs bs' s s'. (bs', s') \<in> results'_gpv (run_state (record t1 bs) s) \<Longrightarrow> fuel s' = fuel s"
  and [dest!]: "\<And>bs s' s n. (bs, s') \<in> results'_gpv (run_state (encryption_key \<eta> n) s) \<Longrightarrow> fuel s' = fuel s"
  and [dest!]: "\<And>bs s' s n' t m2 n. (bs, s') \<in> results'_gpv (run_state (encryption \<eta> n' t m2 n) s) \<Longrightarrow> fuel s' = fuel s"
  and [dest!]: "\<And>bs s' s n. (bs, s') \<in> results'_gpv (run_state (verification_key \<eta> n) s) \<Longrightarrow> fuel s' = fuel s"
  and [dest!]: "\<And>bs s' s n' t m2 n. (bs, s') \<in> results'_gpv (run_state (signature \<eta> n' t m2 n) s) \<Longrightarrow> fuel s' = fuel s"
  shows "fuel s' = fuel s"
apply(rule sym) using assms(1)
apply(induction rule: \<beta>_invariant_p)
apply(auto intro: transp_inv_image[where f="fuel"] transp_equality)
done

lemma randomness_\<beta>_same:
  assumes "(bs, s') \<in> results'_gpv (run_state (\<beta> \<eta> t) s)"
  and [dest!]: "\<And>bs s' s n. (bs, s') \<in> results'_gpv (run_state (pnonce \<eta> n) s) \<Longrightarrow> randomness s' = randomness s"
  and [dest!]: "\<And>t1 t2 t3 bs bs' s s'. (bs', s') \<in> results'_gpv (run_state (record t1 bs) s) \<Longrightarrow> randomness s' = randomness s"
  and [dest!]: "\<And>bs s' s n. (bs, s') \<in> results'_gpv (run_state (encryption_key \<eta> n) s) \<Longrightarrow> randomness s' = randomness s"
  and [dest!]: "\<And>bs s' s n' t m2 n. (bs, s') \<in> results'_gpv (run_state (encryption \<eta> n' t m2 n) s) \<Longrightarrow> randomness s' = randomness s"
  and [dest!]: "\<And>bs s' s n. (bs, s') \<in> results'_gpv (run_state (verification_key \<eta> n) s) \<Longrightarrow> randomness s' = randomness s"
  and [dest!]: "\<And>bs s' s n' t m2 n. (bs, s') \<in> results'_gpv (run_state (signature \<eta> n' t m2 n) s) \<Longrightarrow> randomness s' = randomness s"
  shows "randomness s' = randomness s"
apply(rule sym) using assms(1)
apply(induction rule: \<beta>_invariant_p)
apply(auto intro: transp_inv_image[where f="randomness"] transp_equality)
done

end

locale comp_impl_sims = comp_impl A
  for A :: implementation

end