theory IND_CCA2_SUF_CMA imports
  "$AFP/Game_Based_Crypto/IND_CCA2"
  "$AFP/Game_Based_Crypto/SUF_CMA"
begin

lemma rel_option_into_ord_option: "rel_option A x y \<Longrightarrow> ord_option A x y"
by(cases x; cases y; simp)

context includes lifting_syntax begin
lemma pmf_parametric [transfer_rule]: (* Move *)
  assumes "bi_unique A"
  shows "(rel_pmf A ===> A ===> op =) pmf pmf"
proof(rule rel_funI antisym)+
  fix p q x y
  assume pq: "rel_pmf A p q" and xy: "A x y "
  have "pmf p x \<le> measure_pmf.prob q {y. A x y}"
    using rel_pmf_measureD[where A="{x}", OF pq] by(simp add: pmf.rep_eq)
  also have "{y. A x y} = {y}" using xy by(auto dest: bi_uniqueDr[OF assms])
  finally show "pmf p x \<le> pmf q y" by(simp add: pmf.rep_eq)
  
  from pq have "rel_pmf A\<inverse>\<inverse> q p" by(simp add: pmf.rel_conversep)
  from rel_pmf_measureD[where A="{y}", OF this]
  have "pmf q y \<le> measure_pmf.prob p {x. A x y}" by(simp add: pmf.rep_eq)
  also have "{x. A x y} = {x}" using xy by(auto dest: bi_uniqueDl[OF assms])
  finally show "pmf q y \<le> pmf p x" by(simp add: pmf.rep_eq)
qed

lemma spmf_parametric: (* Move *)
  assumes [transfer_rule]: "bi_unique A"
  shows "(rel_spmf A ===> A ===> op =) spmf spmf"
by transfer_prover

lemma bind_spmf_parametric_ord_spmf: "(rel_spmf A ===> (A ===> ord_spmf B) ===> ord_spmf B) bind_spmf bind_spmf"
proof -
  have [transfer_rule]: "(rel_spmf B ===> (A ===> ord_spmf B) ===> rel_option A ===> ord_spmf B) case_option case_option"
    by(auto simp add: rel_fun_def rel_option_into_ord_option split: option.split elim: pmf.rel_mono_strong)
  show ?thesis unfolding bind_spmf_def[abs_def] by transfer_prover
qed

lemma ord_spmf_bindI':
  "\<lbrakk> rel_spmf A p q; \<And>a b. A a b \<Longrightarrow> ord_spmf B (f a) (g b) \<rbrakk> \<Longrightarrow> ord_spmf B (p \<bind> f) (q \<bind> g)"
using bind_spmf_parametric_ord_spmf[of A B] by(simp add: rel_fun_def)

end

lemma map_sum_eq_Inl [simp]: "map_sum f g x = Inl y \<longleftrightarrow> (\<exists>x'. x = Inl x' \<and> f x' = y)" (* Move *)
by(cases x) simp_all

lemma map_sum_eq_Inr [simp]: "map_sum f g x = Inr y \<longleftrightarrow> (\<exists>x'. x = Inr x' \<and> g x' = y)" (* Move *)
by(cases x) simp_all

lemma notin_range_Inr: "x \<notin> range Inr \<longleftrightarrow> x \<in> range Inl" (* Move *)
by(cases x) auto

lemma notin_range_Inl: "x \<notin> range Inl \<longleftrightarrow> x \<in> range Inr" (* Move *)
by(cases x) auto

lemma vimage2p_apply: "BNF_Def.vimage2p f g R x y \<longleftrightarrow> R (f x) (g y)"
by(simp add: vimage2p_def)

definition rel_prod_swap :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> ('c \<Rightarrow> 'd \<Rightarrow> bool) \<Rightarrow> 'a \<times> 'c \<Rightarrow> 'd \<times> 'b \<Rightarrow> bool"
where "rel_prod_swap P Q = BNF_Def.vimage2p (\<lambda>x. x) (\<lambda>(d, b). (b, d)) (rel_prod P Q)"

lemma rel_prod_swap_apply [code, simp]: "rel_prod_swap R1 R2 (a, b) (c, d) \<longleftrightarrow> R1 a d \<and> R2 b c"
by(simp add: rel_prod_swap_def vimage2p_apply)

subsection \<open>Combination of IND-CCA2 and SUF-CMA game\<close>

locale ind_cca2_suf_cma =
  ind_cca2: ind_cca2 key_gen_enc encrypt decrypt valid_plain +
  suf_cma: suf_cma key_gen_sig sign verify valid_message
  for key_gen_enc :: "security \<Rightarrow> ('ekey \<times> 'dkey) spmf" -- {*  probabilistic *}
  and encrypt :: "security \<Rightarrow> 'ekey \<Rightarrow> 'plain \<Rightarrow> 'cipher spmf"  -- {*  probabilistic *}
  and decrypt :: "security \<Rightarrow> 'dkey \<Rightarrow> 'cipher \<Rightarrow> 'plain option" -- {*  deterministic, but not used *}
  and valid_plain :: "security \<Rightarrow> 'plain \<Rightarrow> bool" -- {* checks whether a plain text is valid, i.e., has the right format *}
  and key_gen_sig :: "security \<Rightarrow> ('vkey \<times> 'sigkey) spmf"
  and sign :: "security \<Rightarrow> 'sigkey \<Rightarrow> 'message \<Rightarrow> 'signature spmf"
  and verify :: "security \<Rightarrow> 'vkey \<Rightarrow> 'message \<Rightarrow> 'signature \<Rightarrow> bool" -- \<open>verification is deterministic\<close>
  and valid_message :: "security \<Rightarrow> 'message \<Rightarrow> bool"
begin

subsubsection \<open>Single-user setting\<close>

text \<open>We only do the type definitions for the single-user setting\<close>

type_synonym ('ekey', 'dkey', 'cipher', 'vkey', 'sigkey', 'message', 'signature') state_oracle\<^sub>1
  = "('ekey', 'dkey', 'cipher') ind_cca2.state_oracle \<times> ('vkey', 'sigkey', 'message', 'signature') suf_cma.state_oracle"

type_synonym ('plain', 'cipher', 'message') call\<^sub>1 
  = "('plain', 'cipher') ind_cca2.call\<^sub>1 + 'message' suf_cma.call\<^sub>1"
type_synonym ('ekey', 'plain', 'cipher', 'vkey', 'signature') ret
  = "('ekey', 'plain', 'cipher') ind_cca2.ret\<^sub>1 + ('vkey', 'signature') suf_cma.ret\<^sub>1"

definition oracle\<^sub>1 :: "bool \<Rightarrow> security \<Rightarrow>
  (('ekey, 'dkey, 'cipher, 'vkey, 'sigkey, 'message, 'signature) state_oracle\<^sub>1,
   ('plain, 'cipher, 'message) call\<^sub>1,
   ('ekey, 'plain, 'cipher, 'vkey, 'signature) ret) oracle'"
where "oracle\<^sub>1 b \<eta> = parallel_oracle (ind_cca2.oracle\<^sub>1 b \<eta>) (suf_cma.oracle\<^sub>1 \<eta>)"

subsubsection \<open>Multi-user setting\<close>

type_synonym ('i, 'ekey', 'dkey', 'cipher', 'j, 'vkey', 'sigkey', 'message', 'signature') state_oracle\<^sub>n
  = "('i \<Rightarrow> ('ekey', 'dkey', 'cipher') ind_cca2.state_oracle) \<times> ('j \<Rightarrow> ('vkey', 'sigkey', 'message', 'signature') suf_cma.state_oracle)"

type_synonym ('i, 'plain', 'cipher', 'j, 'message') call\<^sub>n
  = "'i \<times> ('plain', 'cipher') ind_cca2.call\<^sub>1 + 'j \<times> 'message' suf_cma.call\<^sub>1"
-- \<open>@{typ "('ekey', 'plain', 'cipher', 'vkey', 'signature') ret"} works for both settings\<close>

definition oracle\<^sub>n :: "bool \<Rightarrow> security \<Rightarrow>
  (('i, 'ekey, 'dkey, 'cipher, 'j, 'vkey, 'sigkey, 'message, 'signature) state_oracle\<^sub>n,
   ('i, 'plain, 'cipher, 'j, 'message) call\<^sub>n,
   ('ekey, 'plain, 'cipher, 'vkey, 'signature) ret) oracle'"
where "oracle\<^sub>n b \<eta> = parallel_oracle (ind_cca2.oracle\<^sub>n b \<eta>) (suf_cma.oracle\<^sub>n \<eta>)"

lemma oracle\<^sub>n_simps [simp]:
  "oracle\<^sub>n b \<eta> (s1, s2) (Inl x) = map_spmf (map_prod Inl (\<lambda>s1'. (s1', s2))) (ind_cca2.oracle\<^sub>n b \<eta> s1 x)"
  "oracle\<^sub>n b \<eta> (s1, s2) (Inr y) = map_spmf (map_prod Inr (\<lambda>s2'. (s1, s2'))) (suf_cma.oracle\<^sub>n \<eta> s2 y)"
by(simp_all add: oracle\<^sub>n_def)

type_synonym ('i, 'ekey', 'plain', 'cipher', 'j, 'vkey', 'message', 'signature') adversary\<^sub>n' =
  "(bool + 'j \<times> 'message' \<times> 'signature', 
    ('i, 'plain', 'cipher', 'j, 'message') call\<^sub>n,
    ('ekey', 'plain', 'cipher', 'vkey', 'signature') ret) gpv"
type_synonym ('i, 'ekey', 'plain', 'cipher', 'j, 'vkey', 'message', 'signature') adversary\<^sub>n =
  "security \<Rightarrow> ('i, 'ekey', 'plain', 'cipher', 'j, 'vkey', 'message', 'signature') adversary\<^sub>n'"

definition ind_cca2_suf_cma\<^sub>n
  :: "('i, 'ekey, 'plain, 'cipher, 'j, 'vkey, 'message, 'signature) adversary\<^sub>n
  \<Rightarrow> security \<Rightarrow> (bool + bool) spmf"
where
  "\<And>log. 
   ind_cca2_suf_cma\<^sub>n \<A> \<eta> = do {
      b \<leftarrow> coin_spmf;
      (x, (\<sigma>1, \<sigma>2)) \<leftarrow> exec_gpv (oracle\<^sub>n b \<eta>) (\<A> \<eta>) (\<lambda>_. None, \<lambda>_. None);
      return_spmf (map_sum 
        (\<lambda>guess. guess = b)
        (\<lambda>(j, m, sig). case \<sigma>2 j of None \<Rightarrow> False
           | Some (vkey, skey, log) \<Rightarrow> verify \<eta> vkey m sig \<and> (m, sig) \<notin> set log)
        x)
  }"

definition advantage\<^sub>n
  :: "('i, 'ekey, 'plain, 'cipher, 'j, 'vkey, 'message, 'signature) adversary\<^sub>n
  \<Rightarrow> security \<Rightarrow> real"
where
  "advantage\<^sub>n \<A> \<eta> = max
    \<bar>spmf (try_spmf (cond_spmf (ind_cca2_suf_cma\<^sub>n \<A> \<eta>) (range Inl)) (map_spmf Inl coin_spmf)) (Inl True) - 1 / 2\<bar>
    (spmf (try_spmf (cond_spmf (ind_cca2_suf_cma\<^sub>n \<A> \<eta>) (range Inr)) (return_pmf None)) (Inr True))"
text \<open>Using @{const try_spmf} is problematic here, because if @{term "\<A>"} fails to produce an
  answer, then we still get a probability distribution in the encryption case.\<close>

lemma advantage\<^sub>n_nonneg: "advantage\<^sub>n \<A> \<eta> \<ge> 0"
by(simp add: advantage\<^sub>n_def pmf_nonneg)


abbreviation secure_for\<^sub>n
  :: "('i, 'ekey, 'plain, 'cipher, 'j, 'vkey, 'message, 'signature) adversary\<^sub>n \<Rightarrow> bool"
where "secure_for\<^sub>n \<A> \<equiv> negligible (advantage\<^sub>n \<A>)"

definition red_ind_cca2
  :: "bool \<Rightarrow> ('i, 'ekey, 'plain, 'cipher, 'j, 'vkey, 'message, 'signature) adversary\<^sub>n
  \<Rightarrow> ('i, 'ekey, 'plain, 'cipher) ind_cca2.adversary\<^sub>n"
  -- \<open>The additional boolean flag indicates whether the response of the adversary should be negated.
    This is necessary for the reduction to work in case the adversary aims at guessing wrong.\<close>
where
  "red_ind_cca2 b \<A> \<eta> = bind_gpv
    (inline (\<lambda>\<sigma> q. case q of Inl x \<Rightarrow> Pause x (\<lambda>r. Done (Inl r, \<sigma>))
                   | Inr y \<Rightarrow> lift_spmf (map_spmf (apfst Inr) (suf_cma.oracle\<^sub>n \<eta> \<sigma> y))) (\<A> \<eta>) (\<lambda>_. None))
    (\<lambda>(x, \<sigma>). case x of Inl b' \<Rightarrow> Done (b = b') | Inr _ \<Rightarrow> lift_spmf coin_spmf)"

definition red_suf_cma
  :: "('i, 'ekey, 'plain, 'cipher, 'j, 'vkey, 'message, 'signature) adversary\<^sub>n
  \<Rightarrow> ('j, 'vkey, 'message, 'signature) suf_cma.adversary\<^sub>n"
where
  "red_suf_cma \<A> \<eta> = bind_gpv
    (lift_spmf coin_spmf)
    (\<lambda>b. bind_gpv
    (inline (\<lambda>\<sigma> q. case q of Inr y \<Rightarrow> Pause y (\<lambda>r. Done (Inr r, \<sigma>))
                   | Inl y \<Rightarrow> lift_spmf (map_spmf (apfst Inl) (ind_cca2.oracle\<^sub>n b \<eta> \<sigma> y))) (\<A> \<eta>) (\<lambda>_. None))
    (\<lambda>(x, \<sigma>). case x of Inr msig \<Rightarrow> Done msig | Inl _ \<Rightarrow> undefined))"
(*
lemma "advantage\<^sub>n \<A> \<eta> \<le> max (max (ind_cca2.advantage\<^sub>n (red_ind_cca2 True \<A>) \<eta>) (ind_cca2.advantage\<^sub>n (red_ind_cca2 False \<A>) \<eta>)) (suf_cma.advantage\<^sub>n (red_suf_cma \<A>) \<eta>)"
proof -
  have "spmf (ind_cca2_suf_cma\<^sub>n \<A> \<eta>) (Inl True) = spmf (ind_cca2_suf_cma\<^sub>n \<A> \<eta> \<upharpoonleft> range Inl) (Inl True)"
    by simp
  also let ?game1 = "(do { 
    b \<leftarrow> coin_spmf;
    (x, _) \<leftarrow> exec_gpv (oracle\<^sub>n b \<eta>) (\<A> \<eta>) (Map.empty, Map.empty);
    _ :: unit \<leftarrow> assert_spmf (isl x);
    return_spmf (projl x = b) })"
  { let ?R = "\<lambda>x y. isl x \<and> projl x = y"
    have "bi_unique ?R" unfolding bi_unique_def by(auto simp add: sum.isl_def)
    moreover have "rel_spmf ?R (ind_cca2_suf_cma\<^sub>n \<A> \<eta> \<upharpoonleft> range Inl) ?game1"
      by(auto simp del: spmf_restrict_spmf_inside simp add: ind_cca2_suf_cma\<^sub>n_def restrict_bind_spmf split_beta restrict_return_spmf notin_range_Inl intro!: rel_spmf_bind_reflI)
    ultimately have "spmf (ind_cca2_suf_cma\<^sub>n \<A> \<eta> \<upharpoonleft> range Inl) (Inl True) = spmf ?game1 True"
      by(rule spmf_parametric[THEN rel_funD, THEN rel_funD]) simp }
  also have "ord_spmf op = ?game1 (ind_cca2.ind_cca2\<^sub>n (red_ind_cca2 True \<A>) \<eta>)"
    apply(simp add: ind_cca2_suf_cma\<^sub>n_def ind_cca2.ind_cca2\<^sub>n_def red_ind_cca2_def exec_gpv_bind exec_gpv_inline bind_map_spmf)
    apply(rule ord_spmf_bind_reflI)
    apply(rule ord_spmf_bindI'[where A="rel_prod op = (rel_prod_swap op = op =)"])
     prefer 2
     apply(clarsimp split: sum.split)
    apply(rule exec_gpv_parametric[THEN rel_funD, THEN rel_funD, THEN rel_funD, where ?CALL3="op ="]; simp add: gpv.rel_eq)
    apply(rule rel_funI)+
    apply(clarsimp)
    subgoal for b s1 s2 query
      by(cases query)(auto simp add: spmf_rel_map map_spmf_conv_bind_spmf[symmetric] split_def intro: rel_spmf_reflI)
    done
  hence "spmf ?game1 True \<le> spmf (ind_cca2.ind_cca2\<^sub>n (red_ind_cca2 True \<A>) \<eta>) True" by(rule ord_spmf_eq_leD)
  finally have "\<bar>spmf (ind_cca2_suf_cma\<^sub>n \<A> \<eta>) (Inl True) - 1 / 2\<bar> \<le> \<bar>spmf (local.ind_cca2.ind_cca2\<^sub>n (red_ind_cca2 True \<A>) \<eta>) True - 1 / 2\<bar>" sorry

  have "\<bar>spmf (ind_cca2_suf_cma\<^sub>n \<A> \<eta>) (Inl True) - 1 / 2\<bar> \<le>
    max \<bar>spmf (local.ind_cca2.ind_cca2\<^sub>n (red_ind_cca2 True \<A>) \<eta>) True - 1 / 2\<bar>
        \<bar>spmf (local.ind_cca2.ind_cca2\<^sub>n (red_ind_cca2 False \<A>) \<eta>) True - 1 / 2\<bar>"
  proof(cases "spmf (ind_cca2_suf_cma\<^sub>n \<A> \<eta>) (Inl True) \<le> 1 / 2")
    case False
    have "spmf (ind_cca2_suf_cma\<^sub>n \<A> \<eta>) (Inl True) = spmf (ind_cca2_suf_cma\<^sub>n \<A> \<eta> \<upharpoonleft> range Inl) (Inl True)"
      by simp
    also let ?game1 = "(do { 
      b \<leftarrow> coin_spmf;
      (x, _) \<leftarrow> exec_gpv (oracle\<^sub>n b \<eta>) (\<A> \<eta>) (Map.empty, Map.empty);
      _ :: unit \<leftarrow> assert_spmf (isl x);
      return_spmf (projl x = b) })"
    { let ?R = "\<lambda>x y. isl x \<and> projl x = y"
      have "bi_unique ?R" unfolding bi_unique_def by(auto simp add: sum.isl_def)
      moreover have "rel_spmf ?R (ind_cca2_suf_cma\<^sub>n \<A> \<eta> \<upharpoonleft> range Inl) ?game1"
        by(auto simp del: spmf_restrict_spmf_inside simp add: ind_cca2_suf_cma\<^sub>n_def restrict_bind_spmf split_beta restrict_return_spmf notin_range_Inl intro!: rel_spmf_bind_reflI)
      ultimately have "spmf (ind_cca2_suf_cma\<^sub>n \<A> \<eta> \<upharpoonleft> range Inl) (Inl True) = spmf ?game1 True"
        by(rule spmf_parametric[THEN rel_funD, THEN rel_funD]) simp }
    also have "ord_spmf op = ?game1 (ind_cca2.ind_cca2\<^sub>n (red_ind_cca2 True \<A>) \<eta>)"
      apply(simp add: ind_cca2_suf_cma\<^sub>n_def ind_cca2.ind_cca2\<^sub>n_def red_ind_cca2_def exec_gpv_bind exec_gpv_inline bind_map_spmf)
      apply(rule ord_spmf_bind_reflI)
      apply(rule ord_spmf_bindI'[where A="rel_prod op = (rel_prod_swap op = op =)"])
       prefer 2
       apply(clarsimp split: sum.split)
      apply(rule exec_gpv_parametric[THEN rel_funD, THEN rel_funD, THEN rel_funD, where ?CALL3="op ="]; simp add: gpv.rel_eq)
      apply(rule rel_funI)+
      apply(clarsimp)
      subgoal for b s1 s2 query
        by(cases query)(auto simp add: spmf_rel_map map_spmf_conv_bind_spmf[symmetric] split_def intro: rel_spmf_reflI)
      done
    hence "spmf ?game1 True \<le> spmf (ind_cca2.ind_cca2\<^sub>n (red_ind_cca2 True \<A>) \<eta>) True" by(rule ord_spmf_eq_leD)
    finally have "\<bar>spmf (ind_cca2_suf_cma\<^sub>n \<A> \<eta>) (Inl True) - 1 / 2\<bar> \<le> \<bar>spmf (local.ind_cca2.ind_cca2\<^sub>n (red_ind_cca2 True \<A>) \<eta>) True - 1 / 2\<bar>"
      using False by simp          
    thus ?thesis by simp
  next
    case True
    
    

  show ?thesis
    apply(simp add: advantage\<^sub>n_def ind_cca2.advantage\<^sub>n_def)
*)

end


end