theory LList_Space imports
  "$AFP/Coinductive/Coinductive_Stream"
  "$AFP/CryptHOL/Misc_CryptHOL"
begin

lemma restrict_space_space_same [simp]: "restrict_space N (space N) = N"
using sets.space_closed[of N]
by(intro measure_eqI)(auto simp add: sets_restrict_space emeasure_restrict_space)

lemma pred_disjCI:
  assumes Q [measurable]: "Measurable.pred M (\<lambda>x. Q x)"
  and P: "Measurable.pred (restrict_space M {x \<in> space M. \<not> Q x}) (\<lambda>x. P x)"
  shows "Measurable.pred M (\<lambda>x. P x \<or> Q x)"
proof -
  have "{x \<in> space M. Q x} \<in> sets M" using Q by simp
  moreover
  have "{x \<in> space (restrict_space M {x \<in> space M. \<not> Q x}). P x} \<in> sets (restrict_space M {x \<in> space M. \<not> Q x})"
    using P by(simp add: pred_def)
  then obtain A where A: "A \<inter> {x \<in> space M. \<not> Q x} = {x \<in> space (restrict_space M {x \<in> space M. \<not> Q x}). P x}"
    and "A \<in> sets M" by(auto simp add: sets_restrict_space)
  have "A \<inter> {x \<in> space M. \<not> Q x} \<in> sets M" by(intro sets.Int sets.sets_Collect_neg) fact+
  ultimately have "{x \<in> space M. Q x} \<union> A \<inter> {x \<in> space M. \<not> Q x} \<in> sets M" by(rule sets.Un)
  also have "{x \<in> space M. Q x} \<union> A \<inter> {x \<in> space M. \<not> Q x} = {x \<in> space M. P x \<or> Q x}"
    using A by(auto simp add: space_restrict_space)
  finally show ?thesis by simp
qed

lemma pred_disjCI2:
  "\<lbrakk> Measurable.pred M (\<lambda>x. Q x);
     Measurable.pred (restrict_space M {x \<in> space M. \<not> Q x}) (\<lambda>x. P x) \<rbrakk>
  \<Longrightarrow> Measurable.pred M (\<lambda>x. Q x \<or> P x)"
by(subst disj_commute)(rule pred_disjCI)

lemma pred_conj_cong1:
  assumes [measurable]: "Measurable.pred M (\<lambda>x. P x)"
  and Q: "Measurable.pred (restrict_space M {x \<in> space M. P x}) (\<lambda>x. Q x)"
  shows "Measurable.pred M (\<lambda>x. P x \<and> Q x)"
proof -
  have "{x \<in> space M. P x} \<in> sets M" using Q by simp
  moreover
  have "{x \<in> space (restrict_space M {x \<in> space M. P x}). Q x} \<in> sets (restrict_space M {x \<in> space M. P x})"
    using Q by(simp add: pred_def)
  then obtain A where A: "A \<inter> {x \<in> space M. P x} = {x \<in> space (restrict_space M {x \<in> space M. P x}). Q x}"
    and "A \<in> sets M" by(auto simp add: sets_restrict_space)
  have "A \<inter> {x \<in> space M. P x} \<in> sets M" by(intro sets.Int sets.sets_Collect_neg) fact+
  ultimately have "{x \<in> space M. P x} \<inter> (A \<inter> {x \<in> space M. P x}) \<in> sets M" by(rule sets.Int)
  also have "{x \<in> space M. P x} \<inter> (A \<inter> {x \<in> space M. P x}) = {x \<in> space M. P x \<and> Q x}"
    using A by(auto simp add: space_restrict_space)
  finally show ?thesis by simp
qed

lemma pred_conj_cong2:
  "\<lbrakk> Measurable.pred M (\<lambda>x. P x); 
     Measurable.pred (restrict_space M {x \<in> space M. P x}) (\<lambda>x. Q x) \<rbrakk>
  \<Longrightarrow> Measurable.pred M (\<lambda>x. Q x \<and> P x)"
by(subst conj_commute)(rule pred_conj_cong1)


(* Move to Coinductive_List *)
coinductive_set llists :: "'a set \<Rightarrow> 'a llist set"
for A :: "'a set"
where
  LCons: "\<lbrakk> x \<in> A; xs \<in> llists A \<rbrakk> \<Longrightarrow> LCons x xs \<in> llists A"
| LNil [intro!, simp]: "LNil \<in> llists A"

lemma in_llistsI: "lset xs \<subseteq> A \<Longrightarrow> xs \<in> llists A"
by(coinduction arbitrary: xs)(auto simp add: neq_LNil_conv)

inductive_simps llists_LCons [iff]: "LCons x xs \<in> llists A"

lemma in_llistsE:
  assumes "xs \<in> llists A"
  obtains "lset xs \<subseteq> A"
proof(rule that subsetI)+
  fix x
  assume "x \<in> lset xs"
  thus "x \<in> A" using \<open>xs \<in> llists A\<close>
    by induction auto
qed

lemma llists_alt_def: "llists A = {xs. lset xs \<subseteq> A}"
by(auto intro: in_llistsI elim: in_llistsE)

lemma llists_empty [simp]: "llists {} = {LNil}"
by(auto simp add: llists_alt_def)

lemma llists_mono: "A \<subseteq> B \<Longrightarrow> llists A \<subseteq> llists B"
by(auto simp add: llists_alt_def)

lemma llists_ltlI: "xs \<in> llists A \<Longrightarrow> ltl xs \<in> llists A"
by(auto simp add: llists_alt_def dest: in_lset_ltlD)

lemma in_llists_lnthD:
  "\<lbrakk> xs \<in> llists A; enat i < llength xs \<rbrakk> \<Longrightarrow> lnth xs i \<in> A"
by(auto simp add: llists_alt_def lset_conv_lnth)

lemma llists_UNIV [simp]: "llists UNIV = UNIV"
by(simp add: llists_alt_def)

lemma in_llists_UNIV[intro!]: "xs \<in> llists UNIV"
by simp

lemma llists_lhdD: "\<lbrakk> xs \<in> llists M; xs \<noteq> LNil \<rbrakk> \<Longrightarrow> lhd xs \<in> M"
by(auto simp add: llists_alt_def neq_LNil_conv)

lemma llists_expand: "(xs \<noteq> LNil \<Longrightarrow> lhd xs \<in> M \<and> ltl xs \<in> llists M) \<Longrightarrow> xs \<in> llists M"
by(cases xs)(auto simp add: llists_alt_def)
(* Move end *)



definition llist_cylinder :: "'a set \<Rightarrow> 'a llist set \<Rightarrow> 'a llist set"
where "llist_cylinder X A = { LCons s x | s x. s \<in> X \<and> x \<in> A}"

inductive_set llist_cylinders :: "'a set \<Rightarrow> 'a set set \<Rightarrow> 'a llist set set"
  for \<Omega> :: "'a set" and \<A> :: "'a set set"
where
  base: "llists \<Omega> \<in> llist_cylinders \<Omega> \<A>"
| step:
  "\<lbrakk> X \<in> \<A>; A \<in> llist_cylinders \<Omega> \<A> \<rbrakk>
  \<Longrightarrow> llist_cylinder X A \<in> llist_cylinders \<Omega> \<A>"

hide_fact (open) base step

lemma list_cylinders_mono:
  assumes "A \<in> llist_cylinders \<Omega> \<A>" "\<A> \<subseteq> \<B>"
  shows "A \<in> llist_cylinders \<Omega> \<B>"
using assms by induct(auto intro: llist_cylinders.intros)

lemma [simp]:
  shows llist_cylinder_empty1: "llist_cylinder {} A = {}"
  and llist_cylinder_empty2: "llist_cylinder A {} = {}"
by(auto simp: llist_cylinder_def)

lemma llist_cylinder_iff_empty [simp]:
  "llist_cylinder A B = {} \<longleftrightarrow> A = {} \<or> B = {}"
by(auto simp: llist_cylinder_def)

lemma in_llist_cylinder_iff:
  "xs \<in> llist_cylinder X A \<longleftrightarrow> \<not> lnull xs \<and> lhd xs \<in> X \<and> ltl xs \<in> A"
by(auto simp add: llist_cylinder_def neq_LNil_conv lnull_def)

lemma llist_cylinders_subset_llists:
  assumes "\<A> \<subseteq> Pow \<Omega>" "A \<in> llist_cylinders \<Omega> \<A>"
  shows "A \<subseteq> llists \<Omega>"
using assms(2, 1)
by (induct A)(auto dest: sets.sets_into_space simp: llist_cylinder_def)

lemma llist_cylinders_subset_Pow:
  "\<A> \<subseteq> Pow \<Omega> \<Longrightarrow> llist_cylinders \<Omega> \<A> \<subseteq> Pow (llists \<Omega>)"
using llist_cylinders_subset_llists by blast

lemma llist_cylinders_Int_stable:
  assumes subset: "\<A> \<subseteq> Pow \<Omega>"
  assumes "Int_stable \<A>"
  shows "Int_stable (llist_cylinders \<Omega> \<A>)"
proof -
  let ?C = "llist_cylinders \<Omega> \<A>"
  { fix A B assume "A \<in> ?C" "B \<in> ?C"
    then have "A \<inter> B \<in> ?C"
    proof (induct A arbitrary: B)
      case base with subset show ?case
        by (simp add: llist_cylinders_subset_llists Int_absorb1)
    next
      case (step X A) note A = this
      from `B \<in> ?C` show ?case
      proof cases
        case base
        with llist_cylinders_subset_llists[OF subset, of A] subset step
        have "llist_cylinder X A \<inter> B = llist_cylinder X A"
          by (auto simp: llist_cylinder_def dest: sets.sets_into_space)
        with step show ?thesis
          by (simp add: Int_absorb2 llist_cylinders.step)
      next
        case (step Y B')
        then have *: "llist_cylinder X A \<inter> B = llist_cylinder (X \<inter> Y) (A \<inter> B')"
          by (auto simp: llist_cylinder_def dest: sets.sets_into_space)
        with step A `Int_stable \<A>` show ?thesis
          by (auto intro: llist_cylinders.intros simp: Int_stable_def)
      qed
    qed }
  then show ?thesis
    by (auto simp: Int_stable_def)
qed

lemma LNil_in_llist_cylinderD: "\<lbrakk> A \<in> llist_cylinders \<Omega> \<A>; LNil \<in> A \<rbrakk> \<Longrightarrow> A = llists \<Omega>"
by(induction rule: llist_cylinders.induct)(simp_all add: in_llist_cylinder_iff)

lemma LNil_in_llist_cylinder_iff: "A \<in> llist_cylinders \<Omega> \<A> \<Longrightarrow> LNil \<in> A \<longleftrightarrow> A = llists \<Omega>"
by(auto dest: LNil_in_llist_cylinderD)





definition llist_space :: "'a measure \<Rightarrow> 'a llist measure"
where "llist_space M = measure_of (llists (space M)) (llist_cylinders (space M) (sets M)) (\<lambda>_. 0)"

lemma
  shows space_llist_space: "space (llist_space M) = llists (space M)"
  and sets_llist_space: "sets (llist_space M) = sigma_sets (llists (space M)) (llist_cylinders (space M) (sets M))"
by (auto simp: llist_space_def llist_cylinders_subset_Pow sets.space_closed)

lemma llist_space_LNil [simp, intro]: "LNil \<in> space (llist_space M)"
by (simp add: space_llist_space)

lemma space_llist_space_not_empty [simp]: "space (llist_space M) \<noteq> {}"
by(blast)

lemma llist_space_LCons:
  "LCons st tr \<in> space (llist_space M) \<longleftrightarrow> st \<in> space M \<and> tr \<in> space (llist_space M)"
by(simp add: space_llist_space)

lemma llists_llist_space [intro]: "llists (space M) \<in> sets (llist_space M)"
using sets.top[of "llist_space M"] by (simp add: space_llist_space)

lemma cylinders_llist_space [intro]: 
  "A \<in> llist_cylinders (space M) (sets M) \<Longrightarrow> A \<in> sets (llist_space M)"
by (simp add: llist_space_def llist_cylinders_subset_Pow sets.space_closed)

lemma measurable_llist_spaceI:
  assumes f: "f \<in> space N \<rightarrow> space (llist_space M)"
  assumes sets: "\<And>A. A \<in> llist_cylinders (space M) (sets M) \<Longrightarrow> {\<omega>\<in>space N. f \<omega> \<in> A } \<in> sets N"
  shows "f \<in> measurable N (llist_space M)"
unfolding llist_space_def measurable_distr_eq2
proof (rule measurable_measure_of)
  show "f \<in> space N \<rightarrow>  llists (space M)"
    using f by (auto simp: space_llist_space)
  fix A assume "A \<in> llist_cylinders (space M) (sets M)"
  moreover have "f -` A \<inter> space N = {\<omega>\<in>space N. f \<omega> \<in> A }"
    by auto
  ultimately show "f -` A \<inter> space N \<in> sets N"
    by (auto intro: sets)
qed (rule sets.space_closed[THEN llist_cylinders_subset_Pow])

lemma LNil_in_llist_space [measurable]: "{LNil} \<in> sets (llist_space M)"
proof -
  have "llist_cylinder (space M) (llists (space M)) \<in> llist_cylinders (space M) (sets M)"
    by(auto intro: llist_cylinders.intros)
  hence "llist_cylinder (space M) (llists (space M)) \<in> sets (llist_space M)"
    by(simp add: sets_llist_space)
  hence "llists (space M) - llist_cylinder (space M) (llists (space M)) \<in> \<dots>"
    using sets.compl_sets[of _ "llist_space M"]
    by(simp add: sets_llist_space space_llist_space)
  also have "llists (space M) - llist_cylinder (space M) (llists (space M)) = {LNil}"
    by(auto simp add: in_llist_cylinder_iff dest: llists_lhdD intro: llists_ltlI)
  finally show ?thesis .
qed

lemma measurable_lnull [measurable]: "Measurable.pred (llist_space M) lnull"
proof -
  have "{x \<in> space (llist_space M). lnull x} = {LNil}" by auto
  also note LNil_in_llist_space
  finally show ?thesis by(unfold pred_def)
qed

lemma measurable_ltl [measurable]:
  "ltl \<in> measurable (llist_space M) (llist_space M)"
proof (rule measurable_llist_spaceI)
  show "ltl \<in> space (llist_space M) \<rightarrow> space (llist_space M)"
    by (auto simp: space_llist_space llists_ltlI)

  fix A
  assume A: "A \<in> llist_cylinders (space M) (sets M)"
  show "{\<omega> \<in> space (llist_space M). ltl \<omega> \<in> A} \<in> sets (llist_space M)"
  proof(cases "LNil \<in> A")
    case False
    then have "{\<omega> \<in> space (llist_space M). ltl \<omega> \<in> A} = llist_cylinder (space M) A"
      using llist_cylinders_subset_llists[OF sets.space_closed A]
      by(auto 4 3 simp: space_llist_space in_llist_cylinder_iff lnull_def dest: llists_lhdD intro: llists_expand)
    also have "\<dots> \<in> sets (llist_space M)"
      by (auto intro: llist_cylinders.intros A)
    finally show ?thesis .
  next
    case True
    with A have "A = llists (space M)" by(rule LNil_in_llist_cylinderD)
    hence "{\<omega> \<in> space (llist_space M). ltl \<omega> \<in> A} = llists (space M)"
      by(auto simp add: space_llist_space llists_ltlI)
    thus ?thesis using sets.top[of "llist_space M"]
      by(simp add: space_llist_space)
  qed
qed

text {* @{const lhd} is only measurable on the subspace of non-empty llists *}

lemma measurable_lhd [measurable]:
  "lhd \<in> measurable (restrict_space (llist_space M) {xs \<in> space (llist_space M). \<not> lnull xs}) M"
  unfolding measurable_def space_restrict_space sets_restrict_space
proof(intro CollectI conjI strip Pi_I)
  fix xs
  assume "xs \<in> {xs \<in> space (llist_space M). \<not> lnull xs} \<inter> space (llist_space M)"
  then show "lhd xs \<in> space M" using sets.space_closed[of M]
    by(cases xs)(auto simp add: space_llist_space)
next
  fix A
  assume A: "A \<in> sets M"
  with sets.space_closed[of M]
  have "lhd -` A \<inter> ({xs \<in> space (llist_space M). \<not> lnull xs} \<inter> space (llist_space M)) =
    {xs \<in> space (llist_space M). \<not> lnull xs} \<inter> llist_cylinder A (llists (space M))"
    by(fastforce simp add: space_llist_space in_llist_cylinder_iff llists_ltlI intro: llists_expand)
  also have "\<dots> \<in> op \<inter> {xs \<in> space (llist_space M). \<not> lnull xs} ` sets (llist_space M)"
    by (auto intro: llist_cylinders.intros A)
  finally show "lhd -` A \<inter> ({xs \<in> space (llist_space M). \<not> lnull xs} \<inter> space (llist_space M)) \<in> \<dots>" .
qed

lemma measurable_llist_space_coinduct[consumes 1, case_names measurable, case_conclusion measurable lnull step]:
  assumes "R f"
  assumes R: "\<And>f. R f \<Longrightarrow>
    {x\<in>space N. lnull (f x)} \<in> sets N \<and>
    (\<exists>f'. (\<exists>g \<in> measurable (restrict_space N {x\<in>space N. \<not> lnull (f x)}) M. 
           \<exists>n \<in> measurable (restrict_space N {x\<in>space N. \<not> lnull (f x)}) N.
           \<forall>x \<in> space N. \<not> lnull (f x) \<longrightarrow> g x = lhd (f x) \<and> f' (n x) = ltl (f x)) \<and> 
          (R f' \<or> f' \<in> measurable N (llist_space M)))"
  shows "f \<in> measurable N (llist_space M)"
proof (rule measurable_llist_spaceI)
  { fix f' x'
    assume "R f'" "x' \<in> space N"
    then have "f' x' \<in> llists (space M)"
    proof (coinduction arbitrary: x' f')
      case (llists x f) 
      with R[OF `R f`] show ?case
        by(fastforce simp add: neq_LNil_conv space_llist_space space_restrict_space dest: measurable_space)
    qed }
  note F_space = this
  with \<open>R f\<close> show f: "f \<in> space N \<rightarrow> space (llist_space M)"
    by (auto simp: space_llist_space)

  fix C assume C: "C \<in> llist_cylinders (space M) (sets M)"
  show "{xs \<in> space N. f xs \<in> C} \<in> sets N"
    using C \<open>R f\<close>
  proof (induction arbitrary: f)
    case base
    with F_space have "{xs \<in> space N. f xs \<in> llists (space M)} = space N"
      by auto
    then show ?case by auto
  next
    case (step X A)
    note [measurable] = \<open>X \<in> sets M\<close>
    from R[OF \<open>R f\<close>]
    obtain f' g n
    where [measurable]: "g \<in> measurable (restrict_space N {x\<in>space N. \<not> lnull (f x)}) M"
      and [measurable]: "n \<in> measurable (restrict_space N {x\<in>space N. \<not> lnull (f x)}) N"
      and lnull: "{x \<in> space N. lnull (f x)} \<in> sets N"
      and unfold: "\<And>x. \<lbrakk> x \<in> space N; \<not> lnull (f x) \<rbrakk> \<Longrightarrow> g x = lhd (f x) \<and> f' (n x) = ltl (f x)"
      and f': "R f' \<or> f' \<in> measurable N (llist_space M)"
      by blast
    note cylinders_llist_space[measurable]
    from f' step(2,3) have [measurable]: "{\<omega> \<in> space N. f' \<omega> \<in> A} \<in> sets N"
      by (cases "R f'") auto
    from lnull have [measurable]: "{x \<in> space N. \<not> lnull (f x)} \<in> sets N" by simp
    have "{x \<in> space N. f x \<in> llist_cylinder X A} = 
      {x \<in> space (restrict_space N {x \<in> space N. \<not> lnull (f x)}). g x \<in> X} \<inter> 
      {x \<in> space (restrict_space N {x \<in> space N. \<not> lnull (f x)}). f' (n x) \<in> A}"
      by(auto simp add: in_llist_cylinder_iff space_restrict_space dest: unfold)
    also have "\<dots> \<in> sets (restrict_space N {x \<in> space N. \<not> lnull (f x)})"
      by simp
    finally show ?case by(auto simp add: sets_restrict_space)
  qed
qed

lemma measurable_llist_unfold [measurable]:
  assumes [measurable]: "NIL \<in> measurable M' (count_space UNIV)" 
  and [intro]: "LHD \<in> measurable (restrict_space M' {x \<in> space M'. \<not> NIL x}) M"
  and [intro]: "LTL \<in> measurable (restrict_space M' {x \<in> space M'. \<not> NIL x}) M'"
  shows "unfold_llist NIL LHD LTL \<in> measurable M' (llist_space M)"
by(coinduction rule: measurable_llist_space_coinduct)(auto 4 4)

lemma measurable_LCons[measurable (raw)]:
  assumes [intro]: "f \<in> measurable N M"
  assumes [intro]: "g \<in> measurable N (llist_space M)"
  shows "(\<lambda>x. LCons (f x) (g x)) \<in> measurable N (llist_space M)"
by(coinduction rule: measurable_llist_space_coinduct)(auto 4 4 intro!: exI[of _ g] bexI[of _ f])

lemma measurable_lmap[measurable]: 
  assumes [measurable]: "f \<in> measurable N M"
  shows "lmap f \<in> measurable (llist_space N) (llist_space M)"
proof(coinduction rule: measurable_llist_space_coinduct)
  case measurable
  have ?lnull by(simp add: measurable_lnull[unfolded pred_def])
  moreover have "?step"
    by(auto 4 3 intro!: rev_bexI[where x="f \<circ> lhd"] simp add: measurable_restrict_space1)
  ultimately show "?case" ..
qed

lemma measurable_restrict_space_mono:
  assumes f: "f \<in> measurable (restrict_space M \<Omega>) N"
  and \<Omega>'\<Omega>: "\<Omega>' \<subseteq> \<Omega>"
  and \<Omega>: "\<Omega> \<inter> space M \<in> sets M"
  and \<Omega>': "\<Omega>' \<inter> space M \<in> sets M"
  shows "f \<in> measurable (restrict_space M \<Omega>') N"
proof -
  have "f \<in> measurable (restrict_space (restrict_space M \<Omega>) \<Omega>') N"
    using f by(rule measurable_restrict_space1)
  also have "restrict_space (restrict_space M \<Omega>) \<Omega>' = restrict_space M \<Omega>'"
    using \<Omega> \<Omega>' \<Omega>'\<Omega> by(simp add: Int_absorb1 restrict_restrict_space)
  finally show ?thesis .
qed

lemma Collect_space_conj_subset1:
  "{x \<in> space M. P x} \<subseteq> {x \<in> space M. R x} \<Longrightarrow> {x \<in> space M. P x \<and> Q x} \<subseteq> {x \<in> space M. R x}"
by blast

lemma Collect_space_conj_subset2:
  "{x \<in> space M. Q x} \<subseteq> {x \<in> space M. R x} \<Longrightarrow> {x \<in> space M. P x \<and> Q x} \<subseteq> {x \<in> space M. R x}"
by blast

lemma measurable_ltake [measurable]:
  "ltake n \<in> measurable (llist_space M) (llist_space M)"
proof(coinduction arbitrary: n rule: measurable_llist_space_coinduct)
  case (measurable n)
  have ?lnull by simp
  moreover have "?step"
    apply(intro exI conjI rev_bexI[where x=lhd] rev_bexI[where x=ltl])
    apply(rule measurable_restrict_space_mono, measurable)
    apply(auto simp add: measurable_restrict_space1 ltl_ltake)
    done
  ultimately show ?case ..
qed

lemma eSuc_less: "n < epred m \<Longrightarrow> eSuc n < m"
by(cases n m rule: enat_coexhaust[case_product enat_coexhaust])simp_all

lemma measurable_llength [measurable]: "llength \<in> measurable (llist_space M) (count_space UNIV)"
unfolding measurable_def
proof(intro CollectI conjI strip)
  show "llength \<in> space (llist_space M) \<rightarrow> space (count_space UNIV)" by simp

  fix A :: "enat set"
  assume "A \<in> sets (count_space UNIV)"
  have "llength -` A \<inter> space (llist_space M) = (\<Union>n. {xs \<in> space (llist_space M). llength xs = n \<and> n \<in> A})"
    by(auto)
  also have "\<dots> \<in> sets (llist_space M)"
  proof(rule sets.countable_UN'')
    fix n
    show "{xs \<in> space (llist_space M). llength xs = n \<and> n \<in> A} \<in> sets (llist_space M)"
    proof(cases "n \<in> A")
      case True
      let ?M = "\<lambda>n. {xs \<in> space (llist_space M). llength xs = n}"
      { fix m :: nat
        have "?M m \<in> sets (llist_space M)"
        proof(induction m)
          case 0
          thus ?case by(simp add: zero_enat_def[symmetric])
        next
          case (Suc n)
          have "?M (Suc n) = ltl -` ?M n \<inter> space (llist_space M) - {LNil}"
            by(auto simp add: space_llist_space neq_LNil_conv eSuc_enat[symmetric] llength_ltl intro: llists_ltlI)
          also have "\<dots> \<in> sets (llist_space M)"
            by(rule sets.Diff)(rule measurable_sets[OF _ Suc.IH], simp_all)
          finally show ?case .
        qed }
      note * = this
      have "?M n \<in> sets (llist_space M)"
      proof(cases n)
        case infinity
        then have "?M n = space (llist_space M) - (\<Union>n :: nat. ?M n)" by auto
        also have "\<dots> \<in> sets (llist_space M)"
          by(intro sets.compl_sets sets.countable_UN'' *) simp
        finally show ?thesis .
      qed(simp add: *)
      with True show ?thesis by simp
    qed simp
  qed simp
  finally show "llength -` A \<inter> space (llist_space M) \<in> sets (llist_space M)" .
qed

lemma measurable_lnth [measurable]:
  "(\<lambda>xs. lnth xs n) \<in> measurable (restrict_space (llist_space M) {xs. enat n < llength xs}) M"
  (is "?f \<in> measurable ?LM _")
unfolding measurable_def
proof(intro CollectI conjI strip)
  show "?f \<in> space ?LM \<rightarrow> space M"
    by(auto simp add: space_restrict_space space_llist_space intro: in_llists_lnthD)

  fix A
  assume A: "A \<in> sets M"
  have "?f -` A \<inter> space ?LM \<in> llist_cylinders (space M) (sets M)" (is "?lhs n \<in> ?rhs")
  proof(induction n)
    case 0
    have "?lhs 0 = llist_cylinder A (llists (space M))"
      using sets.space_closed[of M, THEN subsetD, OF A]
      by(auto 4 3 simp add: space_restrict_space space_llist_space zero_enat_def[symmetric] in_llist_cylinder_iff lnth_0_conv_lhd intro: llists_ltlI llists_expand)
    also from A have "\<dots> \<in> ?rhs" by rule rule
    finally show ?case .
  next
    case (Suc n)
    have "?lhs (Suc n) = llist_cylinder (space M) (?lhs n)"
      by(auto simp add: space_restrict_space in_llist_cylinder_iff space_llist_space lnth.simps Suc_ile_eq eSuc_enat[symmetric] llength_ltl split: llist.splits intro: llists_expand eSuc_less)
    also from sets.top Suc.IH have "\<dots> \<in> ?rhs" ..
    finally show ?case .
  qed
  hence "?lhs n \<in> sets (llist_space M)" ..
  then show "?f -` A \<inter> space ?LM \<in> sets ?LM"
    by(simp add: sets_restrict_space_iff)(auto simp add: space_restrict_space)
qed

lemma measurable_in_lset [measurable (raw generic)]:
  assumes "y \<in> space M \<Longrightarrow> {y} \<in> sets M"
  shows "Measurable.pred (llist_space M) (\<lambda>xs. y \<in> lset xs)"
proof(cases "y \<in> space M")
  case True
  have "{xs \<in> space (llist_space M). y \<in> lset xs} = 
    (\<Union>n. (\<lambda>xs. lnth xs n) -` {y} \<inter> space (restrict_space (llist_space M) {xs. enat n < llength xs}))"
    (is "?lhs = (\<Union>n. ?A n)")
    by(auto simp add: in_lset_conv_lnth space_restrict_space)
  also have "\<dots> \<in> sets (llist_space M)"
  proof(rule sets.countable_UN'')
    fix n
    have "?A n \<in> sets (restrict_space (llist_space M) {xs. enat n < llength xs})"
      by(rule measurable_sets[OF measurable_lnth])(rule assms[OF True])
    then show "?A n \<in> sets (llist_space M)"
      by(simp add: sets_restrict_space_iff)
  qed simp
  finally show ?thesis by(simp add: pred_def)
next
  case False
  hence "{xs \<in> space (llist_space M). y \<in> lset xs} = {}"
    by(auto simp add: space_llist_space llists_alt_def)
  thus ?thesis by(simp add: pred_def del: Collect_empty_eq)
qed

lemma measurable_ltakeWhile [measurable]:
  assumes [measurable]: "Measurable.pred M P"
  shows "ltakeWhile P \<in> measurable (llist_space M) (llist_space M)"
proof(coinduction rule: measurable_llist_space_coinduct)
  case measurable
  { note [measurable] = pred_disjCI2
    have "?lnull" by(simp del: lnull_ltakeWhile) }
  moreover {
    note [measurable] = pred_conj_cong1
    have "lhd \<in> measurable (restrict_space (llist_space M) {x \<in> space (llist_space M). \<not> lnull x \<and> P (lhd x)}) M"
      by(rule measurable_restrict_space_mono)(measurable, auto) }
  then have "?step" by(fastforce intro: measurable_restrict_space1)
  ultimately show ?case ..
qed

lemma measurable_llist_of_stream [measurable]:
  "llist_of_stream \<in> measurable (stream_space M) (llist_space M)"
by(coinduction rule: measurable_llist_space_coinduct) fastforce

lemma measurable_stakeWhile [measurable]:
  assumes [measurable]: "Measurable.pred M P"
  shows "stakeWhile P \<in> measurable (stream_space M) (llist_space M)"
by(simp add: stakeWhile_def[abs_def])


text {* Definition of measures on @{term "llist_space"} by definition on cylinders *}

type_synonym 'a llist_cylinder = "'a set list"

context fixes \<Omega> :: "'a set" begin

fun llist_cylinder_of :: "'a llist_cylinder \<Rightarrow> 'a llist set"
where
  "llist_cylinder_of [] = llists \<Omega>"
| "llist_cylinder_of (A # cyl) = llist_cylinder A (llist_cylinder_of cyl)"

lemma llist_cylinder_of_in_llist_cylinders:
  assumes "set cyl \<subseteq> \<A>"
  shows "llist_cylinder_of cyl \<in> llist_cylinders \<Omega> \<A>"
using assms by(induction cyl)(auto intro: llist_cylinders.intros)

lemma llist_cylinders_is_llist_cylinder_of:
  assumes "A \<in> llist_cylinders \<Omega> \<A>"
  obtains cyl where "set cyl \<subseteq> \<A>" "A = llist_cylinder_of cyl"
proof(atomize_elim)
  from assms show "\<exists>cyl. set cyl \<subseteq> \<A> \<and> A = local.llist_cylinder_of cyl"
    by(induction arbitrary: thesis)(auto 4 3 intro: exI[where x="[]"] exI[where x="X # cyl" for X cyl])
qed

lemma llist_cylinder_of_lists [simp]:
  "llist_cylinder_of ` lists \<A> = llist_cylinders \<Omega> \<A>"
by(blast intro: llist_cylinder_of_in_llist_cylinders elim: llist_cylinders_is_llist_cylinder_of)

end

definition extend_llist_space :: "'a measure \<Rightarrow> ('a llist_cylinder \<Rightarrow> ereal) \<Rightarrow> 'a llist measure"
where
  "extend_llist_space \<mu> f = 
   extend_measure (llists (space \<mu>)) (lists (sets \<mu>)) (llist_cylinder_of (space \<mu>)) f"

lemma space_extend_llist_space: "space (extend_llist_space \<mu> f) = llists (space \<mu>)"
by(simp add: extend_llist_space_def space_extend_measure llist_cylinders_subset_Pow sets.space_closed)

lemma sets_extend_llist_space: "sets (extend_llist_space \<mu> f) = sets (llist_space \<mu>)"
by(simp add: extend_llist_space_def sets_llist_space sets_extend_measure sets.space_closed llist_cylinders_subset_Pow)

thm emeasure_extend_measure[OF refl]

lemma emeasure_extend_llist_space:
  assumes "\<Omega> = space \<mu>"
  and "set xs \<subseteq> sets \<mu>"
  shows "emeasure (extend_llist_space \<mu> f) (llist_cylinder_of \<Omega> xs) = f xs"
apply(simp add: extend_llist_space_def assms)
apply(rule emeasure_extend_measure[OF refl])
defer
apply(simp add: llist_cylinders_subset_Pow sets.space_closed)
apply(fold extend_llist_space_def)
apply(simp add: sets_extend_llist_space)
prefer 2
apply(simp add: sets_extend_llist_space)
prefer 2
apply(simp add: positive_def)
defer
oops


end

