theory Length_Regularity imports 
  "$AFP/CryptHOL/Computational_Model"
begin

subsection \<open>Length regularity\<close>

consts length_reg :: "'a \<Rightarrow> 'b \<Rightarrow> bool"

overloading lr \<equiv> "length_reg :: 'a list \<Rightarrow> 'b list \<Rightarrow> bool" begin
definition length_reg_list_def [simp]: "lr xs ys \<longleftrightarrow> length xs = length ys"
end

lemma length_reg_listI [intro!]: "length xs = length ys \<Longrightarrow> length_reg xs ys"
by simp

lemma length_reg_listD: "length_reg xs ys \<Longrightarrow> length xs = length ys"
by simp

overloading lr \<equiv> "length_reg :: 'a option \<Rightarrow> 'a option \<Rightarrow> bool" begin
definition length_reg_option_def:
  "lr x y = (case (x, y) of (Some x', Some y') \<Rightarrow> length_reg x' y' | _ \<Rightarrow> True)"
  -- \<open>Do we need something stronger such as @{const rel_option}? \<close>
end

lemma length_reg_option_simps[iff]:
  "\<And>y :: _ option. length_reg None y"
  "\<And>x :: _ option. length_reg x None"
  "length_reg (Some x) (Some y) \<longleftrightarrow> length_reg x y"
by(simp_all add: length_reg_option_def split: option.split)

lemma length_reg_optionI:
  "(\<And>x' y'. \<lbrakk> x = Some x'; y = Some y' \<rbrakk> \<Longrightarrow> length_reg x' y') \<Longrightarrow> length_reg x y"
by(cases x y rule: option.exhaust[case_product option.exhaust]) simp_all

lemma length_reg_optionD:
  "\<lbrakk> length_reg x y; x = Some x'; y = Some y' \<rbrakk> \<Longrightarrow> length_reg x' y'"
by simp

overloading lr \<equiv> "length_reg :: ('a list \<Rightarrow> 'b) \<Rightarrow> ('c list \<Rightarrow> 'd) \<Rightarrow> bool" begin
definition length_reg_list_fun_def [simp]:
  "lr f g = (\<forall>xs ys. length xs = length ys \<longrightarrow> length_reg (f xs) (g ys))"
end

lemma length_reg_list_funI:
  "(\<And>xs ys. length xs = length ys \<Longrightarrow> length_reg (f xs) (g ys)) \<Longrightarrow> length_reg f g"
by simp

lemma length_reg_list_funD:
  "\<lbrakk> length_reg f g; length xs = length ys \<rbrakk> \<Longrightarrow> length_reg (f xs) (g ys)"
by simp

overloading lr \<equiv> "length_reg :: (security \<Rightarrow> 'a) \<Rightarrow> (security \<Rightarrow> 'b) \<Rightarrow> bool" begin
definition length_reg_security_def [simp]: 
  "lr f g = (\<forall>\<eta> :: security. length_reg (f \<eta>) (g \<eta>))"
end

lemma length_reg_securityI [intro!]:
  "(\<And>\<eta> :: security. length_reg (f \<eta>) (g \<eta>)) \<Longrightarrow> length_reg f g"
by simp

overloading lr \<equiv> "length_reg :: (bool \<Rightarrow> 'a) \<Rightarrow> (bool \<Rightarrow> 'b) \<Rightarrow> bool" begin
definition length_reg_bool_def [simp]: "lr f g = (\<forall>\<eta> :: bool. length_reg (f \<eta>) (g \<eta>))"
end

lemma length_reg_boolI [intro!]:
  "(\<And>\<eta> :: bool. length_reg (f \<eta>) (g \<eta>)) \<Longrightarrow> length_reg f g"
by simp

lemma length_reg_boolD: fixes b :: bool shows "length_reg f g \<Longrightarrow> length_reg (f b) (g b)"
by simp

abbreviation length_regular :: "'a \<Rightarrow> bool"
where "length_regular x \<equiv> length_reg x x"

end