theory Markov_Search imports
  SPMF
  "Stream_Space_More"
  "LList_Space"
  "$AFP/Markov_Models/Discrete_Time_Markov_Chain"
begin


lemma (in MC_syntax) enabled_sdropI: "enabled s \<omega> \<Longrightarrow> enabled ((s ## \<omega>) !! n) (sdrop n \<omega>)"
proof(induction n arbitrary: s \<omega>)
  case (Suc n)
  thus ?case by(cases \<omega> rule: stream.exhaust)(auto elim: enabled.cases)
qed simp

lemma (in MC_syntax) enabled_Stream: "enabled s (x ## \<omega>) \<longleftrightarrow> x \<in> set_pmf (K s) \<and> enabled x \<omega>"
by(auto elim: enabled.cases intro: enabled.intros)

lemma not_suntil_conv_ev: "not P suntil P = ev P" (is "?lhs = ?rhs")
proof(intro ext iffI)
  fix \<omega>
  assume "?lhs \<omega>"
  thus "?rhs \<omega>" by induction auto
next
  fix \<omega>
  assume "?rhs \<omega>"
  thus "?lhs \<omega>" by induction(auto intro: suntil.intros)
qed

lemma suntil_ev_redundant:
  assumes *: "\<And>\<omega>. P \<omega> \<Longrightarrow> Q \<omega>"
  shows "(not Q suntil (Q aand ev P)) = ev P" (is "?\<phi> = _")
proof(intro ext iffI)
  fix \<omega>
  assume "?\<phi> \<omega>"
  hence "ev (Q aand ev P) \<omega>" by(rule ev_suntil)
  hence "ev (ev P) \<omega>" by(rule ev_mono) simp
  thus "ev P \<omega>" by simp
next
  fix \<omega>
  assume "ev P \<omega>"
  thus "?\<phi> \<omega>" by(induction rule: ev.induct)(auto simp add: * intro: suntil.intros)
qed




context notes [[inductive_internals]] begin

inductive_set path :: "('a \<times> 'a) set \<Rightarrow> ('a \<times> 'a list) set"
  for r
where
  "(x, []) \<in> path r"
| "\<lbrakk> (x, y) \<in> r; (y, xs) \<in> path r \<rbrakk> \<Longrightarrow> (x, y # xs) \<in> path r"

end

definition paths_from :: "('a \<times> 'a) set \<Rightarrow> 'a \<Rightarrow> 'a list set"
where "paths_from r x = {ys. (x, ys) \<in> path r}"

context
  fixes r :: "('a \<times> 'a) set" and F
  defines "F \<equiv> (\<lambda>pf x. insert [] (\<Union>y\<in>r `` {x}. op # y ` pf y))"
begin

lemma paths_from_unfold: "paths_from r x = F (paths_from r) x"
unfolding paths_from_def F_def by(subst path.simps) force

lemma continuous_paths_from: "Order_Continuity.sup_continuous F"
unfolding Order_Continuity.sup_continuous_def F_def by(auto simp add: fun_eq_iff)

lemma paths_from_conv_lfp: "paths_from r = lfp F" (is "?lhs = ?rhs")
proof(rule antisym)
  have g: "mono F" using continuous_paths_from by(rule sup_continuous_mono)
  let ?f = "\<lambda>p x1 x2. x2 = [] \<or> (\<exists>y xs. x2 = y # xs \<and> (x1, y) \<in> r \<and> p y xs)"
  have "mono ?f"
    by(tactic {* resolve_tac @{context} (@{thm monoI} :: @{thm le_funI} :: @{thm le_boolI'} :: Inductive.get_monos @{context}) 1 *})+
      (blast elim: le_funE)
  then have "lfp ?f \<le> (\<lambda>x ys. ys \<in> ?rhs x)"
    by(rule lfp_induct)(subst (2) lfp_unfold[OF g], auto 4 3 simp add: F_def le_fun_def intro: rev_bexI rev_image_eqI)
  then show "?lhs \<le> ?rhs"
    unfolding paths_from_def[abs_def] path_def pathp_def
    by(auto simp add: le_fun_def)
  show "?rhs \<le> ?lhs" using g
    by(rule lfp_induct)(subst (2) paths_from_unfold[abs_def], auto simp add: le_fun_def F_def)
qed

end

lemma countable_paths_from: "(\<And>x. countable (r `` {x})) \<Longrightarrow> countable (paths_from r x)"
unfolding paths_from_conv_lfp by(rule countable_lfp_apply)(simp, rule continuous_paths_from)

lemma Nil_in_paths_from [iff]: "[] \<in> paths_from r x"
by(subst paths_from_unfold) simp

lemma Cons_in_paths_from_iff:
  "x # xs \<in> paths_from r y \<longleftrightarrow> (y, x) \<in> r \<and> xs \<in> paths_from r x"
by(subst paths_from_unfold) auto

lemma append_in_paths_from:
  "xs @ ys \<in> paths_from r x \<longleftrightarrow> xs \<in> paths_from r x \<and> ys \<in> paths_from r (last (x # xs))"
by(induction xs arbitrary: x)(auto simp add: Cons_in_paths_from_iff)


lemma path_mono: 
  assumes *: "(x, xs) \<in> path R"
  and RS: "R \<subseteq> S"
  shows "(x, xs) \<in> path S"
using * apply(induction) using RS by(auto intro: path.intros)

lemma paths_from_mono:
  "\<lbrakk> xs \<in> paths_from R x; R \<subseteq> S \<rbrakk> \<Longrightarrow> xs \<in> paths_from S x"
unfolding paths_from_def by(auto intro: path_mono)

lemma path_restrict1:
  "\<lbrakk> (x, xs) \<in> path R; \<And>y. y \<in> set (butlast (x # xs)) \<Longrightarrow> P y \<rbrakk>
  \<Longrightarrow> (x, xs) \<in> path {(x, y) \<in> R. P x}"
by(induction rule: path.induct)(auto intro: path.intros split: if_split_asm)

lemma paths_from_restrict1:
  "\<lbrakk> xs \<in> paths_from R x; \<And>y. y \<in> set (butlast (x # xs)) \<Longrightarrow> P y \<rbrakk>
  \<Longrightarrow> xs \<in> paths_from {(x, y)\<in>R. P x} x"
unfolding paths_from_def by(auto elim: path_restrict1)

lemma path_into_rtrancl: "(x, xs) \<in> path r \<Longrightarrow> (x, last (x # xs)) \<in> r\<^sup>*"
by(induction rule: path.induct) simp_all

lemma paths_from_into_rtrancl: "xs \<in> paths_from r x \<Longrightarrow> (x, last (x # xs)) \<in> r\<^sup>*"
by(auto simp add: paths_from_def dest: path_into_rtrancl)



fun search_path :: "('s \<Rightarrow> bool) \<Rightarrow> 's \<times> 's list \<Rightarrow> bool"
where "search_path P (x, xs) \<longleftrightarrow> P (last (x # xs)) \<and> (\<forall>y\<in>set (butlast (x # xs)). \<not> P y)"

lemma search_path_intros [intro?]:
  shows search_path_empty: "P x \<Longrightarrow> search_path P (x, [])"
  and search_path_cons: "\<lbrakk> \<not> P y; search_path P (x, xs) \<rbrakk> \<Longrightarrow> search_path P (y, x # xs)"
by auto

lemma search_path_simps [simp]:
  shows search_path_empty_iff: "search_path P (x, []) \<longleftrightarrow> P x"
  and search_path_cons_iff: "search_path P (y, x # xs) \<longleftrightarrow> \<not> P y \<and> search_path P (x, xs)"
by auto

lemmas search_path_iff = search_path.simps
declare search_path.simps [simp del]

lemma search_path_cases [elim?, cases pred: search_path]:
  assumes "search_path P (x, xs)"
  obtains (empty) "xs = []" "P x"
  | (cons) y xs' where "xs = y # xs'" "\<not> P x" "search_path P (y, xs')"
using assms by(cases xs) auto

lemma search_path_append:
  "ys \<noteq> [] \<Longrightarrow> search_path P (x, xs @ ys) \<longleftrightarrow> \<not> P x \<and> (\<forall>x\<in>set xs. \<not> P x) \<and> search_path P (last (x # xs), ys)"
by(cases ys)(auto simp add: search_path.simps butlast_append)


context MC_syntax begin

fun prob_path :: "'s \<Rightarrow> 's list \<Rightarrow> real"
where
  "prob_path x [] = 1"
| "prob_path x (y # ys) = pmf (K x) y * prob_path y ys"

lemma prob_path_nonneg: "prob_path x xs \<ge> 0"
by(induction rule: prob_path.induct)(simp_all add: pmf_nonneg)

lemma prob_path_le_1: "prob_path x xs \<le> 1"
by(induction rule: prob_path.induct)(simp_all add: pmf_le_1 mult_le_one prob_path_nonneg)

lemma prob_path_le_0_iff [simp]: "prob_path x xs \<le> 0 \<longleftrightarrow> prob_path x xs = 0"
using prob_path_nonneg[of x xs] by simp

lemma emeasure_T_sprefix:
  assumes [measurable]: "Measurable.pred S P"
  shows
  "emeasure (T s) {\<omega> \<in> space (T s). sprefix xs \<omega> \<and> P (sdrop (length xs) \<omega>)} =
   prob_path s xs * emeasure (T (last (s # xs))) {\<omega> \<in> space (T (last (s # xs))). P \<omega>}"
  (is "?lhs s xs = ?rhs s xs")
proof(induction s xs rule: prob_path.induct)
  case 1 thus ?case by(simp add: one_ereal_def[symmetric])
next
  case (2 s s' xs)
  have "?lhs s (s' # xs) = \<integral>\<^sup>+ t. indicator {s'} t * ?lhs s' xs \<partial>measure_pmf (K s)"
    by(subst emeasure_Collect_T)(auto intro: nn_integral_cong split: split_indicator)
  also have "\<dots> = pmf (K s) s' * ?lhs s' xs"
    by(simp add: nn_integral_multc emeasure_pmf_single)
  also have "?lhs s' xs = ?rhs s' xs" by(rule "2.IH")
  also have "pmf (K s) s' * ?rhs s' xs = ?rhs s (s' # xs)"
    by(simp add: T.emeasure_eq_measure ennreal_mult[symmetric] prob_path_nonneg)
  finally show ?case .
qed

lemma measure_T_sprefix:
  "Measurable.pred S P
  \<Longrightarrow> measure (T s) {\<omega> \<in> space (T s). sprefix xs \<omega> \<and> P (sdrop (length xs) \<omega>)} =
      prob_path s xs * measure (T (last (s # xs))) {\<omega> \<in> space (T (last (s # xs))). P \<omega>}"
using emeasure_T_sprefix[of P s xs] by(simp add: T.emeasure_eq_measure prob_path_nonneg ennreal_mult[symmetric])

lemma path_prob_conv_prob_sprefix: "prob_path x xs = \<P>(\<omega> in T x. sprefix xs \<omega>)"
using measure_T_sprefix[where P="\<lambda>_. True", of x xs]
by(simp del: space_T add: T.prob_space)


abbreviation MC_paths_from :: "'s \<Rightarrow> 's list set"
where "MC_paths_from \<equiv> paths_from (SIGMA s:UNIV. set_pmf (K s))"

lemma countable_MC_paths_from: "countable (MC_paths_from s)"
by(rule countable_paths_from)(simp add: Sigma_Image)

lemma prob_path_eq_0_iff_notin_paths_from:
  "prob_path s xs = 0 \<longleftrightarrow> xs \<notin> MC_paths_from s"
by(induction s xs rule: prob_path.induct)(auto simp add: Cons_in_paths_from_iff pmf_eq_0_set_pmf)

context fixes P :: "'s \<Rightarrow> bool" begin
partial_function (spmf) search' :: "'s \<Rightarrow> 's list spmf"
where "search' s =
  (if P s then return_spmf [] else bind_pmf (K s) (\<lambda>s'. map_spmf (op # s') (search' s')))"
end

lemma search'_simps [simp]:
  "P s \<Longrightarrow> search' P s = return_spmf []"
  "\<not> P s \<Longrightarrow> search' P s = bind_pmf (K s) (\<lambda>s'. map_spmf (op # s') (search' P s'))"
by(simp_all add: search'.simps)

lemmas search'_fixp_induct = search'.fixp_induct[case_names adm bottom step]

lemma set_spmf_search'_subset_path:
  "set_spmf (search' P s) \<subseteq> {xs \<in> MC_paths_from s. search_path P (s, xs)}"
proof(induction arbitrary: s rule: search'_fixp_induct)
  case adm
  note [cont_intro] = ccpo_class.admissible_leI
  show ?case by(intro cont_intro)
next
  case (step search')
  then show ?case
    apply(clarsimp simp add: set_spmf_bind_pmf)
    apply(fastforce simp add: Cons_in_paths_from_iff)
    done
qed simp

lemma in_set_spmf_search'D_search_path:
  "xs \<in> set_spmf (search' P s) \<Longrightarrow> xs \<in> MC_paths_from s \<and> search_path P (s, xs)"
using set_spmf_search'_subset_path[of P s] by blast

lemma set_spmf_search':
  "set_spmf (search' P s) = {xs \<in> MC_paths_from s. search_path P (s, xs)}"
proof -
  { fix xs
    assume "xs \<in> MC_paths_from s" "search_path P (s, xs)"
    then have "xs \<in> set_spmf (search' P s)"
      by(induction s xs rule: prob_path.induct)(auto simp add: Cons_in_paths_from_iff set_spmf_bind_pmf) }
  thus ?thesis by(auto dest: in_set_spmf_search'D_search_path)
qed

lemma in_set_spmf_search'_iff [simp]:
  "xs \<in> set_spmf (search' P s) \<longleftrightarrow> xs \<in> MC_paths_from s \<and> search_path P (s, xs)"
by(simp add: set_spmf_search')

lemma spmf_search'_empty: "spmf (search' P s) [] = indicator {s. P s} s"
apply(subst search'.simps)
apply(clarsimp split: split_indicator simp add: pmf_bind)
apply(subst spmf_map_outside)
apply auto
done

lemma spmf_search'_conv':
  "search_path P (s, xs) \<Longrightarrow> spmf (search' P s) xs = prob_path s xs"
proof(induction s xs rule: prob_path.induct)
  case 1 thus ?case by(simp add: spmf_search'_empty split: split_indicator)
next
  case (2 x y ys)
  have "spmf (search' P x) (y # ys) = \<integral> x. spmf (map_spmf (op # x) (search' P x)) (y # ys) \<partial>measure_pmf (K x)"
    using "2.prems" by (simp add: pmf_bind)
  also have "\<dots> = \<integral> x. indicator {y} x * spmf (map_spmf (op # y) (search' P y)) (y # ys) \<partial>measure_pmf (K x)"
    by(rule Bochner_Integration.integral_cong)(auto intro: spmf_map_outside split: split_indicator)
  also have "\<dots> = pmf (K x) y * spmf (search' P y) ys"
    by(simp add: measure_pmf_single spmf_map_inj')
  finally show ?case using 2 by simp
qed

lemma spmf_search': "spmf (search' P s) xs = (if search_path P (s, xs) then prob_path s xs else 0)"
by(auto simp add: spmf_search'_conv' spmf_eq_0_set_spmf dest: in_set_spmf_search'D_search_path)

lemma spmf_search'_simps [simp]:
  shows spmf_search_path: "search_path P (s, xs) \<Longrightarrow> spmf (search' P s) xs = prob_path s xs"
  and spmf_search_eq_0: "\<not> search_path P (s, xs) \<Longrightarrow> spmf (search' P s) xs = 0"
by(auto simp add: spmf_search')


definition search :: "('s \<Rightarrow> bool) \<Rightarrow> 's \<Rightarrow> 's spmf"
where "search P s = map_spmf (\<lambda>xs. last (s # xs)) (search' P s)"

lemma search_unfold: "search P s = (if P s then return_spmf s else bind_pmf (K s) (search P))"
by(simp add: search_def[abs_def] map_bind_pmf pmf.map_comp option.map_comp o_def)

lemma search_simps[simp]: 
  "P s \<Longrightarrow> search P s = return_spmf s"
  "\<not> P s \<Longrightarrow> search P s = bind_pmf (K s) (search P)"
by(simp_all add: search_unfold)

lemma search_reachableI:
  assumes *: "(s, s') \<in> (SIGMA s:{s. \<not> P s}. set_pmf (K s))\<^sup>*"
  and "P s'"
  shows "s' \<in> set_spmf (search P s)"
using * by(induction rule: converse_rtrancl_induct)(auto simp add: set_spmf_bind_pmf \<open>P s'\<close>)

lemma search_reachableD:
  assumes "s' \<in> set_spmf (search P s)"
  shows "(s, s') \<in> (SIGMA s:{s. \<not> P s}. set_pmf (K s))\<^sup>*" "P s'"
proof -
  from assms obtain xs where s': "s' = last (s # xs)"
    and xs: "xs \<in> set_spmf (search' P s)"
    by(clarsimp simp add: search_def simp del: last.simps)
  from xs have "xs \<in> MC_paths_from s" "search_path P (s, xs)" by simp_all
  then have "xs \<in> paths_from {(s, s')\<in>(SIGMA s:UNIV. set_pmf (K s)). \<not> P s} s"
    by(intro paths_from_restrict1)(simp_all add: search_path.simps)
  hence "(s, s') \<in> {(s, s')\<in>(SIGMA s:UNIV. set_pmf (K s)). \<not> P s}\<^sup>*"
    unfolding s' by(rule paths_from_into_rtrancl)
  also have "{(s, s')\<in>(SIGMA s:UNIV. set_pmf (K s)). \<not> P s} = (SIGMA s:{s. \<not> P s}. set_pmf (K s))"
    by auto
  finally show "(s, s') \<in> \<dots>\<^sup>*" .
  show "P s'" using s' \<open>search_path P (s, xs)\<close> by(simp add: search_path.simps)
qed

lemma set_spmf_search:
  "set_spmf (search P s) = {s'. (s, s') \<in> (SIGMA s:{s. \<not> P s}. set_pmf (K s))\<^sup>* \<and> P s'}"
by(auto intro: search_reachableI dest: search_reachableD)

lemma in_set_spmf_search_iff [simp]:
  "s' \<in> set_spmf (search P s) \<longleftrightarrow> P s' \<and> (s, s') \<in> (SIGMA s:{s. \<not> P s}. set_pmf (K s))\<^sup>*"
by(auto simp add: set_spmf_search)


lemma emeasure_suntil_HLD_search:
  assumes [measurable]: "Measurable.pred S P"
  shows "emeasure (T s) {\<omega> \<in> space (T s). (not (HLD A) suntil (HLD A aand P)) (s ## \<omega>)} =
    \<integral>\<^sup>+ s'. emeasure (T s') {\<omega> \<in> space (T s). P (s' ## \<omega>)} \<partial>measure_spmf (search (\<lambda>x. x \<in> A) s)"
   (is "?lhs = ?rhs" is "emeasure _ {\<omega> \<in> _. ?\<phi> \<omega>} = _")
proof -
  let ?A = "\<lambda>x. x \<in> A"
  have "?lhs = emeasure (T s) {\<omega> \<in> space (T s). enabled s \<omega> \<and> ?\<phi> \<omega>}"
    using AE_T_enabled by(rule emeasure_eq_AE[OF AE_mp]) simp_all
  also {
    note [simp del] = split_paired_Ex
    fix \<omega>
    assume "?\<phi> \<omega>" and "enabled s \<omega>"
    then have "\<exists>xs. xs \<in> MC_paths_from s \<and> search_path ?A (s, xs) \<and> sprefix xs \<omega>"
    proof(induction \<omega>'\<equiv>"s ## \<omega>" arbitrary: s \<omega>)
      case (base s \<omega>)
      thus ?case
        by(cases \<omega> rule: stream.exhaust)(auto simp add: HLD_iff)
    next
      case (step s \<omega>)
      from step.hyps(3)[of "shd \<omega>" "stl \<omega>"] step.hyps(1-2) step.prems enabled_iff[of s \<omega>] show ?case
        by(cases \<omega> rule: stream.exhaust)(auto 4 3 intro: exI[where x="x # xs" for x xs] simp add: Cons_in_paths_from_iff)
    qed }
  then have "{\<omega> \<in> space (T s). enabled s \<omega> \<and> ?\<phi> \<omega>} = (\<Union>xs \<in> {xs \<in> MC_paths_from s. search_path ?A (s, xs)}. {\<omega> \<in> space (T s). sprefix xs \<omega> \<and> search_path (\<lambda>x. x \<in> A) (s, xs) \<and> enabled s \<omega> \<and> ?\<phi> \<omega>})"
    (is "_ = (\<Union>xs\<in>?M. ?X xs)")  by auto
  also have "emeasure (T s) \<dots> = \<integral>\<^sup>+ xs. emeasure (T s) (?X xs) \<partial>count_space ?M"
  proof(rule emeasure_UN_countable)
    show "countable ?M" by(rule countable_Collect)(rule countable_MC_paths_from)
    show "disjoint_family_on ?X ?M"
      apply(clarsimp simp add: disjoint_family_on_def, safe)
      apply(drule (1) sprefix_down_linear, erule exE, erule disjE)
      apply(simp_all add: search_path_append, simp_all add: search_path.simps)
      done
  qed simp
  also have "\<dots> = \<integral>\<^sup>+ xs. spmf (search' ?A s) xs * emeasure (T (last (s # xs))) {\<omega> \<in> space (T (last (s # xs))). enabled (last (s # xs)) \<omega> \<and> P (last (s # xs) ## \<omega>)} \<partial>count_space ?M"
    (is "_ = \<integral>\<^sup>+ xs. _ * ?rest xs \<partial>_")
  proof(rule nn_integral_cong)
    fix xs
    assume "xs \<in> space (count_space ?M)"
    hence xs: "xs \<in> MC_paths_from s" "search_path ?A (s, xs)" by auto

    have "emeasure (T s) (?X xs) =
      emeasure (T s) {\<omega> \<in> space (T s). sprefix xs \<omega> \<and> enabled (last (s # xs)) (sdrop (length xs) \<omega>) \<and> P (last (s # xs) ## sdrop (length xs) \<omega>)}"
      (is "_ = emeasure _ ?Y")
    proof(intro arg_cong2[where f=emeasure] subsetI equalityI)
      fix \<omega>
      assume "\<omega> \<in> ?X xs"
      hence "?\<phi> \<omega>" "sprefix xs \<omega>" "search_path ?A (s, xs)" by auto
      hence "P (last (s # xs) ## sdrop (length xs) \<omega>)"
      proof(induction \<omega>'\<equiv>"s ## \<omega>" arbitrary: s \<omega> xs)
        case step thus ?case by(cases \<omega> rule: stream.exhaust)(cases xs, simp_all)
      qed(simp add: search_path.simps split: if_split_asm)
      moreover have "last (s # xs) = (s ## \<omega>) !! length xs" using \<open>sprefix xs \<omega>\<close>
        unfolding sprefix_def by(auto simp add: Stream_snth shift_snth last_conv_nth split: nat.split)
      with \<open>\<omega> \<in> ?X xs\<close> have "enabled (last (s # xs)) (sdrop (length xs) \<omega>)"
        by(auto intro: enabled_sdropI)
      ultimately show "\<omega> \<in> ?Y" using \<open>\<omega> \<in> ?X xs\<close> by auto
    next
      fix \<omega>
      assume "\<omega> \<in> ?Y"
      with xs show "\<omega> \<in> ?X xs"
      proof(induction s xs arbitrary: \<omega> rule: prob_path.induct)
        case (2 s xs)
        from "2.IH"[of "stl \<omega>"] "2.prems" show ?case
          by(cases \<omega> rule: stream.exhaust)(auto intro: enabled.intros suntil.step simp add: Cons_in_paths_from_iff space_stream_space)
      qed(simp add: suntil.base)
    qed simp
    also have "\<dots> = prob_path s xs * ?rest xs"
      by(rule emeasure_T_sprefix) simp
    also have "\<dots> = spmf (search' ?A s) xs * ?rest xs"
      using xs by simp
    finally show "emeasure (T s) (?X xs) = \<dots>" .
  qed
  also have "\<dots> = \<integral>\<^sup>+ xs. spmf (search' ?A s) xs * ?rest xs \<partial>count_space UNIV"
    by(auto simp add: nn_integral_count_space_indicator spmf_search' prob_path_eq_0_iff_notin_paths_from[symmetric] intro: nn_integral_cong split: split_indicator simp del: last.simps)
  also have "\<dots> = \<integral>\<^sup>+ xs. ?rest xs \<partial>measure_spmf (search' ?A s)"
    by(subst nn_integral_measure_spmf) simp
  also have "\<dots> = \<integral>\<^sup>+ s'. emeasure (T s') {\<omega> \<in> space (T s'). enabled s' \<omega> \<and> P (s' ## \<omega>)} \<partial>measure_spmf (search ?A s)"
    unfolding search_def by(simp add: measure_map_spmf_conv_distr nn_integral_distr)
  also have "\<dots> = ?rhs" apply(rule nn_integral_cong)
    using AE_T_enabled by(rule emeasure_eq_AE[OF AE_mp]) simp_all
  finally show ?thesis .
qed

lemma weight_spmf_search: "weight_spmf (search P s) = \<P>(\<omega> in T s. ev (holds P) (s ## \<omega>))"
  (is "?lhs = ?rhs")
proof -
  have "?lhs= \<integral>\<^sup>+ s'. emeasure (T s') (space (T s')) \<partial>measure_spmf (search P s)"
    unfolding T.emeasure_space_1 by(simp add: measure_spmf.emeasure_eq_measure space_measure_spmf weight_spmf_def)
  also have "\<dots> = emeasure (T s) {\<omega> \<in> space (T s). (not (HLD {x. P x}) suntil (HLD {x. P x} aand nxt (\<lambda>_. True))) (s ## \<omega>)}"
    by(subst emeasure_suntil_HLD_search) simp_all
  also have "\<dots> = ?rhs"
    by(simp add: T.emeasure_eq_measure not_suntil_conv_ev HLD_def holds.simps[abs_def])
  finally show ?thesis by simp
qed

lemma weight_spmf_search': "weight_spmf (search' P s) = \<P>(\<omega> in T s. ev (holds P) (s ## \<omega>))"
using weight_spmf_search[of P s] unfolding search_def by simp

lemma emeasure_ev_HLD_search:
  assumes AP: "\<And>x. x \<in> A \<Longrightarrow> P x"
  shows
    "emeasure (T s) {\<omega> \<in> space (T s). ev (HLD A) (s ## \<omega>)} = 
     \<integral>\<^sup>+ s'. emeasure (T s') {\<omega> \<in> space (T s'). ev (HLD A) (s' ## \<omega>)} \<partial>measure_spmf (search P s)"
   (is "?lhs = ?rhs")
proof -
  have "?lhs = emeasure (T s) {\<omega> \<in> space (T s). (not (HLD {x. P x}) suntil (HLD {x. P x} aand ev (HLD A))) (s ## \<omega>)}"
    by(subst suntil_ev_redundant)(auto intro: AP simp add: HLD_def)
  also have "\<dots> = ?rhs" apply(subst emeasure_suntil_HLD_search) by simp_all
  finally show ?thesis .
qed

lemma measure_ev_HLD_search:
  "(\<And>x. x \<in> A \<Longrightarrow> P x)
  \<Longrightarrow> \<P>(\<omega> in T s. ev (HLD A) (s ## \<omega>)) = 
      \<integral> s'. \<P>(\<omega> in T s'. ev (HLD A) (s' ## \<omega>)) \<partial>measure_spmf (search P s)"
using emeasure_ev_HLD_search[of A P s] unfolding T.emeasure_eq_measure
by(subst (asm) nn_integral_eq_integral)(simp_all add: measure_spmf.integrable_const_bound[where B=1] measure_nonneg T.measure_le_1)

end

subsection {* Searching in sub-Markov chains *}

locale SMC_syntax =
  fixes K :: "'s \<Rightarrow> 's spmf"
begin
interpretation MC: MC_syntax "case_option (return_pmf None) K" .

definition search' :: "('s \<Rightarrow> bool) \<Rightarrow> 's \<Rightarrow> 's list spmf"
where "search' P s = MC.search' (case_option False P) (Some s) \<bind> (return_spmf \<circ> map the)"

definition search :: "('s \<Rightarrow> bool) \<Rightarrow> 's \<Rightarrow> 's spmf"
where "search P s = map_spmf (\<lambda>xs. last (s # xs)) (search' P s)"

abbreviation SMC_paths_from :: "'s \<Rightarrow> 's list set"
where "SMC_paths_from \<equiv> paths_from (SIGMA s:UNIV. set_spmf (K s))"

lemma MC_paths_from_None [simp]: "MC.MC_paths_from None = range (\<lambda>n. replicate n None)"
proof(rule equalityI; rule subsetI, (erule imageE, hypsubst)?)
  fix xs
  assume "xs \<in> MC.MC_paths_from None"
  then show "xs \<in> range (\<lambda>n. replicate n None)"
    by(induction xs)(auto 4 3 simp add: Cons_in_paths_from_iff intro: image_eqI[where x="Suc _"])
next
  fix n
  show "replicate n None \<in> MC.MC_paths_from None"
    by(induction n)(auto simp add: Cons_in_paths_from_iff)
qed

lemma in_SMC_paths_from: "xs \<in> SMC_paths_from x \<longleftrightarrow> map Some xs \<in> MC.MC_paths_from (Some x)"
by(induction xs arbitrary: x)(auto simp add: Cons_in_paths_from_iff in_set_spmf)

lemma search'_simps [simp]:
  "P s \<Longrightarrow> search' P s = return_spmf []"
  "\<not> P s \<Longrightarrow> search' P s = bind_spmf (K s) (\<lambda>s'. map_spmf (op # s') (search' P s'))"
by(simp_all add: search'_def map_spmf_bind_spmf o_def)(auto 4 4 simp add: search'_def bind_spmf_def bind_assoc_pmf in_set_spmf[symmetric] bind_map_pmf search_path.simps o_def bind_eq_return_pmf simp del: MC.search'_simps split: if_split_asm option.split intro: ccontr intro!: bind_pmf_cong[OF refl])

lemma search'_unfold:
  "search' P s =
  (if P s then return_spmf [] else bind_spmf (K s) (\<lambda>s'. map_spmf (op # s') (search' P s')))"
by simp

lemma search_alt_def:
  "search P s = MC.search (case_option False P) (Some s) \<bind> return_pmf"
by(auto simp add: search_def search'_def MC.search_def bind_map_spmf map_spmf_bind_spmf search_path.simps last_map intro!: bind_spmf_cong[OF refl] split: option.split_asm)

lemma set_spmf_search':
  "set_spmf (search' P s) = {xs \<in> SMC_paths_from s. search_path P (s, xs)}"
proof(rule set_eqI, unfold mem_Collect_eq, intro iffI conjI; (erule conjE)?)
  fix xs
  assume "xs \<in> set_spmf (search' P s)"
  then obtain ys where xs: "xs = map the ys"
    and ys: "ys \<in> MC.MC_paths_from (Some s)"
    and sp: "search_path (case_option False P) (Some s, ys)"
    by(auto simp add: search'_def set_bind_spmf)
  have *: "\<forall>y\<in>set ys. \<exists>x. y = Some x" using ys sp
    by(fastforce simp add: search_path.simps append_in_paths_from Cons_in_paths_from_iff split: if_split_asm option.split_asm dest: split_list)
  with ys show "xs \<in> SMC_paths_from s" unfolding xs by(simp add: in_SMC_paths_from cong: map_cong)
  show "search_path P (s, xs)" using sp xs *
    by(auto 4 3 simp add: search_path.simps last_map map_butlast[symmetric] dest: in_set_butlastD split: if_split_asm option.split_asm)
next
  fix xs
  assume "xs \<in> SMC_paths_from s" 
  hence "map Some xs \<in> MC.MC_paths_from (Some s)" by(simp add: in_SMC_paths_from)
  moreover
  assume "search_path P (s, xs)"
  hence "search_path (case_option False P) (Some s, map Some xs)"
    by(auto simp add: search_path.simps last_map map_butlast[symmetric] o_def split: option.split)
  ultimately show "xs \<in> set_spmf (search' P s)"
    by(auto simp add: search'_def set_bind_spmf o_def intro: rev_bexI)
qed

lemma in_set_spmf_search'_iff [simp]:
  "xs \<in> set_spmf (search' P s) \<longleftrightarrow> xs \<in> SMC_paths_from s \<and> search_path P (s, xs)"
by(simp add: set_spmf_search')

lemma spmf_search'_empty: "spmf (search' P s) [] = indicator {s. P s} s"
apply(subst search'_unfold)
apply(clarsimp split: split_indicator simp add: spmf_bind)
apply(subst spmf_map_outside)
apply auto
done

definition prob_path :: "'s \<Rightarrow> 's list \<Rightarrow> real"
where "prob_path x xs = MC.prob_path (Some x) (map Some xs)"

lemma prob_path_simps [simp]:
  "prob_path x [] = 1"
  "prob_path x (y # ys) = spmf (K x) y * prob_path y ys"
by(simp_all add: prob_path_def)

lemma prob_path_nonneg: "prob_path x xs \<ge> 0"
by(simp add: prob_path_def MC.prob_path_nonneg)

lemma prob_path_le_1: "prob_path x xs \<le> 1"
by(simp add: prob_path_def MC.prob_path_le_1)

lemma spmf_search': "spmf (search' P s) xs = (if search_path P (s, xs) then prob_path s xs else 0)"
  (is "?lhs = ?rhs")
proof -
  let ?f = "\<lambda>ys. if search_path (case_option False P) (Some s, ys) then MC.prob_path (Some s) ys else 0"
  have "ennreal ?lhs = \<integral>\<^sup>+ ys. ennreal (?f ys) * indicator {Some xs} (Some (map the ys)) \<partial>count_space UNIV"
    by(simp add: search'_def MC.spmf_search' ennreal_spmf_bind prob_path_def nn_integral_measure_spmf ennreal_indicator)
  also have "\<dots> = \<integral>\<^sup>+ ys. ennreal (?f ys) * indicator {xs} (map the ys) * indicator (lists (range Some)) ys \<partial>count_space UNIV"
    apply(intro nn_integral_cong)
    apply(auto split: split_indicator option.split_asm simp add: MC.prob_path_eq_0_iff_notin_paths_from search_path.simps notin_range_Some ennreal_eq_0_iff)
    apply(drule split_list, auto split: if_split_asm simp add: append_in_paths_from Cons_in_paths_from_iff)
    done
  also have "\<dots> = \<integral>\<^sup>+ ys. ennreal (?f ys) * indicator {xs} (map the ys) \<partial>count_space (lists (range Some))"
    by(simp add: nn_integral_count_space_indicator)
  also have "\<dots> = \<integral>\<^sup>+ ys. ennreal (?f (map Some ys)) * indicator {xs} (map the (map Some ys)) \<partial>count_space UNIV"
    by(rule nn_integral_bij_count_space[symmetric])(auto simp add: bij_betw_def ex_map_conv image_iff)
  also have "\<dots> = ?f (map Some xs)"
    by(simp add: o_def MC.prob_path_nonneg max_def)
  also have "\<dots> = ennreal ?rhs"
    by(auto simp add: prob_path_def search_path.simps last_map map_butlast[symmetric] split: option.split_asm)
  finally show ?thesis by(simp add: prob_path_nonneg)
qed

lemma spmf_search'_simps [simp]:
  shows spmf_search_path: "search_path P (s, xs) \<Longrightarrow> spmf (search' P s) xs = prob_path s xs"
  and spmf_search_eq_0: "\<not> search_path P (s, xs) \<Longrightarrow> spmf (search' P s) xs = 0"
by(auto simp add: spmf_search')

lemma search_unfold: "search P s = (if P s then return_spmf s else bind_spmf (K s) (search P))"
by(simp add: search_def[abs_def] map_spmf_bind_spmf pmf.map_comp o_def option.map_comp)

lemma search_simps[simp]: 
  "P s \<Longrightarrow> search P s = return_spmf s"
  "\<not> P s \<Longrightarrow> search P s = bind_spmf (K s) (search P)"
by(simp_all add: search_unfold)

lemma search_reachableI:
  assumes *: "(s, s') \<in> (SIGMA s:{s. \<not> P s}. set_spmf (K s))\<^sup>*"
  and "P s'"
  shows "s' \<in> set_spmf (search P s)"
using * by(induction rule: converse_rtrancl_induct)(auto simp add: set_bind_spmf \<open>P s'\<close>)

lemma search_reachableD:
  assumes "s' \<in> set_spmf (search P s)"
  shows "(s, s') \<in> (SIGMA s:{s. \<not> P s}. set_spmf (K s))\<^sup>*" "P s'"
proof -
  from assms obtain xs where s': "s' = last (s # xs)"
    and xs: "xs \<in> set_spmf (search' P s)"
    by(clarsimp simp add: search_def simp del: last.simps)
  from xs have "xs \<in> SMC_paths_from s" "search_path P (s, xs)" by simp_all
  then have "xs \<in> paths_from {(s, s')\<in>(SIGMA s:UNIV. set_spmf (K s)). \<not> P s} s"
    by(intro paths_from_restrict1)(simp_all add: search_path.simps)
  hence "(s, s') \<in> {(s, s')\<in>(SIGMA s:UNIV. set_spmf (K s)). \<not> P s}\<^sup>*"
    unfolding s' by(rule paths_from_into_rtrancl)
  also have "{(s, s')\<in>(SIGMA s:UNIV. set_spmf (K s)). \<not> P s} = (SIGMA s:{s. \<not> P s}. set_spmf (K s))"
    by auto
  finally show "(s, s') \<in> \<dots>\<^sup>*" .
  show "P s'" using s' \<open>search_path P (s, xs)\<close> by(simp add: search_path.simps)
qed

lemma set_spmf_search:
  "set_spmf (search P s) = {s'. (s, s') \<in> (SIGMA s:{s. \<not> P s}. set_spmf (K s))\<^sup>* \<and> P s'}"
by(auto intro: search_reachableI dest: search_reachableD)

lemma in_set_spmf_search_iff [simp]:
  "s' \<in> set_spmf (search P s) \<longleftrightarrow> P s' \<and> (s, s') \<in> (SIGMA s:{s. \<not> P s}. set_spmf (K s))\<^sup>*"
by(auto simp add: set_spmf_search)

subsubsection {* Trace space *}

abbreviation L :: "'s llist measure" where "L \<equiv> llist_space (count_space UNIV)"

text {* 
  Traces end when the transition goes to @{const None}, i.e., in a state @{term s},
  the probability of reaching the end is @{term "pmf (K s) None"}.
*}

definition T :: "'s \<Rightarrow> 's llist measure"
where
  "T s = distr (MC.T (Some s)) L (lmap the \<circ> stakeWhile (\<lambda>x. ~ Option.is_none x))"

lemma space_T_conv_llist_space: "space (T s) = space L"
by(simp add: T_def)

lemma space_T: "space (T s) = UNIV"
by(simp add: space_T_conv_llist_space space_llist_space)

lemma sets_T [measurable_cong]: "sets (T s) = sets L"
by(simp add: T_def)

lemma prob_space_T: "prob_space (T s)"
unfolding T_def by(rule MC.T.prob_space_distr) simp

interpretation T: prob_space "T s" for s
by(fact prob_space_T)

lemma subprob_space_T [simp]: "subprob_space (T s)"
by unfold_locales

lemma nn_integral_T:
  assumes [measurable]: "f \<in> borel_measurable L"
  shows
  "(\<integral>\<^sup>+ \<omega>. f \<omega> \<partial>T s) = 
   max 0 (f LNil) * pmf (K s) None + \<integral>\<^sup>+ s'. \<integral>\<^sup>+ \<omega>. f (LCons s' \<omega>) \<partial>T s' \<partial>measure_spmf (K s)"
  (is "?lhs = ?LNil + ?LCons")
proof -
  let ?f = "\<lambda>\<omega>. lmap the (stakeWhile (\<lambda>x. \<not> Option.is_none x) \<omega>)"
  let ?M = "measure_pmf (K s)"
  have "?lhs = \<integral>\<^sup>+ \<omega>. f (?f \<omega>) \<partial>MC.T (Some s)"
    unfolding T_def by(simp add: nn_integral_distr)
  also have "\<dots> = \<integral>\<^sup>+ s'. (if Option.is_none s' then max 0 (f LNil) else \<integral>\<^sup>+ \<omega>. f (LCons (the s') (?f \<omega>)) \<partial>MC.T s') \<partial>?M"
    by(subst MC.nn_integral_T)(auto intro!: nn_integral_cong simp add: MC.T.emeasure_space_1[simplified] max_def)
  also have "\<dots> = \<integral>\<^sup>+ s'. max 0 (f LNil) * indicator {None} s' + (\<integral>\<^sup>+ \<omega>. f (LCons (the s') (?f \<omega>)) \<partial>MC.T s') * indicator (range Some) s' \<partial>?M"
    by(rule nn_integral_cong)(auto simp add: Option.is_none_def)
  also have "\<dots> = max 0 (f LNil) * emeasure ?M {None} + \<integral>\<^sup>+ s'. (\<integral>\<^sup>+ \<omega>. f (LCons (the s') (?f \<omega>)) \<partial>MC.T s') * indicator (range Some) s' \<partial>?M"
    by(subst nn_integral_add)(simp_all)
  also have "\<dots> = ?LNil + \<integral>\<^sup>+ s'. (\<integral>\<^sup>+ \<omega>. f (LCons (the s') (?f \<omega>)) \<partial>MC.T s') \<partial>restrict_space ?M (range Some)"
     (is "_ = _ + ?ltl") by(simp add: emeasure_pmf_single nn_integral_restrict_space)
  also have "?ltl = \<integral>\<^sup>+ s'. \<integral>\<^sup>+ \<omega>. f (LCons s' (?f \<omega>)) \<partial>MC.T (Some s') \<partial>measure_spmf (K s)"
    by(auto intro!: nn_integral_cong simp add: space_restrict_space nn_integral_measure_spmf_conv_measure_pmf)
  also have "\<dots> = ?LCons" by(rule nn_integral_cong)(clarsimp simp add: T_def nn_integral_distr)
  finally show ?thesis .
qed

lemma emeasure_Collect_T':
  assumes [measurable]: "Measurable.pred L P"
  shows "emeasure (T s) {\<omega> \<in> space (T s). P \<omega>} =
    ennreal (pmf (K s) None) * indicator {\<omega>. P \<omega>} LNil +
    (\<integral>\<^sup>+ s'. emeasure (T s') {\<omega> \<in> space (T s'). P (LCons s' \<omega>)} \<partial>measure_spmf (K s))"
apply(subst (1 2) nn_integral_indicator[symmetric])
  apply simp
 apply simp
apply(subst nn_integral_T; simp add: max_def indicator_def[abs_def] space_T)
done

lemma emeasure_Collect_T:
  "Measurable.pred L P
  \<Longrightarrow> emeasure (T s) {\<omega>. P \<omega>} =
      ennreal (pmf (K s) None) * indicator {\<omega>. P \<omega>} LNil +
      (\<integral>\<^sup>+ s'. emeasure (T s') {\<omega>. P (LCons s' \<omega>)} \<partial>measure_spmf (K s))"
using emeasure_Collect_T'[of P s] by(simp add: space_T)

lemma measure_Collect_T':
  assumes [measurable]: "Measurable.pred L P"
  shows "\<P>(\<omega> in T s. P \<omega>) = pmf (K s) None * indicator {\<omega>. P \<omega>} LNil + (\<integral> s'. \<P>(\<omega> in T s'. P (LCons s' \<omega>)) \<partial>measure_spmf (K s))"
  (is "?lhs = ?rhs")
proof -
  have "ennreal ?lhs = emeasure (T s) {\<omega> \<in> space (T s). P \<omega>}" by(simp add: T.emeasure_eq_measure)
  also have "\<dots> = ennreal (pmf (K s) None) * indicator {\<omega>. P \<omega>} LNil +
    \<integral>\<^sup>+ s'. ennreal (measure (T s') {\<omega> \<in> space (T s'). P (LCons s' \<omega>)}) \<partial>measure_spmf (K s)"
    by(subst emeasure_Collect_T')(simp_all add: T.emeasure_eq_measure split_def)
  also have "\<dots> = ennreal ?rhs"
    by(simp add: nn_integral_eq_integral measure_spmf.integrable_const_bound[where B=1] measure_spmf.subprob_measure_le_1 measure_nonneg ennreal_mult ennreal_indicator)
  finally show ?thesis by(simp del: ennreal_plus)
qed

lemma measure_Collect_T:
  "Measurable.pred L P
  \<Longrightarrow> measure (T s) {\<omega>. P \<omega>} =
      pmf (K s) None * indicator {\<omega>. P \<omega>} LNil +
      (\<integral> s'. measure (T s') {\<omega>. P (LCons s' \<omega>)} \<partial>measure_spmf (K s))"
using measure_Collect_T'[of P s] by(simp add: space_T)

end

definition bind_spmf_llist :: "'a spmf \<Rightarrow> ('a \<Rightarrow> 'b llist measure) \<Rightarrow> 'b llist measure"
where
  "bind_spmf_llist p f = 
   measure_pmf p \<bind>
   (\<lambda>x. case x of None \<Rightarrow> Giry_Monad.return (llist_space (count_space UNIV)) LNil
    | Some y \<Rightarrow> f y)"

lemma space_bind_spmf_llist:
  "(\<And>x. sets (f x) = sets (llist_space (count_space UNIV)))
  \<Longrightarrow> space (bind_spmf_llist p f) = UNIV"
by(simp add: bind_spmf_llist_def space_bind[where N="llist_space (count_space UNIV)"] space_llist_space split: option.split)

lemma sets_bind_spmf_llist [measurable_cong]:
  "(\<And>x. sets (f x) = sets (llist_space (count_space UNIV)))
  \<Longrightarrow> sets (bind_spmf_llist p f) = sets (llist_space (count_space UNIV))"
by(simp add: bind_spmf_llist_def sets_bind[where N="llist_space (count_space UNIV)"] split: option.split)

lemma emeasure_bind_spmf_llist:
  assumes A: "A \<in> sets (llist_space (count_space UNIV))"
  and f: "f \<in> measurable (count_space UNIV) (subprob_algebra (llist_space (count_space UNIV)))"
  shows
  "emeasure (bind_spmf_llist p f) A =
   ennreal (pmf p None) * indicator A LNil + \<integral>\<^sup>+ x. emeasure (f x) A \<partial>measure_spmf p"
  (is "?lhs = ?rhs")
proof -
  from f have [simp, measurable_cong]: "\<And>x. sets (f x) = sets (llist_space (count_space UNIV))"
    by(auto simp add: space_subprob_algebra)
  interpret f: subprob_space "f x" for x using f by(auto simp add: space_subprob_algebra)
  let ?f = "\<lambda>x. case x of None \<Rightarrow> Giry_Monad.return (llist_space (count_space UNIV)) LNil | Some y \<Rightarrow> f y"
  have "?lhs = \<integral>\<^sup>+ x. emeasure (?f x) A \<partial>measure_pmf p" using assms unfolding bind_spmf_llist_def
    by(subst emeasure_bind[where N="llist_space (count_space UNIV)"])
      (auto 4 3 simp add: space_subprob_algebra subprob_space_return_ne f.subprob_space_distr split: option.splits)
  also have "\<dots> = \<integral>\<^sup>+ x. indicator A LNil * indicator {None} x + 
                        emeasure (f (the x)) A * indicator (range Some) x \<partial>measure_pmf p"
    by(rule nn_integral_cong)(auto simp add: emeasure_distr A split: option.split)
  also have "\<dots> = (\<integral>\<^sup>+ x. indicator A LNil * indicator {None} x \<partial>measure_pmf p) + 
           \<integral>\<^sup>+ x. emeasure (f (the x)) A * indicator (range Some) x \<partial>measure_pmf p"
    by(subst nn_integral_add)(simp_all)
  also have "\<dots> = ?rhs"
    by(simp add: emeasure_pmf_single max_def mult.commute nn_integral_measure_spmf_conv_measure_pmf nn_integral_restrict_space)
  finally show ?thesis .
qed

lemma return_in_space_subprob_algebra [simp]:
  "space M \<noteq> {} \<Longrightarrow> Giry_Monad.return M x \<in> space (subprob_algebra M)"
by(simp add: space_subprob_algebra subprob_space_return_ne)

lemma prob_space_bind_spmf_llist:
  "\<lbrakk> \<And>x. f x \<in> space (subprob_algebra (llist_space (count_space UNIV)));
     \<And>x. prob_space (f x) \<rbrakk>
   \<Longrightarrow> prob_space (bind_spmf_llist p f)"
unfolding bind_spmf_llist_def
by(rule measure_pmf.prob_space_bind[where S="llist_space (count_space UNIV)"])
  (auto 4 3 simp add: AE_measure_pmf_iff prob_space_return split: option.split)

lemma bind_spmf_llist_return_None [simp]:
  "(\<And>x. f x \<in> space (subprob_algebra (llist_space (count_space UNIV))))
  \<Longrightarrow> bind_spmf_llist (return_pmf None) f = Giry_Monad.return (llist_space (count_space UNIV)) LNil"
apply(simp add: bind_spmf_llist_def return_pmf.rep_eq)
apply(subst bind_return[where N="llist_space (count_space UNIV)"])
apply(auto 4 3 split: option.split)
done

lemma bind_spmf_llist_return_spmf [simp]:
  "(\<And>x. f x \<in> space (subprob_algebra (llist_space (count_space UNIV))))
  \<Longrightarrow> bind_spmf_llist (return_spmf x) f = f x"
apply(simp add: bind_spmf_llist_def return_pmf.rep_eq)
apply(subst bind_return[where N="llist_space (count_space UNIV)"])
apply(auto 4 3 split: option.split)
done

lemma bind_spmf_llist_bind_spmf [simp]:
  "(\<And>x. g x \<in> space (subprob_algebra (llist_space (count_space UNIV))))
  \<Longrightarrow> bind_spmf_llist (p \<bind> f) g = bind_spmf_llist p (\<lambda>x. bind_spmf_llist (f x) g)"
apply(simp add: bind_spmf_llist_def bind_spmf_def measure_pmf_bind)
apply(subst bind_assoc[where N="count_space UNIV" and R="llist_space (count_space UNIV)"])
apply(auto 4 3 simp add: return_pmf.rep_eq measure_subprob split: option.split intro!: bind_cong)
apply(subst bind_return[where N="llist_space (count_space UNIV)"])
apply(auto 4 3 simp add: return_pmf.rep_eq measure_subprob split: option.split)
done

lemma bind_spmf_llist_map_spmf [simp]:
  "(\<And>x. g x \<in> space (subprob_algebra (llist_space (count_space UNIV))))
  \<Longrightarrow> bind_spmf_llist (map_spmf f p) g = bind_spmf_llist p (g \<circ> f)"
by(simp add: map_spmf_conv_bind_spmf o_def)

context SMC_syntax begin

lemma T_unfold: "T s = bind_spmf_llist (K s) (\<lambda>s. distr (T s) L (LCons s))"
  (is "?lhs = ?rhs")
proof(rule measure_eqI)
  show "sets ?lhs = sets ?rhs"
    by(simp add: sets_T sets_bind_spmf_llist)

  fix A
  assume A: "A \<in> sets (T s)"
  have "\<And>x. LCons x -` A \<inter> space (T x) = {\<omega>. LCons x \<omega> \<in> A}" by(auto simp add: space_T)
  moreover
  from A have "Measurable.pred (llist_space (count_space UNIV)) (\<lambda>\<omega>. \<omega> \<in> A)" by simp
  note emeasure_Collect_T[OF this, of s] A
  ultimately show "emeasure ?lhs A = emeasure ?rhs A"
    by(subst emeasure_bind_spmf_llist)(auto simp add: emeasure_distr space_subprob_algebra sets_T space_T subprob_space.subprob_space_distr[OF subprob_space_T])
qed


definition T' :: "'s spmf \<Rightarrow> 's llist measure"
where "T' I = bind_spmf_llist I (\<lambda>s. distr (T s) L (LCons s))"

lemma distr_LCons_in_subprob [simp]: "distr (T s) L (LCons s') \<in> space (subprob_algebra L)"
by(simp add: space_subprob_algebra subprob_space.subprob_space_distr[OF subprob_space_T])

lemma sets_T': "sets (T' I) = sets L"
  by (simp add: T'_def sets_bind_spmf_llist)

lemma prob_space_T': "prob_space (T' I)"
unfolding T'_def
by(intro prob_space_bind_spmf_llist prob_space.prob_space_distr[OF prob_space_T]) simp_all

lemma T_eq_T': "T s = T' (K s)"
by(subst T_unfold)(simp add: T'_def)

end

end