theory Stream_Space_More
imports
  "$AFP/CryptHOL/Misc_CryptHOL"
begin

lemma stream_eq_conv_all_nth: "(xs = ys) \<longleftrightarrow> (\<forall>n. snth xs n = snth ys n)"
by (metis stream.map_cong0 stream_smap_nats)

lemma restrict_space_UNIV [simp]: "restrict_space M UNIV = M"
by(rule measure_eqI)(auto simp add: sets_restrict_space emeasure_restrict_space)

lemma fromN_Suc_eq_smap: "fromN (Suc n) = smap Suc (fromN n)"
  by (coinduction arbitrary: n) auto
  
lemma smap_streams:
  "X \<in> streams A \<Longrightarrow> f \<in> A \<rightarrow> B \<Longrightarrow> smap f X \<in> streams B"
  by (coinduction arbitrary: X)
     (blast intro: stream.map elim: streams.cases)

lemmas streams_SCons = streams_Stream

lemma streams_UNIV[simp]: "X \<in> streams UNIV"
  by (coinduction arbitrary: X) (blast intro: stream.exhaust)


lemma to_stream_nth: "to_stream X !! n = X n"
unfolding to_stream_def by simp

lemma snth_to_stream [simp]: "snth (to_stream x) = x"
by(simp add: fun_eq_iff to_stream_nth)

lemma to_stream_snth [simp]: "to_stream (snth x) = x"
by(simp add: stream_eq_conv_all_nth)

lemma stream_iso_sequence: "type_definition snth to_stream (UNIV :: (nat \<Rightarrow> 'a) set)"
by(unfold_locales) simp_all

lemma to_stream_inject [simp]: "to_stream f = to_stream g \<longleftrightarrow> f = g"
by(auto simp add: type_definition.Abs_inject[OF stream_iso_sequence])

lemma inj_to_stream [simp]: "inj to_stream"
by(simp add: inj_on_def)

lemma range_to_stream [simp]: "range to_stream = UNIV"
by(rule type_definition.Abs_image[OF stream_iso_sequence])

lemma snth_inject [simp]: "snth xs = snth ys \<longleftrightarrow> xs = ys"
by(rule type_definition.Rep_inject[OF stream_iso_sequence])

lemma inj_snth [simp]: "inj snth"
by(simp add: inj_on_def)



subsection {* Prefixes of streams *}

definition sprefix :: "'a list \<Rightarrow> 'a stream \<Rightarrow> bool"
where "sprefix xs zs \<longleftrightarrow> (\<exists>ys. zs = xs @- ys)"

lemma sprefix_Nil [simp]: "sprefix [] zs"
by(simp add: sprefix_def)

lemma sprefix_Cons_SCons: "sprefix (x # xs) (z ## zs) \<longleftrightarrow> x = z \<and> sprefix xs zs"
by(auto simp add: sprefix_def)

lemma sprefix_Cons [simp]: "sprefix (x # xs) zs \<longleftrightarrow> shd zs = x \<and> sprefix xs (stl zs)"
by(cases zs rule: stream.exhaust)(auto simp add: sprefix_Cons_SCons)

lemma sprefix_conv_stake: "sprefix xs zs \<longleftrightarrow> stake (length xs) zs = xs"
by(induction xs arbitrary: zs) simp_all

lemma sprefix_shift_same [simp]: "sprefix xs (xs @- zs)"
by(simp add: sprefix_def)

lemma sprefixD: "sprefix xs zs \<Longrightarrow> \<exists>ys. zs = xs @- ys"
by(auto simp add: sprefix_def)

lemma sprefix_append: "sprefix (xs @ ys) zs \<longleftrightarrow> sprefix xs zs \<and> sprefix ys (sdrop (length xs) zs)"
by(induct xs arbitrary: zs) simp_all

lemma sprefix_shift_stake [simp]:
  assumes "length xs \<le> n"
  shows "sprefix xs (stake n ys @- zs) \<longleftrightarrow> sprefix xs ys"
proof -
  from assms have f1: "\<And>w v. stake n (xs @- w) = v \<longrightarrow> take (length xs) v = xs"
    by(simp add: append_eq_conv_conj stake_shift)
  have "\<And>v w. stake (length (v :: 'a list)) (v @- w) = v" by (simp add: stake_shift)
  thus ?thesis using f1 
    by (metis (no_types) append_Nil2 append_eq_conv_conj length_stake sprefixD sprefix_shift_same stake_sdrop stake_shift)
qed

lemma sprefix_stake_same [simp]: "sprefix (stake n xs) xs"
by(induct n arbitrary: xs) simp_all

lemma sprefix_same_imp_eq:
  assumes "sprefix xs ys" "sprefix xs' ys"
  and "length xs = length xs'"
  shows "xs = xs'"
using assms(3,1,2) by(induct arbitrary: ys rule: list_induct2) auto

lemma sprefix_snthD: "\<lbrakk> sprefix xs ys; i < length xs \<rbrakk> \<Longrightarrow> xs ! i = ys !! i"
by(auto simp add: sprefix_def)

lemma sprefix_snthI: "(\<And>i. i < length xs \<Longrightarrow> xs ! i = ys !! i) \<Longrightarrow> sprefix xs ys"
by(auto simp add: sprefix_def shift_snth sdrop_snth intro!: exI[where x="sdrop (length xs) ys"] snth_inject[THEN iffD1, OF ext])

lemma sprefix_down_linear: 
  "\<lbrakk> sprefix xs zs; sprefix ys zs \<rbrakk>
  \<Longrightarrow> \<exists>us. xs = ys @ us \<or> ys = xs @ us"
apply(rule exI[where x="if length xs \<le> length ys then drop (length xs) ys else drop (length ys) xs"])
apply(auto simp add: append_eq_conv_conj sprefix_conv_stake stake_shift)
apply(metis append_Nil2 append_take_drop_id drop_all length_drop list.size(3) sprefix_conv_stake sprefix_def stake.simps(1) stake_shift)
apply(metis append_Nil2 append_take_drop_id drop_eq_Nil length_drop linear sprefix_conv_stake sprefix_def stake_shift)
done



(*
definition "stream_cylinder X A = { s ## x | s x. s \<in> X \<and> x \<in> A}"

inductive_set stream_cylinders for \<Omega> :: "'a set" and \<A> :: "'a set set" where
  base: "streams \<Omega> \<in> stream_cylinders \<Omega> \<A>"
| step: "X \<in> \<A> \<Longrightarrow> A \<in> stream_cylinders \<Omega> \<A> \<Longrightarrow> stream_cylinder X A \<in> stream_cylinders \<Omega> \<A>"

hide_fact (open) base step

lemma UNIV_in_stream_cylinders [simp]: "UNIV \<in> stream_cylinders UNIV UNIV"
using stream_cylinders.base[of UNIV UNIV] by simp

lemma stream_cylinders_mono:
  assumes "A \<in> stream_cylinders \<Omega> \<A>" "\<A> \<subseteq> \<B>"
  shows "A \<in> stream_cylinders \<Omega> \<B>"
using assms by induct (auto intro: stream_cylinders.intros)

lemma stream_cylinder_empty1[simp]: "stream_cylinder {} A = {}"
by (auto simp: stream_cylinder_def)

lemma stream_cylinder_empty2[simp]: "stream_cylinder A {} = {}"
by (auto simp: stream_cylinder_def)

lemma stream_cylinder_iff_empty[simp]: "stream_cylinder A B = {} \<longleftrightarrow> A = {} \<or> B = {}"
by (auto simp: stream_cylinder_def)

lemma stream_cylinders_subset_streams:
  assumes subset: "\<A> \<subseteq> Pow \<Omega>"
  assumes A: "A \<in> stream_cylinders \<Omega> \<A>"
  shows "A \<subseteq> streams \<Omega>"
using A subset by (induct A) (auto dest: sets.sets_into_space simp: stream_cylinder_def)

lemma stream_cylinders_subset_Pow: "\<A> \<subseteq> Pow \<Omega> \<Longrightarrow> stream_cylinders \<Omega> \<A> \<subseteq> Pow (streams \<Omega>)"
using stream_cylinders_subset_streams by blast

lemma stream_cylinders_Int_stable:
  assumes subset: "\<A> \<subseteq> Pow \<Omega>"
  assumes "Int_stable \<A>"
  shows "Int_stable (stream_cylinders \<Omega> \<A>)"
proof -
  let ?C = "stream_cylinders \<Omega> \<A>"
  { fix A B assume "A \<in> ?C" "B \<in> ?C"
    then have "A \<inter> B \<in> ?C"
    proof (induct A arbitrary: B)
      case base with subset show ?case
        by (simp add: stream_cylinders_subset_streams Int_absorb1)
    next
      case (step X A) note A = this
      from `B \<in> ?C` show ?case
      proof cases
        case base
        with stream_cylinders_subset_streams[OF subset, of A] subset step
        have "stream_cylinder X A \<inter> B = stream_cylinder X A"
          by (auto simp: stream_cylinder_def dest: sets.sets_into_space)
        with step show ?thesis
          by (simp add: Int_absorb2 stream_cylinders.step)
      next
        case (step Y B')
        then have *: "stream_cylinder X A \<inter> B = stream_cylinder (X \<inter> Y) (A \<inter> B')"
          by (auto simp: stream_cylinder_def dest: sets.sets_into_space)
        with step A `Int_stable \<A>` show ?thesis
          by (auto intro: stream_cylinders.intros simp: Int_stable_def)
      qed
    qed }
  then show ?thesis
    by (auto simp: Int_stable_def)
qed

lemma shift_stream_cylindersI:
  assumes "A \<in> stream_cylinders \<Omega> \<A>"
  and "set xs \<subseteq> \<Omega>" "\<And>x. x \<in> set xs \<Longrightarrow> {x} \<in> \<A>"
  shows "shift xs ` A \<in> stream_cylinders \<Omega> \<A>"
using assms(2-3)
proof(induction xs)
  case Nil
  with assms(1) show ?case by(simp add: shift.simps[abs_def])
next
  case (Cons x xs)
  have "shift (x # xs) ` A = stream_cylinder {x} (shift xs ` A)"
    by(auto simp add: stream_cylinder_def)
  also from Cons have "shift xs ` A \<in> stream_cylinders \<Omega> \<A>" by auto
  with _ have "stream_cylinder {x} (shift xs ` A) \<in> stream_cylinders \<Omega> \<A>" 
    by(rule stream_cylinders.step)(simp add: Cons.prems)
  finally show ?case .
qed

definition to_stream :: "(nat \<Rightarrow> 'a) \<Rightarrow> 'a stream"
where "to_stream X = smap X nats"
*)
(*
definition stream_space :: "'a measure \<Rightarrow> 'a stream measure"
where "stream_space M = distr (\<Pi>\<^sub>M i\<in>UNIV. M) (sigma (streams (space M)) (stream_cylinders (space M) (sets M))) to_stream"
*)
lemma sets_stream_space:
  "sets (stream_space M) = sets (vimage_algebra (streams (space M)) snth (\<Pi>\<^sub>M i\<in>UNIV. M))"
by(simp add: stream_space_def)
(*
lemma
  shows sets_stream_space: "sets (stream_space M) = sigma_sets (streams (space M)) (stream_cylinders (space M) (sets M))"
  by (auto simp: stream_space_def stream_cylinders_subset_Pow sets.space_closed)
*)
(*
lemma stream_space_SCons:
  "x ## \<omega> \<in> space (stream_space M) \<longleftrightarrow> x \<in> space M \<and> \<omega> \<in> space (stream_space M)"
  by (simp add: space_stream_space streams_SCons)
*)
(*
lemma cylinders_stream_space[intro]: "A \<in> stream_cylinders (space M) (sets M) \<Longrightarrow> A \<in> sets (stream_space M)"
  by (simp add: stream_space_def stream_cylinders_subset_Pow sets.space_closed)
*)
(*
lemma measurable_stream:
  assumes f: "f \<in> space N \<rightarrow> space (stream_space M)"
  assumes sets: "\<And>A. A \<in> stream_cylinders (space M) (sets M) \<Longrightarrow> {\<omega>\<in>space N. f \<omega> \<in> A } \<in> sets N"
  shows "f \<in> measurable N (stream_space M)"
  apply (simp add: stream_space_def)
proof (rule measurable_measure_of[OF stream_cylinders_subset_Pow])
  show "f \<in> space N \<rightarrow>  streams (space M)"
    using f by (auto simp: space_stream_space)
  fix A assume "A \<in> stream_cylinders (space M) (sets M)"
  moreover have "f -` A \<inter> space N = {\<omega>\<in>space N. f \<omega> \<in> A }"
    by auto
  ultimately show "f -` A \<inter> space N \<in> sets N"
    by (auto intro: sets)
qed (rule sets.space_closed)

lemma measurable_stl[measurable]: "stl \<in> measurable (stream_space M) (stream_space M)"
proof (rule measurable_stream)
  show "stl \<in> space (stream_space M) \<rightarrow> space (stream_space M)"
    by (auto elim: streams.cases simp: space_stream_space)
  fix A assume A: "A \<in> stream_cylinders (space M) (sets M)"
  then have "{\<omega> \<in> space (stream_space M). stl \<omega> \<in> A} = stream_cylinder (space M) A"
    using stream_cylinders_subset_streams[OF sets.space_closed]
    by (auto simp: stream_cylinder_def space_stream_space elim: streams.cases)
  also have "\<dots> \<in> sets (stream_space M)"
    by (auto intro: stream_cylinders.intros A)
  finally show "{\<omega> \<in> space (stream_space M). stl \<omega> \<in> A} \<in> sets (stream_space M)" .
qed
*)
(*
lemma measurable_stream_coinduct[consumes 1, case_names measurable, coinduct set: measurable]:
  assumes "F f"
  assumes F: "\<And>f. F f \<Longrightarrow>
    (\<exists>f'. (\<exists>g\<in>measurable N M. \<exists>n\<in>measurable N N. \<forall>x\<in>space N. f x = g x ## f' (n x)) \<and> (F f' \<or> f' \<in> measurable N (stream_space M)))"
  shows "f \<in> measurable N (stream_space M)"
proof (rule measurable_stream)
  { fix f' x' assume "F f'" "x' \<in> space N" then have "f' x' \<in> streams (space M)"
    proof (coinduction arbitrary: x' f')
      case (streams x f) 
      with F[OF `F f`] obtain f' g n where "g \<in> measurable N M" "n \<in> measurable N N"
        "f x = g x ## f' (n x)" "F f' \<or> f' \<in> measurable N (stream_space M)"
        by blast
      with `x \<in> space N` show ?case
        by (auto dest: measurable_space simp: space_stream_space)
      qed }
  note F_space = this
  with `F f` show f: "f \<in> space N \<rightarrow> space (stream_space M)"
    by (auto simp: space_stream_space)

  fix C assume C: "C \<in> stream_cylinders (space M) (sets M)"
  show "{\<omega> \<in> space N. f \<omega> \<in> C} \<in> sets N"
    using C `F f`
  proof (induct arbitrary: f)
    case base
    with F_space have "{\<omega> \<in> space N. f \<omega> \<in> streams (space M)} = space N"
      by auto
    then show ?case
      by auto
  next
    case (step X A)
    note step(1)[measurable]
    from F `F f` obtain f' g n where f': "F f' \<or> f' \<in> measurable N (stream_space M)"
      and f'_unfold: "\<And>x. x \<in> space N \<Longrightarrow> f x = g x ## f' (n x)"
      and g[measurable]: "g \<in> measurable N M"
      and n[measurable]: "n \<in> measurable N N"
      by blast
    note cylinders_stream_space[measurable]
    from f' step(2,3) have [measurable]: "{\<omega> \<in> space N. f' \<omega> \<in> A} \<in> sets N"
      by (cases "F f'") auto

    have "\<And>x. x \<in> A \<Longrightarrow> x \<in> streams (space M)"
      using step stream_cylinders_subset_streams[OF sets.space_closed] by auto
    then have "{\<omega> \<in> space N. f \<omega> \<in> stream_cylinder X A} = {\<omega> \<in> space N. g \<omega> \<in> X \<and> f' (n \<omega>) \<in> A}"
      unfolding stream_cylinder_def by (auto simp: f'_unfold)
    then show ?case
      by simp
  qed
qed
*)
context fixes SHD :: "'a \<Rightarrow> 'b" and STL :: "'a \<Rightarrow> 'a" begin
primcorec unfold_stream :: "'a \<Rightarrow> 'b stream"
where "unfold_stream x = SHD x ## unfold_stream (STL x)"
end

lemmas stream_unfold = unfold_stream.code

lemma snth_unfold_stream: "unfold_stream h t s !! n = h ((t ^^ n) s)"
proof(induction n arbitrary: s)
  case 0 show ?case by(subst stream_unfold) simp
qed(simp add: funpow_swap1)

lemma measurable_fun_pow [measurable]:
  assumes [measurable]: "f \<in> measurable M M"
  shows "f ^^ n \<in> measurable M M"
by(induction n) simp_all

lemma measurable_stream_unfold [measurable]:
  assumes [measurable]: "n \<in> measurable M' M'" "h \<in> measurable M' M"
  shows "unfold_stream h n \<in> measurable M' (stream_space M)"
by(rule measurable_stream_space2)(simp add: snth_unfold_stream)

lemma measurable_fromN: "(\<lambda>X. smap X (fromN n)) \<in> measurable (\<Pi>\<^sub>M u\<in>UNIV. M) (stream_space M)"
by(rule measurable_stream_space2) simp

lemmas measurable_snth' [measurable] = measurable_snth_PiM

lemma PiM_eq_distr_stream_space: "Pi\<^sub>M UNIV (\<lambda>_. M) = distr (stream_space M) (Pi\<^sub>M UNIV (\<lambda>i. M)) snth"
by(subst stream_space_eq_distr)(simp add: distr_distr o_def)

(*
lemma sets_stream_space_generated:
  assumes M: "sets M = sigma_sets (space M) \<G>" "\<G> \<subseteq> Pow (space M)"
    and C: "space M = \<Union>C" "countable C" "C \<subseteq> \<G>"
  shows "sets (stream_space M) = sigma_sets (streams (space M)) (stream_cylinders (space M) \<G>)"
   (is "_ = ?R")
proof -
  note G_closed = stream_cylinders_subset_Pow[OF M(2)]

  def G \<equiv> "sigma (streams (space M)) (stream_cylinders (space M) \<G>)"
  then have sets_G: "sets G = ?R" and space_G: "space G = streams (space M)"
    by (simp_all add: G_closed)

  have [measurable]: "shd \<in> measurable G M"
  proof (rule measurable_sigma_sets[OF M])
    fix Y assume Y: "Y \<in> \<G>"
    with M have *: "shd -` Y \<inter> space G = stream_cylinder Y (streams (space M))"
      by (auto simp: stream_cylinder_def space_G elim: streams.cases)
    from Y show "shd -` Y \<inter> space G \<in> sets G"
      unfolding * by (auto simp: sets_G intro: stream_cylinders.intros)
  qed (simp add: space_G streams_shd)

  have [measurable]: "stl \<in> measurable G G"
  proof (rule measurable_sigma_sets[OF sets_G G_closed])
    fix Y assume Y: "Y \<in> stream_cylinders (space M) \<G>"
    with G_closed have "stl -` Y \<inter> space G = (\<Union>c\<in>C. stream_cylinder c Y)"
      apply (auto simp: stream_cylinder_def space_G C(1) elim: streams.cases)
      apply (erule_tac streams.cases)
      apply auto
      done
    also have "\<dots> \<in> sets G"
      using Y C(2, 3)
      by (intro sets.countable_UN') (auto simp: sets_G intro: stream_cylinders.intros)
    finally show "stl -` Y \<inter> space G \<in> sets G" .
  qed (simp add: space_G streams_stl)
    
  have "sets (stream_space M) = sigma_sets (space G) (stream_cylinders (space M) (sets M))"
    by (simp add: sets_stream_space space_G)
  also have "\<dots> \<subseteq> sets G"
  proof (safe intro!: sets.sigma_sets_subset)
    fix A assume "A \<in> stream_cylinders (space M) (sets M)"
    then show "A \<in> sets G"
    proof induction
      case (step X A)
      then have "(shd -` X \<inter> space G) \<inter> (stl -` A \<inter> space G) \<in> sets G"
        by measurable
      also have "(shd -` X \<inter> space G) \<inter> (stl -` A \<inter> space G) = stream_cylinder X A"
        using step sets.sets_into_space stream_cylinders_subset_streams[OF sets.space_closed, of A M]
        by (auto simp: stream_cylinder_def space_G elim: streams.cases)
      finally show ?case .
    qed (simp add: space_G[symmetric])
  qed
  finally have "sets (stream_space M) \<subseteq> sets G" .
  moreover have "sets G \<subseteq> sets (stream_space M)"
    unfolding sets_G space_stream_space[symmetric]
  proof (safe intro!: sets.sigma_sets_subset)
    fix A assume "A \<in> stream_cylinders (space M) \<G>"
    then have "A \<in> stream_cylinders (space M) (sets M)"
      by (rule stream_cylinders_mono) (auto simp: M)
    then show "A \<in> sets (stream_space M)"
      by auto
  qed
  finally show ?thesis
    by (simp add: sets_G)
qed
*)
(*
lemma measure_eqI_stream:
  assumes "prob_space A"
  assumes G: "sets M = sigma_sets (space M) \<G>" "\<G> \<subseteq> Pow (space M)" "Int_stable \<G>"
    and C: "space M = \<Union>C" "countable C" "C \<subseteq> \<G>"
  assumes sets_eq: "sets A = sets (stream_space M)" "sets B = sets (stream_space M)"
  assumes eq: "\<And>X. X \<in> stream_cylinders (space M) \<G> \<Longrightarrow> emeasure A X = emeasure B X"
  shows "A = B"
proof (rule measure_eqI_generator_eq)
  interpret A: prob_space A by fact
  show "Int_stable (stream_cylinders (space M) \<G>)" "stream_cylinders (space M) \<G> \<subseteq> Pow (streams (space M))"
    by (intro stream_cylinders_Int_stable stream_cylinders_subset_Pow G)+
  show "sets A = sigma_sets (streams (space M)) (stream_cylinders (space M) \<G>)"
      "sets B = sigma_sets (streams (space M)) (stream_cylinders (space M) \<G>)"
    unfolding sets_eq by (intro sets_stream_space_generated[where C=C] G C)+
  have C_ne: "C \<noteq> {}"
    using C(1) A.not_empty sets_eq_imp_space_eq[OF sets_eq(1)] by (auto simp: space_stream_space streams_empty)
  then have from_nat_into_C: "\<And>i. from_nat_into C i \<in> C"
    by (rule from_nat_into)
  with C show "range (\<lambda>i. stream_cylinder (from_nat_into C i) (streams (space M))) \<subseteq> stream_cylinders (space M) \<G>"
    by (auto intro!: stream_cylinders.intros from_nat_into)
  show "(\<Union>i. stream_cylinder (from_nat_into C i) (streams (space M))) = streams (space M)"
    using from_nat_into_C C C_ne from_nat_into_surj[OF C(2)]
    apply (auto simp: stream_cylinder_def elim: streams.cases)
    apply (erule_tac streams.cases)
    apply auto
    apply meti 
    done
  { fix i show "emeasure A (stream_cylinder (from_nat_into C i) (streams (space M))) \<noteq> \<infinity>"
      by simp }
qed fact

lemma measure_eqI_stream_countable:
  assumes "prob_space A" and "countable S"
  assumes sets_eq: "sets A = sets (stream_space (count_space S))" "sets B = sets (stream_space (count_space S))"
  assumes eq: "\<And>X. X \<in> stream_cylinders S ((\<lambda>s. {s})`S) \<Longrightarrow> emeasure A X = emeasure B X"
  shows "A = B"
proof (rule measure_eqI_stream[OF `prob_space A`])
  show "sets A = sets (stream_space (count_space S))" "sets B = sets (stream_space (count_space S))"
    by fact+
  let ?\<Omega> = "(space (count_space S))" and ?G = "((\<lambda>s. {s}) ` S \<union> {{}})"
  interpret G: sigma_algebra ?\<Omega> "sigma_sets ?\<Omega> ?G"
    by (rule sigma_algebra_sigma_sets) auto
  show "sets (count_space S) = sigma_sets (space (count_space S)) ?G" (is "_ = ?R")
  proof safe
    fix X assume "X \<in> sets (count_space S)"
    then have "X = (\<Union>s\<in>S. if s \<in> X then {s} else {})"
      by (auto split: split_if_asm)
    also have "\<dots> \<in> ?R"
      by (intro G.countable_UN' `countable S`) auto
    finally show "X \<in> ?R" .
  next
    fix X assume "X \<in> ?R"
    from G.sets_into_space[OF this]
    show "X \<in> sets (count_space S)"
      by auto
  qed
  show "countable ((\<lambda>s. {s}) ` S \<union> {{}})"
    by (intro countable_image `countable S` countable_Un countable_empty countable_insert)
  show "(\<lambda>s. {s}) ` S \<union> {{}} \<subseteq> Pow (space (count_space S))" "Int_stable ((\<lambda>s. {s}) ` S \<union> {{}})"
    "space (count_space S) = \<Union>(((\<lambda>s. {s}) ` S) \<union> {{}})"
    by (auto simp: Int_stable_def)
  show "(\<lambda>s. {s}) ` S \<union> {{}} \<subseteq> (\<lambda>s. {s}) ` S \<union> {{}}"
    by auto
next
  fix X assume "X \<in> stream_cylinders (space (count_space S)) ((\<lambda>s. {s}) ` S \<union> {{}})"
  then have "X \<in> stream_cylinders S ((\<lambda>s. {s}) ` S) \<or> X = {}"
    by induct (auto intro: stream_cylinders.intros)
  with eq show "emeasure A X = emeasure B X"
    by auto
qed
*)
lemma measurable_shift [measurable]:
  "set xs \<subseteq> space M \<Longrightarrow> shift xs \<in> measurable (stream_space M) (stream_space M)"
by(induction xs)(simp_all add: shift.simps[abs_def])

lemma (in algebra) finite_INT':
  "\<lbrakk> finite I; \<And>i. i \<in> I \<Longrightarrow> A i \<in> M \<rbrakk> \<Longrightarrow> (\<Inter>i\<in>I. A i) \<inter> \<Omega> \<in> M"
by(cases "I = {}") auto

lemma measurable_sprefix: 
  assumes "\<And>x. x \<in> set xs \<Longrightarrow> {x} \<in> sets M"
  shows "sprefix xs \<in> measurable (stream_space M) (count_space UNIV)"
proof -
  from assms have [measurable]: "\<And>i. i < length xs \<Longrightarrow> {xs ! i} \<in> sets M"
    unfolding in_set_conv_nth by auto
  have "{\<omega> \<in> space (stream_space M). sprefix xs \<omega>} =  
    snth -` ((\<Inter>i\<in>{..<length xs}. {\<omega> \<in> space (Pi\<^sub>M UNIV (\<lambda>_. M)). xs ! i = \<omega> i}) \<inter> space (Pi\<^sub>M UNIV (\<lambda>_. M))) \<inter> space (stream_space M)"
    by(auto dest: measurable_snth_PiM[THEN measurable_space] sprefix_snthD intro: sprefix_snthI)

  also have "\<dots> \<in> sets (stream_space M)"
    unfolding sets_stream_space space_stream_space
    by(rule in_vimage_algebra)(rule sets.finite_INT', simp_all)
  finally show ?thesis by(unfold pred_def)
qed

lemma measurable_prefix_UNIV [measurable]:
  "sprefix xs \<in> measurable (stream_space (count_space UNIV)) (count_space UNIV)"
using measurable_sprefix[of xs "count_space UNIV"] by simp





definition comb_stream :: "nat \<Rightarrow> 'a stream \<Rightarrow> 'a stream \<Rightarrow> 'a stream"
where "comb_stream n xs ys = to_stream (comb_seq n (snth xs) (snth ys))"

lemma comb_stream_alt_def: "comb_stream n xs ys = stake n xs @- ys"
by(simp add: comb_stream_def stream_eq_conv_all_nth to_stream_def comb_seq_def)

lemma comb_stream_0 [simp]: "comb_stream 0 xs ys = ys"
by(simp add: comb_stream_alt_def)

lemma comb_stream_Suc_SCons [simp]: "comb_stream (Suc n) (x ## xs) ys = x ## comb_stream n xs ys"
by(simp add: comb_stream_alt_def)

lemma sdrop_comb_stream2 [simp]: "n \<ge> m \<Longrightarrow> sdrop n (comb_stream m xs ys) = sdrop (n - m) ys"
by(simp add: comb_stream_alt_def sdrop_shift)

lemma sprefix_comb_stream1 [simp]: "length xs \<le> n \<Longrightarrow> sprefix xs (comb_stream n ys zs) \<longleftrightarrow> sprefix xs ys"
by(simp add: comb_stream_alt_def)

lemma measurable_comb_stream [measurable]:
  "(\<lambda>(\<omega>, \<omega>'). comb_stream i \<omega> \<omega>') \<in> measurable (stream_space M \<Otimes>\<^sub>M stream_space M) (stream_space M)"
unfolding comb_stream_def by simp

lemma sdrop_comb_stream: "sdrop n (comb_stream m xs ys) = drop n (stake m xs) @- sdrop (n - m) ys"
by(simp add: comb_stream_alt_def sdrop_shift)

lemma (in prob_space) nn_integral_stream_space_split:
  assumes [measurable]: "f \<in> borel_measurable (stream_space M)"
  shows "(\<integral>\<^sup>+X. f X \<partial>stream_space M) = (\<integral>\<^sup>+Y. (\<integral>\<^sup>+X. f (comb_stream n Y X) \<partial>stream_space M) \<partial>stream_space M)"
proof -
  interpret sequence_space M by unfold_locales
  show ?thesis unfolding comb_stream_def (* could this be done with transfer? *)
    apply(subst stream_space_eq_distr)
    apply(simp add: nn_integral_distr)
    apply(subst nn_integral_split[where i=n], simp)
    apply(subst PiM_eq_distr_stream_space)
    apply(simp add: nn_integral_distr)
    apply(rule nn_integral_cong)
    apply(subst PiM_eq_distr_stream_space)
    apply(simp add: nn_integral_distr)
    done
qed

lemma measurable_component_update':
  assumes "x \<in> space (Pi\<^sub>M I M)" "i \<in> I"
  shows "(\<lambda>v. x(i := v)) \<in> measurable (M i) (Pi\<^sub>M I M)"
proof -
  from assms have "(\<lambda>v. x(i := v)) \<in> measurable (M i) (Pi\<^sub>M (insert i I) M)" by simp
  with \<open>i \<in> I\<close> show ?thesis by(simp add: insert_absorb)
qed

lemma fun_upd_in_space_PiM:
  "\<lbrakk> f \<in> space (Pi\<^sub>M I M); x \<in> I; y \<in> space (M x) \<rbrakk> \<Longrightarrow> f(x := y) \<in> space (Pi\<^sub>M I M)"
by(rule measurable_space)(rule measurable_component_update')

context product_prob_space begin

interpretation Mi: product_prob_space M "insert i I" ..

lemma distr_PiM_insert:
  shows "distr (Pi\<^sub>M I M \<Otimes>\<^sub>M M i) (Pi\<^sub>M (insert i I) M) (\<lambda>(f, y). f(i := y)) = Pi\<^sub>M (insert i I) M"
  (is "?D = _" is "distr ?Mi _ ?f = _")
proof(rule Mi.PiM_eq)
  fix J F
  assume J: "finite J" "J \<subseteq> insert i I"
    and F: "\<And>j. j \<in> J \<Longrightarrow> F j \<in> sets (M j)"
  let ?X = "Mi.emb (insert i I) J (Pi\<^sub>E J F)"

  have "\<And>j x. j \<in> J \<Longrightarrow> x \<in> F j \<Longrightarrow> x \<in> space (M j)"
    using F[THEN sets.sets_into_space] by(auto simp: space_PiM Pi_iff subset_eq)
  with J
  have "?f -` ?X \<inter> space ?Mi = Mi.emb I (J - {i}) (PIE j:J - {i}. F j) \<times> (if i \<in> J then F i else space (M i))"
    (is "_ = ?E \<times> ?F")
    by(fastforce simp add: prod_emb_def PiE_iff space_pair_measure space_PiM Int_absorb1 split: if_split_asm)
  then have "emeasure ?D ?X = emeasure (Pi\<^sub>M I M \<Otimes>\<^sub>M M i) (?E \<times> ?F)" using J F
    by(subst emeasure_distr) auto
  also have "\<dots> = emeasure (Pi\<^sub>M I M) ?E * emeasure (M i) ?F" using J F
    by(intro M.emeasure_pair_measure_Times) auto
  also have "emeasure (Pi\<^sub>M I M) ?E = (\<Prod>j\<in>(J - {i}). emeasure (M j) (F j))" using J F
    by(intro emeasure_PiM_emb) auto
  also have "\<dots> * emeasure (M i) ?F = (\<Prod>j\<in>J. emeasure (M j) (F j))" using J
    using prod.subset_diff[of "{i}" J, symmetric]
    by(auto simp add: M.emeasure_space_1)
  finally show "emeasure ?D ?X = \<dots>" .
qed simp_all

lemma distr_PiM_insert':
  "i \<in> I \<Longrightarrow> distr (Pi\<^sub>M I M \<Otimes>\<^sub>M M i) (Pi\<^sub>M I M) (\<lambda>(f, y). f(i := y)) = Pi\<^sub>M I M"
using distr_PiM_insert[of i] by(simp add: insert_absorb)

lemma nn_integral_PiM_insert:
  assumes [measurable]: "f \<in> borel_measurable (Pi\<^sub>M (insert i I) M)"
  shows "(\<integral>\<^sup>+X. f X \<partial>Pi\<^sub>M (insert i I) M) = (\<integral>\<^sup>+x. (\<integral>\<^sup>+X. f (X(i := x)) \<partial>Pi\<^sub>M I M) \<partial>M i)"
proof -
  let ?PMi = "Pi\<^sub>M (insert i I) M"
  let ?D = "distr (Pi\<^sub>M I M \<Otimes>\<^sub>M M i) (Pi\<^sub>M (insert i I) M) (\<lambda>(f, y). f(i := y))"
  have "(\<integral>\<^sup>+ X. f X \<partial>?PMi) = (\<integral>\<^sup>+ X. f X \<partial>?D)" by(simp add: distr_PiM_insert)
  also have "\<dots> = \<integral>\<^sup>+(g, y). f (g(i := y)) \<partial>(Pi\<^sub>M I M \<Otimes>\<^sub>M M i)"
    by(subst nn_integral_distr)(simp_all add: split_def)
  also have "\<dots> = \<integral>\<^sup>+ y. \<integral>\<^sup>+ g. f (g(i := y)) \<partial>Pi\<^sub>M I M \<partial>M i"
    by(subst pair_sigma_finite.nn_integral_snd[symmetric])(simp_all, intro_locales)
  finally show ?thesis .
qed

lemma nn_integral_PiM_insert':
  "\<lbrakk> f \<in> borel_measurable (Pi\<^sub>M I M); i \<in> I \<rbrakk>
  \<Longrightarrow> (\<integral>\<^sup>+X. f X \<partial>Pi\<^sub>M I M) = (\<integral>\<^sup>+x. \<integral>\<^sup>+X. f (X(i := x)) \<partial>Pi\<^sub>M I M \<partial>M i)"
using nn_integral_PiM_insert[of f i] by(simp add: insert_absorb)

lemma emeasure_PiM_insert:
  assumes X[measurable]: "X \<in> sets (Pi\<^sub>M (insert i I) M)"
  shows "emeasure (Pi\<^sub>M (insert i I) M) X = (\<integral>\<^sup>+y. emeasure (Pi\<^sub>M I M) {f\<in>space (Pi\<^sub>M I M). f(i := y) \<in> X} \<partial>M i)"
proof -
  have eq: "\<And>y f. f \<in> space (Pi\<^sub>M I M) \<Longrightarrow> y \<in> space (M i) \<Longrightarrow>
      indicator X (f(i := y)) = indicator {f\<in>space (Pi\<^sub>M I M). f(i := y) \<in> X } f"
    by (auto split: split_indicator)
  show ?thesis
    using nn_integral_PiM_insert[of "indicator X" i]
    apply (auto intro!: nn_integral_cong)
    apply (subst nn_integral_cong)
    apply (rule eq)
    apply simp_all
    done
qed

lemma emeasure_PiM_insert':
  "\<lbrakk> X \<in> sets (Pi\<^sub>M I M); i \<in> I \<rbrakk>
  \<Longrightarrow> emeasure (Pi\<^sub>M I M) X = (\<integral>\<^sup>+y. emeasure (Pi\<^sub>M I M) {f\<in>space (Pi\<^sub>M I M). f(i := y) \<in> X} \<partial>M i)"
using emeasure_PiM_insert[of X i] by(simp add: insert_absorb)

lemma prob_PiM_insert':
  assumes X[measurable]: "Measurable.pred (Pi\<^sub>M I M) X"
  and "i \<in> I"
  shows "\<P>(f in Pi\<^sub>M I M. X f) = (\<integral>\<^sup>+y. \<P>(f in Pi\<^sub>M I M. X (f(i := y))) \<partial>M i)"
by(simp add: emeasure_eq_measure[symmetric] emeasure_PiM_insert'[OF _ \<open>i \<in> I\<close>] fun_upd_in_space_PiM \<open>i \<in> I\<close> cong: conj_cong nn_integral_cong)

end

context prob_space begin

interpretation S: prob_space "stream_space M" by(rule prob_space_stream_space)

lemma stream_space_P_AE_I_eq_1:
  "\<lbrakk> emeasure (stream_space M) {x \<in> space (stream_space M). P x} = 1;
    {x \<in> space (stream_space M). P x} \<in> sets (stream_space M) \<rbrakk>
  \<Longrightarrow> almost_everywhere (stream_space M) P"
by(fact S.AE_I_eq_1)

end

lemma sprefix_in_sets_stream_space:
  assumes "\<And>x. x \<in> set xs \<Longrightarrow> {x} \<in> sets M"
  shows "{ys \<in> space (stream_space M). sprefix xs ys} \<in> sets (stream_space M)"
using assms
proof(induction xs)
  case Nil thus ?case by simp
next
  case (Cons x xs)
  from Cons.prems have x: "{x} \<in> sets M" and xs: "\<And>x. x \<in> set xs \<Longrightarrow> {x} \<in> sets M" by auto
  have "{ys \<in> space (stream_space M). sprefix (x # xs) ys} = 
    (stl -` {ys \<in> space (stream_space M). sprefix xs ys} \<inter> space (stream_space M)) \<inter>
    (shd -` {x} \<inter> space (stream_space M))"
    by(auto simp add: space_stream_space intro: streams_stl)
  also have "\<dots> \<in> sets (stream_space M)"
    by(rule measurable_sets measurable_stl Cons.IH[OF xs] measurable_shd x sets.Int | assumption)+
  finally show ?case .
qed


end
